(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Ceta_Verifier
imports 
  CPF_Parser 
  QDP_Framework_Impl
  Dependency_Pair_Problem_Impl
  Termination_Problem_Impl
begin

abbreviation drop_white_space :: "string \<Rightarrow> string" where
  "drop_white_space \<equiv> filter (\<lambda>c. c \<notin> set wspace)"

definition eq_white_space :: "string \<Rightarrow> string \<Rightarrow> bool" ("_ =w _" 50) where
  "s1 =w s2 \<equiv> debug ''0''
    (''comparing whether parsed input corresponds to real input'')
    (drop_white_space s1 = s2)"

fun search_len :: "string \<Rightarrow> string \<Rightarrow> nat \<Rightarrow> (string \<times> string) option" where
  "search_len t s n = (if take n t = s
    then Some ([],drop n t) else
    (case t of
      Nil \<Rightarrow> None
    | Cons h tt \<Rightarrow>
    (case search_len tt s n of
      None \<Rightarrow> None
    | Some (bef,aft) \<Rightarrow> Some (h # bef,aft))))"

lemma search_len_sound:
  fixes t s before after::string
  assumes "search_len t s n = Some (before,after)"
  shows "t = (before @ s @ after)"
using assms
proof (induct t arbitrary: before)
  case Nil
  thus ?case 
    by (cases s, auto)
next
  case (Cons h tt)
  show ?case 
  proof (cases "take n (h # tt) = s")
    case True
    with Cons show ?thesis by auto 
  next
    case False with `search_len (h # tt) s n = Some (before,after)` 
    have res: "(case search_len tt s n of None \<Rightarrow> None | Some (bef,aft) \<Rightarrow> Some (h # bef, aft)) = Some (before,after)" by auto
    show ?thesis 
    proof (cases "search_len tt s n")
      case None with res show ?thesis by auto
    next
      case (Some befaft)
      show ?thesis proof (cases befaft)
	case (Pair bef aft)
	with Some res have "search_len tt s n = Some (bef,aft) \<and> before = h # bef \<and> after = aft" by auto
	with Cons show ?thesis by auto
      qed
    qed
  qed
qed

declare search_len.simps[simp del]    

abbreviation search :: "string \<Rightarrow> string \<Rightarrow> (string \<times> string)option"
where "search t s \<equiv> search_len t s (length s)"

definition starts_with :: "string \<Rightarrow> string \<Rightarrow> bool"
where "starts_with t s = (take (length s) (trim t) = s)"

abbreviation s_proof :: string where "s_proof \<equiv> ''<proof>''"
abbreviation s_input :: string where "s_input \<equiv> ''<input>''"
abbreviation e_input :: string where "e_input \<equiv> ''</input>''"


fun extract_input_proof :: "string \<Rightarrow> (string \<times> string \<times> string) option"
where "extract_input_proof s = (
   case search (drop_white_space s) s_input of
      None \<Rightarrow> None
   | Some (_,minput) \<Rightarrow> (
   case search minput e_input of 
      None \<Rightarrow> None
   | Some (the_input,s') \<Rightarrow> (
   case search s' s_proof of
      None \<Rightarrow> None
   | Some (version,prf) \<Rightarrow> Some (the_input,version,prf))))" 


lemma extract_input_proof_sound: 
  fixes append :: "string \<Rightarrow> string \<Rightarrow> string" (infixr "<@>" 65)
  defines append[simp]: "append \<equiv> List.append"
  assumes res: "extract_input_proof s = Some (the_input,version,the_proof)"
  shows "\<exists> bef. s =w (bef <@> ''<input>'' <@> the_input <@> ''</input>'' <@> version <@> ''<proof>'' <@> the_proof)"
proof (cases "search (drop_white_space s) s_input")
  case None
  with res show ?thesis by auto
next
  case (Some bef_minput)
  let ?s = "drop_white_space s"
  show ?thesis
  proof (cases bef_minput)
    case (Pair bef minput)
    with Some have s1: "search ?s s_input = Some (bef, minput)" by auto
    show ?thesis 
    proof (cases "search minput e_input")
      case None
      with res s1 show ?thesis by auto
    next
      case (Some input_sp)
      show ?thesis
      proof (cases input_sp)
	case (Pair tthe_input s')
	with Some s1 have s2: "search minput e_input = Some (tthe_input, s')" by auto
	show ?thesis
	proof (cases "search s' s_proof")
	  case None 
	  with res s1 s2 show ?thesis by auto
	next
	  case (Some vers_proof)
	  show ?thesis 
	  proof (cases vers_proof)
	    case (Pair _ _)
	    with Some s1 s2 res have s3: "search s' s_proof = Some (version, the_proof)" by auto		
            from res s1 s2 s3 have input: "tthe_input = the_input" by simp
	    from s1 have sp1: "?s = (bef @ s_input @ minput)" by (rule search_len_sound)
	    from s2 have sp2: "minput = (tthe_input @ e_input @ s')" by (rule search_len_sound)
	    from s3 have sp3: "s' = (version @ s_proof @ the_proof)" by (rule search_len_sound)            
            from sp1 sp2 sp3 have sp: "?s = (bef @ s_input @ tthe_input @ e_input @ version @ s_proof @ the_proof)"
              (is "_ = ?r")
              by simp
            have eq: "s =w ?r" unfolding eq_white_space_def sp by simp
            show ?thesis
              by (rule exI[of _ bef], unfold append, rule eq[unfolded input])
          qed
        qed
      qed
    qed
  qed
qed

text {* here we choose the implementation *}
(*abbreviation tp_impl where "tp_impl \<equiv> tp_list_impl"*)
abbreviation tp_impl where "tp_impl \<equiv> tp_rbt_impl"

(*abbreviation dpp_impl where "dpp_impl \<equiv> dpp_list_impl"*)
abbreviation dpp_impl where "dpp_impl \<equiv> dpp_rbt_impl"

lemma tp_impl: "tp_spec tp_impl" ..
lemma dpp_impl: "dpp_spec dpp_impl" ..

definition max_tag :: nat where
  "max_tag = 29"

lemma max_tag: assumes ti: "tp_spec ti" and di: "dpp_spec di"
  and res: "certify_cert_problem a ti di cp = Certified tag"
  shows "length tag \<le> max_tag"
proof -
  from certify_cert_problem_with_assms_sound[OF ti di res] have
    tag: "tag = xml_tag cp" by simp
  show ?thesis 
  proof (cases cp)
    case (TRS_Termination_Proof nfs a b c d)
    thus ?thesis unfolding max_tag_def tag by (cases c, auto)
  qed (auto simp: max_tag_def tag)
qed


definition eval_list :: "'a list \<Rightarrow> 'a list sparse_result" where 
  "eval_list x = return x"

(* function which enforces that a list is strictly evaluated within the monad *)
fun eval_list_haskell :: "'a list \<Rightarrow> 'a list sparse_result" where
  "eval_list_haskell (x # xs) = do {ys \<leftarrow> eval_list_haskell xs; return (x # ys)}"
| "eval_list_haskell [] = return []"

lemma eval_list_haskell: "eval_list = eval_list_haskell"
proof 
  fix xs :: "'a list"
  have "eval_list_haskell xs = eval_list xs" unfolding eval_list_def
    by (induct xs, auto)
  thus "eval_list xs = eval_list_haskell xs" by simp
qed

lemma trim_take: "trim (take n (trim xs)) = take n (trim xs)"
proof (induct xs arbitrary: n)
  case (Cons x xs)
  thus ?case
    by (cases "x \<in> set wspace", (cases n, auto)+)
qed simp

definition certify_proof :: "bool \<Rightarrow> string \<Rightarrow> cert_result sparse_result" where
  "certify_proof a problemString \<equiv> try (
      case extract_input_proof problemString of
        None \<Rightarrow> return (Error ''could not find input and proof in string for comparison with parsed result'')
      | Some (the_input, version, the_proof) \<Rightarrow> (      
      do {
       short_prf \<leftarrow> eval_list (take max_tag (trim the_proof)); (* to enable garbage collection of the_proof *)
       cp \<leftarrow> parse_cert_problem problemString;
       return
      (if ((show (xml_cert_problem cp)) =w the_input) 
          then case (certify_cert_problem a tp_impl dpp_impl cp) of
                 Certified tag \<Rightarrow>
                   if starts_with short_prf tag 
                     then Certified tag 
                     else Error (''proven property '' @ tag @ '' does not correspond to proof in input string: '' @ short_prf) 
               | e \<Rightarrow> e
          else Error (concat [''parsed problem does not correspond to input'', [newline], 
            ''input:'',[newline], 
            drop_white_space the_input,[newline],[newline],
            ''parsed:'',[newline],drop_white_space (show (xml_cert_problem cp)), [newline]])
          )})) catch (\<lambda> err. return (Unsupported (''error while parsing'' @ [newline] @ err)))"


text {*
if the certifier gives an answer then
- the problem string contains an input and a proof of some desired property
- moreover, there is a certification problem with some input and some proof where the pretty printer delivers exactly the same input as in the problem string
  and the proof-goals are identical
- and for the input the desired property for which the proof has been given is established
*}
lemma certify_proof_sound: 
  fixes append :: "string \<Rightarrow> string \<Rightarrow> string" (infixr "<@>" 65)
  defines "append \<equiv> List.append"
  assumes answer: "certify_proof a problemString = return (Certified answer)"
  shows
    "\<exists>(cp::(string, nat list, string) cert_problem) the_problem the_input version the_proof. 
    problemString =w the_problem <@> ''<input>'' <@> the_input <@> ''</input>''
      <@> version <@> ''<proof>'' <@> the_proof
    \<and> show (xml_cert_problem cp) =w the_input \<and> starts_with the_proof (xml_tag cp)
    \<and> xml_tag cp = answer \<and> ((\<forall>p\<in>set (cert_assms a cp). holds p) \<longrightarrow> desired_property cp)"
proof -
  note answer = answer[unfolded certify_proof_def]
  from answer obtain triple where ex: "extract_input_proof problemString = Some triple"
    by (cases "extract_input_proof problemString", auto)
  obtain the_input version the_proof where triple: "triple = (the_input, version, the_proof)" by (cases triple, auto)
  note ex = ex[unfolded triple]
  from extract_input_proof_sound[OF ex] obtain bef
        where problem: "problemString =w bef @ s_input @ the_input @ e_input
          @ version @ s_proof @ the_proof" ..
  def short \<equiv> "take max_tag (trim the_proof)"
  note answer = answer[unfolded ex option.simps prod.simps eval_list_def short_def[symmetric] return_def]
  from answer obtain cp where parse: "parse_cert_problem problemString = PResult cp"
    by (cases "parse_cert_problem problemString", auto)
  note answer = answer[unfolded parse]
  from answer have eqw: "show (xml_cert_problem cp) =w the_input" by (cases ?thesis, auto)
  from answer eqw obtain answ where cert: "certify_cert_problem a tp_impl dpp_impl cp = Certified answ"
    by (cases "certify_cert_problem a tp_impl dpp_impl cp", auto)
  from certify_cert_problem_with_assms_sound[OF tp_impl dpp_impl cert]
  have sound: "(\<forall>p\<in>set (cert_assms a cp). holds p) \<longrightarrow> desired_property cp" and xml: "xml_tag cp = answ" by auto
  from max_tag[OF tp_impl dpp_impl cert] have len: "length answ \<le> max_tag" by auto
  from answer cert eqw have start: "starts_with short answ" by (cases ?thesis, auto)
  from answer cert eqw start have answ: "answer = answ" by auto
  have "starts_with the_proof (xml_tag cp)
    = (take (length answ) (trim the_proof) = answ)" unfolding xml short_def starts_with_def by simp
  also have "\<dots> = (take (length answ) (trim the_proof) = take (length answ) (trim short))" using start
    unfolding starts_with_def by simp
  also have "trim short = trim (take max_tag (trim the_proof))" unfolding short_def by simp
  also have "\<dots> = take max_tag (trim the_proof)" unfolding trim_take by simp
  also have "take (length answ) (trim the_proof) = take (length answ) (take max_tag (trim the_proof))"
    by (metis length_take take_all take_take len)
  finally have start: "starts_with the_proof (xml_tag cp)" by simp
  show ?thesis unfolding answ append_def
    by (intro exI conjI, rule problem, rule eqw, rule start, rule xml, rule sound)
qed

end

