(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Check_Completion_Proof
imports Check_TP 
begin

subsection {*Proving completion of TRSs*}

type_synonym ('f,'v)conversion = "(('f,'v)equation \<times> ('f,'v)eseq)"
type_synonym ('f, 'v) subsumption_proof = "('f, 'v)conversion list"

datatype ('f, 'l, 'v) completion_proof =
  SN_WCR_Eq
    "('f, 'l) lab join_info"
    "('f, 'l, 'v) trs_termination_proof"
    "(('f, 'l) lab, 'v)subsumption_proof"
    "(('f, 'l) lab, 'v)subsumption_proof option"

fun
  check_completion_proof ::
    "bool \<Rightarrow> shows \<Rightarrow>
     ('tp, ('f::{show,key,countable}, label_type) lab, string) tp_ops \<Rightarrow>
     ('dpp, ('f, label_type) lab, string) dpp_ops \<Rightarrow> (('f,label_type)lab,string) equation list \<Rightarrow> (('f,label_type)lab,string) rules \<Rightarrow>
     ('f, label_type, string) completion_proof \<Rightarrow>
     shows check" 
where
  "check_completion_proof a i I J E R (SN_WCR_Eq joins_i prf conv1 conv2) = debug (i []) ''SN_WCR_Eq'' (do {
    let tp = tp_ops.mk I False [] R []; (*FIXME: default value*)
    check_trs_termination_proof a (i +@+ shows_string ''.1'') I J tp prf
      <+? (\<lambda>s. i +@+ '': error below strong normalization + wcr'' +#+ shows_nl +@+ indent s);
    check_subsumption_guided R E conv1
      <+? (\<lambda> s. i +@+ shows '': error when showing that rewrite relation can be simulated by equations'' +@+ shows_nl +@+ indent s);
    check_subsumption E R conv2
      <+? (\<lambda> s. i +@+ shows '': error when showing that equations can be simulated by rewrite system'' +@+  shows_nl +@+ indent s);
    check_critical_pairs R (critical_pairs_impl R R) joins_i
      <+? (\<lambda>s. i +@+ shows '': error when proving local confluence of '' +@+ shows_nl +@+ shows_tp I tp +@+ shows_nl +@+ indent s)
  })"

fun completion_assms :: "bool \<Rightarrow> ('f, 'l, 'v) completion_proof \<Rightarrow> (('f, 'l, 'v) snassm + ('f, 'l, 'v) fassm) list" where
  "completion_assms a (SN_WCR_Eq _ p _ _) = assms a p"

lemma check_completion_proof_with_assms_sound: 
  assumes I: "tp_spec I" and J: "dpp_spec J"
    and fin: "\<forall>p\<in>set (completion_assms a prf). holds p"
    and ok: "isOK (check_completion_proof a i I J E R prf)"
  shows "completed_rewrite_system (set E) (set R)"
proof -
  interpret tp_spec I by fact
  from ok fin show ?thesis
  proof (induct "prf" arbitrary: i)
    case (SN_WCR_Eq joins_i prof c1 c2 i)
    from SN_WCR_Eq(1)
    have cp: "isOK (check_critical_pairs R (critical_pairs_impl R R) joins_i)"
      and ok: "isOK (check_trs_termination_proof a (i +@+ shows_string ''.1'') I J (mk False [] R []) prof)" 
      and ok1: "isOK (check_subsumption_guided R E c1)"
      and ok2: "isOK (check_subsumption E R c2)" by (auto simp: Let_def)
    from SN_WCR_Eq(2) have a: "\<forall>a\<in>set (assms a prof). holds a" by auto
    from check_critical_pairs[OF cp] have WCR: "WCR (rstep (set R))" .
    from check_trs_termination_proof_with_assms_sound[OF I J a ok] have SN: "SN (rstep (set R))" by simp
    show ?case
      by (rule completion_via_WCR_SN_simulation[OF check_subsumption_guided[OF ok1] check_subsumption[OF ok2]  WCR SN])
  qed
qed

lemma completion_assms_False[simp]: "completion_assms False prf = []"
  by (induct False "prf" rule: completion_assms.induct) simp_all

lemma check_completion_proof_sound: 
  assumes I: "tp_spec I" and J: "dpp_spec J"
    and ok: "isOK (check_completion_proof False i I J E R prf)"
  shows "completed_rewrite_system (set E) (set R)"
  by (rule check_completion_proof_with_assms_sound[OF I J _ ok], simp)

end