(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Conditional_Rewriting
imports
  Trs
  Substitution
  Reduction_Pair
begin

type_synonym ('f, 'v) condition = "('f, 'v) term \<times> ('f, 'v) term"
type_synonym ('f, 'v) crule = "('f, 'v) rule \<times> ('f, 'v) condition list"
type_synonym ('f, 'v) ctrs = "('f, 'v) crule set"

fun cond_step_n :: "('f, 'v) ctrs \<Rightarrow> nat \<Rightarrow> ('f, 'v) trs"
  where "cond_step_n R 0 = {}"
    |   "cond_step_n R (Suc n) = {(C\<langle>l \<cdot> \<sigma>\<rangle>,C\<langle>r \<cdot> \<sigma>\<rangle>) | C l r \<sigma> cs. ((l, r), cs) \<in> R \<and> 
            (\<forall> (si, ti) \<in> set cs. (si \<cdot> \<sigma>, ti \<cdot> \<sigma>) \<in> (cond_step_n R n)^*)}"

lemma cond_step_nI[intro]: "((l,r),cs) \<in> R \<Longrightarrow>
  (\<And> si ti.  (si,ti) \<in> set cs \<Longrightarrow> (si \<cdot> \<sigma>, ti \<cdot> \<sigma>) \<in> (cond_step_n R n)^*) \<Longrightarrow> s = C \<langle> l \<cdot> \<sigma> \<rangle> 
  \<Longrightarrow> t = C \<langle> r \<cdot> \<sigma> \<rangle> \<Longrightarrow> (s,t) \<in> cond_step_n R (Suc n)"
  unfolding cond_step_n.simps by blast

lemma cond_step_nE[elim]: "(s, t) \<in> cond_step_n R (Suc n) \<Longrightarrow>
  (\<And>l r cs \<sigma> C. \<lbrakk> ((l,r),cs) \<in> R;
  (\<And> si ti.  (si,ti) \<in> set cs \<Longrightarrow> (si \<cdot> \<sigma>, ti \<cdot> \<sigma>) \<in> (cond_step_n R n)^*); s = C \<langle> l \<cdot> \<sigma> \<rangle>;
  t = C \<langle> r \<cdot> \<sigma> \<rangle>\<rbrakk> \<Longrightarrow> P) \<Longrightarrow> P"
  unfolding cond_step_n.simps by blast

lemma cond_step_n_ctxt_subst: assumes step: "(s,t) \<in> cond_step_n R n"
  shows "(D \<langle>s \<cdot> \<sigma> \<rangle>,D \<langle>t \<cdot> \<sigma> \<rangle>) \<in> cond_step_n R n"
  using step
proof (induct n arbitrary: s t D)
  case 0
  thus ?case by simp
next
  case (Suc n s t D)
  from Suc(2) obtain l r cs C \<mu> where lr: "((l,r),cs) \<in> R" and
    cs: "\<And> si ti.  (si,ti) \<in> set cs \<Longrightarrow> (si \<cdot> \<mu>, ti \<cdot> \<mu>) \<in> (cond_step_n R n)^*"
    and s: "s = C \<langle> l \<cdot> \<mu> \<rangle>" and t: "t = C \<langle> r \<cdot> \<mu> \<rangle>" by blast
  show ?case unfolding s t
  proof (rule cond_step_nI[OF lr, of "\<mu> \<circ>s \<sigma>" _ _ "D \<circ>c (C \<cdot>c \<sigma>)"])
    fix si ti 
    {
      fix s t
      assume "(s,t) \<in> (cond_step_n R n)^*"
      hence "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (cond_step_n R n)^*"
      proof (induct)
        case base show ?case by simp
      next
        case (step t u)
        from step(3) Suc(1)[OF step(2), of Hole] show ?case by auto
      qed
    } note main = this
    assume "(si, ti) \<in> set cs" 
    from main[OF cs[OF this]]
    show "(si \<cdot> \<mu> \<circ>s \<sigma>, ti \<cdot> \<mu> \<circ>s \<sigma>) \<in> (cond_step_n R n)^*" by simp
  qed auto
qed

definition cond_step :: "('f,'v)ctrs \<Rightarrow> ('f,'v)trs" 
  where "cond_step R \<equiv> \<Union>n. cond_step_n R n"

definition quasi_reductive :: "('f,'v)ctrs \<Rightarrow> bool"
  where "quasi_reductive R \<equiv> \<exists> S NS. redpair_order S NS \<and> ctxt_closed S \<and> (\<forall> l r cs (\<sigma> :: ('f,'v)subst). ((l,r),cs) \<in> R \<longrightarrow> 
         (\<forall> i. i < length cs 
         \<longrightarrow> (\<forall> j. j < i \<longrightarrow> (fst (cs ! j) \<cdot> \<sigma>, snd (cs ! j) \<cdot> \<sigma>) \<in> S^=) 
         \<longrightarrow> (l \<cdot> \<sigma>, fst (cs ! i) \<cdot> \<sigma>) \<in> (S \<union> op \<rhd>)^+) \<and> 
         ((\<forall> j. j < length cs \<longrightarrow> (fst (cs ! j) \<cdot> \<sigma>, snd (cs ! j) \<cdot> \<sigma>) \<in> S^=)
          \<longrightarrow> (l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> S)
        )"

lemma quasi_reductive_SN: assumes qr: "quasi_reductive (R :: ('f,'v)ctrs)"
  shows "SN (cond_step R)"
proof -
  from qr[unfolded quasi_reductive_def]
  obtain S NS where rp: "redpair_order S NS"
  and mono: "ctxt_closed S"
  and qr: "\<And> l r cs (\<sigma> :: ('f,'v)subst).
             ((l, r), cs) \<in> R \<Longrightarrow>
             (\<forall>j<length cs. (fst (cs ! j) \<cdot> \<sigma>, snd (cs ! j) \<cdot> \<sigma>) \<in> S^=) \<Longrightarrow> (l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> S" by blast
  interpret redpair_order S NS by fact
  show ?thesis
  proof (rule SN_subset[OF SN])
    show "cond_step R \<subseteq> S"
    proof
      fix s t
      assume "(s,t) \<in> cond_step R"
      then obtain n where "(s,t) \<in> cond_step_n R n" unfolding cond_step_def by auto
      thus "(s,t) \<in> S"
      proof (induct n arbitrary: s t)
        case 0 thus ?case by simp
      next
        case (Suc n s t)
        from Suc(2)
        obtain l r cs C \<sigma> where lr: "((l,r),cs) \<in> R" and
          cs: "\<And> si ti.  (si,ti) \<in> set cs \<Longrightarrow> (si \<cdot> \<sigma>, ti \<cdot> \<sigma>) \<in> (cond_step_n R n)^*"
          and s: "s = C \<langle> l \<cdot> \<sigma> \<rangle>" and t: "t = C \<langle> r \<cdot> \<sigma> \<rangle>" by blast
        note qr = qr[OF lr, of \<sigma>]
        have "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> S"
        proof (rule qr[rule_format])
          fix j
          assume j: "j < length cs"
          obtain sj tj where csj: "cs ! j = (sj,tj)" by force
          from j have "(sj,tj) \<in> set cs" unfolding csj[symmetric] set_conv_nth by auto
          from cs[OF this] have steps: "(sj \<cdot> \<sigma>, tj \<cdot> \<sigma>) \<in> (cond_step_n R n)^*" .
          {
            fix s t
            assume "(s,t) \<in> (cond_step_n R n)^*"
            hence "(s,t) \<in> S^="
            proof (induct)
              case base thus ?case by simp
            next
              case (step t u)
              from step(3) Suc(1)[OF step(2)] have "(s,u) \<in> S^= O S" by auto
              thus "(s,u) \<in> S^=" using trans_S_point by blast
            qed
          }
          from this[OF steps]
          show "(fst (cs ! j) \<cdot> \<sigma>, snd (cs ! j) \<cdot> \<sigma>) \<in> S^=" unfolding csj by simp
        qed
        from ctxt_closedE[OF mono this, of C]
        show ?case unfolding s t .
      qed
    qed
  qed
qed
     
locale unraveling =  
  fixes R :: "('f,'v)ctrs"
  and U :: "('f,'v)crule \<Rightarrow> nat \<Rightarrow> ('f,'v)ctxt"
  (* e.g., U ((l,r),s1 = t1, ... sn = tn) i = U_i(\<box>,vars(l) \<union> vars(t1) ... \<union> vars(t i-1) *)
begin

fun lhs_n :: "('f,'v)crule \<Rightarrow> nat \<Rightarrow> ('f,'v)term"
  where "lhs_n (lr,cs) 0 = fst lr"
      | "lhs_n (lr,cs) (Suc n) = (U (lr,cs) n) \<langle>snd (cs ! n)\<rangle>"

fun rhs_n :: "('f,'v)crule \<Rightarrow> nat \<Rightarrow> ('f,'v)term"
  where "rhs_n (lr,cs) n = (if n < length cs then 
         (U (lr,cs) n)\<langle>fst (cs ! n)\<rangle>
         else snd lr)"

definition rules :: "('f,'v)crule \<Rightarrow> ('f,'v)trs"
  where "rules crule \<equiv> (\<lambda> n. (lhs_n crule n, rhs_n crule n)) ` {n . n \<le> length (snd crule)}"

definition UR :: "('f,'v)trs"
  where "UR \<equiv> \<Union> rules ` R"

lemma UR_simulation_main: assumes lr: "((l, r), cs) \<in> R"
  and i: "i \<le> length cs"
  and cs: "\<And> j. j < i \<Longrightarrow> (fst (cs ! j) \<cdot> \<sigma>, snd (cs ! j) \<cdot> \<sigma>) \<in> (rstep UR)^*"
  shows "(l \<cdot> \<sigma>, rhs_n ((l,r),cs) i \<cdot> \<sigma>) \<in> (rstep UR)^+"
proof -
  let ?lr = "((l,r),cs)"
  let ?n = "length cs"
  from lr have lr: "rules ?lr \<subseteq> UR" unfolding UR_def by auto
  hence lr: "\<And> i. i \<le> ?n \<Longrightarrow> (lhs_n ?lr i, rhs_n ?lr i) \<in> UR" unfolding rules_def
    by auto
  show ?thesis
  proof (rule rtrancl_into_trancl2)
    show "(l \<cdot> \<sigma>,rhs_n ?lr 0 \<cdot> \<sigma>) \<in> rstep UR"
      by (rule rstepI[OF lr[of 0], of _ Hole \<sigma>], auto)
    from i cs
    show "(rhs_n ?lr 0 \<cdot> \<sigma>, rhs_n ?lr i \<cdot> \<sigma>) \<in> (rstep UR)^*"
    proof (induct i)
      case 0 thus ?case by simp
    next
      case (Suc i)
      hence i: "i < ?n" and i': "i \<le> ?n" by auto
      let ?f = "U ?lr i"
      let ?U = "\<lambda> t. ?f \<langle> t \<rangle>"
      from Suc(1)[OF i' Suc(3)]
      have "(rhs_n ?lr 0 \<cdot> \<sigma>, rhs_n ?lr i \<cdot> \<sigma>) \<in> (rstep UR)^*" by simp
      also have "rhs_n ?lr i = ?U (fst (cs ! i))"
        using i by simp
      also have "(?U (fst (cs ! i)) \<cdot> \<sigma>, ?U (snd (cs ! i)) \<cdot> \<sigma>) \<in> (rstep UR)^*" 
      proof -
        obtain si ti where csi: "cs ! i = (si,ti)" by force
        from Suc(3)[of i, unfolded csi] have "(si \<cdot> \<sigma>, ti \<cdot> \<sigma>) \<in> (rstep UR)^*" by simp
        from ctxt_closedE[OF ctxt_closed_rtrancl[OF ctxt_closed_rstep] this]
        have "\<And> C. (C\<langle>si \<cdot> \<sigma>\<rangle>, C\<langle>ti \<cdot> \<sigma>\<rangle>) \<in> (rstep UR)^*" .
        from this[of "?f \<cdot>c \<sigma>"]
        show ?thesis unfolding csi by simp
      qed
      also have "?U (snd (cs ! i)) = lhs_n ?lr (Suc i)" 
        unfolding lhs_n.simps by simp
      also have "(lhs_n ?lr (Suc i) \<cdot> \<sigma>, rhs_n ?lr (Suc i) \<cdot> \<sigma>) \<in> (rstep UR)"
        by (rule rstepI[OF lr[OF Suc(2)], of _ Hole \<sigma>], auto)
      finally show ?case .
    qed
  qed
qed

lemma SN_UR_quasi_reductive: assumes SN: "SN (rstep UR)"
  shows "quasi_reductive R"
proof -
  let ?S = "(rstep UR)^+"
  let ?NS = "(rstep UR)^*"
  have id: "?S^= = ?NS" by simp
  show ?thesis unfolding quasi_reductive_def
  proof (rule exI[of _ ?S], rule exI[of _ ?NS], unfold id, intro conjI allI impI)
    show "redpair_order ?S ?NS"
    proof
      show "SN ?S" by (rule SN_imp_SN_trancl[OF SN])
      show "ctxt_closed ?NS" by auto
      show "subst_closed ?NS" by (rule subst_closed_rtrancl, auto) 
      show "subst_closed ?S" by (rule subst_closed_trancl, auto) 
    qed auto
    show "ctxt_closed ?S" by auto
  next
    fix l r cs \<sigma>
    assume lr: "((l, r), cs) \<in> R"
      and cs: "\<forall>j<length cs. (fst (cs ! j) \<cdot> \<sigma>, snd (cs ! j) \<cdot> \<sigma>) \<in> (rstep UR)^*"
    from UR_simulation_main[OF lr, of "length cs" \<sigma>] cs
    show "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> (rstep UR)^+" by simp
  next
    fix l r cs \<sigma> i
    assume lr: "((l, r), cs) \<in> R"
    and i: "i < length cs"
    and cs: "\<forall>j<i. (fst (cs ! j) \<cdot> \<sigma>, snd (cs ! j) \<cdot> \<sigma>) \<in> (rstep UR)^*"
    from UR_simulation_main[OF lr, of i \<sigma>] i cs
    have lm: "(l \<cdot> \<sigma>, rhs_n ((l, r), cs) i \<cdot> \<sigma>) \<in> (rstep UR)^+" (is "(?l,?m) \<in> _") by simp
    have mr: "?m \<unrhd> fst (cs ! i) \<cdot> \<sigma>" (is "_ \<unrhd> ?r")
      by (rule supteq_stable, insert i, auto)
    let ?R = "(rstep UR)^+ \<union> op \<rhd>"
    from mr have "(?m,?r) \<in> (op \<rhd>)^=" by auto
    hence mr: "(?m,?r) \<in> ?R^*" by auto
    have lm: "(?l,?m) \<in> ?R^+"
      by (rule trancl_mono[OF lm], auto)
    from lm mr show lr: "(?l,?r) \<in> ?R^+" by auto
  qed
qed

lemma UR_complete: assumes step: "(s,t) \<in> cond_step R"
  shows "(s,t) \<in> (rstep UR)^+"
proof -
  from step[unfolded cond_step_def] obtain n where "(s,t) \<in> cond_step_n R n" ..
  thus ?thesis
  proof (induct n arbitrary: s t)
    case 0 thus ?case by simp
  next
    case (Suc n) note IH = this
    from IH(2) obtain l r cs C \<sigma> where lr: "((l,r),cs) \<in> R" and
    cs: "\<And> si ti.  (si,ti) \<in> set cs \<Longrightarrow> (si \<cdot> \<sigma>, ti \<cdot> \<sigma>) \<in> (cond_step_n R n)^*"
    and s: "s = C \<langle> l \<cdot> \<sigma> \<rangle>" and t: "t = C \<langle> r \<cdot> \<sigma> \<rangle>" by blast
    let ?lr = "((l,r),cs)"
    let ?n = "length cs"
    have rsig: "r \<cdot> \<sigma> = rhs_n ?lr ?n \<cdot> \<sigma>" by auto
    have "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> (rstep UR)^+" unfolding rsig
    proof (rule UR_simulation_main[OF lr, of "length cs" \<sigma>])
      fix j
      assume j: "j < length cs"
      obtain sj tj where csj: "cs ! j = (sj,tj)" by force
      from j have "(sj,tj) \<in> set cs" unfolding csj[symmetric] set_conv_nth by auto
      from cs[OF this] have steps: "(sj \<cdot> \<sigma>, tj \<cdot> \<sigma>) \<in> (cond_step_n R n)^*" .
      {
        fix s t
        assume "(s,t) \<in> (cond_step_n R n)^*"
        hence "(s,t) \<in> (rstep UR)^*"
        proof (induct)
          case base thus ?case by auto
        next
          case (step t u)
          from step(3) IH(1)[OF step(2)] show ?case by auto
        qed
      }
      from this[OF steps]
      show "(fst (cs ! j) \<cdot> \<sigma>, snd (cs ! j) \<cdot> \<sigma>) \<in> (rstep UR)^*" 
        unfolding csj by auto
    qed simp
    from ctxt_closedE[OF ctxt_closed_trancl[OF ctxt_closed_rstep] this, of C]
    show ?case unfolding s t .
  qed
qed

end

definition eqs where "eqs \<mu> = map (\<lambda>(x, t). (t, Var x)) \<mu>"

fun
  solve_conditions ::
    "(('f, 'v) term \<Rightarrow> ('f, 'v) term list option) \<Rightarrow>
    ('f, 'v) condition list \<Rightarrow>
    ('v \<times> ('f, 'v) term) list \<Rightarrow>
    ('v \<times> ('f, 'v) term) list list option" where
  "solve_conditions f [] \<mu> = Some [\<mu>]" |
  "solve_conditions f ((s, t) # cs) \<mu> = do {
    us \<leftarrow> f (s \<cdot> mk_subst Var \<mu>);
    fold (\<lambda>u res. do {
      mus \<leftarrow> res;
      (case match_list ((t \<cdot> mk_subst Var \<mu>, u) # eqs \<mu>) of
        None \<Rightarrow> Some mus
      | Some \<tau> \<Rightarrow> do {
        mus' \<leftarrow> solve_conditions f cs \<tau>;
        Some (mus @ mus')
      })
    }) us (Some [])
  }"

(* mono_option problem 
find_theorems "match_list"

lemma match_term_list_eqs [simp]:
  "match_term_list (eqs \<mu>) = Some \<mu>"
  by (induct \<mu>) (auto simp: eqs_def)

lemma "match_list t es = Some bs \<Longrightarrow> match_list t (eqs bs) = Some bs"
  apply (induct es arbitrary: bs)
  apply (simp add: eqs_def)
  apply (auto simp add:  split: option.splits)

lemma
  assumes "match_term_list es1 = Some \<tau>"
    and "match_term_list (eqs \<tau> @ es2) = Some \<mu>"
  shows "match_term_list (es1 @ es2) = Some \<mu>"
  using assms
  apply (induct es1 arbitrary: \<tau> \<mu> es2)
  apply (simp add: eqs_def)
  apply (simp add: eqs_def)
  apply auto
  apply (case_tac b)
  apply simp
  apply force
  apply (case_tac a)
  apply simp
  apply simp
  apply auto

lemma
  assumes "match_term_list (ss @ ts) = Some \<tau>"
  shows "\<exists>\<mu>. match_term_list ss = Some \<mu> \<and> match_term_list (ts @ eqs \<mu>) = Some \<tau>"
  using assms
  apply (induct ss arbitrary: ts \<tau>)
  apply (simp add: eqs_def)
  apply simp
  apply (case_tac a)
  apply (case_tac b)
  apply auto

find_theorems "match_term_list"

lemma solve_conditions_sound:
  assumes f_sound:
    "\<And>s us t. \<lbrakk>f s = Some us; t \<in> set us\<rbrakk> \<Longrightarrow>
      (s, t) \<in> (cond_step (set R))^*"
    and some: "solve_conditions f cs \<sigma> = Some mus"
    and "\<mu> \<in> set mus"
    and "(s, t) \<in> set cs"
  shows "(s \<cdot> mk_subst \<mu>, t \<cdot> mk_subst \<mu>) \<in> (cond_step (set R))^*"

partial_function (option)
  cond_rewrite :: "('f, 'v) crule list \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term list option"
where
  [code]: "cond_rewrite R s = Some (do {
    ((l, r), cs) \<leftarrow> R;
    p \<leftarrow> poss_list s;
    \<sigma> \<leftarrow> option_list (match (s |_ p) l);
    (si, ti) \<leftarrow> cs;
    us \<leftarrow> the (cond_rewrite R si);
    []
  })"
*)

hide_const (open) eqs;

end
