(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Decreasing_Diagrams
imports
  "$AFP/Abstract-Rewriting/Abstract_Rewriting"
  "~~/src/HOL/Library/Multiset"
begin
  
(*Appendix*)
definition remove :: "'a multiset \<Rightarrow> 'a set \<Rightarrow> 'a multiset" ("(_-s_)" [999] 998)
 where "M -s S = Multiset.filter (\<lambda>x. x \<notin> S) M"

lemma remove_from_empty: "{#}-s N = {#}" unfolding remove_def by auto

definition intersect :: "'a multiset \<Rightarrow> 'a set \<Rightarrow> 'a multiset" ("(_\<inter>s_)" [999] 998)
 where "M \<inter>s S = Multiset.filter (\<lambda>x. x \<in> S) M"

lemma submultiset_implies_subset: assumes "M <= N" shows "set_of M \<subseteq> set_of N"  proof 
 fix x assume "x \<in> set_of M" thus "x \<in> set_of N" unfolding set_of_def mset_le_def 
by (metis (full_types) assms mem_Collect_eq mset_leD neq0_conv)
qed

lemma subset_implies_remove_empty: assumes "set_of M \<subseteq> A" shows "M -s A = {#}" using assms proof (induct M)
  case empty thus ?case unfolding remove_def by auto 
 next
  case (add N n) hence IH1: "set_of N \<subseteq> A \<Longrightarrow> N -s A = {#}" and IH2: "set_of (N + {#n#}) \<subseteq> A" by auto
  from IH1 IH2 have a: "N -s A = {#}" by auto 
  from IH2 have "n \<in> A" by auto
  thus ?case using a unfolding remove_def by auto
qed

lemma remove_empty_implies_subset: assumes "M -s A = {#}" shows "set_of M \<subseteq> A" proof
 fix x assume A: "x \<in> set_of M" 
 hence "x \<in># (M -s A) \<equiv> False" using assms unfolding mem_set_of_iff[symmetric] remove_def by auto
 thus "x \<in> A" using A unfolding mem_set_of_iff[symmetric] remove_def by auto
qed

(*changed order of parameters*)
lemma lemmaA_3_4: "(M + N) \<inter>s S = (M \<inter>s S) + (N \<inter>s S)"  unfolding intersect_def filter_union by simp
lemma lemmaA_3_5: "(S \<inter> (T - U)) = (S \<inter> T) - (S \<inter> U)" by auto 
lemma lemmaA_3_6: "(S \<inter> T) - U = (S - U) \<inter> (T - U)" by auto
lemma lemmaA_3_7: "(M + N) -s S = (M -s S) + (N -s S)" unfolding remove_def by (rule multiset_eqI) simp
lemma lemmaA_3_8:  "(M + N) -s S = (M -s S) + (N -s S)" unfolding remove_def by (rule multiset_eqI) simp 
lemma lemmaA_3_9: "(((M -s S) -s T)::'a multiset) = M -s (S \<union> T)" unfolding remove_def by (rule multiset_eqI) simp
lemma lemmaA_3_10: "M = (M \<inter>s S) + (M -s S)" 
 using multiset_partition unfolding remove_def intersect_def filter_def  by auto
lemma lemmaA_3_11: "(M -s T) \<inter>s S = (M \<inter>s S) -s T" unfolding remove_def intersect_def
 by (rule multiset_eqI) simp

(*Definition 2.5(1)*)
definition downset_elt :: "('a \<times>'a) set \<Rightarrow> 'a \<Rightarrow> 'a set" ("(de _ _)" [1000] 999)
 where "downset_elt r x = {y | y. (y,x) \<in> r}"

definition downset_set :: "('a \<times> 'a) set \<Rightarrow> 'a set \<Rightarrow> 'a set" ("(ds _ _)" [1000] 999)
 where "downset_set R s = {y | x y. x \<in> s \<and> y \<in> downset_elt R x}"

definition downset_multiset :: "('a \<times> 'a) set \<Rightarrow> 'a multiset \<Rightarrow> 'a set" ("(dm _ _)" [1000] 999)
 where "downset_multiset R M = downset_set R (Multiset.set_of M)"

definition downset_list :: "('a \<times> 'a) set \<Rightarrow> 'a list \<Rightarrow> 'a set" ("(dl _ _)" [1000] 999)
 where "downset_list R xs = downset_set R (List.set xs)"

lemma downset_set_singleton: "ds r {a} = de r a" unfolding downset_elt_def downset_set_def by auto
lemma downset_multiset_singleton: "dm r {#a#} = de r a" unfolding downset_multiset_def set_of_single downset_set_singleton by auto 
lemma downset_list_singleton:"dl r [a] = de r a" unfolding downset_list_def set.simps(2) set.simps(1) downset_set_singleton by auto 

lemma downset_list_empty: "dl r [] = {}" unfolding downset_list_def downset_set_def by auto

(*Definition 2.5(2)*)

(*strict order (mult) is used from Multiset.thy*)
definition mult_eq :: "('a \<times> 'a) set => ('a multiset \<times> 'a multiset) set" where
  "mult_eq r = (mult1 r)\<^sup>*"

lemma mult_eq_downset: 
assumes t: "trans r" and "\<forall>k\<in> set_of K. \<exists>j\<in> set_of J. (k,j) \<in> r" 
shows "set_of K \<subseteq> dm r J" 
proof
 fix x assume a: "x\<in> set_of K" show "x \<in> dm r J" 
 proof - 
  from a assms(2) obtain j where "j\<in> set_of J" and "(x,j)\<in> r" by fast 
  thus ?thesis unfolding downset_multiset_def downset_set_def downset_elt_def by auto 
 qed
qed

lemma mult_eq_downset_inv:
assumes t: "trans r" and "set_of K \<subseteq> dm r J"
shows "\<forall>k\<in>set_of K.\<exists>j\<in>set_of J.(k,j)\<in> r" 
proof
 fix k assume a: "k\<in> set_of K" show "\<exists>j\<in>set_of J. (k,j)\<in> r" proof -
  from assms a have "k\<in> dm r J" by auto
  thus ?thesis unfolding downset_multiset_def downset_set_def downset_elt_def by auto
 qed
qed

lemma mult_imp_mult_eq: assumes "(x,y) \<in> mult r" shows "(x,y) \<in> mult_eq r" 
 using assms trancl_into_rtrancl unfolding mult_def mult_eq_def by auto 

lemma mult_eq_and_not_mult_implies_equal: assumes "(M,N) \<in> mult_eq r" and "(M,N) \<notin> mult r" shows "M = N" 
 using assms unfolding mult_eq_def mult_def by (metis rtrancl_eq_or_trancl)

text {* to closure-free representation *}
lemma mult_eq_implies_one_or_zero_step: 
assumes "trans r" and "(M,N) \<in> mult_eq r" 
shows "\<exists>I J K. N = I + J \<and> M = I + K \<and> (\<forall>k \<in> set_of K. \<exists>j \<in> set_of J. (k, j) \<in> r)" 
proof (cases "(M,N) \<in> mult r")
  case True thus ?thesis using mult_implies_one_step[OF assms(1)] by blast 
 next
  case False from assms mult_eq_and_not_mult_implies_equal False have A: "M = N" by auto  
  hence "N = N + {#} \<and> M = M + {#} \<and> (\<forall>k\<in>set_of {#}. \<exists>j\<in>set_of {#}. (k, j) \<in> r)" by auto
  thus ?thesis unfolding A by fast 
qed

(*
lemma mult_implies_one_step:
assumes "trans r" and "(M,N) \<in> mult r"
shows "\<exists> I J K. N = I + J \<and> M = I + K \<and> (\<forall>k \<in> set_of K. \<exists>j \<in> set_of J.(k,j) \<in> r) \<and> J \<noteq>{#}" 
*)

text {* from closure-free representation *}
lemma one_step_implies_mult_eq: 
assumes "trans r" and "\<forall>k\<in> set_of K.\<exists>j\<in>set_of J. (k,j)\<in>r" 
shows "(I+K,I+J)\<in>mult_eq r"
proof (cases "set_of J = {}")
  case True 
  hence "set_of K = {}" using assms by (metis all_not_in_conv emptyE)
  hence "I+K = I" and "I+J=I" using True by auto 
  thus ?thesis unfolding mult_eq_def by auto 
 next
  case False
   hence h:"J \<noteq> {#}" using set_of_eq_empty_iff by auto
   hence "(I+K,I+J)\<in> mult r" using assms one_step_implies_mult by fast 
   thus ?thesis unfolding mult_eq_def mult_def by auto
 qed

(*Lemma 2.6*)

(*Lemma 2.6.1*)
lemma lemma2_6_1_set: "ds r (S \<union> T) = (ds r S) \<union> (ds r T)" unfolding downset_set_def by auto

lemma lemma2_6_1_list: "dl r (xs@ys) = dl r xs \<union> dl r ys" 
 unfolding downset_list_def using lemma2_6_1_set by (metis set_append)

lemma lemma2_6_1_list_elt: "dl r (\<alpha>#\<sigma>) = (de r \<alpha>) \<union> (dl r \<sigma>)" proof - 
 have "dl r (\<alpha>#\<sigma>) = dl r ([\<alpha>] @ \<sigma>)" by simp
 thus ?thesis unfolding lemma2_6_1_list downset_list_def downset_set_def by auto 
qed

lemma lemma2_6_1_multiset: "dm r (M + N) = dm r M \<union> dm r N"
 unfolding downset_multiset_def set_of_union using lemma2_6_1_set by metis 

lemma lemma2_6_1_diff: "(dm r M) - ds r S \<subseteq> dm r (M -s S)" proof (rule subsetI)
 fix x assume a: "x\<in> (dm r M) - ds r S" show "x \<in> dm r (M-s S)" proof -
 from a have b: "x\<in> dm r M" and c: "x \<notin> ds r S" by auto
 from b obtain y where d: "y\<in> set_of M" and e: "(x,y) \<in> r" unfolding downset_multiset_def downset_set_def downset_elt_def by auto
 from c have "\<not>(\<exists>y\<in>S.(x,y)\<in> r)" unfolding downset_multiset_def downset_set_def downset_elt_def by auto
 hence "y \<notin> S" using e by auto 
 hence "y \<in> set_of (M-s S)" using d unfolding remove_def by auto
 thus ?thesis using e unfolding downset_multiset_def downset_set_def downset_elt_def by auto 
qed
qed

(*Lemma 2.6.2*)

lemma downset_mem_imp_exists: assumes "x \<in> ds r S" shows "\<exists> y \<in> S. (x,y) \<in> r" 
 using assms unfolding downset_set_def downset_elt_def by auto 

lemma downset_exists_imp_mem: assumes "\<exists> y \<in> S. (x,y) \<in> r" shows "x \<in> ds r S" 
 using assms unfolding downset_set_def downset_elt_def by auto
 
lemma mult_eq_not_equal_imp_elt: 
 assumes t: "trans r" and "(M,N) \<in> mult_eq r" and "y \<in> set_of M" and "\<not>(y \<in> set_of N)"
 shows "\<exists> z \<in> set_of N. (y,z) \<in> r"
proof - from assms(1) obtain I J K where 
  F: "N = I + J \<and> M = I + K \<and> (\<forall>k \<in> set_of K. \<exists>j \<in> set_of J. (k, j) \<in> r)" 
  using mult_eq_implies_one_or_zero_step[OF t assms(2)] by auto 
 have F1: "N = I + J" and F2: "M = I + K" and F3: "(\<forall>k \<in> set_of K. \<exists>j \<in> set_of J. (k, j) \<in> r)" using F by auto
 have F4: "y \<notin> set_of I" using F1 assms(4) by auto
 have K: "y \<in> set_of K" using assms(3) F2 F4 by auto 
 have J: "y \<notin> set_of J" using assms(4) F1 F4 by auto
 thus ?thesis using K J F1 F3 by auto
qed

lemma lemma2_6_2_a: assumes t: "trans r" and "M \<le> N" shows "(M,N) \<in> mult_eq r" proof -
 from assms(2) obtain J where eq: "N=M+J" unfolding mset_le_def multiset_eq_iff by (metis assms(2) mset_le_exists_conv) 
 hence "\<forall>k\<in>set_of {#}.\<exists>j\<in>set_of J. (k,j)\<in>r" by auto
 hence "(M+{#},M+J)\<in> mult_eq r" using one_step_implies_mult_eq[OF t] by fast 
 thus ?thesis using eq by auto 
qed

lemma lemma2_6_2_list_b: assumes "trans r" and "(M,N) \<in> mult_eq r" shows "dm r M \<subseteq> dm r N"
using assms apply auto
proof - 
 fix x assume A1: "trans r" and A2: "(M,N) \<in> mult_eq r" and A3: "x \<in> dm r M" 
       show "x \<in> dm r N" proof -
 have F1: "x \<in> ds r set_of M" using A1 A2 A3 unfolding downset_multiset_def by auto
 from this obtain y where F2: "y \<in> set_of M" and F3: "(x,y) \<in> r" 
  using downset_mem_imp_exists by fast 
 hence "\<exists> z \<in> set_of N. (x,z) \<in> r" proof (cases "y \<in> set_of N")
  case True thus ?thesis using F3 downset_exists_imp_mem by auto 
 next
  case False have "\<exists> z \<in> set_of N. (y,z) \<in> r" using mult_eq_not_equal_imp_elt[OF assms(1) A2 F2 False] by simp
  thus ?thesis using F3 A1 trans_def by fast 
 qed
 thus ?thesis using downset_exists_imp_mem unfolding downset_multiset_def by fast
qed 
qed

(*Lemma 2.6.3*)
lemma extract_elt: assumes t: "trans r" and wf: "wf r" and P: "\<And>k. k \<in> set_of K \<Longrightarrow> \<exists> j \<in> set_of J. (k,j) \<in> r"
  and K: "k \<in> set_of (K - (J#\<inter>K))"
  shows "\<exists> j \<in> set_of (J - (J#\<inter>K)). ((k,j) \<in> r)" 
proof -
  have "J - K = J - (J #\<inter> K)" by (intro multiset_eqI, simp)
  moreover have "k \<in># K"  using K by simp
  moreover have "\<And>k. k \<in># K \<Longrightarrow> \<exists> j \<in> set_of (J - K). (k, j) \<in> r"
  proof -
    fix k
    assume k: "k \<in># K"
    {
      fix K'
      from k have "K' \<le> K \<Longrightarrow> \<exists> j \<in> set_of (J - K'). (k, j) \<in> r"
      proof (induct K' arbitrary: k rule: multiset_induct)
        case empty thus ?case using P [of "k"] by auto
      next
        case (add Q q)
        from add(2) have "Q \<le> K" using order_trans [of "Q" "Q + {#q#}" "K"] by auto
        from add(2) have "\<forall>a. count (Q+{#q#}) a \<le> count K a" unfolding mset_le_def by auto
        hence "q \<in># K" 
        proof 
          assume "count (Q + {#q#}) q \<le> count K q"
          thus "q \<in># K" by simp
        qed
        obtain j where "j\<in>set_of (J - Q)" and "(k, j) \<in> r"
          using `Q \<le> K` and add(1) [of "k"] add(3) by auto
        show ?case
        proof (cases "j \<in># J - (Q + {#q#})")
          case True thus ?thesis using `(k, j) \<in> r` `j\<in>set_of (J - Q)`
            by (intro bexI [of _ "j"], auto)
        next
          case False
          from False `(k, j) \<in> r` `j \<in> set_of (J - Q)` have "q = j"
            by (cases "q = j") auto
          obtain j2 where "j2\<in>set_of (J - Q)" and "(j, j2) \<in> r"
            using `Q \<le> K` `q = j` `q \<in># K` add(1) [of "j"] by auto
          thus ?thesis
          proof (intro bexI [of _ "j2"])
            from t `(k, j) \<in> r` `(j, j2) \<in> r` transD [of r k j j2]
            show "(k, j2) \<in> r" by simp
          next
            have "j2 \<noteq> q" using `q = j` `(j, j2) \<in> r` wf by auto
            thus "j2 \<in> set_of (J - (Q + {#q#}))"
              using `j2\<in>set_of (J - Q)` by auto
          qed
        qed
      qed
    }
    thus "\<exists> j \<in> set_of (J - K). (k, j) \<in> r" by simp
  qed
  ultimately show ?thesis by auto
qed

lemma mult_eq_imp_disjoint: 
 assumes t: "trans r" and w: "wf r" and "(M,N) \<in> mult_eq r" shows
 "\<exists> I J K. N = I + J \<and> M = I + K \<and> J#\<inter>K = {#} \<and> (\<forall>k\<in> set_of K. \<exists>j\<in> set_of J. (k,j) \<in> r)"
 proof - from assms obtain I J K where 
 f: "N = I + J \<and> M = I + K \<and> (\<forall>k\<in> set_of K. \<exists>j\<in> set_of J. (k,j) \<in> r)" 
 using mult_eq_implies_one_or_zero_step[OF assms(1) assms(3)] by auto
 hence A: "N = (I + (J #\<inter> K)) + (J - (J #\<inter> K))"
  by (metis diff_union_cancelL inf_le2 multiset_diff_union_assoc multiset_inter_commute union_commute union_lcomm)
 hence B: "M = (I + (J #\<inter> K)) + (K - (J #\<inter> K))" 
 by (metis diff_le_self diff_union_cancelL f multiset_diff_union_assoc multiset_inter_commute multiset_inter_def union_assoc)
 have C: "(J-(J#\<inter>K)) #\<inter> (K-(J#\<inter>K)) = {#}" by (rule multiset_eqI) auto
 hence D: "\<forall>k\<in> set_of (K-(J#\<inter>K)). \<exists>j \<in> set_of (J-(J#\<inter>K)). (k,j) \<in> r" using extract_elt[OF assms(1,2)] f by fast 
 have "N = (I+(J#\<inter>K))+(J-(J#\<inter>K)) \<and> M = (I+(J#\<inter>K))+(K-(J#\<inter>K)) \<and> (J-(J#\<inter>K)) #\<inter> (K-(J#\<inter>K)) = {#}  \<and> (\<forall>k\<in> set_of (K-(J#\<inter>K)). \<exists>j \<in> set_of (J-(J#\<inter>K)). (k,j) \<in> r)" using A B C D by auto
 thus ?thesis by metis
qed 


(*Lemma 2.6.4*)
lemma lemma2_6_4: assumes t: "trans r" and "G \<noteq> {#}" and "set_of F \<subseteq> dm r G" shows "(F,G) \<in> mult r" 
 using assms mult_eq_downset_inv one_step_implies_mult empty_neutral(1) by metis 

(*
lemma lemma2_6_4b: assumes t: "trans r" and "set_of F \<subseteq> dm r G" shows "(F,G) \<in> mult_eq r"
proof (cases "G = {#}")
 case True hence "downset_multiset r G = {}" unfolding True downset_multiset_def set_of_def downset_set_def by auto 
 hence "F = {#}" using assms(2) by auto
 thus ?thesis using True unfolding mult_eq_def by auto
next
 case False 
 thus ?thesis using lemma2_6_4[OF t False assms(2)] unfolding mult_eq_def mult_def by auto
qed
*)

(*Lemma 2.6.5*)

(*does lemma2_6_5_simplified suffice?*)
lemma lemma2_6_5:
 assumes t: "trans r" and "ds r S \<subseteq> S" and "(M,N) \<in> mult_eq r" shows "(M -s S, N -s S) \<in> mult_eq r" proof -
 from mult_eq_implies_one_or_zero_step[OF assms(1,3)] mult_eq_downset[OF assms(1)] obtain I J K where
  a: "N=I+J \<and> M=I+K \<and> set_of K \<subseteq> dm r J" by fast 
  from a have A: "N-sS = (I-sS)+(J-sS)" using lemmaA_3_7 by metis
  from a have B: "M-sS = (I-sS)+(K-sS)" using lemmaA_3_7 by metis
  have "set_of (K-sS) \<subseteq> dm r (J-sS)" proof -
  have 1: "set_of (K-sS) \<subseteq> set_of (K-s (ds r S))" proof (rule subsetI) 
   fix x assume b: "x \<in> set_of (K-sS)" show "x\<in> set_of(K-s ds r S)" proof - 
   from b have "x \<in> set_of K" and "x \<notin> S" unfolding remove_def by auto 
   hence "x \<notin> ds r S" using assms(2) by auto 
   hence "x \<in> set_of (K-s ds r S)" using `x\<in> set_of K` unfolding remove_def by auto 
   thus ?thesis by auto
  qed
  qed
  have 2: "set_of(K-s (ds r S)) \<subseteq> (dm r J) - (ds r S)" using a unfolding remove_def by auto 
  have 3: "(dm r J) - (ds r S) \<subseteq> dm r (J -s S)" using lemma2_6_1_diff by fast 
  thus ?thesis using 1 2 3 by auto 
 qed
 hence C: "\<forall>k\<in>set_of(K-sS).\<exists>j\<in>set_of(J-sS).(k,j)\<in>r" using mult_eq_downset_inv[OF assms(1)] by fast 
 show ?thesis using A B C using one_step_implies_mult_eq[OF t] by metis 
qed

lemma ds_ds_subseteq_ds: assumes t: "trans r" shows "ds r (ds r S) \<subseteq> ds r S" proof (rule subsetI)
 fix x assume A: "x \<in> ds r (ds r S)" show "x \<in> ds r S" proof -
 from A obtain y where B: "y \<in> (ds r S)" and C: "(x,y) \<in> r" using downset_mem_imp_exists by fast
 from B obtain z where D: "z \<in> S" and E: "(y,z) \<in> r" using downset_mem_imp_exists by fast 
 have "(x,z) \<in> r" using t C E trans_def by metis
 hence "\<exists> z \<in> S. (x,z) \<in> r" using D by auto  
 thus ?thesis using downset_exists_imp_mem by fast 
qed 
qed

lemma lemma2_6_5_simplified: assumes t: "trans r" and "(M,N) \<in> mult_eq r" shows "(M -s ds r S, N -s ds r S) \<in> mult_eq r" 
 proof -
 have "ds r (ds r S) \<subseteq> ds r S" using ds_ds_subseteq_ds[OF t] by auto
 thus ?thesis using assms lemma2_6_5[OF t] by force 
qed

(*Lemma 2.6.6*)
lemma mult_eq_add_left: assumes t: "trans r" and "(M,N) \<in> mult_eq r" shows "(P + M,P + N) \<in> mult_eq r" proof -
 from assms 
 have "\<exists> I J K. N = I + J \<and> M = I + K \<and> (\<forall>k \<in> set_of K . \<exists> j \<in> set_of J . (k,j) \<in> r)" 
 using mult_eq_implies_one_or_zero_step by auto  
 hence "\<exists> I J K. P + N = P + I + J \<and> P + M = P + I + K \<and> (\<forall>k \<in> set_of K . \<exists> j \<in> set_of J . (k,j) \<in> r)" by (metis union_assoc) 
 hence "\<exists> I J K. P + N = I + J \<and> P + M = I + K \<and> (\<forall>k \<in> set_of K . \<exists> j \<in> set_of J . (k,j) \<in> r)" by (metis union_lcomm union_assoc )
 from this obtain I J K where "P+N=I+J" and "P+M=I+K" and "\<forall>k\<in>set_of K.\<exists>j\<in>set_of J.(k,j)\<in> r" by auto
 hence "(I+K,I+J)\<in> mult_eq r" using one_step_implies_mult_eq[OF t] by auto 
 thus ?thesis using `P+N=I+J` `P+M=I+K` by auto 
qed

lemma mult_eq_add_right: assumes "trans r" and "(M,N) \<in> mult_eq r" shows "(M + P, N + P) \<in> mult_eq r" 
 using assms mult_eq_add_left unfolding union_commute by auto 

lemma intersection_not_empty: assumes "q \<in># M" and "q \<in># N" and "M #\<inter> N = {#}" shows False proof -
 from assms have "count M q > 0" and "count N q > 0" by auto 
 hence "count (M #\<inter> N) q > 0" by auto
 thus ?thesis using assms(3) by auto
qed

lemma add_left_one: 
 assumes  "\<exists> I J K. {#q#} + N = I + J \<and> {#q#} + M = I + K \<and> (J#\<inter>K={#}) \<and> (\<forall>k \<in> set_of K. \<exists>j \<in> set_of J. (k, j) \<in> r)" 
 shows "\<exists> I2 J K. N = I2 + J \<and> M = I2 + K \<and> (\<forall>k \<in> set_of K. \<exists>j \<in> set_of J. (k, j) \<in> r)"
 proof -
 from assms obtain I J K where 
  A: "{#q#} + N = I + J \<and> {#q#} + M = I + K \<and> (J #\<inter> K = {#}) \<and> (\<forall>k \<in> set_of K. \<exists>j \<in> set_of J. (k, j) \<in> r)" by auto
  have "q\<in>#I" proof (cases "q \<in># I")
   case True thus ?thesis by auto
  next
   case False 
   from A have "{#q#} + N = I + J" by auto
   hence "q \<in># J" using False
    by (metis Nat.add_0_right count_union gr0I union_commute union_single_eq_member) 
   from A have "{#q#} + M = I + K" by simp
   hence "q \<in># K" using False 
    by (metis Nat.add_0_right count_union gr0I union_commute union_single_eq_member)
   have "J#\<inter>K = {#}" using A by auto
   hence False using `J #\<inter> K = {#}` `q \<in>#J` `q \<in>#K` using intersection_not_empty by auto 
  thus ?thesis by auto
 qed
  hence "\<exists> I2. I = {#q#} + I2" by (metis multi_member_split union_commute)
  hence "\<exists> I2. {#q#} + N = ({#q#} + I2) + J \<and> {#q#} + M = ({#q#} + I2) + K \<and> (\<forall>k \<in> set_of K. \<exists>j \<in> set_of J. (k, j) \<in> r)" 
   using A by auto
  hence "\<exists> I2. N = (I2 + J) \<and> M = (I2 + K) \<and> (\<forall>k \<in> set_of K. \<exists>j \<in> set_of J. (k, j) \<in> r)"
   by (metis union_assoc diff_union_cancelR  union_commute)
 thus ?thesis by auto
qed

lemma lemma2_6_6_b_one : assumes t: "trans r" and "wf r" and "({#q#} + M,  {#q#} + N) \<in> mult_eq r" shows "(M,N) \<in> mult_eq r"
 proof -
 have "\<exists> I J K. {#q#} + N = I + J \<and> {#q#} + M = I + K \<and> (J #\<inter> K = {#} \<and> (\<forall>k\<in> set_of K. \<exists>j\<in> set_of J. (k,j) \<in> r))" 
 using mult_eq_imp_disjoint[OF assms] by auto
 hence "\<exists> I2 J K. N = I2 + J \<and> M = I2 + K \<and> (\<forall>k\<in> set_of K. \<exists>j\<in> set_of J. (k,j) \<in> r)" using add_left_one by simp
 thus ?thesis using one_step_implies_mult_eq[OF t] by auto
qed

lemma lemma2_6_6_b : assumes "trans r" and "wf r" and "(P + M, P + N) \<in> mult_eq r" shows "(M,N) \<in> mult_eq r" 
 using assms proof (induct P arbitrary: M N)
 case empty thus ?case by auto  
 next
 case (add Q q) 
  hence "(Q + ({#q#} + M), Q + {#q#} + N) \<in> mult_eq r" unfolding union_assoc by auto
  hence "({#q#} + M, {#q#} + N) \<in> mult_eq r" using assms(1) add(1)[where M="{#q#}+M" and N="{#q#}+N"]
by (metis add(3) union_commute union_lcomm) 
 thus ?case using assms(1) lemma2_6_6_b_one assms(1,2) by auto
 qed 

lemma mul_eq_extend_right: assumes t: "trans r" and "(M,N) \<in> mult_eq r" shows "(M,N+P) \<in> mult_eq r" proof -
 from assms obtain I J K where "N = I + J \<and> M = I + K \<and> (\<forall>k \<in> set_of K. \<exists> j \<in> set_of J. (k,j) \<in> r)" 
  using mult_eq_implies_one_or_zero_step[OF t assms(2)] by auto
 hence "N+P = I + J+P \<and> M = I + K \<and> (\<forall>k \<in> set_of K. \<exists> j \<in> set_of (J+P). (k,j) \<in> r)" by auto
 hence "N+P = I + (J+P) \<and> M = I + K \<and> (\<forall>k \<in> set_of K. \<exists> j \<in> set_of (J+P). (k,j) \<in> r)" using union_commute union_lcomm by metis 
 thus ?thesis using one_step_implies_mult_eq[OF t] by metis 
qed

lemma mul_eq_drop_left: assumes t: "trans r" and "wf r" and "(M+P,N) \<in> mult_eq r" shows "(M,N) \<in> mult_eq r" proof - 
 have "(M+P,N+P) \<in> mult_eq r" using mul_eq_extend_right[OF assms(1,3)] by simp
 thus ?thesis using lemma2_6_6_b[OF t assms(2)] unfolding union_commute  by fast
qed

lemma submultiset_implies_mult_eq: assumes t: "trans r" and w: "wf r" and "M \<le> N" shows "(M,N) \<in> mult_eq r" proof -
 from assms(3) obtain C where A: "M + C = N" by (metis mset_le_exists_conv) 
 hence "(M + C, N) \<in> mult_eq r" unfolding mult_eq_def by auto
 thus ?thesis using mul_eq_drop_left[OF t w] by fast 
qed 

(*
lemma mult_implies_not_empty: assumes "trans r" and "(M,N) \<in> mult r" shows "N \<noteq> {#}" proof -
 show ?thesis using assms mult_implies_one_step by auto 
qed
*)

lemma mult_add_left:
assumes t: "trans r" and "(M,N) \<in> mult r"
shows "(P+M,P+N) \<in> mult r" proof -
 from assms obtain I J K where F1:"N = I + J" and F2:"M = I + K" and F3:"\<forall>k\<in>set_of K.\<exists>j\<in>set_of J.(k,j) \<in> r" and F4:"J \<noteq> {#}" using mult_implies_one_step assms by metis
 hence "N \<noteq>{#}" by auto
  have P1:"P + N = P + I + J" using F1 union_commute union_lcomm union_left_cancel by metis 
  have P2:"P + M = P + I + K" using F2 union_commute union_lcomm union_left_cancel by metis
  show ?thesis using P1 P2 F2 F3 F4 using one_step_implies_mult assms by auto 
qed

lemma mult_add_right: assumes t: "trans r" and "(M,N) \<in> mult r" shows "(M+P,N+P) \<in> mult r" 
 using mult_add_left[OF assms] unfolding union_commute by auto 

lemma mult_and_mult_ge_implies_mult: 
 assumes t: "trans r" and "(M,N) \<in> mult r" and "(N,P) \<in> mult_eq r" 
 shows "(M,P) \<in> mult r" using assms unfolding mult_def mult_eq_def by auto 

(*Lemma 2.6.7*)
lemma lemma2_6_7: assumes t: "trans r" and "set_of P \<subseteq> dm r N - dm r M" and "(M,N) \<in> mult_eq r" 
                  shows "(P+M,N) \<in> mult_eq r" proof -
 from assms obtain I J K where A: "N=I+J \<and> M=I+K \<and> (\<forall>k\<in>set_of K.\<exists>j\<in> set_of J. (k,j) \<in> r)"
  using mult_eq_implies_one_or_zero_step[OF t] by fast 
 hence "set_of P \<subseteq> dm r (I+J) - dm r (I+K)" using assms by auto
 hence 0: "set_of P \<subseteq> dm r J - dm r K" unfolding downset_multiset_def downset_set_def downset_elt_def by auto
 have a: "dm r K = ds r (set_of K)" unfolding downset_multiset_def downset_set_def downset_elt_def by auto
 hence "dm r J - dm r K \<subseteq> dm r (J-s (set_of K))" using lemma2_6_1_diff unfolding a by metis 
 hence "set_of P \<subseteq> dm r (J-s (set_of K))" using 0 by auto 
 hence "set_of P \<subseteq> dm r J" unfolding remove_def downset_multiset_def downset_set_def downset_elt_def by auto 
 hence "\<forall>p\<in>set_of P.\<exists>j\<in>set_of J.(p,j)\<in> r" using mult_eq_downset_inv[OF t] by auto 
 hence "\<forall>k\<in>set_of(P+K).\<exists>j\<in>set_of J.(k,j)\<in>r" using A by auto
 hence "(I+(P+K),I+J) \<in> mult_eq r" using one_step_implies_mult_eq[OF t] by fast
 thus ?thesis using A union_assoc union_commute by metis 
qed

(*Def 3.1: lexicographic maximum measure*)
fun lexmax :: "('a \<times> 'a) set \<Rightarrow> 'a list \<Rightarrow> 'a multiset" ("(_|_|)") where 
   "lexmax r [] = {#}"
 | "lexmax r (\<alpha>#\<sigma>) =  {#\<alpha>#} + ((lexmax r \<sigma>) -s (de r \<alpha>))" 

lemma lexmax_singleton: "lexmax r [\<alpha>] = {#\<alpha>#}" unfolding lexmax.simps unfolding remove_def by auto

(*Lemma 3.2 *)
lemma empty_union_empty_is_empty : "({#}::'a multiset) + {#} = {#}" by auto
lemma empty_subset_single: "{#} \<subset># {#x#}" unfolding less_multiset_def mult_def mult1_def by auto 
lemma empty_subseteq_single: "{#} \<subseteq># {#x#}" using empty_subset_single unfolding le_multiset_def by auto

lemma filter: "M = Multiset.filter (\<lambda>x. True) M" unfolding filter_def by (simp add: count_inverse)

lemma obvious: "x \<notin> {} = True" by auto


lemma key_inverse: assumes "trans r" and "(x,z) \<in> r" and "z \<in> ds r S" shows "x \<in> ds r S" proof -
 from assms obtain y where mem_y: "y\<in>S" and rel_y: "(z,y) \<in> r" unfolding downset_set_def downset_elt_def by auto
 have "(x,y) \<in> r" using assms(2) rel_y assms(1) trans_def by metis
 thus ?thesis unfolding downset_set_def downset_elt_def using mem_y by auto
qed

text {* mising *}
lemma key: assumes "trans r" and "(x,z) \<in> r" and "x \<notin> ds r S" shows "z \<notin> ds r S" using assms key_inverse by metis

text {* missing *}
lemma key2: assumes t: "trans r" and "x \<in> set \<sigma>" shows "(\<exists>y \<in> set_of (lexmax r \<sigma>). (x,y) \<in> r^*)" 
 using assms proof (induct \<sigma>) 
 case Nil thus ?case by auto 
 next
 case (Cons w ws) thus ?case proof (cases "x=w") 
  case True thus ?thesis by auto 
  next
  case False hence "x \<in> set ws" using Cons by auto 
  from this obtain z where a: "z \<in> set_of (lexmax r ws)" and b: "(x,z) \<in> r^*" using Cons by auto
  thus ?thesis proof (cases "z \<in> de r w")
   case False thus ?thesis using a b unfolding lexmax.simps remove_def by auto  
  next
   case True hence "(z,w) \<in> r^*" unfolding downset_elt_def by auto
   thus ?thesis unfolding lexmax.simps remove_def using b t by auto 
  qed 
 qed 
qed

(*Lemma 3.2(1)*)
lemma lemma3_2_1: assumes t: "trans r" shows "dm r (lexmax r \<sigma>) = dl r \<sigma>" proof (induct \<sigma>)
 case Nil show ?case unfolding downset_multiset_def downset_list_def by auto
 next
 case (Cons \<alpha> \<sigma>) hence IH: "dm r r|\<sigma>| = dl r \<sigma>" by auto
  show "dm r (lexmax r (\<alpha>#\<sigma>)) = dl r (\<alpha>#\<sigma>)" (is "?L = ?R") proof -
  have 1: "?L \<subseteq> ?R" proof 
   fix x assume A: "x\<in> dm r (lexmax r (\<alpha>#\<sigma>))" show "x \<in> dl r (\<alpha>#\<sigma>)" proof -
   from A have "x \<in> dm r ({#\<alpha>#} + (lexmax r \<sigma> -s de r \<alpha>))" using lexmax.simps by auto 
   hence B: "x \<in> (dm r {#\<alpha>#} \<union> dm r (lexmax r \<sigma> -s de r \<alpha>))" 
     using lemma2_6_1_multiset unfolding downset_multiset_def downset_set_def downset_elt_def by auto 
   hence "x\<in> (dm r {#\<alpha>#} \<union> dm r (lexmax r \<sigma>))" 
   proof (cases "x\<in>dm r {#\<alpha>#}")
    case True thus ?thesis by auto
    next
    case False hence "x \<in> dm r (lexmax r \<sigma>)" using B 
     unfolding remove_def downset_multiset_def downset_set_def downset_elt_def by auto 
    thus ?thesis by auto
   qed
   thus ?thesis unfolding downset_multiset_singleton downset_list_def IH downset_set_def by auto 
   qed
   qed
    have 2: "?R \<subseteq> ?L" proof
    fix x assume A: "x \<in> dl r (\<alpha>#\<sigma>)" show "x \<in> dm r (lexmax r (\<alpha>#\<sigma>))"
     proof (cases "x \<in> dl r [\<alpha>]")
      case True thus ?thesis unfolding downset_multiset_def downset_list_def downset_set_def downset_elt_def by auto
     next
      case False hence "x \<in> dl r \<sigma>" using A 
       unfolding downset_multiset_def downset_list_def downset_set_def downset_elt_def by auto
      from this obtain y where mem_y: "y\<in> set \<sigma>" and rel_y:"(x,y) \<in> r" 
       unfolding downset_list_def downset_set_def downset_elt_def by auto
      from `y \<in> set \<sigma>` obtain z where a: "z \<in> set_of (lexmax r \<sigma>)" and b: "(y,z) \<in> r^*" using key2[OF t] by fast
      from False have "y \<noteq> \<alpha>" using `(x,y)\<in>r` unfolding downset_list_def downset_set_def downset_elt_def by auto 
      hence "(x,z) \<in> r^+" using rel_y b by auto
      hence "(x,z) \<in> r" using t by auto
      hence F1: "z \<notin> dl r [\<alpha>]" using key[OF t `(x,z)\<in>r`] False unfolding downset_list_def by auto
      hence F2: "z \<in> set_of ((lexmax r \<sigma>) -s (de r \<alpha>))" using a
       unfolding remove_def downset_list_def downset_set_def downset_elt_def remove_def by auto 
      hence "x \<in> dm r ((lexmax r \<sigma> -s (de r \<alpha>)))" using `(x,z)\<in>r` F1 
       unfolding downset_multiset_def downset_set_def downset_elt_def by auto
      hence "x \<in> dm r({#\<alpha>#} + ((lexmax r \<sigma> -s (de r \<alpha>))))"
      by (metis F2 `(x, z) \<in> r` downset_exists_imp_mem downset_multiset_def mset_le_add_right set_mp submultiset_implies_subset)
      thus ?thesis by auto
    qed
   qed
  show ?case using 1 2 by auto
 qed
 qed

(* Lemma 3.2(2)*)
lemma lemma3_2_2: "lexmax r (\<sigma>@\<tau>) = lexmax r \<sigma> + ((lexmax r \<tau>) -s (dl r \<sigma>))"
proof(induct \<sigma>)
 case Nil thus ?case unfolding downset_list_empty obvious remove_def using filter by auto 
 next
 case (Cons \<alpha> \<sigma>)
  have "lexmax r (\<alpha>#\<sigma>@\<tau>) = {#\<alpha>#} + (remove (lexmax r (\<sigma>@\<tau>)) (de r \<alpha>))" (is "?L = _") by auto
  hence "?L = {#\<alpha>#} + (remove (lexmax r \<sigma> + (remove (lexmax r \<tau>) (dl r \<sigma>))) (de r \<alpha>))" 
   using Cons unfolding remove_def by auto 
  hence "?L = {#\<alpha>#} + (remove (lexmax r \<sigma>) (de r \<alpha>)) + (remove (remove (lexmax r \<tau>) (dl r \<sigma>)) (de r \<alpha>))"
   unfolding remove_def add_assoc by simp (*no lemmaA_3_8 needed*)
 hence "?L = lexmax r (\<alpha>#\<sigma>) + (remove (remove (lexmax r \<tau>) (dl r \<sigma>)) (de r \<alpha>))" by simp
  hence tmp: "?L= lexmax r (\<alpha>#\<sigma>) + (remove (lexmax r \<tau>) (dl r \<sigma> \<union> dl r [\<alpha>]))" 
   unfolding lemmaA_3_9 downset_list_def downset_set_def by simp
  hence "?L = lexmax r (\<alpha>#\<sigma>) + (remove (lexmax r \<tau>) (dl r [\<alpha>] \<union> dl r \<sigma>))" using Un_commute[where A="dl r \<sigma>"] by auto
  hence "?L = lexmax r (\<alpha>#\<sigma>) + (remove (lexmax r \<tau>) (dl r (\<alpha>#\<sigma>)))" 
   unfolding lemma2_6_1_list_elt downset_list_empty (*different lemma2_6_1 needed*) by auto
 thus ?case unfolding remove_def by simp 
qed

(*Definition 3.3*)

definition decreasing :: "('a \<times> 'a) set \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> bool" where
 "decreasing r \<tau> \<sigma> \<sigma>' \<tau>' = (
  (lexmax r (\<sigma>@\<tau>'), lexmax r \<tau> + lexmax r \<sigma>) \<in> mult_eq r \<and> 
  (lexmax r (\<tau>@\<sigma>'), lexmax r \<tau> + lexmax r \<sigma>) \<in> mult_eq r)"

lemma dec_eq1: assumes "trans r" and "wf r" and "decreasing r \<tau> \<sigma> \<sigma>' \<tau>'" shows "((lexmax r \<tau>') -s (dl r \<sigma>), lexmax r \<tau>) \<in> mult_eq r" 
  using assms unfolding decreasing_def lemma3_2_2 using lemma2_6_6_b union_commute by metis

lemma dec_eq2: assumes "trans r" and "wf r" and "decreasing r \<tau> \<sigma> \<sigma>' \<tau>'" shows "((lexmax r \<sigma>') -s (dl r \<tau>), lexmax r \<sigma>) \<in> mult_eq r"
 using assms unfolding decreasing_def lemma3_2_2 using union_commute lemma2_6_6_b by metis

lemma dec_inv: assumes "trans r" and "wf r" and "((lexmax r \<tau>') -s (dl r \<sigma>), lexmax r \<tau>) \<in> mult_eq r" 
                                            and "((lexmax r \<sigma>') -s (dl r \<tau>), lexmax r \<sigma>) \<in> mult_eq r"
 shows "decreasing r \<tau> \<sigma> \<sigma>' \<tau>'" 
 using assms unfolding decreasing_def lemma3_2_2 using mult_eq_add_left[OF assms(1)] union_commute by metis 

(*Proposition 3.4*)

definition ld1 :: "('a \<times> 'a) set \<Rightarrow> 'a \<Rightarrow> 'a \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> bool" where
 "ld1 r \<beta> \<alpha> \<sigma>_1 \<sigma>_2 \<sigma>_3 = (
   (\<forall> e \<in> set \<sigma>_1. (e,\<beta>) \<in> r) \<and>
   (length \<sigma>_2 <= 1) \<and> (\<forall> e \<in> set \<sigma>_2. e = \<alpha>) \<and>
   (\<forall> e \<in> set \<sigma>_3. ((e,\<alpha>)\<in> r) \<or> (e,\<beta>) \<in> r)
)"

definition locally_decreasing :: "('a \<times> 'a) set \<Rightarrow> 'a \<Rightarrow> 'a \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> bool" where
 "locally_decreasing r \<beta> \<alpha> \<sigma>_1 \<sigma>_2 \<sigma>_3 \<tau>_1 \<tau>_2 \<tau>_3 =
  ((ld1 r \<beta> \<alpha> \<sigma>_1 \<sigma>_2 \<sigma>_3) \<and> (ld1 r \<alpha> \<beta> \<tau>_1 \<tau>_2 \<tau>_3))"

lemma locally_decreasing_symmetric:
assumes "locally_decreasing r \<beta> \<alpha> \<sigma>_1 \<sigma>_2 \<sigma>_3 \<tau>_1 \<tau>_2 \<tau>_3"
shows "locally_decreasing r \<alpha> \<beta> \<tau>_1 \<tau>_2 \<tau>_3 \<sigma>_1 \<sigma>_2 \<sigma>_3"
using assms unfolding locally_decreasing_def ld1_def by auto 

lemma mult_eq_trans:
assumes "trans r" and "(M,N) \<in> mult_eq r" and "(N,P) \<in> mult_eq r" 
shows "(M,P) \<in> mult_eq r" using assms unfolding mult_eq_def mult_def mult1_def by auto 

lemma mult_eq_remove_left_id: assumes t: "trans r" shows "(M-s A,M) \<in> mult_eq r" using assms
 by (metis lemma2_6_2_a lemmaA_3_10 mset_le_exists_conv union_commute) 

lemma mult_eq_remove_left: assumes t: "trans r" and "(M,N) \<in> mult_eq r" shows "(M-s A,N) \<in> mult_eq r"
 using assms mult_eq_remove_left_id mult_eq_trans by metis 

lemma mult_remove_left: assumes t: "trans r" and "(M,N) \<in> mult r" shows "(M-s A,N) \<in> mult r" proof -
 have "(M-s A,M) \<in> mult_eq r" using mult_eq_remove_left_id[OF t] by auto
 thus ?thesis using assms(2) unfolding mult_eq_def mult_def by auto
qed

lemma mult_eq_list: assumes "trans r" and "length \<tau> <= 1" and "\<forall>k\<in>(set \<tau>). k = e" shows "(multiset_of \<tau>,{#e#}) \<in> mult_eq r" 
 using assms proof (cases "\<tau>")
 case Nil thus ?thesis unfolding Nil multiset_of.simps mult_eq_def mult_def mult1_def by auto
 next
 case (Cons x xs) hence e: "\<tau> = [x]"  using assms by auto 
 thus ?thesis using assms(3) unfolding e multiset_of.simps mult_eq_def mult_def mult1_def by auto 
qed

lemma submultiset_lexmax: assumes t: "trans r" shows "lexmax r \<sigma> <= multiset_of \<sigma>" proof (induct "\<sigma>")
 case Nil thus ?case unfolding lexmax.simps by auto
  next
 case (Cons s \<tau>) hence IH: "lexmax r \<tau> <= multiset_of \<tau>" by auto
 hence "lexmax r \<tau> -s downset_elt r s <= multiset_of \<tau>" (is "?l <= ?r") 
 unfolding remove_def downset_elt_def 
by (metis  Cons downset_elt_def lemmaA_3_10 mset_le_add_right order_trans remove_def)
 hence "{#s#} + ?l <= {#s#} + ?r" by auto
 thus ?case unfolding lexmax.simps multiset_of.simps union_commute by auto  
qed

lemma mult_eq_lexmax: assumes t: "trans r" shows "(lexmax r \<sigma>,multiset_of \<sigma>) \<in> mult_eq r" 
 using assms submultiset_lexmax lemma2_6_2_a by fast

lemma ld1_prop3_4:
 assumes t: "trans r" and "wf r" and "locally_decreasing r \<beta> \<alpha> \<sigma>_1 \<sigma>_2 \<sigma>_3 \<tau>_1 \<tau>_2 \<tau>_3" 
 shows "((lexmax r (\<sigma>_1@\<sigma>_2@\<sigma>_3)) -s (dl r [\<beta>]),lexmax r [\<alpha>]) \<in> mult_eq r"
proof -

 from assms have "set \<sigma>_1 \<subseteq> downset_elt r \<beta>" unfolding locally_decreasing_def ld1_def downset_elt_def by auto
 hence "set_of (lexmax r \<sigma>_1) \<subseteq> downset_elt r \<beta>" using submultiset_implies_subset[OF submultiset_lexmax[OF t]] subset_trans by auto
 hence one: "lexmax r \<sigma>_1 -s downset_elt r \<beta> = {#}" (is "?l = _") using subset_implies_remove_empty by auto

 have "set \<sigma>_3 \<subseteq> downset_multiset r {#\<alpha>,\<beta>#}" using assms 
   unfolding locally_decreasing_def ld1_def downset_multiset_def downset_set_def downset_elt_def by auto
 hence "set_of (lexmax r \<sigma>_3) \<subseteq> downset_multiset r {#\<alpha>,\<beta>#}" using submultiset_implies_subset[OF submultiset_lexmax[OF t]] subset_trans by auto
 hence "lexmax r \<sigma>_3 -s downset_multiset r {#\<alpha>,\<beta>#} = {#}" using subset_implies_remove_empty by auto 
 hence pre3: "lexmax r \<sigma>_3 -s (downset_elt r \<alpha> \<union> downset_elt r \<beta>) = {#}" unfolding lemma2_6_1_multiset using downset_multiset_singleton by metis
show ?thesis
proof (cases "\<sigma>_2 = []")
 case True 
  hence two:"(lexmax r \<sigma>_2 -s downset_list r \<sigma>_1) -s downset_elt r \<beta> = {#}" (is "(?l2 = _)") unfolding remove_def by auto 
  from pre3 have "(lexmax r \<sigma>_3 -s downset_elt r \<beta>) -s downset_elt r \<alpha> = {#}" using lemmaA_3_9 Un_commute by metis
  hence "(lexmax r \<sigma>_3 -s downset_elt r \<beta>,{#\<alpha>#}) \<in> mult r" using t remove_empty_implies_subset lemma2_6_4
   by (metis single_not_empty downset_multiset_singleton)
  hence "(((lexmax r \<sigma>_3 -s downset_elt r \<beta>) -s downset_list r \<sigma>_1) -s downset_list r \<sigma>_2,{#\<alpha>#}) \<in> mult r" using mult_remove_left[OF t] by auto
 hence three:"(((lexmax r \<sigma>_3 -s downset_list r \<sigma>_2) -s downset_list r \<sigma>_1) -s downset_elt r \<beta>,{#\<alpha>#}) \<in> mult r" (is "(?l3,_) \<in> _") 
 unfolding lemmaA_3_9 using Un_commute Un_assoc by metis 
  show ?thesis unfolding lemma3_2_2 lexmax_singleton lemmaA_3_8 downset_list_singleton one two empty_neutral using three mult_imp_mult_eq by auto
 next
 case False 
  from assms have "(multiset_of \<sigma>_2,{#\<alpha>#}) \<in> mult_eq r" unfolding locally_decreasing_def ld1_def using mult_eq_list[OF t] by auto
 hence "(lexmax r \<sigma>_2,{#\<alpha>#}) \<in> mult_eq r" using  mult_eq_trans[OF t] mult_eq_lexmax[OF t] by auto
 hence "(lexmax r \<sigma>_2 -s downset_elt r \<beta>,{#\<alpha>#}) \<in> mult_eq r" 
  by (metis lemma2_6_2_a lemmaA_3_10 mset_le_exists_conv mult_eq_trans t union_commute)  
 hence "((lexmax r \<sigma>_2 -s downset_elt r \<beta>) -s downset_list r \<sigma>_1, {#\<alpha>#}) \<in> mult_eq r" 
  by (metis assms(2)  lemmaA_3_10 mul_eq_drop_left t union_commute)
 hence two: "((lexmax r \<sigma>_2 -s downset_list r \<sigma>_1) -s downset_elt r \<beta>,{#\<alpha>#}) \<in> mult_eq r"
  (is "(?l2,?r2) \<in> _") unfolding lemmaA_3_9  by (metis Un_commute) 

  from assms have "length \<sigma>_2 <= 1" unfolding locally_decreasing_def ld1_def by auto
  hence "length \<sigma>_2 = 1" using False by (metis One_nat_def length_0_conv less_Suc0 nat_less_le) 
  from this obtain s2 where eq2a: "\<sigma>_2 = [s2]" by (metis False append_Nil append_butlast_last_id butlast_conv_take diff_self_eq_0 take_0)
  from this assms have eq2b: "s2 = \<alpha>" unfolding locally_decreasing_def ld1_def by auto
  (*prep done*)
  from pre3 have "((lexmax r \<sigma>_3 -s downset_list r \<sigma>_2) -s downset_elt r \<beta> = {#})" 
   using eq2a eq2b unfolding lemmaA_3_9[symmetric] downset_list_def downset_set_def by auto
  hence "(((lexmax r \<sigma>_3 -s downset_list r \<sigma>_2) -s downset_elt r \<beta>) -s downset_list r \<sigma>_1 = {#})" unfolding remove_def by auto
  hence three: "(((lexmax r \<sigma>_3 -s downset_list r \<sigma>_2) -s downset_list r \<sigma>_1) -s downset_elt r \<beta>) = {#}" (is "?l3 = _") 
   unfolding lemmaA_3_9 using Un_commute Un_assoc by metis
 thus ?thesis unfolding lemma3_2_2 lexmax_singleton lemmaA_3_8 downset_list_singleton one three empty_neutral using two by auto
qed
qed

lemma proposition3_4:
assumes t: "trans r" and w: "wf r" and ld: "locally_decreasing r \<alpha> \<beta> \<sigma>_1 \<sigma>_2 \<sigma>_3 \<tau>_1 \<tau>_2 \<tau>_3"
shows "decreasing r [\<alpha>] [\<beta>] (\<sigma>_1@\<sigma>_2@\<sigma>_3) (\<tau>_1@\<tau>_2@\<tau>_3)"
proof -
 have 1:"(lexmax r (\<sigma>_1 @ \<sigma>_2 @ \<sigma>_3)-s (downset_list r [\<alpha>]), lexmax r [\<beta>]) \<in> mult_eq r" using ld1_prop3_4[OF t w ld] by auto 
 have ld_symm: "locally_decreasing r \<beta> \<alpha> \<tau>_1 \<tau>_2 \<tau>_3 \<sigma>_1 \<sigma>_2 \<sigma>_3" using locally_decreasing_symmetric[OF ld] by auto
 have 2:"(lexmax r (\<tau>_1 @ \<tau>_2 @ \<tau>_3)-s (downset_list r [\<beta>]), lexmax r [\<alpha>]) \<in> mult_eq r" using ld1_prop3_4[OF t w ld_symm] by auto 
 show ?thesis using dec_inv[OF t w] using 1 2 by auto 
qed


(*Lemma 3.5*)
lemma lemma3_5_1: 
 assumes t: "trans r" and w: "wf r" and D1: "decreasing r \<tau> \<sigma> \<sigma>' \<tau>'" and D2: "decreasing r v \<sigma>' \<sigma>'' v'" 
 shows "(lexmax r (\<tau> @ v @ \<sigma>''), lexmax r (\<tau> @ v) + lexmax r \<sigma>) \<in> mult_eq r" proof -
 have "lexmax r (\<tau> @ v @ \<sigma>'') = (lexmax r (\<tau> @ v) + ((lexmax r \<sigma>'') -s (dl r (\<tau>@v))))" (is "?L = ?R")
  by (metis append_assoc lemma3_2_2) 
 hence x: "?L = lexmax r (\<tau>@v) + (((lexmax r \<sigma>'') -s dl r v) -s dl r \<tau>)" (is "?L = ?R")
 unfolding lemmaA_3_9 lemma2_6_1_list using Un_commute by metis 
  have A: "((lexmax r \<sigma>'') -s (dl r v),lexmax r \<sigma>') \<in> mult_eq r" (is "(?l,?r) \<in> _ ") using assms dec_eq2 by blast
  hence "(?l -s (dl r \<tau>), ?r -s (dl r \<tau>)) \<in> mult_eq r" (is "(?l,?r) \<in> _") 
   using lemma2_6_5_simplified[OF t A] downset_list_def by metis 
  hence "(lexmax r (\<tau>@v) + ?l,lexmax r (\<tau>@v) + ?r) \<in> mult_eq r" using t mult_eq_add_left by blast 
 hence y: "(?L,lexmax r (\<tau>@v) + ((lexmax r \<sigma>') -s (dl r \<tau>))) \<in> mult_eq r" using x by auto
  have "((lexmax r \<sigma>') -s (dl r \<tau>),lexmax r \<sigma>) \<in> mult_eq r" (is "(?l,?r) \<in> _") using t w D1 dec_eq2 by blast
  hence "(lexmax r (\<tau>@v) + ?l, lexmax r (\<tau>@v) + ?r) \<in> mult_eq r" using t mult_eq_add_left by auto
 hence "(?L,lexmax r (\<tau>@v) + lexmax r \<sigma>) \<in> mult_eq r"
 using y unfolding mult_eq_def using rtrancl_trans by auto  
 thus ?thesis by auto
 qed 

lemma q: assumes t: "trans r" shows "(lexmax r (\<sigma>@\<tau>), lexmax r \<sigma> + lexmax r \<tau>) \<in> mult_eq r" proof -
 have "lexmax r (\<sigma>@\<tau>) = lexmax r \<sigma> + (lexmax r \<tau> -s (dl r \<sigma>))" using lemma3_2_2 by auto
 hence "(lexmax r (\<sigma>@\<tau>)) \<le> (lexmax r \<sigma> + lexmax r \<tau>)" unfolding remove_def mset_le_def by auto 
 thus ?thesis using lemma2_6_2_a[OF t] by auto
qed

lemma claim1: 
 assumes t: "trans r" and  w: "wf r" and D1: "decreasing r \<tau> \<sigma> \<sigma>' \<tau>'"
 shows "(lexmax r (\<sigma>@\<tau>') + (((lexmax r v' -s dl r (\<sigma>@\<tau>')) \<inter>s dl r \<tau>)),lexmax r \<sigma> + lexmax r \<tau>) \<in> mult_eq r" (is "(?F+?H,?G) \<in> _")
 proof -
  have 1: "(?F,?G) \<in> mult_eq r" using D1 t dec_eq1 by (metis decreasing_def union_commute)
  have 2: "set_of ?H \<subseteq> (dm r ?G) - (dm r ?F)" (is "(?L \<subseteq> _)") 
   proof -
    have "?L= set_of ((lexmax r v' \<inter>s dl r \<tau>) -s dl r (\<sigma>@\<tau>'))" 
     using lemmaA_3_11[where M="lexmax r v'" and T = "dl r (\<sigma>@\<tau>')"] by simp
    hence "?L \<subseteq> (dl r \<tau> - dl r (\<sigma>@\<tau>'))" unfolding remove_def intersect_def by auto 
    hence "?L \<subseteq> ((dl r \<sigma> \<union> dl r \<tau>) - dl r (\<sigma>@\<tau>'))" by auto
    hence "?L \<subseteq> (dl r (\<sigma>@\<tau>) - dl r (\<sigma>@\<tau>'))" unfolding lemma2_6_1_list lemmaA_3_11 by auto 
    hence A: "?L \<subseteq> (dm r (lexmax r (\<sigma>@\<tau>)) - dm r (lexmax r (\<sigma>@\<tau>')))" unfolding lemma3_2_1[OF t] by auto
    hence "dm r (lexmax r (\<sigma>@\<tau>)) \<subseteq> dm r (lexmax r \<sigma> + lexmax r \<tau>)" using q[OF t] lemma2_6_2_list_b[OF t] 
     unfolding downset_multiset_def by metis 
    thus ?thesis using A by auto
   qed
  show ?thesis using lemma2_6_7[OF t 2 1] union_commute by metis
 qed

lemma detailed_step_aux: 
 assumes t: "trans r" shows "dm r (lexmax r \<sigma>' -s dl r \<tau>) \<supseteq> dl r \<sigma>' - dl r \<tau>" 
by (metis (full_types) Diff_mono assms downset_multiset_def ds_ds_subseteq_ds lemma2_6_1_diff lemma3_2_1 order_refl order_trans)

lemma detailed_step: 
assumes t: "trans r" and  w: "wf r" and D1: "decreasing r \<tau> \<sigma> \<sigma>' \<tau>'"
shows "dl r (\<sigma>@\<tau>) \<supseteq> dl r \<sigma>' \<union> dl r \<tau>" (is "?L \<supseteq> ?R")  proof -
 have "dl r (\<sigma>@\<tau>) = dl r \<sigma> \<union> dl r \<tau>" unfolding downset_list_def lemma2_6_1_set set_append by auto 
 hence "?L = dm r (lexmax r \<sigma>) \<union> dl r \<tau>" using lemma3_2_1[OF t]  by auto
 hence "?L \<supseteq> (dm r (lexmax r \<sigma>' -s dl r \<tau>)) \<union> dl r \<tau>" using t w D1 dec_eq2 lemma2_6_2_list_b by fast
 hence "?L \<supseteq> ((dl r \<sigma>' - dl r \<tau>) \<union> dl r \<tau>)" using detailed_step_aux[OF t] 
 by (metis Un_Diff_cancel le_sup_iff subset_trans sup_commute) 
thus ?thesis by auto
qed

text {* missing *}
lemma downset_monotone: "dl r (\<sigma>@\<tau>) \<subseteq> dl r (\<sigma>@\<tau>'@\<tau>)" (is "(?L \<subseteq> _)") proof -
 have 1: "dl r (\<sigma>@\<tau>) = dl r \<sigma> \<union> dl r \<tau>" unfolding lemma2_6_1_list by auto 
 hence 2: "... \<subseteq> dl r \<sigma> \<union> dl r \<tau>' \<union> dl r \<tau>" by auto
 hence 3: "... \<subseteq> dl r (\<sigma>@\<tau>'@\<tau>)" using lemma2_6_1_list by (metis append_assoc subset_refl)
 thus ?thesis using 1 2 by auto
qed

text {* missing?; similar to lemma_2.6.2? *}
lemma zz: assumes t: "trans r" and w: "wf r" and "A \<subseteq> B" shows "(M -s B,M -s A) \<in> mult_eq r" proof -
 from assms(3) have "(M -s B) \<le> (M -s A)" unfolding remove_def 
  by (metis Un_absorb2 Un_commute lemmaA_3_10 lemmaA_3_9 mset_le_add_right remove_def) 
 thus ?thesis using submultiset_implies_mult_eq[OF t w] by auto
qed

lemma claim2_mult_eq: 
 assumes t: "trans r" and  w: "wf r" and D1: "decreasing r \<tau> \<sigma> \<sigma>' \<tau>'"
 shows "((lexmax r v' -s dl r (\<sigma>@\<tau>')) -s dl r \<tau>, ((lexmax r v' -s dl r \<sigma>') -s dl r \<tau>)) \<in> mult_eq r" (is "(?L,_) \<in> mult_eq r")
 proof -
  have "?L = lexmax r v' -s (dl r (\<sigma>@\<tau>') \<union> dl r \<tau>)" using lemmaA_3_9 by auto 
  hence "?L = lexmax r v' -s (dl r (\<sigma>@\<tau>'@\<tau>))" using lemma2_6_1_list[symmetric, where xs="\<sigma>@\<tau>'"] unfolding append_assoc by simp
  hence A: "(?L,lexmax r v' -s dl r (\<sigma>@\<tau>)) \<in> mult_eq r" using downset_monotone using zz[OF t w] by metis
   from t w D1 detailed_step have "dl r (\<sigma>@\<tau>) \<supseteq> dl r \<sigma>' \<union> dl r \<tau>" by fast
  hence "(?L,lexmax r v' -s (dl r \<sigma>' \<union> dl r \<tau>)) \<in> mult_eq r" using A t zz[OF t w] by (metis mult_eq_trans) 
  thus ?thesis using lemmaA_3_9[symmetric] by metis 
 qed

lemma lemma3_5_2: assumes t: "trans r" and w: "wf r" and D1: "decreasing r \<tau> \<sigma> \<sigma>' \<tau>'" and D2: "decreasing r v \<sigma>' \<sigma>'' v'" 
 shows "(lexmax r (\<sigma> @ \<tau>' @ v'), lexmax r \<sigma> + lexmax r (\<tau>@v)) \<in> mult_eq r" 
proof -
 have "lexmax r (\<sigma> @ \<tau>' @ v') = lexmax r (\<sigma>@\<tau>') + (lexmax r v' -s dl r (\<sigma>@\<tau>'))" (is "?L = _") using lemma3_2_2 
  by (metis append_assoc) 
 hence 1: "?L = lexmax r (\<sigma>@\<tau>') + ((lexmax r v' -s dl r (\<sigma>@\<tau>')) \<inter>s dl r \<tau>) + ((lexmax r v' -s dl r (\<sigma>@\<tau>')) -s dl r \<tau>)" 
  using lemmaA_3_10  by (metis union_assoc) 
 have 2: "(?L, lexmax r \<sigma> + lexmax r \<tau> + ((lexmax r v' -s dl r (\<sigma>@\<tau>')) -s dl r \<tau>)) \<in> mult_eq r" unfolding 1 
  using claim1 D1 t w mult_eq_add_right[where P="(r|v'|-sdl r \<sigma> @ \<tau>')-sdl r \<tau>"] by fast
 have "((lexmax r \<sigma> + lexmax r \<tau>) + ((lexmax r v' -s dl r (\<sigma>@\<tau>')) -s dl r \<tau>), (lexmax r \<sigma> + lexmax r \<tau>) + (((lexmax r v' -s dl r \<sigma>') -s dl r \<tau>))) \<in> mult_eq r" 
  using t w D1 claim2_mult_eq mult_eq_add_left[where P="r|\<sigma>| + r|\<tau>|"] by fast
 hence 3: "(?L, lexmax r \<sigma> + lexmax r \<tau> + ((lexmax r v' -s dl r \<sigma>') -s dl r \<tau>)) \<in> mult_eq r" 
 using t 2 mult_eq_trans[where M="?L"] by force 
 have "((lexmax r v') -s dl r \<sigma>', lexmax r v) \<in> mult_eq r" using D2 t w using dec_eq1 by force 
 hence "(((lexmax r v') -s dl r \<sigma>') -s dl r \<tau> , lexmax r v -s dl r \<tau>) \<in> mult_eq r" 
  using lemma2_6_5_simplified[OF t] downset_list_def by metis
 hence 4: "(lexmax r \<sigma> + lexmax r \<tau> + ((lexmax r v') -s dl r \<sigma>') -s dl r \<tau> , lexmax r \<sigma> + lexmax r \<tau> + lexmax r v -s dl r \<tau>) \<in> mult_eq r" (is "(_,?R) \<in> _") 
  using mult_eq_add_left t by auto 
 hence "(?L,?R) \<in> mult_eq r" using 3 t mult_eq_trans by auto 
 thus ?thesis using lemma3_2_2 union_assoc by metis 
qed

lemma lemma3_5: 
 assumes t: "trans r" and w: "wf r" and D1: "decreasing r \<tau> \<sigma> \<sigma>' \<tau>'" and D2: "decreasing r v \<sigma>' \<sigma>'' v'" 
 shows "decreasing r (\<tau>@v) \<sigma> \<sigma>'' (\<tau>'@v')"
 proof -
 from assms have 1:"(lexmax r (\<tau>@v@\<sigma>''),lexmax r (\<tau>@v) + lexmax r \<sigma>) \<in> mult_eq r" using t lemma3_5_1 by fast
 from assms have 2:"(lexmax r (\<sigma>@\<tau>'@v'),lexmax r \<sigma> + lexmax r (\<tau>@v)) \<in> mult_eq r" using t lemma3_5_2 by fast
 from 1 2 show ?thesis  unfolding decreasing_def by (metis append_eq_appendI union_commute) 
qed

lemma not_empty_implies_lexmax_not_empty: assumes "\<sigma> \<noteq> []" shows "lexmax r \<sigma> \<noteq> {#}" proof -
 from assms obtain x xs where "\<sigma>=x#xs" by (metis list.exhaust) 
 thus ?thesis by auto
qed 

lemma lemma3_6: 
 assumes t: "trans r" and w: "wf r" and ne: "\<tau> \<noteq> []" and D: "decreasing r \<tau> \<sigma> \<sigma>' \<tau>'" 
 shows "(lexmax r \<sigma>' + lexmax r v, lexmax r \<sigma> + (lexmax r (\<tau>@v))) \<in> mult r" (is "(?L,?R) \<in> _") proof -
 have L1: "lexmax r \<sigma>' + lexmax r v = ((lexmax r \<sigma>' + lexmax r v) \<inter>s dl r \<tau>) + ((lexmax r \<sigma>' + lexmax r v) -s dl r \<tau>)"  (is "?F = ?R1") using lemmaA_3_10 by auto
  have F1: "set_of (?F \<inter>s dl r \<tau>) \<subseteq> dm r (lexmax r \<tau>)" unfolding intersect_def using lemma3_2_1[OF t] by auto
  have F2: "lexmax r \<tau> \<noteq> {#}" using not_empty_implies_lexmax_not_empty[OF ne] by auto
 have a: "(?F\<inter>sdl r \<tau>,lexmax r \<tau>) \<in> mult r" using lemma2_6_4[OF t F2] F1 
  unfolding downset_multiset_def downset_list_def by auto 
 have b: "lexmax r \<tau> + (?F-sdl r \<tau>) = lexmax r \<tau> + ((lexmax r \<sigma>'-s dl r \<tau>) + (lexmax r v -s dl r \<tau>))" 
  using lemmaA_3_8 by auto 
 have c: "... = lexmax r (\<tau>@\<sigma>') + (lexmax r v -s dl r \<tau>)" unfolding lemma3_2_2 using union_assoc by metis 
 have d: "(lexmax r (\<tau>@\<sigma>'),lexmax r \<sigma> + lexmax r \<tau>) \<in> mult_eq r" 
  using D union_commute unfolding decreasing_def by metis 
 have e: "lexmax r \<tau> + (lexmax r v -s dl r \<tau>) = lexmax r (\<tau>@v)" using lemma3_2_2[symmetric] by auto 
 
 have L2: "(?R1,lexmax r \<tau> + (?F -s dl r \<tau>)) \<in> mult r" (is "(_,?R2) \<in> _") using a mult_add_right[OF t] by auto
 have L3: "?R2 = lexmax r \<tau> + (lexmax r \<sigma>' -s dl r \<tau>) + (lexmax r v -s dl r \<tau>)" (is "_ = ?R3") 
  using b union_assoc by metis 
 have L4: "?R3 = lexmax r (\<tau>@\<sigma>') + (lexmax r v -s dl r \<tau>)" (is "_=?R4") using c union_assoc by metis 
 have L5: "(?R4,lexmax r \<sigma> + lexmax r \<tau> + (lexmax r v -s dl r \<tau>)) \<in> mult_eq r" (is "(_,?R5) \<in> _") 
  using d mult_eq_add_right[OF t] by auto
 have L6: "?R5 = ?R" using e union_assoc by metis
 have I1: "(?L,?R2) \<in> mult r" using L1 L2 by auto 
 have I2: "(?R2,?R) \<in> mult_eq r" using L3 L4 L5 L6 by auto 
  show ?thesis using I1 I2 mult_and_mult_ge_implies_mult[OF t] by fast  
qed 

lemma decreasing_mirror:
 assumes t:"trans r" and w: "wf r" and "decreasing r \<tau> \<sigma> \<sigma>' \<tau>'"
 shows "decreasing r \<sigma> \<tau> \<tau>' \<sigma>'" using dec_eq1[OF t w assms(3)] dec_eq2[OF t w assms(3)] dec_inv[OF t w] by simp 

lemma lemma3_6_v: 
 assumes t: "trans r" and w: "wf r" and ne: "\<sigma> \<noteq> []" and D: "decreasing r \<tau> \<sigma> \<sigma>' \<tau>'" 
 shows "(lexmax r \<tau>' + lexmax r v, lexmax r \<tau> + (lexmax r (\<sigma>@v))) \<in> mult r" (is "(?L,?R) \<in> _")
proof -
 have D_mirror: "decreasing r \<sigma> \<tau> \<tau>' \<sigma>'" using decreasing_mirror[OF t w D] by auto 
 show ?thesis using lemma3_6[OF t w ne D_mirror] by auto
qed
 
(*Theorem 3.7*)

definition fst3 :: "('a\<times>'b\<times>'c) \<Rightarrow> 'a" where "fst3 a = fst a"
definition snd3 :: "('a\<times>'b\<times>'c) \<Rightarrow> 'b" where "snd3 a = fst (snd a)"
definition thd3 :: "('a\<times>'b\<times>'c) \<Rightarrow> 'c" where "thd3 a = snd (snd a)"

inductive_set
  sequence :: "('a\<times>'b\<times>'a) set \<Rightarrow> ('a\<times>(('b\<times>'a)list)) set"
  for A
where
 "(a,[]) \<in> sequence A " |
 "(a,\<alpha>,b) \<in> A \<Longrightarrow> (b,ss) \<in> sequence A \<Longrightarrow>  (a,(\<alpha>,b) # ss) \<in> sequence A"

type_synonym ('a,'b) seq = "('a \<times> ('b\<times>'a)list)"

definition last_elt :: "('a,'b) seq \<Rightarrow> 'a"
 where "last_elt ss = (if snd ss = [] then fst ss else snd (last (snd ss)))"

type_synonym ('a,'b) pea = "('a,'b) seq \<times> ('a,'b) seq"
type_synonym ('a,'b) dia = "('a,'b) seq \<times> ('a,'b) seq \<times> ('a,'b) seq \<times> ('a,'b) seq"

definition diagram :: "('a\<times>'b\<times>'a) set \<Rightarrow> (('a,'b) dia) set"
where
"diagram A = {d. (let \<tau> = fst d in let \<sigma> = fst (snd d) in let \<sigma>' = fst (snd (snd d)) in let \<tau>' = snd (snd (snd d)) in
   (\<sigma> \<in> sequence A \<and> \<tau> \<in> sequence A \<and> \<sigma>' \<in> sequence A \<and> \<tau>' \<in> sequence A \<and>
   (fst \<sigma> = fst \<tau>) \<and> (last_elt \<sigma> = fst \<tau>') \<and> (last_elt \<tau> = fst \<sigma>') \<and> (last_elt \<sigma>' = last_elt \<tau>')))}"

definition local_diagram :: "('a\<times>'b\<times>'a) set \<Rightarrow> (('a,'b) dia) set"
 where "local_diagram A = {d \<in> diagram A. (length (snd (fst d)) = 1) \<and> (length (snd (fst (snd d))) = 1)}"

definition labels :: "('a,'b) seq \<Rightarrow> 'b list"
 where "labels ss = map fst (snd ss)"

definition Decreasing :: "('b\<times>'b) set \<Rightarrow> ('a,'b) dia \<Rightarrow> bool"
 where "Decreasing r d = decreasing r (labels (fst d)) (labels (fst (snd d))) (labels (fst (snd (snd d)))) (labels (snd (snd (snd d))))"
  
definition peak :: "('a \<times> 'b \<times> 'a) set \<Rightarrow> (('a,'b)seq \<times> ('a,'b) seq) set"
 where "peak ars = {P. (fst P) \<in> sequence ars \<and> (snd P) \<in> sequence ars \<and> (fst (fst P) = fst(snd P))}"

definition local_peak :: "('a\<times>'b\<times>'a) set \<Rightarrow> (('a,'b) seq \<times> ('a,'b) seq) set"
 where "local_peak ars = {P \<in> peak ars. ((length (snd (fst P)) = 1) \<and> (length (snd (snd P)) = 1))}"

definition PDecreasing :: "('a\<times>'b\<times>'a) set \<Rightarrow> ('b\<times>'b) set \<Rightarrow> ('a,'b) pea \<Rightarrow> bool"
 where "PDecreasing ars r P = (\<exists>\<sigma>'\<in>sequence ars.\<exists>\<tau>'\<in>sequence ars. (fst P,snd P,\<sigma>',\<tau>') \<in> diagram ars \<and> Decreasing r (fst P, snd P, \<sigma>', \<tau>'))"

definition D_D :: "('a\<times>'b\<times>'a) set \<Rightarrow> ('b\<times>'b) set \<Rightarrow> bool"
 where "D_D ars r = (\<forall>P \<in> peak ars.(PDecreasing ars r P))"

definition LD_D :: "('a\<times>'b\<times>'a) set \<Rightarrow> ('b\<times>'b) set \<Rightarrow> bool"
 where "LD_D ars r = (\<forall>P \<in> local_peak ars. PDecreasing ars r P)"
 
definition measure :: "('b\<times>'b) set \<Rightarrow> ('a,'b) pea \<Rightarrow> 'b multiset"
 where "measure r P = lexmax r (labels (fst P)) + lexmax r (labels (snd P))"

definition rel :: "('b\<times>'b) set \<Rightarrow> (('a,'b) pea \<times> ('a,'b) pea) set"
 where "rel r = {P12. ((measure r (fst P12),measure r (snd P12)) \<in> mult r)}"

lemma readable: assumes "LD_D ars r" shows "P \<in> local_peak ars \<longrightarrow> PDecreasing ars r P"
using assms using LD_D_def by (metis (full_types) mem_Collect_eq)

lemma sequence_tail1 : assumes seq: "(s,x#xs) \<in> sequence ars" 
 shows "(snd x,xs) \<in> sequence ars" 
   and "(s,fst x,snd x) \<in> ars"
   and "last_elt (s,x#xs) = last_elt (snd x,xs)" 
proof -
 show "(snd x,xs)\<in>sequence ars" 
 using assms proof (cases)
  fix \<alpha> b assume A:"x = (\<alpha>,b)" and B:"(s,\<alpha>,b) \<in> ars" and C:"(b,xs) \<in> sequence ars" show ?thesis
  proof - from A B C show ?thesis by auto 
qed
qed 
next 
 show "(s,fst x,snd x) \<in> ars"  using assms proof (cases) 
  fix \<alpha> b assume A:"x = (\<alpha>,b)" and B:"(s,\<alpha>,b) \<in> ars" and C:"(b,xs) \<in> sequence ars" show ?thesis
  proof - from A B C show ?thesis by auto
qed
qed
next
 show "last_elt (s,x#xs) = last_elt (snd x,xs)" using assms proof (cases)
  fix \<alpha> b assume A:"x = (\<alpha>,b)" and B:"(s,\<alpha>,b) \<in> ars" and C:"(b,xs) \<in> sequence ars" show ?thesis
  proof - from A B C show ?thesis unfolding last_elt_def by auto
qed
qed
qed

lemma sequence_tail: assumes "(s,ss@ts) \<in> sequence ars" shows "(last_elt(s,ss),ts) \<in> sequence ars"
 using assms proof (induct ss arbitrary:s)
  case Nil thus ?case unfolding last_elt_def by auto
 next
  case (Cons x xs) 
  from Cons have "(s,x#xs@ts)\<in>sequence ars" by auto
  hence "(snd x,xs@ts) \<in> sequence ars" using sequence_tail1(1) by fast 
  hence "(last_elt (snd x,xs),ts) \<in> sequence ars" using Cons(1) by auto
  thus ?case unfolding last_elt_def by auto
qed

lemma sequence_chop: assumes "(s,ss@ts) \<in> sequence ars" shows "(s,ss) \<in> sequence ars"
 using assms proof (induct ss arbitrary: s) 
 case Nil show ?case using sequence.intros(1) by fast 
 next
 case (Cons x xs) 
 from Cons have "(s,(x#xs)@ts) \<in> sequence ars" by auto
 hence k:"(s,x#(xs@ts)) \<in> sequence ars" by auto
 hence "(snd x,xs@ts) \<in> sequence ars" using sequence_tail1(1) by fast 
 hence "(snd x,xs) \<in> sequence ars" using Cons(1) by auto 
 hence "(s,(x#xs)) \<in> sequence ars" using  sequence.intros(2)[OF sequence_tail1(2)[OF k]] by auto 
 thus ?case by auto
qed

lemma sequence_tail2: assumes "(s,ss@ts) \<in> sequence ars" shows "last_elt (s,ss@ts) = last_elt (last_elt(s,ss),ts)"
 using assms proof (induct ss arbitrary:s)
  case Nil thus ?case unfolding last_elt_def by auto
 next
  case (Cons x xs) 
  from Cons have fact1: "(s,x#xs@ts) \<in> sequence ars" by auto 
  hence fact2: "(s,x#xs) \<in> sequence ars" using sequence_chop by force 
  from fact1 have  "(snd x,xs@ts) \<in> sequence ars" using sequence_tail1(1) by fast 
  hence key: "last_elt(snd x,xs@ts) = last_elt (last_elt(snd x,xs),ts)" using Cons(1) by auto
  have "last_elt(s,x#xs@ts) = last_elt(snd x,xs@ts)" using sequence_tail1(3)[OF fact1] by auto 
  thus ?case using key using sequence_tail1(3)[OF fact2] by auto 
qed

lemma sequence_concat2:
 assumes "(s,ls) \<in> sequence ars" and "ss2 \<in> sequence ars" and "last_elt (s,ls) = fst ss2"
 shows "(s,ls@snd ss2) \<in> sequence ars \<and> (last_elt (s,ls@snd ss2) = last_elt ss2)"
using assms
proof (induct ls arbitrary: s ss2 rule:list.induct)
 case Nil 
  have "s = fst ss2" using Nil(3) unfolding last_elt_def by auto 
  thus ?case using Nil by auto
 next
 case (Cons x xs) 
 from Cons have ss1: "(s,x#xs) \<in> sequence ars" by auto
 have xs:"(snd x,xs) \<in> sequence ars" and
     mem:"(s,fst x,snd x) \<in> ars" and
     lst:"last_elt (snd x,xs) = fst ss2" using sequence_tail1[OF Cons(2)] Cons(4) by auto
 hence "(snd x,xs@snd ss2) \<in> sequence ars" and x:"last_elt (snd x,xs@snd ss2) = last_elt ss2" 
  using Cons(1)[OF xs Cons(3) lst] by auto 
 hence o: "(s,x#xs@snd ss2) \<in> sequence ars" using sequence.intros(2)[OF mem] by auto 
 hence t: "last_elt (s,x#xs@snd ss2) = last_elt ss2" using x unfolding last_elt_def by auto
 show ?case using o t by auto
qed

lemma sequence_concat:
 assumes "ss1 \<in> sequence ars" and "ss2 \<in> sequence ars" and "last_elt ss1 = fst ss2"
 shows "(fst ss1,snd ss1@snd ss2) \<in> sequence ars" "(last_elt (fst ss1,snd ss1@snd ss2) = last_elt ss2)"
proof -
 from assms(1) obtain s ls where ss1: "ss1 = (s,ls)" by (metis surjective_pairing) 
 show "(fst ss1,snd ss1@snd ss2) \<in> sequence ars" using assms ss1 using sequence_concat2 by force 
 next
 from assms(1) obtain s ls where ss1: "ss1 = (s,ls)" by (metis surjective_pairing) 
 show "(last_elt (fst ss1,snd ss1@snd ss2) = last_elt ss2)" using assms ss1 using sequence_concat2 by force
qed

lemma pasting_h: assumes "(\<tau>,\<sigma>,\<sigma>',\<tau>') \<in> diagram ars" and "(\<nu>,\<sigma>',\<sigma>'',\<nu>') \<in> diagram ars"
 shows "((fst \<tau>,snd \<tau>@snd \<nu>),\<sigma>,\<sigma>'',(fst \<tau>'),snd \<tau>'@snd \<nu>') \<in> diagram ars" proof -
 from assms have tau: "\<tau> \<in> sequence ars" and nu: "\<nu> \<in> sequence ars" unfolding diagram_def by auto 
 from assms have o1: "last_elt \<tau> = fst \<sigma>'" and o2: "fst \<nu> = fst \<sigma>'" unfolding diagram_def by auto
 have o: "last_elt \<tau> = fst \<nu>" using o1 o2 by auto
 hence h1: "(fst \<tau>,snd \<tau> @ snd \<nu>) \<in> sequence ars" (is "?h1 \<in> _") using sequence_concat(1)[OF tau nu o] by auto
 have h1': "last_elt ?h1 = last_elt \<nu>"  using sequence_concat(2)[OF tau nu o] by auto 
 from assms have tau': "\<tau>' \<in> sequence ars" and nu': "\<nu>' \<in> sequence ars" unfolding diagram_def by auto
 from assms have l1: "last_elt \<tau>' = last_elt \<sigma>'" and l2: "last_elt \<sigma>' = fst \<nu>'" 
  unfolding diagram_def by auto 
 have l: "last_elt \<tau>' = fst \<nu>'" unfolding l1 l2 by simp 
 hence h2: "(fst \<tau>',snd \<tau>'@snd \<nu>') \<in> sequence ars" (is "?h2 \<in> _") using sequence_concat[OF tau' nu' l] by auto
 hence h2': "(last_elt ?h2 = last_elt \<nu>')" using sequence_concat(2)[OF tau' nu' l] by auto
 thus ?thesis using h1 h2 h1' h2' using assms unfolding diagram_def by auto  
qed

lemma pasting_v: assumes "(\<tau>,\<sigma>,\<sigma>',\<tau>') \<in> diagram ars" and "(\<tau>',\<nu>,\<nu>',\<tau>'') \<in> diagram ars"
 shows "(\<tau>,(fst \<sigma>,snd \<sigma>@snd \<nu>),(fst \<sigma>',snd \<sigma>'@snd \<nu>'),\<tau>'') \<in> diagram ars"
proof -
 from assms(1) have d1: "(\<sigma>,\<tau>,\<tau>',\<sigma>') \<in> diagram ars" unfolding diagram_def by auto
 from assms(2) have d2: "(\<nu>,\<tau>',\<tau>'',\<nu>') \<in> diagram ars" unfolding  diagram_def by auto
 have "((fst \<sigma>,snd \<sigma>@snd \<nu>),\<tau>,\<tau>'',(fst \<sigma>',snd \<sigma>'@snd \<nu>')) \<in> diagram ars" using pasting_h[OF d1 d2] by auto
 thus ?thesis unfolding diagram_def Let_def by auto 
qed

lemma Decreasing_h: 
 assumes t:"trans r" and w:"wf r" and "Decreasing r (\<tau>,\<sigma>,\<sigma>',\<tau>')" and "Decreasing r (\<nu>,\<sigma>',\<sigma>'',\<nu>')"
 shows "Decreasing r ((fst \<tau>,snd \<tau>@snd \<nu>),\<sigma>,\<sigma>'',(fst \<tau>'),snd \<tau>'@snd \<nu>')" 
 proof -
 from assms(3,4) have d1: "decreasing r (labels \<tau>) (labels \<sigma>) (labels \<sigma>') (labels \<tau>')" and 
                 d2: "decreasing r (labels \<nu>) (labels \<sigma>') (labels \<sigma>'') (labels \<nu>')" unfolding Decreasing_def by auto
 have "decreasing r (labels \<tau> @ labels \<nu>) (labels \<sigma>) (labels \<sigma>'') (labels \<tau>' @ labels \<nu>')" using lemma3_5[OF t w d1 d2] by auto
 hence "decreasing r (labels (fst \<tau>, snd \<tau> @ snd \<nu>)) (labels \<sigma>) (labels \<sigma>'') (labels (fst \<tau>', snd \<tau>' @ snd \<nu>'))" unfolding labels_def by auto 
 thus ?thesis unfolding Decreasing_def by auto 
qed

lemma mult_trans: assumes t:"trans r" and "(M,N) \<in> mult r" and "(N,P) \<in> mult_eq r" shows "(M,P) \<in> mult r"
 using assms unfolding mult_def mult_eq_def by auto

lemma Decreasing_v: 
 assumes t: "trans r" and w:"wf r" and "Decreasing r (\<tau>,\<sigma>,\<sigma>',\<tau>')" and "Decreasing r (\<tau>',\<nu>,\<nu>',\<tau>'')"
 shows "Decreasing r (\<tau>, (fst \<sigma>,snd \<sigma>@snd \<nu>),(fst \<sigma>',snd \<sigma>'@snd \<nu>'),\<tau>'')" proof -
 from assms(3,4) have D1:"Decreasing r (\<sigma>,\<tau>,\<tau>',\<sigma>')" and D2: "Decreasing r (\<nu>,\<tau>',\<tau>'',\<nu>')" 
  unfolding Decreasing_def using decreasing_mirror[OF t w] by auto
 hence "Decreasing r ((fst \<sigma>,snd \<sigma>@snd \<nu>),\<tau>,\<tau>'',(fst \<sigma>',snd \<sigma>'@snd \<nu>'))" using Decreasing_h[OF t w] by fast
 thus ?thesis unfolding Decreasing_def using decreasing_mirror[OF t w] by auto 
qed

lemma trivial_diagram: assumes "\<sigma> \<in> sequence ars" shows "(\<sigma>,(fst \<sigma>,[]),(last_elt \<sigma>,[]),\<sigma>) \<in> diagram ars"
 using assms sequence.intros(1) unfolding diagram_def Let_def last_elt_def by auto

lemma trivial_Decreasing: assumes "\<sigma> \<in> sequence ars" 
 shows "Decreasing r (\<sigma>,(fst \<sigma>,[]),(last_elt \<sigma>,[]),\<sigma>)"
 unfolding Decreasing_def using assms unfolding decreasing_def labels_def mult_eq_def by auto


lemma indirect_step: assumes "(C,D) \<in> rel r" shows "(measure r C,measure r D) \<in> mult r"
using assms unfolding rel_def by auto 

lemma sn_construct: assumes "\<not> SN r" shows "\<exists>S.\<forall>i. (S i, S (Suc i)) \<in> r" proof -
 from assms obtain elt where "\<not> SN_on r {elt}" unfolding SN_def by fast 
 thus ?thesis  unfolding SN_on_def by auto 
qed

lemma strange_pre: assumes "\<not> wf r" shows "\<not> SN (r\<inverse>)" using assms unfolding SN_iff_wf converse_converse by auto

lemma strange2: 
  assumes "rel r = relr" and "\<not> wf (relr)" 
  shows "\<not> SN ((relr)\<inverse>)" using assms(2) 
  apply (rule contrapos_nn) using assms(1) by (metis strange_pre) 

lemma wf_indirect2: assumes "rel r = relr" and "\<not> wf (relr)" shows "\<not> wf (mult r)" proof -
 have f: "\<not> SN ((relr)\<inverse>)" using strange2[OF assms] by auto
 from this obtain s where "\<forall>i. (s i, s (Suc i)) \<in> relr\<inverse>" using sn_construct[OF f] by auto
 hence nm:"\<forall>i. (measure r (s i), measure r (s (Suc i))) \<in> (mult r)\<inverse>" using indirect_step assms(1) by blast
 hence "\<not> SN ((mult r)\<inverse>)" using chain_imp_not_SN_on[OF nm] unfolding SN_on_def by auto
 thus ?thesis unfolding SN_iff_wf converse_converse .
qed

lemma wf: assumes "wf r" shows "wf (rel r)" proof -
 have "wf (mult r)" using wf_mult[OF assms] by auto
 thus ?thesis using wf_indirect2 by auto
qed
lemma P_induct:
 assumes wf: "wf r" and step: "\<And>b.(\<And>a.(a,b) \<in> r \<Longrightarrow> P a) \<Longrightarrow> P b" shows "P b"
proof - 
 show ?thesis using wf
 proof induct
 case (less a)
 with step show ?case by auto
qed
qed

lemmas P_induct_rule = P_induct[consumes 1, case_names IH, induct pred: peak]

lemma main: assumes t: "trans r" and w: "wf r" and d: "LD_D lars r" shows "D_D lars r"
 proof -
 have wf: "wf (rel r)" using wf[OF w] . show ?thesis unfolding D_D_def proof
 fix P assume A: "P \<in> peak lars" thus "PDecreasing lars r P" proof (induct rule:P_induct_rule[OF wf])
  case (1 P)
 hence IH1: "\<And>a.(a, P) \<in> rel r \<and> a \<in> peak lars \<Longrightarrow> PDecreasing lars r a"
 and P: "P \<in> peak lars" by auto
   from P obtain \<tau> \<sigma>  where decompose: "P = (\<tau>,\<sigma>)" unfolding peak_def using prod_cases3 by metis 
   from decompose have tau: "\<tau> \<in> sequence lars" using P unfolding peak_def by auto 
   from decompose have sigma: "\<sigma> \<in> sequence lars" using P unfolding peak_def by auto
   have tau_sigma: "fst \<tau> = fst \<sigma>" using P unfolding decompose peak_def by auto
  hence P': "(\<tau>,\<sigma>) \<in> peak lars" using P unfolding decompose by auto 
  show  ?case
  proof (cases "snd \<tau>")
   case Nil 
     have d: "(\<sigma>,(fst \<sigma>,[]),(last_elt \<sigma>,[]),\<sigma>) \<in> diagram lars" (is "?D \<in> diagram lars") using trivial_diagram[OF sigma] by auto
     have D: "Decreasing r ?D" using trivial_Decreasing[OF sigma] by auto
     have d: "((fst \<sigma>,[]),\<sigma>,\<sigma>,(last_elt \<sigma>,[])) \<in> diagram lars" (is "?D \<in> _") using d unfolding diagram_def Let_def by auto 
     have D: "Decreasing r ?D" using D unfolding Decreasing_def using decreasing_mirror[OF t w] by auto 
     have d: "(\<tau>,\<sigma>,\<sigma>,(last_elt \<sigma>,[])) \<in> diagram lars" (is "?D \<in> _") using d tau_sigma Nil surjective_pairing by metis
     have D: "Decreasing r ?D" using D tau_sigma Nil surjective_pairing by metis
     from d D show ?thesis unfolding PDecreasing_def decompose using sigma sequence.intros(1) apply simp by force
   next
   case (Cons \<beta>_step \<nu>_step) hence C1: "snd \<tau> = \<beta>_step#\<nu>_step" by auto show ?thesis
   proof (cases "snd \<sigma>")
    case Nil 
     have d: "(\<tau>,(fst \<tau>,[]),(last_elt \<tau>,[]),\<tau>) \<in> diagram lars" (is "?D \<in> diagram lars") using trivial_diagram[OF tau] by auto
     have D: "Decreasing r ?D" using trivial_Decreasing[OF tau] by auto
     have d: "(\<tau>,\<sigma>,(last_elt \<tau>,[]),\<tau>) \<in> diagram lars" (is "?D \<in> _") using d tau_sigma Nil surjective_pairing by metis 
     have D: "Decreasing r ?D" using D tau_sigma Nil surjective_pairing by metis
     from d D show ?thesis unfolding PDecreasing_def decompose using tau sequence.intros(1) apply simp by force 
    next
    case (Cons \<alpha>_step \<rho>_step) hence C2: "snd \<sigma> = \<alpha>_step#\<rho>_step" by auto
    have tau_decompose: "(fst \<tau>,[\<beta>_step]@\<nu>_step) \<in> sequence lars" using tau C1 by 
     (metis append_Cons append_Nil pair_collapse) 
    hence beta: "(fst \<tau>,[\<beta>_step]) \<in> sequence lars" (is "?\<beta> \<in> _") using sequence_chop by metis 
    have sigma_decompose: "(fst \<sigma>,[\<alpha>_step]@\<rho>_step) \<in> sequence lars" using sigma C2 by (metis append_Cons append_Nil pair_collapse)
    hence alpha: "(fst \<sigma>,[\<alpha>_step]) \<in> sequence lars" (is "?\<alpha> \<in> _") using sequence_chop by metis

    from tau_sigma have LP: "(?\<beta>,?\<alpha>) \<in> local_peak lars" using beta alpha unfolding local_peak_def peak_def by auto 
    hence P1: "(?\<beta>,?\<alpha>) \<in> peak lars" unfolding peak_def local_peak_def by auto 
   
   have "PDecreasing lars r (?\<beta>,?\<alpha>)" using LP d unfolding LD_D_def by best 
   from this obtain \<kappa> \<mu> where kappa:"\<kappa>\<in>sequence lars" and mu:"\<mu>\<in>sequence lars" and
      d1:"(?\<beta>,?\<alpha>,\<kappa>,\<mu>) \<in> diagram lars" and D1:"Decreasing r (?\<beta>,?\<alpha>,\<kappa>,\<mu>)" unfolding PDecreasing_def by auto 
   hence d1:"(?\<beta>,?\<alpha>,\<kappa>,\<mu>) \<in> diagram lars" and D1:"Decreasing r (?\<beta>,?\<alpha>,\<kappa>,\<mu>)" by auto

   from d1 have l1: "last_elt ?\<beta> = fst \<kappa>" unfolding diagram_def Let_def fst_conv snd_conv by force 
   from C1 obtain s where s1: "\<tau> = (s,\<beta>_step#\<nu>_step)" by (metis surjective_pairing)
   hence "(s,\<beta>_step#\<nu>_step) \<in> sequence lars" using tau by auto
   hence "(s,[\<beta>_step]@\<nu>_step) \<in> sequence lars" by auto
   hence "(last_elt (s,[\<beta>_step]),\<nu>_step) \<in> sequence lars" using sequence_tail by fast
   hence nu: "(fst \<kappa>, \<nu>_step) \<in> sequence lars" (is "?\<nu> \<in>sequence lars") using l1 unfolding s1 fst_conv by auto 

   from C2 obtain s2 where s2: "\<sigma> = (s2,\<alpha>_step#\<rho>_step)" by (metis surjective_pairing)
   have P2: "(?\<nu>,\<kappa>) \<in> peak lars" using nu kappa unfolding peak_def by auto
   have beta_ne: "labels (fst \<tau>,[\<beta>_step]) \<noteq> []" unfolding labels_def by auto
   have dec1: "decreasing r (labels ?\<beta>) (labels ?\<alpha>) (labels \<kappa>) (labels \<mu>)" using D1 unfolding Decreasing_def by auto

   have "( lexmax r (labels ?\<nu>) + lexmax r (labels \<kappa>), lexmax r (labels (fst \<tau>, [\<beta>_step]) @ labels ?\<nu>) + lexmax r (labels (fst \<sigma>, [\<alpha>_step]))) \<in> mult r"
    unfolding decompose tau_decompose rel_def measure_def using lemma3_6[OF t w beta_ne dec1] using union_commute by metis
   hence x1:"((?\<nu>,\<kappa>), (\<tau>,?\<alpha>)) \<in> (rel r)" unfolding rel_def measure_def decompose labels_def s1 by auto 
   
   have "(lexmax r (labels (fst \<sigma>, [\<alpha>_step])),lexmax r (labels \<sigma>)) \<in> mult_eq r" 
    unfolding s2 labels_def snd_conv map.simps lexmax.simps remove_from_empty by (metis empty_neutral(1) lemma2_6_2_a mset_le_exists_conv t union_commute) 
   hence "(lexmax r (labels \<tau>) + lexmax r (labels (fst \<sigma>, [\<alpha>_step])), lexmax r (labels \<tau>) + lexmax r (labels \<sigma>)) \<in> mult_eq r" 
     using mult_eq_add_left[OF t] by auto
   hence "((?\<nu>,\<kappa>),P) \<in> rel r" using x1 mult_trans[OF t] unfolding rel_def measure_def mult_eq_def mult_def decompose by simp  
   hence "PDecreasing lars r (?\<nu>,\<kappa>)" using IH1 P2 unfolding decompose by auto  
   from this obtain \<kappa>' \<nu>' where kappa':"\<kappa>'\<in>sequence lars" and nu': "\<nu>'\<in>sequence lars" and 
      d2: "(?\<nu>,\<kappa>,\<kappa>',\<nu>') \<in> diagram lars" and D2: "Decreasing r (?\<nu>,\<kappa>,\<kappa>',\<nu>')" unfolding PDecreasing_def by auto
   hence d2: "(?\<nu>,\<kappa>,\<kappa>',\<nu>') \<in> diagram lars" and D2: "Decreasing r (?\<nu>,\<kappa>,\<kappa>',\<nu>')" by auto
   
   have "last_elt \<mu> = last_elt \<kappa>" using d1 unfolding diagram_def Let_def by auto
   hence "last_elt \<mu> = fst \<nu>'" using d2 unfolding diagram_def Let_def by auto
   hence tau': "(fst \<mu>,snd \<mu>@(snd \<nu>')) \<in> sequence lars" (is "?\<tau>' \<in> _") using sequence_concat(1)[OF mu nu'] by auto  
  
   from d1 have l2: "last_elt ?\<alpha> = fst \<mu>" unfolding diagram_def Let_def by auto 
   from C2 obtain s where s: "\<sigma> = (s,\<alpha>_step#\<rho>_step)" by (metis surjective_pairing)
   hence "(s,\<alpha>_step#\<rho>_step) \<in> sequence lars" using sigma by auto
   hence "(s,[\<alpha>_step]@\<rho>_step) \<in> sequence lars" by auto
   hence "(last_elt (s,[\<alpha>_step]),\<rho>_step) \<in> sequence lars" using sequence_tail by fast 
   hence rho: "(fst \<mu>,\<rho>_step) \<in> sequence lars" (is "?\<rho> \<in> _") using l2 unfolding s by auto 
   have P3: "(?\<tau>',?\<rho>) \<in> peak lars" using tau' rho unfolding peak_def by auto  
  
have D12: "Decreasing r (\<tau>,?\<alpha>,\<kappa>',?\<tau>')" using Decreasing_h[OF t w D1 D2] C1 
   apply auto by (metis surjective_pairing) 
 
   have dec: "decreasing r (labels \<tau>) (labels ?\<alpha>) (labels \<kappa>') (labels ?\<tau>')"
    using D12 unfolding Decreasing_def by simp
   have alpha_ne: "labels ?\<alpha> \<noteq> []" unfolding labels_def by simp 
   have "((?\<tau>',?\<rho>),P) \<in> rel r" using lemma3_6_v[OF t w alpha_ne dec, where v="labels ?\<rho>"] 
    unfolding rel_def measure_def decompose labels_def s2 by simp 

   hence "PDecreasing lars r (?\<tau>',?\<rho>)" using IH1 P3 by simp 
   from this obtain \<rho>' \<tau>'' where rho':"\<rho>'\<in>sequence lars" and tau'': "\<tau>''\<in>sequence lars" and
       d3: "(?\<tau>',?\<rho>,\<rho>',\<tau>'') \<in> diagram lars" and D3: "Decreasing r (?\<tau>',?\<rho>,\<rho>',\<tau>'')"
    unfolding PDecreasing_def by auto 

 have d12: "(\<tau>,?\<alpha>,\<kappa>',?\<tau>') \<in> diagram lars" using pasting_h[OF d1 d2] C1 s1 tau_sigma fst_conv s1 by auto 
 have x1: "last_elt \<kappa>' = last_elt \<nu>'" using d2 unfolding diagram_def Let_def by auto
 from tau' have "last_elt ?\<tau>' = last_elt (last_elt (fst \<mu>,snd \<mu>),snd \<nu>')"  using sequence_tail2 by force
 hence x2:"last_elt ?\<tau>' = last_elt \<nu>'" by (metis `last_elt \<mu> = fst \<nu>'` surjective_pairing) 
 have x3:"last_elt ?\<tau>' = fst \<rho>'" using d3  unfolding diagram_def Let_def by auto  
 have "last_elt \<kappa>' = fst \<rho>'" using x1 x2 x3 by auto 
 hence sigma': "(fst \<kappa>',snd \<kappa>'@snd \<rho>') \<in> sequence lars" (is "?\<sigma>' \<in> _") 
  using sequence_concat[OF kappa' rho'] by auto 

 have d123: "(\<tau>,\<sigma>,?\<sigma>',\<tau>'') \<in> diagram lars" using pasting_v[OF d12 d3] C2 unfolding s by auto 
 have D123:"Decreasing r (\<tau>,\<sigma>,?\<sigma>',\<tau>'')" using Decreasing_v[OF t w D12 D3] C2 unfolding s by auto 
show ?thesis using d123 D123 sigma' tau'' unfolding PDecreasing_def decompose by auto
qed
qed
qed
qed
qed

(*DD implies CR*)
definition CR_l :: "('a\<times>'b\<times>'a) set \<Rightarrow> bool"
 where "CR_l ars = CR {(a,c) | a b c. (a,b,c) \<in> ars}"

definition label :: "('a\<times>'a) set \<Rightarrow> ('a\<times>'b\<times>'a) set"
 where "label ars = {(a,b,c) | a b c. (a,c) \<in> ars}"

lemma step_imp_sequence: assumes "(a,b) \<in> ars" and "label ars = lars" shows 
"\<exists>ss \<in> sequence lars. fst ss = a \<and> last_elt ss = b" 
proof -
 obtain \<alpha> where step:"(a,\<alpha>,b) \<in> lars" using assms unfolding label_def by auto
 hence ss: "(a,[(\<alpha>,b)]) \<in> sequence lars" (is "?ss \<in> _") 
  using Collect_mem_eq sequence.intros by metis
 have fst: "fst ?ss = a" by auto
 have lst: "last_elt ?ss = b" unfolding last_elt_def by auto
 show ?thesis using ss fst lst unfolding last_elt_def by fast 
qed

lemma labeled_step_imp_step: assumes "(a,b,c) \<in> label ars" shows "(a,c) \<in> ars" 
 using assms unfolding label_def by auto 

lemma steps_imp_sequence:
 assumes "(a,b) \<in> ars^*" and "label ars = lars"
 shows "\<exists>ss \<in> sequence lars. fst ss = a \<and> last_elt ss = b"
 using assms(1) proof
 fix n assume A: "(a,b) \<in> ars^^n" thus ?thesis using assms(2) proof (induct n arbitrary: a b ars)
 case 0
   have eq: "a = b" using 0 by auto
   have "(a,[]) \<in> sequence lars" using sequence.intros(1) by fast 
  thus ?case using fst_conv last_elt_def eq by (metis fst_eqD snd_conv) 
 next
 case (Suc m) 
  obtain c where steps: "(a,c) \<in> ars^^m" and step: "(c,b) \<in> ars" using Suc by auto
  obtain ss where ss1: "ss \<in> sequence lars" and ss2:"fst ss = a" and ss3:"last_elt ss = c" 
   using Suc steps by fast  
  obtain ts where ts1: "ts \<in> sequence lars" and ts2:"fst ts = c" and ts3:"last_elt ts = b" 
   using step_imp_sequence[OF step Suc(3)] by auto 
 hence eq : "last_elt ss = fst ts" using ss3 ts2 by auto
 show ?case using sequence_concat[OF ss1 ts1 eq] ss2 ts3 fst_eqD by metis 
qed
qed
  
lemma sequence_imp_steps:
 assumes "ss \<in> sequence lars" and "fst ss = a" and "last_elt ss = b" and "lars = label ars" shows "(a,b) \<in> ars^*"
 proof -
 from assms obtain ls where A1: "(a,ls) \<in> sequence lars" and A2: "last_elt (a,ls) = b" using surjective_pairing by metis 
 thus ?thesis proof (induct ls arbitrary: a b rule:list.induct)
 case Nil thus ?case unfolding last_elt_def by auto 
 next
 case (Cons x xs) 
  from Cons have A: "(snd x,xs) \<in> sequence lars" by (metis sequence_tail1(1)) 
  from Cons have B: "last_elt (snd x, xs) = b" by (metis sequence_tail1(3)) 
  have fst: "(snd x,b) \<in> ars^*" using Cons(1)[OF A B] by auto 
  from Cons(2) have "(a,fst x,snd x) \<in> lars" by (metis sequence_tail1(2) surjective_pairing) 
  hence "(a,snd x) \<in> ars" using labeled_step_imp_step assms(4) by fast 
 thus ?case using fst by auto
qed
qed

definition CR_elt :: "('a\<times>'a)set \<Rightarrow> 'a \<Rightarrow> bool"
 where "CR_elt ars a = True"
 

lemma decreasing_imp_cr: assumes a1: "D_D lars r" and a2: "label ars = lars" shows "CR ars" proof 
 fix a show "\<And>b c. (a, b) \<in> ars\<^sup>* \<Longrightarrow> (a, c) \<in> ars\<^sup>* \<Longrightarrow> (b, c) \<in> ars\<^sup>\<down>" proof -
 fix b show "\<And>c. (a, b) \<in> ars\<^sup>* \<Longrightarrow> (a, c) \<in> ars\<^sup>* \<Longrightarrow> (b, c) \<in> ars\<^sup>\<down>" proof -
 fix c show "(a,b) \<in> ars^* \<Longrightarrow> (a,c) \<in> ars^* \<Longrightarrow> (b,c) \<in> ars\<^sup>\<down>" proof -
 assume B: "(a,b) \<in> ars^*" show "(a,c) \<in> ars^* \<Longrightarrow> (b,c) \<in> ars\<^sup>\<down>" proof -
 assume C: "(a,c) \<in> ars^*" thus "(b,c) \<in> ars\<^sup>\<down>" proof -
 obtain ss1 ss2 where "(ss1,ss2) \<in> peak lars" and b: "last_elt ss1 = b" and c: "last_elt ss2 = c" 
 unfolding assms(2) using peak_def steps_imp_sequence[OF B a2] steps_imp_sequence[OF C a2] 
 by (metis (lifting, mono_tags) fst_eqD mem_Collect_eq snd_conv) 
 from this have "PDecreasing lars r (ss1,ss2)" using assms(1) unfolding D_D_def by auto
 from this obtain ss3 ss4 where dia: "(ss1,ss2,ss3,ss4) \<in> diagram lars" 
  unfolding PDecreasing_def by auto 
 from dia have ss3: "ss3 \<in> sequence (label ars)" and ss4: "ss4 \<in> sequence (label ars)" 
  unfolding diagram_def using assms(2) by auto
 from dia obtain d where ss3_1: "fst ss3 = b" and ss3_2: "last_elt ss3 = d" and 
                         ss4_1: "fst ss4 = c" and ss4_2:"last_elt ss4 = d"
  using b c unfolding diagram_def Let_def by auto  
 hence bd: "(b,d) \<in> ars^*" unfolding assms(2) using sequence_imp_steps ss3 by fast 
 have "(c,d) \<in> ars^*" unfolding assms(2) using sequence_imp_steps ss4 ss4_1 ss4_2 by fast
 thus ?thesis using bd by auto 
qed
qed
qed
qed
qed
qed

lemma soundness: assumes t: "trans r" and w: "wf r" and d: "LD_D (label ars) r" 
 shows "CR ars" using  main[OF assms] decreasing_imp_cr by auto

end


