theory Forbidden_Pattern_Loops
imports
  Trs
  "~~/src/HOL/Library/Monad_Syntax"
  Forbidden_Patterns
  Outermost_Loops
begin

definition fploop :: "('f,'v)forb_patterns \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)oloop \<Rightarrow> bool"
  where "fploop \<Pi> R fpl \<equiv> case fpl of (C,\<mu>,len,ts,ps) \<Rightarrow>
  (\<forall> i \<le> len. (\<exists> \<sigma> lr. (ts i, ts (Suc i)) \<in> rstep_r_p_s R lr (ps i) \<sigma>)) \<and> 
  (\<forall> i \<le> len. \<forall> n. fpstep_cond \<Pi> (hole_pos C ^ n <#> ps i) (ctxt_subst C \<mu> n (ts i))) \<and>
  (ts (Suc len) = C\<langle>ts 0 \<cdot> \<mu>\<rangle>)"

lemma fploop_cond: assumes loop: "fploop \<Pi> R (C,\<mu>,len,ts,ps)"
  and i: "i \<le> len"
  shows "fpstep_cond \<Pi> (hole_pos C ^ n <#> ps i) (ctxt_subst C \<mu> n (ts i))"
  using assms unfolding fploop_def by blast

lemma fploop_fpstep_p: assumes fploop: "fploop \<Pi> R (C,\<mu>,len,ts,ps)"
  and i: "i \<le> len"
  shows "(ctxt_subst C \<mu> n (ts i), ctxt_subst C \<mu> n (ts (Suc i))) \<in> fpstep_p \<Pi> R (hole_pos C ^ n <#> ps i)"
proof -
  note loop = fploop[unfolded fploop_def]
  from loop i obtain \<sigma> lr where step: "(ts i, ts (Suc i)) \<in> rstep_r_p_s R lr (ps i) \<sigma>" by auto
  show ?thesis
    by (rule fpstep_pI[OF ctxt_subst_step[OF step]], rule fploop_cond[OF fploop], insert i, auto)
qed

lemma fploop_end: assumes fploop: "fploop \<Pi> R (C,\<mu>,len,ts,ps)"
  shows "ctxt_subst C \<mu> n (ts (Suc len)) = ctxt_subst C \<mu> (Suc n) (ts 0)"
proof -
  note loop = fploop[unfolded fploop_def]
  from loop have ts: "ts (Suc len) = C\<langle>ts 0 \<cdot> \<mu>\<rangle>" by auto
  show ?thesis unfolding ts unfolding ctxt_subst_Suc ..
qed

lemma fploop_iterate: assumes fploop: "fploop \<Pi> R (C,\<mu>,len,ts,ps)"
  shows "(ctxt_subst C \<mu> n (ts 0), ctxt_subst C \<mu> (Suc n) (ts 0)) \<in> (fpstep \<Pi> R)^+"
proof -
  let ?t = "\<lambda> i. ctxt_subst C \<mu> n (ts i)"
  from fploop_fpstep_p[OF fploop] have 
   steps_i: "\<And> i. i \<le> len \<Longrightarrow> (?t i, ?t (Suc i)) \<in> fpstep \<Pi> R" unfolding fpstep_def by blast
  have steps: "(?t 0, ?t len) \<in> (fpstep \<Pi> R)^*"
    unfolding rtrancl_fun_conv
    by (rule exI[of _ ?t], rule exI[of _ len], insert steps_i, auto)
  have step: "(?t len, ctxt_subst C \<mu> (Suc n) (ts 0)) \<in> fpstep \<Pi> R"
    using steps_i[of len]
    unfolding fploop_end[OF fploop, of n] by simp
  from steps step show ?thesis by auto
qed

lemma fploop_imp_not_SN: assumes fploop: "fploop \<Pi> R (C,\<mu>,len,ts,ps)"
  shows "\<not> SN_on (fpstep \<Pi> R) {ts 0}"
proof
  assume SN: "SN_on (fpstep \<Pi> R) {ts 0}"
  let ?O = "(fpstep \<Pi> R)^+"
  from SN have SN: "SN_on ?O {ts 0}" by (rule SN_on_trancl)
  let ?t = "\<lambda> i. ctxt_subst C \<mu> i (ts 0)"
  from fploop_iterate[OF fploop] have steps: "\<And> i. (?t i, ?t (Suc i)) \<in> ?O" .
  have steps: "\<exists> f. f 0 \<in> {ts 0} \<and> (\<forall> i. (f i, f (Suc i)) \<in> ?O)" 
    by (rule exI[of _ ?t], insert steps, auto)
  with SN show False unfolding SN_defs by blast
qed

definition fpstep_ctxt_subst_cond :: "('f,'v)forb_pattern \<Rightarrow> ('f,'v)ctxt \<Rightarrow> ('f,'v)subst \<Rightarrow> pos \<Rightarrow> ('f,'v)term \<Rightarrow> bool"
  where "fpstep_ctxt_subst_cond pt C \<mu> p t \<equiv> (\<forall> n. fpstep_cond_single pt (hole_pos C ^ n <#> p) (ctxt_subst C \<mu> n t))"

lemma fploopI: assumes steps: "\<And> i. i \<le> len \<Longrightarrow> \<exists> \<sigma> lr. (ts i, ts (Suc i)) \<in> rstep_r_p_s R lr (ps i) \<sigma>"
  and len: "ts (Suc len) = C\<langle>ts 0 \<cdot> \<mu>\<rangle>"
  and fpcond: "\<And> pt i. i \<le> len \<Longrightarrow> ps i \<in> poss (ts i) \<Longrightarrow> pt \<in> \<Pi> \<Longrightarrow> fpstep_ctxt_subst_cond pt C \<mu> (ps i) (ts i)"
  shows "fploop \<Pi> R (C,\<mu>,len,ts,ps)"
  unfolding fploop_def split_def fst_conv snd_conv
proof (intro conjI allI impI)
  fix i
  assume "i \<le> len"
  from steps[OF this]
  show "\<exists> \<sigma> lr. (ts i, ts (Suc i)) \<in> rstep_r_p_s R lr (ps i) \<sigma>" by blast
next
  show "ts (Suc len) = C\<langle>ts 0 \<cdot> \<mu>\<rangle>" by (rule len)
next
  fix i n
  assume i: "i \<le> len"
  from steps[OF i] have "ps i \<in> poss (ts i)" unfolding rstep_r_p_s_def' by auto
  from fpcond[OF i this, unfolded fpstep_ctxt_subst_cond_def]
  show "fpstep_cond \<Pi> (hole_pos C ^ n <#> ps i) (ctxt_subst C \<mu> n (ts i))" 
    unfolding fpstep_cond_def by blast
qed

section {* decision procedure for forbidden pattern loops *}

context fixed_subst 
begin

definition n0 :: "pos \<Rightarrow> pos \<Rightarrow> pos \<Rightarrow> nat"
where "n0 p q oo \<equiv> nat (ceiling (rat_of_nat (size oo - size q) / (rat_of_nat (size p))))"

definition pos_dec :: "pos \<Rightarrow> pos \<Rightarrow> pos \<Rightarrow> (nat \<times> pos) option"
where "pos_dec p q oo \<equiv>
 if p = \<epsilon> then 
  (case pos_suffix oo q of None \<Rightarrow> None | Some r \<Rightarrow> Some (0,r))
 else
  let n0 = nat (ceiling (rat_of_nat (size oo - size q) / (rat_of_nat (size p)))) in
  (case (pos_suffix oo ((p^n0) <#> q)) of None  \<Rightarrow> None | Some r \<Rightarrow> Some (n0,r))"

lemma pos_dec_sound: 
 assumes l:"pos_dec p q oo = Some (n,r)"
 shows "p^n <#> q = r <#> oo"
proof(cases p, simp)
 case Empty
  with l have "pos_suffix oo q = Some r" unfolding pos_dec_def using option.distinct(1) by (cases "pos_suffix oo q",simp,force)
  with Empty show "q = r <#> oo" using suffix_sound by auto
next
 case (PCons i p')
  (*let ?n0 = "((size oo) - (size q) + (size (i <# p')) - 1) div (size (i <# p'))"*)
  let ?n0 = "nat (ceiling (rat_of_nat (size oo - size q) / (rat_of_nat (size (i <# p')))))"
  let ?q = "(i <# p')^?n0 <#> q"
  from PCons l have n:"?n0 = n  \<and> pos_suffix oo ?q = Some r" unfolding pos_dec_def using option.distinct(1) by (cases "pos_suffix oo ?q",simp,force)
  hence "pos_suffix oo ((i <# p')^n <#> q) = Some r" by auto
  with PCons have "(i <# p')^n <#> q = r <#> oo" using suffix_sound[symmetric] by auto
  with PCons show ?thesis by blast
qed


lemma ceiling_mult_le:"int a * ceiling (b::rat) \<ge> ceiling (rat_of_nat a * b::rat)" 
proof(induct a, simp)
  case (Suc a)
  have "int (Suc a) * ceiling b = int a * ceiling b + ceiling b" by (metis comm_semiring_1_class.normalizing_semiring_rules(2) comm_semiring_1_class.normalizing_semiring_rules(24) of_nat_Suc)
  with Suc have x:"int (Suc a) * ceiling b \<ge> ceiling (rat_of_nat a * b) + ceiling b" by auto
  have "int (Suc a) * ceiling b \<ge> ceiling (rat_of_nat a * b  + b)" using ceiling_add_le by (metis x order_trans)
  thus ?case by (metis comm_semiring_1_class.normalizing_semiring_rules(24) comm_semiring_1_class.normalizing_semiring_rules(3) of_nat_Suc)
qed


lemma pos_dec_complete:
 assumes rn:"p^n <#> q = r <#> oo"
 shows "\<exists> r' n'. pos_dec p q oo = Some (n',r') \<and> (n \<ge> n')"
proof(cases p)
  case Empty
    with rn have "q = r <#> oo" by simp
    then obtain r' where "r' <#> oo = q" and "pos_suffix oo q = Some r'" using suffix_complete by blast
    with Empty show ?thesis using pos_dec_def by auto
  next
  case (PCons i p)
    let ?p = "i <# p"
    let ?n0 = "nat (ceiling (rat_of_nat (size oo - size q) / (rat_of_nat (size ?p))))"
      from PCons have nonzero:"size ?p > 0" by simp
    have n_ge_n0:"n \<ge> ?n0"
    proof-
      from PCons rn have sum:"(size ?p) * n + size q = size r + (size oo)" using power_size by (metis size_append)
      from sum have sum':"size r + (size oo) - size q = (size ?p) * n" by (metis add_diff_cancel_left' nat_add_commute)
      hence rp:"rat_of_nat (size r + (size oo) - size q) = (rat_of_nat (size ?p)) *  rat_of_nat n" by (metis of_nat_mult)
      from nonzero have nonzero':"rat_of_nat (size ?p) > 0.0" using of_nat_0_less_iff by auto
      with sum' rp have key:"nat (ceiling (rat_of_nat (size r + size oo - size q) / (rat_of_nat (size ?p)))) = n" by (metis nonzero ceiling_of_nat ceiling_zero comm_semiring_1_class.normalizing_semiring_rules(12) divide_1 gr_implies_not0 int_eq_0_conv mult_divide_mult_cancel_left nat_int)
      have "rat_of_nat (size r + size oo - size q) / (rat_of_nat (size ?p)) \<ge> (rat_of_nat (size oo - size q)) / (rat_of_nat (size ?p))" by (metis divide_le_cancel le_add1 le_diff_conv nat_add_commute nat_mult_commute of_nat_le_iff of_nat_less_0_iff sum sum')
      with key show ?thesis by (metis ceiling_mono nat_mono)
    qed
    have pn0q_ge_o:"size (?p^?n0 <#> q) \<ge> size oo"
    proof-
      have 0:"(size ?p) * ?n0 = nat (int (size ?p)) * ?n0" by (metis nat_int)
      have 1:"nat (int (size ?p)) * ?n0 = nat ((int (size ?p)) * ceiling ((rat_of_nat (size oo - size q) / (rat_of_nat (size ?p)))))" by (metis nat_mult_distrib of_nat_0_le_iff)
      have 2:"nat ((int (size ?p)) * ceiling ((rat_of_nat (size oo - size q) / (rat_of_nat (size ?p))))) \<ge> nat (ceiling ((rat_of_nat (size ?p)) * (rat_of_nat (size oo - size q) / (rat_of_nat (size ?p)))))" using ceiling_mult_le nat_mono by blast
      have 3:"nat (ceiling ((rat_of_nat (size ?p)) * (rat_of_nat (size oo - size q) / (rat_of_nat (size ?p))))) \<ge> nat (ceiling (rat_of_nat (size oo - size q)))" using nonzero by auto
      have 4:"nat (ceiling (rat_of_nat (size oo - size q))) = size oo - size q" by simp
      with 0 1 2 3 have "size ?p * ?n0 \<ge> size oo - size q" by auto
      thus ?thesis using power_size size_append by auto
    qed 
    from n_ge_n0 obtain k where "?p^n = ?p^k <#> ?p^?n0" by (metis le_add_diff_inverse2 power_append_distr)
    with rn PCons have pk:"?p^k <#> ?p^?n0 <#> q = r <#> oo" using append_assoc by auto
    hence "size (?p^k) + size (?p^?n0 <#> q) = size r + size oo" using size_append by metis
    with pn0q_ge_o have kr:"size (?p^k) \<le> size r" by fastforce
    from pk have "?p^k \<le> r <#> oo" using less_eq_simps(1) by metis
    from kr this have "?p^k \<le> r" unfolding  less_eq_pos_def  by (metis diff_is_0_eq' diff_less_mono2 less_eq_simps(1) less_imp_of_nat_less less_pos_def of_nat_0 of_nat_less_0_iff pos_append_cases prefix_smaller)
    then obtain r' where "r = ?p^k <#> r'" using prefix_exists by auto
    with pk have "?p^?n0 <#> q  = r' <#> oo" by auto
    hence "pos_suffix oo (?p^?n0 <#> q) = Some r'" using suffix_complete by force
    with n_ge_n0 show ?thesis using pos_dec_def by (metis PCons option.simps(5) pos.distinct(1))
qed 

definition h_match_probs :: 
 "('f, 'v) term \<Rightarrow> pos \<Rightarrow> pos \<Rightarrow> ('f, 'v) ctxt \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) match_prob set" 
 where "h_match_probs l oo q C t \<equiv> 
  (case (pos_dec (hole_pos C) q oo) of 
    None \<Rightarrow> {}
  | Some (n,o') \<Rightarrow> {((ctxt_subst C \<mu> n t) |_ o', l)})"


lemma fp_h_match_probs_sound:
  assumes qt: "q \<in> poss t" 
    and ol: "oo  \<in> poss l"
    and mp: "mp \<in> h_match_probs l oo q C t"
    and sol: "match_solution mp (n, \<sigma>)"
  shows "\<not> fpstep_ctxt_subst_cond (ctxt_of_pos_term oo l,l|_oo,Forbidden_Patterns.H) C \<mu> q t"
proof-
 from mp have "\<exists> n o'. pos_dec (hole_pos C) q oo = Some (n,o')" using h_match_probs_def by (cases "pos_dec (hole_pos C) q oo", auto) 
 then obtain k o' where sm:"pos_dec (hole_pos C) q oo = Some (k,o')" by auto
 let ?q = "(hole_pos C)^k <#> q"
 let ?pt = "(ctxt_of_pos_term oo l,l|_oo,Forbidden_Patterns.H)"
 let ?t = "ctxt_subst C \<mu> (n + k) t"
 let ?q' = "(hole_pos C)^(n + k) <#> q"
 have qpos:"?q  \<in> poss (ctxt_subst C \<mu> k t)" using qt ctxt_subst_hole_pos by auto
 have ppos:"(hole_pos C)^n \<in> poss ?t " using ctxt_subst_hole_pos by auto
 from sm have q:"?q = o' <#> oo" using pos_dec_sound by auto 
 from qt q have o':"o' \<in> poss (ctxt_subst C \<mu> k t)" using ctxt_subst_hole_pos by (metis ctxt_subst_subt_at poss_append_poss poss_imp_subst_poss)
 hence opos:"o' \<in> poss ((ctxt_subst C \<mu> k t) \<cdot> \<mu> ^^ n)" using poss_imp_subst_poss by auto
 have "((ctxt_subst C \<mu> k t) |_ o') \<cdot> \<mu> ^^ n = l \<cdot> \<sigma>" using mp sol sm  h_match_probs_def match_solution_def by auto 
 with o' have eq:"((ctxt_subst C \<mu> k t) \<cdot> \<mu> ^^ n) |_ o' = l \<cdot> \<sigma>" using subst_subt_at_distr by auto
 with qpos have "?t |_((hole_pos C)^n) |_(hole_pos C)^k <#> q = l \<cdot> \<sigma>|_oo"  using ctxt_subst_add subt_at_subst q by auto
 hence "?t |_ ?q' = l \<cdot> \<sigma>|_oo" using subt_at_append power_append_distr by auto
 from sm have pos:"(hole_pos C) ^ k <#> q = o' <#> oo" by (rule pos_dec_sound)
 have l:"(ctxt_of_pos_term oo l)\<langle>l|_oo \<rangle> = l" using ctxt_supt_id ol by auto
 have l':"hole_pos (ctxt_of_pos_term oo l) = oo" by (metis hole_pos_ctxt_of_pos_term ol)
 have tt:"?t|_(hole_pos C ^n) = (ctxt_subst C \<mu> k t) \<cdot> \<mu> ^^ n" using ctxt_subst_ctxt by auto
 let ?C' = "ctxt_of_pos_term (hole_pos C ^ n <#> o') ?t"
 from eq have "?t|_(hole_pos C ^ n) |_o' = l \<cdot> \<sigma>" using ctxt_subst_ctxt by auto
 hence t_ctxt:"?t = ?C'\<langle>l \<cdot> \<sigma>\<rangle>" using ppos opos tt by (metis ctxt_supt_id pos_append_poss subt_at_append)
 have "hole_pos ?C' = (hole_pos C ^ n <#> o')" by (metis hole_pos_ctxt_of_pos_term opos pos_append_poss ppos tt)
 with t_ctxt q have "\<exists> C' \<sigma>.( ?t = C'\<langle>l \<cdot> \<sigma>\<rangle> \<and> ?q' = hole_pos C' <#> oo)" by (metis append_assoc power_append_distr)
 hence "\<not> fpstep_cond_single ?pt ?q' ?t" using fpstep_cond_single_def[of ?pt ?q'] l l' split_conv by force
 thus ?thesis  using fpstep_ctxt_subst_cond_def[of ?pt C \<mu> q t] by blast
qed

lemma fp_h_match_probs_complete:
  assumes nfp:"\<not>fpstep_cond_single (L, u,Forbidden_Patterns.H ) (hole_pos C ^ n <#> q) (ctxt_subst C \<mu> n t)"
  and qt: "q \<in> poss t"
  shows "\<exists> mp n \<sigma>. mp \<in> h_match_probs L\<langle>u\<rangle> (hole_pos L) q C t \<and> match_solution mp (n, \<sigma>)"
proof-
  let ?ct = "ctxt_subst C \<mu> n t"
  let ?o = "hole_pos L"
  let ?p = "hole_pos C"
  from nfp obtain C' \<sigma> where c:"?ct = C'\<langle>(L \<cdot>c \<sigma>)\<langle>u \<cdot> \<sigma>\<rangle>\<rangle> \<and> ?p ^ n <#> q = (hole_pos C') <#> ?o"
    unfolding fpstep_cond_single_def by auto
  hence oLs:"?o = hole_pos (L \<cdot>c \<sigma>)" using hole_pos_subst by auto
 let ?o' = "hole_pos C'"
  from c oLs have pos_eq0:"?p^n <#> q = ?o' <#> ?o" by fast
  then obtain o'' n' where pd:"pos_dec ?p q ?o = Some (n',o'')" and n_ge_n':"n \<ge> n'" using pos_dec_complete by blast
  then obtain k where nkn:"n = k + n'" by (metis le_iff_add nat_add_commute)
  let ?mp = "((ctxt_subst C \<mu> n' t) |_ o'', L\<langle>u\<rangle>)"
  from pd have pos_eq1:"?p^n' <#> q = o'' <#> ?o" by (rule pos_dec_sound)
  hence "?p^k <#> ?p^n' <#> q = ?p^k <#>  o'' <#> ?o" by metis
  with pos_eq1 have "?o' <#> ?o = ?p^k <#>  o'' <#> ?o" using nkn pos_eq0 power_append_distr by auto
  hence pos_eq2:"?o' = ?p^k <#> o''" using same_append_eq' append_assoc by metis
  have "?p^n' \<in> poss (ctxt_subst C \<mu> n' t)" using ctxt_subst_hole_pos by auto
  hence p_pos:"?p^n' <#> q \<in> poss (ctxt_subst C \<mu> n' t)" using ctxt_subst_subt_at qt by (metis pos_append_poss poss_imp_subst_poss)
  from c have ls:"L\<langle>u\<rangle> \<cdot> \<sigma> = (ctxt_subst C \<mu> n t) |_ ?o'" using subst_apply_term_ctxt_apply_distrib subt_at_hole_pos by auto
  also have "\<dots> = ?ct |_ (?p^k <#> o'')" using pos_eq2 by auto  
  also have "\<dots> =  (ctxt_subst C \<mu> k (ctxt_subst C \<mu> n' t)) |_  (?p^k <#> o'')" using nkn using ctxt_subst_add by auto
  also have "\<dots> =  (ctxt_subst C \<mu> n' t) \<cdot> \<mu> ^^ k |_ o''" by (metis ctxt_subst_hole_pos ctxt_subst_subt_at subt_at_append) 
  also have "\<dots> = (ctxt_subst C \<mu> n' t) |_ o'' \<cdot> \<mu> ^^ k" using p_pos by (metis pos_eq1 poss_append_poss subst_subt_at_distr)
  finally have sol:"match_solution ?mp (k,\<sigma>)" using match_solution_def by force
  from pd n_ge_n' have "h_match_probs L\<langle>u\<rangle> (hole_pos L) q C t = {?mp}" unfolding h_match_probs_def using split by (metis option.simps(5))
  with sol have "?mp \<in> h_match_probs L\<langle>u\<rangle> (hole_pos L) q C t \<and> match_solution ?mp (k, \<sigma>)" by auto
  thus ?thesis by fast
qed


definition a_match_probs :: 
 "('f, 'v) term \<Rightarrow> pos \<Rightarrow> pos \<Rightarrow> ('f, 'v) ctxt \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) match_prob set" 
 where "a_match_probs l oo q C t \<equiv>
  case t|_ q of Var x \<Rightarrow> {}
   | Fun f ls \<Rightarrow>
   {(u, l). \<exists>o'' q'. u = (ctxt_subst C \<mu> (n0 (hole_pos C) q oo) t) |_ o'' \<and>  (o'' \<le> (hole_pos C)^(n0 (hole_pos C) q oo) <#> q <#> q' \<and> q <#> q' \<in> poss t \<and> q' > \<epsilon> \<and> ((hole_pos C)^ (n0 (hole_pos C) q oo) <#> q) < (o'' <#> oo)) } 
   \<union>
   {(u,l). \<exists> x. x \<in> W (t |_ q)\<and> ((Var x) \<cdot> \<mu> \<unrhd> u)}"

lemma fp_a_match_probs_sound:
  assumes qt: "q \<in> poss t" 
    and ol: "oo  \<in> poss l"
    and mp: "(u,l) \<in> a_match_probs l oo q C t"
    and sol: "match_solution (u,l) (n, \<sigma>)"
  shows "\<not> fpstep_ctxt_subst_cond (ctxt_of_pos_term oo l,l|_oo,Forbidden_Patterns.A) C \<mu> q t"
proof(cases "t |_ q")
  case (Var x)
    hence "a_match_probs l oo q C t = {}" unfolding a_match_probs_def by auto
    with mp show ?thesis by blast
  next
  case (Fun f ls)
   hence tqf: "t |_ q = Fun f ls" by auto
   let ?p = "hole_pos C"
   let ?n0 = "n0 ?p q oo"
   let ?pt = "(ctxt_of_pos_term oo l,l|_oo,Forbidden_Patterns.A)"
   from mp[unfolded a_match_probs_def] have 
   ab:"(u,l) \<in> {(u, l). \<exists>o'' q'. u = (ctxt_subst C \<mu> ?n0 t) |_ o'' \<and>  (o'' \<le> ?p^?n0 <#> q <#> q' \<and> q <#> q' \<in> poss t \<and> q' > \<epsilon> \<and> (?p^?n0 <#> q) < (o'' <#> oo)) } \<or>
   (u,l) \<in> {(u,l). \<exists> x. x \<in> W (t |_ q)\<and> ((Var x) \<cdot> \<mu> \<unrhd> u)}" (is "?A \<or> ?B") using Un_iff tqf by auto
   have l:"(ctxt_of_pos_term oo l)\<langle>l|_oo \<rangle> = l" using ctxt_supt_id ol by auto
   have l':"hole_pos (ctxt_of_pos_term oo l) = oo" using ol by (metis hole_pos_ctxt_of_pos_term)
   show ?thesis
   proof(cases ?A)
    case True
    assume ?A
    with sol[unfolded match_solution_def] obtain o'' q' where  "((ctxt_subst C \<mu> ?n0 t) |_ o'') \<cdot> \<mu> ^^ n = l \<cdot> \<sigma>" and 
           ole:"o'' \<le> ?p ^ ?n0 <#> q <#> q'" and qt:"q <#> q' \<in> poss t" and qe:"q' > \<epsilon>" and pp:"?p ^ ?n0 <#> q < o'' <#> oo" by auto
    then obtain k where ls:"((ctxt_subst C \<mu> ?n0 t) |_ o'') \<cdot> (\<mu> ^^ k) =  l \<cdot> \<sigma>" by auto
    let ?o' = "?p^k <#> o''"
    let ?t = "ctxt_subst C \<mu> (k + ?n0) t"
    have "?p ^ ?n0 \<in> poss (ctxt_subst C \<mu> ?n0 t)" using ctxt_subst_hole_pos ctxt_subst_add by auto
    with qt have "?p ^ ?n0 <#> q <#> q' \<in> poss (ctxt_subst C \<mu> ?n0 t)" using  poss_append_poss poss_imp_subst_poss by force
    hence opos: "o'' \<in> poss (ctxt_subst C \<mu> ?n0 t)" using ole less_eq_pos_def poss_append_poss by metis
    have "?p^k \<in> poss (ctxt_subst C \<mu> (k + ?n0) t)" using ctxt_subst_hole_pos ctxt_subst_add by auto
    hence ot:"?o' \<in> poss ?t" using opos using ctxt_subst_add ctxt_subst_subt_at pos_append_poss poss_imp_subst_poss by auto
    hence "?t |_ ?o' = ((ctxt_subst C \<mu> k (ctxt_subst C \<mu> ?n0 t)) |_ ?p^k |_ o'')" using ctxt_subst_add subt_at_append by auto
    also have "\<dots> = (ctxt_subst C \<mu> ?n0 t) \<cdot> \<mu> ^^ k |_ o''" using ctxt_subst_subt_at by auto
    finally have eq:"?t |_ ?o' =  l \<cdot> \<sigma>" using ls opos by auto
    from pp have posc:"?p^(k + ?n0) <#> q  < ?o' <#> oo" using power_append_distr by auto
    let ?q = "?p^(k + ?n0) <#> q"
    let ?C' = "ctxt_of_pos_term ?o' ?t"
    from eq ot have t_ctxt:"?t = ?C'\<langle>l \<cdot> \<sigma>\<rangle>" by (metis ctxt_supt_id)
    have "hole_pos ?C' = ?o'" by (metis hole_pos_ctxt_of_pos_term ot)
    with t_ctxt posc have "\<exists>C' \<sigma>. ?t = C'\<langle>l \<cdot> \<sigma>\<rangle> \<and> ?p^(k + ?n0) <#> q  < (hole_pos C') <#> oo" by metis
    hence "\<not>fpstep_cond_single ?pt ?q ?t" using fpstep_cond_single_def[of ?pt] l l' split_conv by force
    thus ?thesis using fpstep_ctxt_subst_cond_def[of ?pt C \<mu> q t] by blast
   next
    case False
    hence ?B using ab by auto
    then obtain x where "x \<in> W (t |_ q)" and supt_u:"(Var x) \<cdot> \<mu> \<unrhd> u" and ul:"u \<cdot> \<mu> ^^ n = l \<cdot> \<sigma>" using sol match_solution_def by auto
    then obtain i where "x \<in> vars_term (t |_ q \<cdot> \<mu> ^^ i)" unfolding W_def by auto
    then obtain o'' where tx:"(t |_ q \<cdot> \<mu> ^^ i) |_ o'' = Var x" and ot:"o'' \<in> poss (t |_q \<cdot> \<mu> ^^ i)" 
     using supteq_imp_subt_at vars_term_supteq(1)[of x] by blast
    obtain r where u:"(Var x) \<cdot> \<mu> |_r = u" and r:"r \<in> poss ((Var x)  \<cdot> \<mu>)" using supteq_imp_subt_at[OF supt_u] by auto
    let ?n = "Suc i + n"
    let ?o' = "?p ^ ?n <#> q <#> o'' <#> r"
    let ?t = "ctxt_subst C \<mu> ?n t"
    have eq:"?t |_ ?o' = l \<cdot> \<sigma>"
     proof-
      have ppos: "?p ^ ?n \<in> poss (ctxt_subst C \<mu> ?n t)" using ctxt_subst_hole_pos by blast
      hence t_o':"?t |_ ?o' = ?t |_ ?p^?n |_ (q <#> o'' <#> r)" using ctxt_subst_hole_pos by auto
     have "\<dots> = t \<cdot> \<mu> ^^ ?n |_ (q <#> o'' <#> r)" using ctxt_subst_subt_at[of C \<mu> ?n] by auto
     also have "\<dots> = t |_q \<cdot> \<mu> ^^ ?n |_ (o'' <#> r)" using qt by auto
     also have "\<dots> = t |_q \<cdot> (\<mu> ^^ i \<circ>s \<mu> ^^ (Suc n)) |_ (o'' <#> r)" using subst_power_compose_distrib add_Suc_shift by metis
     also have lm:"\<dots> = t |_q \<cdot> \<mu> ^^ i \<cdot> \<mu> ^^ (Suc n) |_ (o'' <#> r)" using subst_compose_apply_term_distrib by auto
     also have "\<dots> = t |_q \<cdot> \<mu> ^^ i |_ o'' \<cdot> \<mu> ^^ (Suc n) |_ r" using lm ot by simp
     also have "\<dots> = (Var x) \<cdot> \<mu> \<cdot> \<mu> ^^ n |_ r" using tx subst_monoid_mult.power_Suc2 by auto 
     finally show ?thesis using r u subt_at_subst ul t_o' by auto
     qed
    have cc:"ctxt_subst C \<mu> ((Suc i) + n) t = ctxt_subst C \<mu> (Suc i) (ctxt_subst C \<mu> n t)" using ctxt_subst_add by auto
    have ot: "?o' \<in> poss ?t"
     proof-
      from tx u have  "(t |_ q \<cdot> \<mu> ^^ i) \<cdot> \<mu> |_ o'' = (Var x) \<cdot> \<mu>" using subt_at_subst ot by force
      with r ot have "o'' <#> r \<in> poss (t |_q \<cdot> \<mu> ^^ i \<cdot> \<mu>)" using pos_append_poss[OF ot] by auto 
      with qt have "q <#> o'' <#> r \<in> poss (t \<cdot> \<mu> ^^ i \<cdot> \<mu>)" using poss_append_poss by auto
      hence qor:"q <#> o'' <#> r \<in> poss (t \<cdot> \<mu> ^^ (Suc i))" using subst_compose_apply_term_distrib subst_power_Suc by metis
      let ?s = "ctxt_subst (C \<cdot>c (\<mu> ^^ (Suc i) )) \<mu> n (t \<cdot> \<mu> ^^ (Suc i))"
      have ax:"?s |_ ?p^n = (t \<cdot> \<mu> ^^ (Suc i)) \<cdot> \<mu> ^^ n" using ctxt_subst_subt_at hole_pos_subst by metis
      have ps:"?p^n \<in> poss ?s" using hole_pos_subst ctxt_subst_hole_pos by metis
      with qor ax have pqor:"?p ^ n <#> (q <#> o'' <#> r) \<in> poss ?s" using 
       pos_append_poss[OF ps] poss_imp_subst_poss by metis
      have ce:"(ctxt_subst C \<mu> n t) \<cdot> (\<mu> ^^ (Suc i)) = ?s" using ctxt_subst_subst_pow by blast
      with pqor have pqor:"?p ^ n <#> (q <#> o'' <#> r) \<in> poss ((ctxt_subst C \<mu> n t) \<cdot> (\<mu> ^^ (Suc i)))" by auto
      from cc have tp:"?t |_ ?p^(Suc i) = (ctxt_subst C \<mu> n t) \<cdot> (\<mu> ^^ (Suc i))" using ctxt_subst_subt_at[of C \<mu> "Suc i"] by auto
      with pqor have pqor:"?p ^ n <#> (q <#> o'' <#> r) \<in> poss (?t |_ ?p^(Suc i))" by auto
      have pit:"?p^(Suc i) \<in> poss ?t" using ctxt_subst_add ctxt_subst_hole_pos by auto
      with pqor show ?thesis using  pos_append_poss[OF pit pqor] power_append_distr by auto
     qed
    have o'':"o'' > \<epsilon>" 
     proof-
      from tx have "is_Var (t \<cdot> \<mu> ^^ i |_ (q <#> o''))" 
       using is_VarI poss_imp_subst_poss qt subt_at_append subt_at_subst by metis
      with tqf show ?thesis using Term.term.distinct(1) append_Empty less_pos_def' subst_Var_imp_Var subt_at.simps(1) tx by metis
    qed
    let ?q = "?p^?n <#> q"
    from o'' have posc:"?o' <#> oo > ?q" using append_assoc less_pos_def' less_simps(5) by auto
    let ?C' = "ctxt_of_pos_term ?o' ?t"
    from eq ot have t_ctxt:"?t = ?C'\<langle>l \<cdot> \<sigma>\<rangle>" by (metis ctxt_supt_id)
    have "hole_pos ?C' = ?o'" by (metis hole_pos_ctxt_of_pos_term ot)
    with t_ctxt posc have "\<exists>C' \<sigma>. ?t = C'\<langle>l \<cdot> \<sigma>\<rangle> \<and> ?q  < (hole_pos C') <#> oo" by auto
    hence "\<not> fpstep_cond_single ?pt ?q ?t" unfolding fpstep_cond_single_def[of ?pt ?q ?t] split_conv l l' by force
    thus ?thesis using fpstep_ctxt_subst_cond_def[of ?pt C \<mu> ?q t] by (metis fpstep_ctxt_subst_cond_def)
  qed
qed

definition b_match_probs :: "('f, 'v) term \<Rightarrow> pos \<Rightarrow> pos \<Rightarrow> ('f, 'v) ctxt \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) match_prob set" where
  "b_match_probs l oo q C t \<equiv> \<Union> {h_match_probs l oo q' C t | q'. q' < q}"

definition n0b :: "pos \<Rightarrow> pos \<Rightarrow> pos \<Rightarrow> nat" where
 "n0b p'' p oo \<equiv> Min { n | n. n \<le> (size oo) \<and> size p'' + n * (size p) > size oo \<and> p'' <#> p^n > oo }"

definition b_ematch_probs :: "('f, 'v) term \<Rightarrow> pos \<Rightarrow>  pos \<Rightarrow> ('f, 'v) ctxt \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) ematch_prob set" where
  "b_ematch_probs l oo q C t \<equiv> {(C |_c p', l, C \<cdot>c \<mu>, (ctxt_subst C \<mu> n t) \<cdot> \<mu>) | p' p'' n. 
   p' <#> p'' = hole_pos C \<and> (n = n0 p'' (hole_pos C) oo) \<and> p' < hole_pos C }"

lemma b_match_probs_sound:
  assumes qt: "q \<in> poss t" 
    and ol: "oo  \<in> poss l"
    and sol: "match_solution mp (n, \<sigma>)"
    and mp: "mp \<in> b_match_probs l oo q C t"
  shows "\<not> fpstep_ctxt_subst_cond (ctxt_of_pos_term oo l,l|_oo,Forbidden_Patterns.B) C \<mu> q t"
proof -
  let ?p = "hole_pos C"
  let ?l = "(ctxt_of_pos_term oo l)\<langle>l|_oo\<rangle>"
  let ?pt = "(ctxt_of_pos_term oo l,l|_oo,Forbidden_Patterns.H)"
  let ?pt' = "(ctxt_of_pos_term oo l,l|_oo,Forbidden_Patterns.B)"
  from mp[unfolded b_match_probs_def] obtain q' where q': "q' < q" and mp: "mp \<in> h_match_probs l oo q' C t" by auto
  have l:"(ctxt_of_pos_term oo l)\<langle>l|_oo \<rangle> = l" using ctxt_supt_id ol by auto
  have l':"hole_pos (ctxt_of_pos_term oo l) = oo" using hole_pos_ctxt_of_pos_term ol by metis
  from l have ls:" \<forall> \<sigma>. (ctxt_of_pos_term oo l \<cdot>c \<sigma>)\<langle>l |_ oo \<cdot> \<sigma>\<rangle> = l  \<cdot> \<sigma>" by (metis subst_apply_term_ctxt_apply_distrib)
  from qt q' have "q' \<in> poss t" using less_pos_def' poss_append_poss by force
  with mp have "\<not> fpstep_ctxt_subst_cond ?pt C \<mu> q' t" using fp_h_match_probs_sound qt ol sol by auto
  then obtain n where " \<not>fpstep_cond_single ?pt (?p ^ n <#> q') (ctxt_subst C \<mu> n t)" using fpstep_ctxt_subst_cond_def by blast
  hence "\<exists>C' \<sigma>. ctxt_subst C \<mu> n t = C'\<langle>(ctxt_of_pos_term oo l \<cdot>c \<sigma>)\<langle>l |_ oo \<cdot> \<sigma>\<rangle>\<rangle> \<and>
         ?p ^ n <#> q' = hole_pos C' <#> oo" unfolding fpstep_cond_single_def using l' by auto
  then obtain C' \<sigma> where c':"ctxt_subst C \<mu> n t = C'\<langle>l \<cdot> \<sigma>\<rangle>" and st:"(?p ^ n <#> q') = (hole_pos C') <#> oo" using ls l' by metis
  with q' have "(?p ^ n <#> q) > (hole_pos C') <#> oo" using less_simps(2) by metis
  with c' have "\<exists> C' \<sigma>. (ctxt_subst C \<mu> n t) = C'\<langle>?l \<cdot> \<sigma>\<rangle> \<and> ?p ^ n <#> q > (hole_pos C') <#> oo" by (metis l)
  hence "\<not>fpstep_cond_single ?pt' (?p ^ n <#> q) (ctxt_subst C \<mu> n t)" using l' fpstep_cond_single_def by force
  thus ?thesis  using fpstep_ctxt_subst_cond_def by blast
qed

definition fp_match_probs :: "('f, 'v) term \<Rightarrow> pos \<Rightarrow> pos \<Rightarrow> ('f, 'v) ctxt \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) match_prob set" where
  "fp_match_probs l oo q C t \<equiv> 
    (h_match_probs l oo q C t) \<union> (a_match_probs l oo q C t) \<union>  (b_match_probs l oo q C t)" 

(*
lemma fp_match_probs_complete:
  assumes nfp:"\<not> fpstep_ctxt_subst_cond (L,u,loc) C \<mu> q t"
  and qt: "q \<in> poss t"
  shows "(\<exists> mp n \<sigma>. mp \<in> fp_match_probs L\<langle>u\<rangle> (hole_pos L) q C t \<and> match_solution mp (n, \<sigma>)) \<or>
         (\<exists> ep n \<sigma> k. ep \<in> fp_b_ematch_probs L\<langle>u\<rangle> (hole_pos L) q C t \<and> ematch_solution ep (n, k, \<sigma>))" 
proof
  from nfp obtain n where "\<not>fpstep_cond_single (L, u, loc) (hole_pos C ^ n <#> q) (ctxt_subst C \<mu> n t)" 
    unfolding fpstep_ctxt_subst_cond_def by auto
  oops*)



end

end
