(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Innermost_Usable_Rules_Impl
imports
  Innermost_Usable_Rules 
  Icap_Impl
  QDP_Framework_Impl
  "$AFP/Transitive-Closure-II/RTrancl"
begin

fun is_ur_closed_term_impl where
  "is_ur_closed_term_impl NFQ e_cap R U S (Var x) = True"
| "is_ur_closed_term_impl NFQ e_cap R U S (Fun f ts) = 
    ((\<forall> t \<in> set ts. is_ur_closed_term_impl NFQ e_cap R U S t) \<and>                
    (\<forall>(l, r) \<in> set R. (l, r) \<in> set U \<or>
      (case mgu_class (Fun f (map e_cap ts)) l of
        None \<Rightarrow> True
      | Some \<mu> \<Rightarrow> \<not> (\<forall> u \<in> set (args l). NFQ (mv_yvar u \<cdot> \<mu>)) \<or> \<not> (\<forall> u \<in> set S. NFQ (u \<cdot> \<mu>)))))"

(* TODO: clarify with Alex why one cannot use n within the zip-call (then the function package does not terminate) *)
fun is_ur_closed_term_af_impl 
  where "is_ur_closed_term_af_impl NFQ e_cap \<pi> R U S (Var x) = True"
     |  "is_ur_closed_term_af_impl NFQ e_cap \<pi> R U S (Fun f ts) = (let n = length ts; \<pi>f = \<pi> (f,n) in
              ((\<forall> (i,t) \<in> set (zip [0 ..< length ts] ts). \<pi>f i \<longrightarrow> is_ur_closed_term_af_impl NFQ e_cap \<pi> R U S t) \<and>                
              (\<forall> (l,r) \<in> set R. (l,r) \<in> set U \<or> (case mgu_class (Fun f (map e_cap ts)) l of
                 None \<Rightarrow> True | Some \<mu> \<Rightarrow> \<not> (\<forall> u \<in> set (args l). NFQ (mv_yvar u \<cdot> \<mu>)) \<or> \<not> (\<forall> u \<in> set S. NFQ (u \<cdot> \<mu>))))))"

declare is_ur_closed_term_af_impl.simps[code del]
lemma [code]: "is_ur_closed_term_af_impl NFQ e_cap \<pi> R U S (Fun f ts) = (let n = length ts; \<pi>f = \<pi> (f,n) in
              ((\<forall> (i,t) \<in> set (zip [0 ..< n] ts). \<pi>f i \<longrightarrow> is_ur_closed_term_af_impl NFQ e_cap \<pi> R U S t) \<and>                
              (\<forall> (l,r) \<in> set R. (l,r) \<in> set U \<or> (case mgu_class (Fun f (map e_cap ts)) l of
                 None \<Rightarrow> True | Some \<mu> \<Rightarrow> \<not> (\<forall> u \<in> set (args l). NFQ (mv_yvar u \<cdot> \<mu>)) \<or> \<not> (\<forall> u \<in> set S. NFQ (u \<cdot> \<mu>))))))"
               "is_ur_closed_term_af_impl NFQ e_cap \<pi> R U S (Var x) = True"
  unfolding is_ur_closed_term_af_impl.simps Let_def by auto

lemma is_ur_closed_term_impl: fixes R :: "('f,string)rules" and Q :: "('f,string)term list"
  assumes NFQ: "NFQ = (\<lambda> t. t \<in> NF_terms (set Q))"
  and e_cap: "e_cap = icap (set R) (set Q) (set S)"
  shows "is_ur_closed_term_impl NFQ e_cap R U S = is_ur_closed_term' (set R) (set U) (set Q) icap full_af (set S)"
proof -
  let ?is_ur_closed_term = "is_ur_closed_term' (set R) (set U) (set Q) icap full_af"
  {
    fix t
    have "is_ur_closed_term_impl NFQ e_cap R U S t = ?is_ur_closed_term (set S) t"
    proof (induct t)
      case (Fun f ts)
      hence ball: "Ball (set ts) (is_ur_closed_term_impl NFQ e_cap R U S) = 
         (\<forall>i<length ts.
        full_af (f, length ts) i \<longrightarrow> ?is_ur_closed_term (set S) (ts ! i))" unfolding full_af_def[of "(f,length ts)"] set_conv_nth[of ts] by auto
      have all: "\<And> p q. \<lbrakk> \<And> l r. p l r = q l r\<rbrakk> \<Longrightarrow> (\<forall> (l,r) \<in> set R. (l,r) \<in> set U \<or> p l r) = (\<forall> (l,r) \<in> set R. (l,r) \<in> set U \<or> q l r)" by auto
      have ecap: "icap (set R) (set Q) (set S) = e_cap" unfolding e_cap ..
      have conj: "\<And> x y z u. x = y \<Longrightarrow> z = u \<Longrightarrow> (x \<and> z) = (y \<and> u)" by auto
      {
        fix x f 
        have "(case x of None \<Rightarrow> True | Some y \<Rightarrow> f y) = (\<not> (\<exists> y. x = Some y \<and> \<not> f y))" by (cases x, auto)
      } note id = this
      show ?case 
        unfolding is_ur_closed_term_impl.simps is_ur_closed_term'.simps
        by (rule conj[OF ball all], unfold ecap NFQ id, simp) blast
    qed simp
  } note main = this
  show ?thesis
    by (rule ext, rule main)
qed

lemma is_ur_closed_term_af_impl: fixes R :: "('f,string)rules" and Q :: "('f,string)term list"
  assumes NFQ: "NFQ = (\<lambda> t. t \<in> NF_terms (set Q))"
  and e_cap: "e_cap = icap (set R) (set Q) (set S)"
  shows "is_ur_closed_term_af_impl NFQ e_cap \<pi> R U S = is_ur_closed_term' (set R) (set U) (set Q) icap \<pi> (set S)"
proof -
  let ?is_ur_closed_term = "is_ur_closed_term' (set R) (set U) (set Q) icap \<pi>"
  {
    fix t
    have "is_ur_closed_term_af_impl NFQ e_cap \<pi> R U S t = ?is_ur_closed_term (set S) t"
    proof (induct t)
      case (Fun f ts)
      hence ball: "(\<forall> i < length ts. \<pi> (f,length ts) i \<longrightarrow> is_ur_closed_term_af_impl NFQ e_cap \<pi> R U S (ts ! i)) = 
         (\<forall>i<length ts.
        \<pi> (f, length ts) i \<longrightarrow> ?is_ur_closed_term (set S) (ts ! i))" (is "?l' = ?r")
        using Fun[unfolded set_conv_nth[of ts]] by auto
      have id: "?l' = (\<forall> (i,t) \<in> set (zip [0 ..< length ts] ts). \<pi> (f,length ts) i \<longrightarrow> is_ur_closed_term_af_impl NFQ e_cap \<pi> R U S t)" unfolding set_zip by force
      note ball = ball[unfolded id]
      have all: "\<And> p q. \<lbrakk> \<And> l r. p l r = q l r\<rbrakk> \<Longrightarrow> (\<forall> (l,r) \<in> set R. (l,r) \<in> set U \<or> p l r) = (\<forall> (l,r) \<in> set R. (l,r) \<in> set U \<or> q l r)" by auto
      have ecap: "icap (set R) (set Q) (set S) = e_cap" unfolding e_cap ..
      have conj: "\<And> x y z u. x = y \<Longrightarrow> z = u \<Longrightarrow> (x \<and> z) = (y \<and> u)" by auto
      {
        fix x f 
        have "(case x of None \<Rightarrow> True | Some y \<Rightarrow> f y) = (\<not> (\<exists> y. x = Some y \<and> \<not> f y))" by (cases x, auto)
      } note id = this
      show ?case 
        unfolding is_ur_closed_term_af_impl.simps is_ur_closed_term'.simps Let_def
        by (rule conj[OF ball all], unfold ecap NFQ id, simp) blast
    qed simp
  } note main = this
  show ?thesis
    by (rule ext, rule main)
qed

definition is_ur_closed_term_impl_mv 
  where "is_ur_closed_term_impl_mv NFQ e_cap R U S \<equiv> let urc = is_ur_closed_term_impl NFQ e_cap R U (map mv_xvar S) in (\<lambda> t. urc (mv_xvar t))"

definition is_ur_closed_term_af_impl_mv 
  where "is_ur_closed_term_af_impl_mv NFQ e_cap \<pi> R U S \<equiv> let urc = is_ur_closed_term_af_impl NFQ e_cap \<pi> R U (map mv_xvar S) in (\<lambda> t. urc (mv_xvar t))"

lemma is_ur_closed_term_impl_mv: fixes R :: "('f,string)rules" and Q :: "('f,string)term list"
  assumes NFQ: "NFQ = (\<lambda> t. t \<in> NF_terms (set Q))"
  and e_cap: "e_cap = icap (set R) (set Q) (set (map mv_xvar S))"
  shows "is_ur_closed_term_impl_mv NFQ e_cap R U S t = is_ur_closed_term_mv' (set R) (set U) (set Q) icap' full_af (set S) t" 
  unfolding
    is_ur_closed_term_impl_mv_def 
    is_ur_closed_term_impl[OF NFQ e_cap]
    set_map
    is_ur_closed_term_mv_icap' Let_def ..

lemma is_ur_closed_term_af_impl_mv: fixes R :: "('f,string)rules" and Q :: "('f,string)term list"
  assumes NFQ: "NFQ = (\<lambda> t. t \<in> NF_terms (set Q))"
  and e_cap: "e_cap = icap (set R) (set Q) (set (map mv_xvar S))"
  shows "is_ur_closed_term_af_impl_mv NFQ e_cap \<pi> R U S t = is_ur_closed_term_mv' (set R) (set U) (set Q) icap' \<pi> (set S) t" 
  unfolding
    is_ur_closed_term_af_impl_mv_def 
    is_ur_closed_term_af_impl[OF NFQ e_cap]
    set_map
    is_ur_closed_term_mv_icap' Let_def ..

definition is_ur_closed_impl_dpp_mv   :: "('d, 'f :: {key,show}, string) dpp_ops \<Rightarrow> 'd
  \<Rightarrow> ('f,string)rules \<Rightarrow> ('f,string)term list \<Rightarrow> ('f,string)term \<Rightarrow> bool"
  where "is_ur_closed_impl_dpp_mv I d \<equiv>
  let ic = icap_impl_dpp I d;
      qnf = dpp_ops.is_QNF I d;
      r = dpp_ops.rules I d;
      urc = (\<lambda> S. is_ur_closed_term_impl qnf (ic S) r)
    in (\<lambda> U S. let S' = map mv_xvar S in (\<lambda> t. urc S' U S' (mv_xvar t)))"

definition is_ur_closed_af_impl_dpp_mv   :: "('d, 'f :: {key,show}, string) dpp_ops \<Rightarrow> 'd \<Rightarrow> 'f af
  \<Rightarrow> ('f,string)rules \<Rightarrow> ('f,string)term list \<Rightarrow> ('f,string)term \<Rightarrow> bool"
  where "is_ur_closed_af_impl_dpp_mv I d \<pi> \<equiv>
  let ic = icap_impl_dpp I d;
      qnf = dpp_ops.is_QNF I d;
      r = dpp_ops.rules I d;
      urc = (\<lambda> S. is_ur_closed_term_af_impl qnf (ic S) \<pi> r)
    in (\<lambda> U S. let S' = map mv_xvar S in (\<lambda> t. urc S' U S' (mv_xvar t)))"

lemma is_ur_closed_impl_dpp_mv:
  fixes I::"('d, 'f::{show,key}, string) dpp_ops" and d::"'d"
  defines q: "q \<equiv> dpp_ops.Q I d"
      and r: "r \<equiv> dpp_ops.rules I d"
  assumes I: "dpp_spec I"
  shows "is_ur_closed_impl_dpp_mv I d u s t = is_ur_closed_term_mv' (set r) (set u) (set q) icap' full_af (set s) t" 
proof -
  interpret dpp_spec I by fact
  show ?thesis
    unfolding is_ur_closed_term_impl_mv[OF refl refl, symmetric]
      is_ur_closed_impl_dpp_mv_def Let_def is_QNF_sound icap_impl_dpp_icap[OF I] r q 
    unfolding is_ur_closed_term_impl_mv_def Let_def ..
qed    

lemma is_ur_closed_af_impl_dpp_mv:
  fixes I::"('d, 'f::{show,key}, string) dpp_ops" and d::"'d"
  defines q: "q \<equiv> dpp_ops.Q I d"
      and r: "r \<equiv> dpp_ops.rules I d"
  assumes I: "dpp_spec I"
  shows "is_ur_closed_af_impl_dpp_mv I d \<pi> u s t = is_ur_closed_term_mv' (set r) (set u) (set q) icap' \<pi> (set s) t" 
proof -
  interpret dpp_spec I by fact
  show ?thesis
    unfolding is_ur_closed_term_af_impl_mv[OF refl refl, symmetric]
      is_ur_closed_af_impl_dpp_mv_def Let_def is_QNF_sound icap_impl_dpp_icap[OF I] r q 
    unfolding is_ur_closed_term_af_impl_mv_def Let_def ..
qed    

definition
  usable_rules_proc ::
    "('dpp, 'f :: {key,show}, string) dpp_ops \<Rightarrow> 
     ('f,string)rules \<Rightarrow> 'dpp proc"
where
  "usable_rules_proc I U dpp \<equiv> 
    check_return (do {
     check (dpp_ops.NFQ_subset_NF_rules I dpp) (shows ''innermost rewriting required'');
     check (dpp_ops.nfs I dpp \<or> dpp_ops.minimal I dpp \<or> dpp_ops.wwf_rules I dpp) (shows ''normal form subst, minimality or well-formedness required'');
     let P = dpp_ops.pairs I dpp;
     let urc = is_ur_closed_impl_dpp_mv I dpp U;
     let check_urc = (\<lambda> S t. check (urc S t) (shows ''term '' +@+ shows t +@+ shows '' is not closed under usable rules''));
     check_allm (\<lambda>(l, r). do {
        check_subseteq (vars_term_list r) (vars_term_list l) <+? (\<lambda> x. shows ''variable condition in P violated'');
        check_urc [l] r
     }) P;       
     check_allm (\<lambda>(l,r). check_urc (args l) r) U
   })
   (dpp_ops.intersect_rules I dpp U)"


lemma usable_rules_proc: assumes I: "dpp_spec I"
 shows "dpp_spec.sound_proc_impl I (usable_rules_proc I (U :: ('f :: {key,show},string)rules))"
proof -
  from assms interpret dpp_spec I .
  show ?thesis
  proof
    fix d d'    
    assume ok: "usable_rules_proc I U d = Inr d'" and fin: "finite_dpp (dpp d')"
    let ?P = "set (P d)"
    let ?Pw = "set (Pw d)"
    let ?Q = "set (Q d)"
    let ?R = "set (R d)"
    let ?Rw = "set (Rw d)"
    let ?U = "set U"
    let ?Rb = "?R \<union> ?Rw"
    let ?Rb' = "set (rules d)" 
    let ?nfs = "NFS d"
    let ?m = "M d"
    note ok = ok[unfolded usable_rules_proc_def dpp_spec_sound Let_def
      is_ur_closed_impl_dpp_mv[OF I], simplified] 
    from ok have NF: "NF_terms ?Q \<subseteq> NF_trs ?Rb" and NF': "NF_terms ?Q \<subseteq> NF_trs ?Rb'" by auto
    from ok have d': "d' = intersect_rules d U" by auto
    have d': "dpp d' = (?nfs, ?m, ?P, ?Pw, ?Q, ?R \<inter> ?U, ?Rw \<inter> ?U)" 
      unfolding d' unfolding intersect_rules_sound ..
    note fin = fin[unfolded this]
    interpret R_Q_U_ecap ?Rb' ?U ?Q icap'
      by (unfold_locales, rule icap, rule NF')
    have Rb: "?Rb = ?Rb'" by simp
    show "finite_dpp (dpp d)" unfolding dpp_spec_sound
    proof (rule usable_rules_proc[OF fin _ _ _ _])
      show "?Rb \<subseteq> ?Rb'" by simp
    next
      fix l r
      assume "(l,r) \<in> ?P \<union> ?Pw" 
      with ok 
      show "vars_term r \<subseteq> vars_term l" by blast
    next
      fix l r
      assume "(l,r) \<in> ?P \<union> ?Pw" 
      with ok
      show "is_ur_closed_term_mv full_af {l} r" by auto
    next
      fix l r
      assume "(l,r) \<in> ?U"
      with ok
      show "is_ur_closed_term_mv full_af (set (args l)) r" by auto    
    qed (insert ok, auto)
  qed
qed 


fun ur_term :: "('f,string)trs \<Rightarrow> ('f,string)terms \<Rightarrow> ('f,string)cap_fun \<Rightarrow> 'f af \<Rightarrow> ('f,string)terms \<Rightarrow> ('f,string)term \<Rightarrow> ('f,string)trs"
  where "ur_term R Q ecap \<pi> S (Var x) = {}"
     |  "ur_term R Q ecap \<pi> S (Fun f ts) = (let rec = map (ur_term R Q ecap \<pi> S) ts in 
             \<Union> {rec ! i | i. i < length ts \<and> \<pi> (f,length ts) i} \<union> { (l,r) | l r \<mu>. (l,r) \<in> R \<and> mgu_class (Fun f (map (ecap R Q S) ts)) l = Some \<mu> \<and> (S \<union> mv_yvar ` (set (args l))) \<cdot>set \<mu> \<subseteq> NF_terms Q })"

lemma ur_term_subset: "ur_term R Q ecap \<pi> S t \<subseteq> R"
proof (induct t)
  case (Fun f ts)
  {
    fix i
    assume "i < length ts"
    hence "ts ! i \<in> set ts" by auto
    from Fun[OF this] have "ur_term R Q ecap \<pi> S (ts ! i) \<subseteq> R" .
  }
  thus ?case by force
qed auto

lemma is_ur_closed'_ur_term: assumes U: "ur_term R Q ecap \<pi> S t \<subseteq> U"
  shows "is_ur_closed_term' R U Q ecap \<pi> S t"
  using U 
proof (induct t)
  case (Var x) thus ?case by auto
next
  case (Fun f ts)
  let ?P = "\<lambda> l \<mu>. mgu_class (Fun f (map (ecap R Q S) ts)) l = Some \<mu> \<and>
    S \<union> mv_yvar ` (set (args l)) \<cdot>set \<mu> \<subseteq> NF_terms Q"
  {
    fix l r
    assume "(l,r) \<in> R"
    with Fun(2)[simplified]
    have "(l,r) \<in> U \<or> \<not> (\<exists> \<mu>. ?P l \<mu>)" by blast
  } note two = this
  show ?case unfolding is_ur_closed_term'.simps
  proof (intro conjI ballI allI impI)
    fix i
    assume i: "i < length ts"
    and \<pi>: "\<pi> (f,length ts) i"
    from i have mem: "ts ! i \<in> set ts" by auto
    from Fun(2)[simplified] i \<pi> have sub: "map (ur_term R Q ecap \<pi> S) ts ! i \<subseteq> U"
      by blast
    show "is_ur_closed_term' R U Q ecap \<pi> S (ts ! i)"
      by (rule Fun(1)[OF mem], insert sub i, simp)
  qed (insert two, auto)
qed

lemma ur_term_is_ur_closed': assumes U: "is_ur_closed_term' R U Q ecap \<pi> S t"
  shows "ur_term R Q ecap \<pi> S t \<subseteq> U"
  using U
proof (induct t)
  case (Var x) thus ?case by auto
next
  case (Fun f ts)
  let ?P = "\<lambda> l \<mu>. mgu_class (Fun f (map (ecap R Q S) ts)) l = Some \<mu> \<and>
    S \<union> mv_yvar ` (set (args l)) \<cdot>set \<mu> \<subseteq> NF_terms Q"
  {
    fix l r
    assume "(l,r) \<in> \<Union> {map (ur_term R Q ecap \<pi> S) ts ! i | i. i < length ts \<and> \<pi> (f,length ts) i}"
    then obtain i where i: "i < length ts" and \<pi>: "\<pi> (f,length ts) i" and lr: "(l,r) \<in> map (ur_term R Q ecap \<pi> S) ts ! i" by auto
    from lr i have lr: "(l,r) \<in> ur_term R Q ecap \<pi> S (ts ! i)" by auto
    also have "... \<subseteq> U"
      by (rule Fun(1), insert \<pi> i Fun(2), auto)
    finally have "(l,r) \<in> U" .
  } note one = this
  let ?P = "\<lambda> l \<mu>. mgu_class (Fun f (map (ecap R Q S) ts)) l = Some \<mu> \<and>
    S \<union> mv_yvar ` (set (args l)) \<cdot>set \<mu> \<subseteq> NF_terms Q"
  {
    fix l r
    assume "(l,r) \<in> { u. \<exists> l r \<mu>. u = (l,r) \<and> (l,r) \<in> R \<and> ?P l \<mu>}"
    then obtain \<mu> where lr: "(l,r) \<in> R" and P: "?P l \<mu>" by auto
    with Fun(2)
    have "(l,r) \<in> U" by auto
  } note two = this
  show ?case using one two by auto
qed

lemma ur_term_least_is_ur_closed': "ur_term R Q ecap \<pi> S t = 
  (LEAST U. is_ur_closed_term' R U Q ecap \<pi> S t)"
  by (rule Least_equality[symmetric],
    rule is_ur_closed'_ur_term[OF subset_refl],
    rule ur_term_is_ur_closed')

fun ur_term_impl 
  where "ur_term_impl NFQ e_cap R \<pi> S (Var x) = []"
     |  "ur_term_impl NFQ e_cap R \<pi> S (Fun f ts) = (let n = length ts; rec = map (ur_term_impl NFQ e_cap R \<pi> S) ts in 
             remdups (concat (map (\<lambda> (i,urs). if \<pi> (f,n) i then urs else []) (zip [0 ..< n] rec)) @ filter (\<lambda> (l,r). 
  case mgu_class (Fun f (map e_cap ts)) l of
                 None \<Rightarrow> False | Some \<mu> \<Rightarrow> \<forall> u \<in> set (S @ map mv_yvar (args l)). NFQ (u \<cdot> \<mu>)
   ) R))"

lemma ur_term_impl: fixes R Q ecap
  assumes NFQ: "NFQ = (\<lambda> t. t \<in> NF_terms Q)"
  and e_cap: "e_cap = ecap (set R) Q (set S)"
  shows "set (ur_term_impl NFQ e_cap R \<pi> S t) = ur_term (set R) Q ecap \<pi> (set S) t"
proof (induct t)
  case (Var x)
  show ?case by simp
next
  case (Fun f ts)
  let ?impl = "ur_term_impl NFQ e_cap R \<pi> S"
  let ?sem = "ur_term (set R) Q ecap \<pi> (set S)"
  let ?base = "set (filter (\<lambda> (l,r). 
  case mgu_class (Fun f (map e_cap ts)) l of
                 None \<Rightarrow> False | Some \<mu> \<Rightarrow> \<forall> u \<in> set (S @ map mv_yvar (args l)). NFQ (u \<cdot> \<mu>)) R)"
  let ?base2 = " {(l,r). (l,r) \<in> set R \<and> (\<exists> \<mu>.
        mgu_class (Fun f (map (ecap (set R) Q (set S)) ts)) l = Some \<mu> \<and>
        set S \<union> mv_yvar ` set (args l) \<cdot>set \<mu> \<subseteq> NF_terms Q)}"
  have "set (?impl (Fun f ts)) = 
          {x. (\<exists> i. i < length ts \<and> \<pi> (f, length ts) i \<and> x \<in> set (?impl (ts ! i)))}
         \<union> ?base" (is "_ = ?rec \<union> _")
    by (auto simp: Let_def set_zip)
  also have "?rec = {x. (\<exists> i. i < length ts \<and> \<pi> (f, length ts) i \<and> x \<in> ?sem (ts ! i))}" 
    (is "_ = ?rec2")
    using Fun[unfolded set_conv_nth[of ts]] by blast
  also have "?base = ?base2" unfolding NFQ e_cap[symmetric]
    by (force split: option.splits)
  finally have id: "set (?impl (Fun f ts)) = ?rec2 \<union> ?base2" .
  show "set (?impl (Fun f ts)) = ?sem (Fun f ts)"
    unfolding id ur_term.simps Let_def by auto
qed

definition usable_rules_step_mv
  where "usable_rules_step_mv R Q ecap \<pi> \<equiv> { ((l,r),ur) | l r ur. 
     (l,r) \<in> R \<and> ur \<in> ur_term R Q ecap \<pi> (mv_xvar ` set (args l)) (mv_xvar r)}"

definition usable_rules_pair_mv 
  where "usable_rules_pair_mv R Q ecap \<pi> \<equiv> \<lambda> (s,t). (usable_rules_step_mv R Q ecap \<pi>)^* `` ur_term R Q ecap \<pi> {mv_xvar s} (mv_xvar t)"

lemma usable_rules_pair_mv_ur_closed_t: "
   is_ur_closed_term' R (usable_rules_pair_mv R Q ecap \<pi> (s,t)) Q ecap \<pi> (mv_xvar ` {s}) (mv_xvar t)"
  by (rule is_ur_closed'_ur_term,
  unfold usable_rules_pair_mv_def split, auto)

lemma usable_rules_pair_mv_ur_closed_U: assumes lr: "(l,r) \<in> usable_rules_pair_mv R Q ecap \<pi> (s,t)"
   shows "is_ur_closed_term_mv' R (usable_rules_pair_mv R Q ecap \<pi> (s,t)) Q ecap \<pi> (set (args l)) r"
proof (rule is_ur_closed'_ur_term, unfold usable_rules_pair_mv_def split)
  let ?U = "usable_rules_step_mv R Q ecap \<pi>"
  let ?U' = "ur_term R Q ecap \<pi>"
  let ?Ulr = "\<lambda> l r. ?U' (mv_xvar ` set (args l)) (mv_xvar r)"
  let ?Ust = "?U' {mv_xvar s} (mv_xvar t)"
  obtain lr where lr_id: "lr = (l,r)" by auto
  from lr[unfolded usable_rules_pair_mv_def split]
  obtain ur where ur: "ur \<in> ?Ust" and lr: "(ur,lr) \<in> ?U^*" 
    unfolding lr_id by auto
  from lr ur have "?Ulr (fst lr) (snd lr) \<subseteq> ?U^* `` ?Ust"
  proof (induct rule: rtrancl_induct)
    case base
    with ur_term_subset have urR: "ur \<in> R" by blast
    obtain ul urr where ur: "ur = (ul, urr)" by force
    show ?case unfolding ur fst_conv snd_conv
    proof
      fix l r
      assume "(l,r) \<in> ?Ulr ul urr"
      hence "((ul,urr),(l,r)) \<in> ?U" using urR[unfolded ur] unfolding usable_rules_step_mv_def by auto
      with base[unfolded ur] 
      show "(l,r) \<in> ?U^* `` ?Ust" by auto
    qed
  next
    case (step lr vw)
    obtain l r where lr: "lr = (l,r)" by force
    obtain v w where vw: "vw = (v,w)" by force
    from step(3)[OF step(4)]
    have Ulr: "?Ulr l r \<subseteq> ?U^* `` ?Ust" unfolding lr by simp
    from step(2) have lrvw: "((l,r),(v,w)) \<in> ?U" unfolding lr vw .
    from this[unfolded usable_rules_step_mv_def] have vwlr: "(v,w) \<in> ?Ulr l r" by auto
    from vwlr ur_term_subset have vwR: "(v,w) \<in> R" by blast
    obtain UU VV where id: "?U = UU" "?Ust = VV" by auto
    have "?Ulr v w  \<subseteq> ?U^* `` ?Ust" 
    proof
      fix a b
      assume ab: "(a,b) \<in> ?Ulr v w"
      with vwR have "((v,w),(a,b)) \<in> ?U" unfolding usable_rules_step_mv_def by auto
      with vwlr have "(a,b) \<in> ?U `` ?Ulr l r" by auto
      with Ulr have "(a,b) \<in> ?U `` (?U^* `` ?Ust)" by auto
      also have "... \<subseteq> ?U^* `` ?Ust" unfolding id 
      proof(rule subsetI)
        fix x
        assume "x \<in> UU `` (UU^* `` VV)"
        then obtain y v where v: "v \<in> VV" and vx: "(v,y) \<in> UU^*" "(y,x) \<in> UU" by auto
        from vx have "(v,x) \<in> UU^*" by auto
        with v show "x \<in> UU^* `` VV" by auto
      qed
      finally show "(a,b) \<in> ?U^* `` ?Ust" .
    qed      
    thus ?case unfolding vw by simp
  qed
  thus "?Ulr l r  \<subseteq> ?U^* `` ?Ust" unfolding lr_id by simp
qed

definition usable_rules_calculator where
  "usable_rules_calculator NFQ e_cap R \<pi> \<equiv> 
    let
      urt = (\<lambda> S. let S' = map mv_xvar S in ur_term_impl NFQ (e_cap S') R \<pi> S');
      urules = map (\<lambda> (l, r). ((l, r), urt (args l) (mv_xvar r))) R;
      ufun = fun_of_map (ceta_map_of urules) [];
      rtran = mk_rtrancl_list (op =) ufun
    in (\<lambda>(s, t). rtran (urt [s] (mv_xvar t)))"

lemma usable_rules_calculator:
  fixes R Q ecap
  assumes NFQ: "NFQ = (\<lambda> t. t \<in> NF_terms Q)"
    and e_cap: "\<And> S. e_cap S = ecap (set R) Q (set S)"
  shows "set (usable_rules_calculator NFQ e_cap R \<pi> (s,t)) = usable_rules_pair_mv (set R) Q ecap \<pi> (s,t)"
proof -
  let ?urt = "\<lambda> S. ur_term_impl NFQ (e_cap (map mv_xvar S)) R \<pi> (map mv_xvar S)"
  let ?urules = "map (\<lambda> (l,r). ((l,r), ?urt (args l) (mv_xvar r))) R"
  let ?ufun = "fun_of_map (map_of ?urules) []"
  obtain ufun where ufun: "ufun = ?ufun" by auto
  let ?rtran = "mk_rtrancl_list (op =) ufun"
  from ur_term_impl[OF NFQ, of _ ecap, OF e_cap]
  have urt: "\<And> S t. set (?urt S t) = ur_term (set R) Q ecap \<pi> (mv_xvar ` set S) t" by auto
  {
    fix lr    
    have ufun: "set (?ufun lr) = {ur. (lr,ur) \<in> usable_rules_step_mv (set R) Q ecap \<pi>}"
    proof (cases "lr \<in> set R")
      case False
      with map_of_SomeD[of ?urules lr] have "?ufun lr = []" 
        by (cases "map_of ?urules lr", force+)
      hence "set (?ufun lr) = {}" by simp
      thus ?thesis using False unfolding usable_rules_step_mv_def by auto
    next
      case True
      obtain l r where lr: "lr = (l,r)" by force
      from True and map_of_eq_None_iff[of ?urules lr] and lr
        obtain urs where look: "map_of ?urules (l, r) = Some urs"
        by (cases "map_of ?urules lr", force+)
      hence id: "set (?ufun lr) = set urs" using lr by simp
      from map_of_SomeD[OF look] have urs: "urs = ?urt (args l) (mv_xvar r)" by auto
      show ?thesis unfolding id urs urt using True unfolding lr usable_rules_step_mv_def by auto
    qed
  } 
  hence ufun_id: "\<And> lr. set (ufun lr) = {ur. (lr,ur) \<in> usable_rules_step_mv (set R) Q ecap \<pi>}" unfolding ufun .
  have id: "set (usable_rules_calculator NFQ e_cap R \<pi> (s,t)) = set (?rtran (?urt [s] (mv_xvar t)))"
    unfolding usable_rules_calculator_def Let_def split ceta_map_of ufun[symmetric] by auto
  interpret relation_subsumption_list ufun "op ="
  proof
    fix a
    show "finite {b. (a, b) \<in> {(a, b). b \<in> set (ufun a)}\<^sup>*}" (is "finite ?R")
    proof (rule finite_subset[OF _ finite_set[of "a # R"]])
      show "?R \<subseteq> set (a # R)" 
      proof (rule subsetI)
        fix b
        assume "b \<in> ?R"
        hence "(a,b) \<in> {(a,b). b \<in> set (ufun a)}^*" by auto
        thus "b \<in> set (a # R)"
        proof(induct)
          case (step b c)
          from step(2) have "c \<in> set (ufun b)" by simp
          thus ?case
            unfolding ufun_id usable_rules_step_mv_def 
            using ur_term_subset[of "set R" Q ecap \<pi>] by auto
        qed auto
      qed
    qed
  qed auto
  note id = id[unfolded mk_rtrancl_list mk_rtrancl_no_subsumption[OF refl],
  unfolded ufun_id]
  show ?thesis unfolding id unfolding usable_rules_pair_mv_def split
    unfolding urt by auto
qed


definition get_ur_calc_impl_dpp 
  where "get_ur_calc_impl_dpp I d \<pi> \<equiv>
  let ic = icap_impl_dpp I d;
      qnf = dpp_ops.is_QNF I d;
      r = dpp_ops.rules I d
    in usable_rules_calculator qnf ic r \<pi>"

lemma get_ur_calc_impl_dpp:
  fixes I::"('d, 'f::{show,key}, string) dpp_ops" and d::"'d" and \<pi> s t
  defines q: "q \<equiv> dpp_ops.Q I d"
      and R: "R \<equiv> dpp_ops.rules I d"
  assumes I: "dpp_spec I"
      and NF: "NF_terms (set q) \<subseteq> NF_trs (set R)"
      and U: "U = get_ur_calc_impl_dpp I d \<pi> (s,t)"
  shows "\<exists> ecap. R_Q_U_ecap (set R) (set q) ecap \<and>
           (\<forall> l r. (l,r) \<in> set U \<longrightarrow> is_ur_closed_term' (set R) (set U) (set q) ecap \<pi> (mv_xvar ` set (args l)) (mv_xvar r)) \<and>
           is_ur_closed_term' (set R) (set U) (set q) ecap \<pi> (mv_xvar ` {s}) (mv_xvar t)"proof -
  interpret dpp_spec I by fact
  let ?R = "set (rules d)"
  let ?Q = "set (Q d)"
  show ?thesis unfolding R q
  proof (intro exI, rule conjI)
    show "R_Q_U_ecap ?R ?Q icap'"
      by (unfold_locales, rule icap, insert NF R q, auto)
  next
    have icap: "icap_impl_dpp I d = (\<lambda> s. icap ?R ?Q (set s))" unfolding R q
      unfolding icap_impl_dpp_icap[OF I, symmetric] ..
    note U = U[unfolded get_ur_calc_impl_dpp_def Let_def, unfolded is_QNF_sound icap]
    hence U: "set U = set (usable_rules_calculator (\<lambda>t. t \<in> NF_terms ?Q)
      (\<lambda>s. icap ?R ?Q (set s)) (rules d) \<pi> (s, t))" by simp
    also have "... = usable_rules_pair_mv ?R ?Q icap \<pi> (s,t)"
      by (rule usable_rules_calculator, auto)
    finally have U: "set U = usable_rules_pair_mv ?R ?Q icap \<pi> (s,t)" .
    note sU = U[symmetric]
    note closed_U = usable_rules_pair_mv_ur_closed_U[of _ _ ?R ?Q icap \<pi> s t, unfolded sU]
    note closed_t = usable_rules_pair_mv_ur_closed_t[of ?R ?Q icap \<pi> s t, unfolded sU] 
    note id = is_ur_closed_term_af_impl_mv[OF refl refl, unfolded is_ur_closed_term_af_impl_mv_def Let_def]
    let ?one = "(\<forall> l r. (l,r) \<in> set U \<longrightarrow> is_ur_closed_term' ?R (set U) ?Q icap' \<pi> (mv_xvar ` set (args l)) (mv_xvar r))"
    let ?two = "is_ur_closed_term' ?R (set U) ?Q icap' \<pi> (mv_xvar ` {s}) (mv_xvar t)"
    from closed_t have two: ?two unfolding is_ur_closed_term_mv_icap' .
    from closed_U have one: ?one unfolding is_ur_closed_term_mv_icap' by auto
    from one two show "?one \<and> ?two" by simp
  qed
qed

lemma get_ur_calc_impl_dpp_steps:
  fixes I::"('d, 'f::{show,key}, string) dpp_ops" and d::"'d"
  defines q: "q \<equiv> dpp_ops.Q I d"
      and R: "R \<equiv> dpp_ops.rules I d"
      and nfs: "nfs \<equiv> dpp_ops.nfs I d"
      and m: "m \<equiv> dpp_ops.minimal I d"
  assumes I: "dpp_spec I"
      and U: "U = get_ur_calc_impl_dpp I d full_af (s,t)"
      and NF: "NF_terms (set q) \<subseteq> NF_trs (set R)"
      and NFs: "s \<cdot> \<sigma> \<in> NF_terms (set q)"
      and nfsigma: "NF_subst nfs (s,t) \<sigma> (set q)"
      and varsR: "nfs \<or> SN_on (qrstep nfs (set q) (set R)) {t \<cdot> \<sigma>} \<or> wwf_qtrs (set q) (set R)"
      and vars: "\<not> nfs \<Longrightarrow> vars_term t \<subseteq> vars_term s"
      and steps: "(t \<cdot> \<sigma> ,u) \<in> (qrstep nfs (set q) (set R))^*"
  shows "(t \<cdot> \<sigma>, u) \<in> (qrstep nfs (set q) (set U))^*"
proof -
  interpret dpp_spec I by fact
  let ?R = "set (rules d)"
  let ?Q = "set (Q d)"
  let ?nfs = "NFS d"
  from get_ur_calc_impl_dpp[OF I NF[unfolded R q] U]
  obtain ecap where ecap: "R_Q_U_ecap ?R ?Q ecap" 
    and U: "\<And> l r. (l,r) \<in> set U \<Longrightarrow> is_ur_closed_term' ?R (set U) ?Q ecap full_af (mv_xvar ` set (args l)) (mv_xvar r)"
    and t: "is_ur_closed_term' ?R (set U) ?Q ecap full_af (mv_xvar ` {s}) (mv_xvar (t))" 
    by auto
  from NFs have NF: "{s} \<cdot>set \<sigma> \<subseteq> NF_terms ?Q" unfolding q by auto
  have st: "(t \<cdot> \<sigma>, u) \<in> (qrstep ?nfs ?Q ?R)^*" 
    using steps unfolding R q nfs .
  have NFt: "\<sigma> ` vars_term t \<subseteq> NF_terms ?Q"
  proof (cases nfs)
    case True
    with nfsigma show ?thesis unfolding NF_subst_def vars_rule_def q by auto
  next
    case False
    show ?thesis
    proof (rule subsetI)
      fix u
      assume "u \<in> \<sigma> ` vars_term t"
      then obtain x where u: "u = \<sigma> x" and "x \<in> vars_term t" by auto
      with vars[OF False] have "x \<in> vars_term s" by auto
      from supteq_stable[OF supteq_Var[OF this], of \<sigma>] u have "s \<cdot> \<sigma> \<unrhd> u" by auto
      from NF_subterm[OF _ this] NF
      show "u \<in> NF_terms ?Q" unfolding q R by auto
    qed
  qed
  have st: "(t \<cdot> \<sigma>, u) \<in> (qrstep ?nfs ?Q (?R \<inter> set U))^*"
    by (rule R_Q_U_ecap.is_ur_closed_term[OF ecap st NFt NF U subset_refl _ t], insert varsR nfs q R, auto)
  have "qrstep ?nfs ?Q (?R \<inter> set U) \<subseteq> qrstep ?nfs ?Q (set U)" by auto
  from st rtrancl_mono[OF this]
  show ?thesis unfolding nfs q by auto
qed

definition inn_usable_rules_pair :: "('d, 'f :: key, string) dpp_ops
      \<Rightarrow> 'd \<Rightarrow> ('f, string)rule \<Rightarrow> ('f,string)rules"
  where "inn_usable_rules_pair I d \<equiv> 
  let inn = dpp_ops.NFQ_subset_NF_rules I d;
      R = dpp_ops.rules I d;
      qnf = dpp_ops.is_QNF I d;
      calc = get_ur_calc_impl_dpp I d full_af;
      nfs = dpp_ops.nfs I d;
      wwf = dpp_ops.wwf_rules I d;
      m   = dpp_ops.minimal I d
   in (\<lambda> (s,t). if inn \<and> (nfs \<or> (set (vars_term_list t) \<subseteq> set (vars_term_list s))) \<and> (nfs \<or> m \<or> wwf) then
          calc (s,t) else R)"

lemma inn_usable_rules_pair:
  fixes I::"('d, 'f::{show,key}, string) dpp_ops" and d::"'d" 
  defines q: "q \<equiv> dpp_ops.Q I d"
      and R: "R \<equiv> dpp_ops.rules I d"
      and nfs: "nfs \<equiv> dpp_ops.nfs I d"
  assumes I: "dpp_spec I"
      and U: "U = inn_usable_rules_pair I d (s,t)"
      and steps: "(t \<cdot> \<sigma>, u) \<in> (qrstep nfs (set q) (set R))^*"
      and NF: "s \<cdot> \<sigma> \<in> NF_terms (set q)"      
      and nfsigma: "NF_subst nfs (s,t) \<sigma> (set q)"
      and SN: "dpp_ops.minimal I d \<Longrightarrow> SN_on (qrstep nfs (set q) (set R)) {t \<cdot> \<sigma>}"
  shows "(t \<cdot> \<sigma>, u) \<in> (qrstep nfs (set q) (set U))^*"
proof -
  interpret dpp_spec I by fact
  let ?R = "set (rules d)"
  let ?Q = "set (Q d)"
  let ?nfs = "NFS d"
  let ?m = "M d"
  let ?wwf = "wwf_rules d"
  let ?inn = "dpp_ops.NFQ_subset_NF_rules I d"
  show ?thesis 
  proof (cases "?inn \<and> (nfs \<or> set (vars_term_list t) \<subseteq> set (vars_term_list s)) \<and> (nfs \<or> ?m \<or> ?wwf)")
    case False
    hence "U = rules d" unfolding U inn_usable_rules_pair_def Let_def nfs split by auto
    thus ?thesis using steps unfolding R by auto
  next
    case True
    hence U: "U = get_ur_calc_impl_dpp I d full_af (s,t)" unfolding U inn_usable_rules_pair_def Let_def split nfs by auto 
    from True have inn: "NF_terms ?Q \<subseteq> NF_trs ?R" 
      and vars: "\<not> nfs \<Longrightarrow> vars_term t \<subseteq> vars_term s" by auto
    show ?thesis unfolding q nfs
      by (rule get_ur_calc_impl_dpp_steps[OF I U inn _ _ _ vars], insert nfsigma steps SN NF q R True nfs, auto)
  qed
qed

end
