(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Linear_Poly_Order
imports
  Term_Order_Extension
  Term_Impl
  Linear_Polynomial
  Name
  Map_Choice
begin

text {* we first head for polynomial interpretations *}

type_synonym ('f, 'a) lpoly_inter = "'f \<times> nat \<Rightarrow> ('a \<times> 'a list)"

fun
  eval_termI :: "'a \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> ('a \<Rightarrow> 'a) \<Rightarrow> ('f, 'a) lpoly_inter \<Rightarrow> ('v, 'a) p_ass \<Rightarrow> ('f, 'v) term \<Rightarrow> 'a" 
where
  "eval_termI _ _ _ _ pi \<alpha> (Var x) = \<alpha> x" 
| "eval_termI ze pl ti ma pi \<alpha> (Fun f ts) = (let (a,as) = pi (f, length ts) in ma (pl a  (list_sumI ze pl (map (\<lambda> at. ti (fst at) (eval_termI ze pl ti ma pi \<alpha> (snd at))) (zip as ts)))))"

context linear_poly_order_carrier
begin
declare poly_simps[simp] poly_order_simps[simp]

abbreviation eval_term :: "('f,'a)lpoly_inter \<Rightarrow> ('v,'a)p_ass \<Rightarrow> ('f,'v)term \<Rightarrow> 'a" ("_\<guillemotleft>_,_\<guillemotright>" 50)
where "eval_term \<equiv> eval_termI zero plus times max0"

abbreviation wf_lpoly_inter :: "('f,'a)lpoly_inter \<Rightarrow> bool"
where "wf_lpoly_inter pi \<equiv> (\<forall> fn. wf_elem (fst (pi fn)) \<and> (\<forall> a. a \<in> set (snd (pi fn)) \<longrightarrow> wf_elem a \<and> geq a zero))
                         \<and> (\<forall> fn. arc_pos (fst (pi fn)) \<or> (\<exists> a \<in> set (take (snd fn) (snd (pi fn))) . arc_pos a))"
end

locale linear_poly_order = linear_poly_order_carrier +
  fixes pi :: "('f::{show,key},'a)lpoly_inter"
  assumes wf_pi: "wf_lpoly_inter pi"
begin

lemma wf_terms[simp]: assumes wf_ass: "wf_ass \<alpha>"
  shows "wf_elem (pi\<guillemotleft>\<alpha>,t\<guillemotright>)"
proof (induct t)
  case (Var x)
  from wf_ass have "wf_elem (\<alpha> x)" unfolding wf_ass_def by auto
  thus ?case by auto
next
  case (Fun f ts)
  show ?case 
  proof (cases "pi (f,length ts)")
    case (Pair a as)
    have "wf_elem (list_sum (map (\<lambda>at. times (fst at) (pi\<guillemotleft>\<alpha>,snd at\<guillemotright>)) (zip as ts)))" (is "wf_elem ?ls")
    proof (rule wf_list_sum, rule ballI)
      fix p
      assume "p \<in> set (map (\<lambda>at. fst at * (pi\<guillemotleft>\<alpha>,snd at\<guillemotright>)) (zip as ts))" (is "p \<in> set (map _ ?zip)")
      from this obtain a' t where p: "p = a' * (pi\<guillemotleft>\<alpha>,t\<guillemotright>)" and ab: "(a',t) \<in> set ?zip" by auto
      from ab have a': "a' \<in> set as" and t: "t \<in> set ts" using set_zip_leftD[where x = a' and y = t] set_zip_rightD[where y = t and x = a'] by auto
      from a' Pair have "a' \<in> set (snd (pi (f,length ts)))" by auto
      with Fun wf_pi have "wf_elem a'" by auto
      with Fun t p show "wf_elem p" by auto
    qed
    from wf_pi have "wf_elem (fst (pi (f,length ts)))" by auto
    with Pair `wf_elem ?ls` have "wf_elem (a + ?ls)" by auto
    thus ?thesis using Pair by (auto simp: wf_max0 max0_pos)
  qed
qed

lemma pos_term[simp]: assumes wf_ass: "wf_ass \<alpha>"
  and pos_ass: "pos_ass \<alpha>"
  shows "geq (pi\<guillemotleft>\<alpha>,t\<guillemotright>) zero"
proof (induct t)
  case (Var x)
  from wf_ass pos_ass have "wf_elem (\<alpha> x) \<and> geq (\<alpha> x) zero" unfolding wf_ass_def pos_ass_def by auto
  thus ?case by auto
next
  case (Fun f ts)
  show ?case 
  proof (cases "pi (f,length ts)")
    case (Pair a as)
    have "wf_elem (list_sum (map (\<lambda>at. times (fst at) (pi\<guillemotleft>\<alpha>,snd at\<guillemotright>)) (zip as ts)))" (is "wf_elem ?ls")
    proof (rule wf_list_sum, rule ballI)
      fix p
      assume "p \<in> set (map (\<lambda>at. fst at * (pi\<guillemotleft>\<alpha>,snd at\<guillemotright>)) (zip as ts))" (is "p \<in> set (map _ ?zip)")
      from this obtain a' t where p: "p = a' * (pi\<guillemotleft>\<alpha>,t\<guillemotright>)" and ab: "(a',t) \<in> set ?zip" by auto
      from ab have a': "a' \<in> set as" and t: "t \<in> set ts" using set_zip_leftD[where x = a' and y = t] set_zip_rightD[where y = t and x = a'] by auto
      from a' Pair have "a' \<in> set (snd (pi (f,length ts)))" by auto
      with Fun wf_pi have "wf_elem a'" by auto
      with Fun t p wf_ass wf_pi show "wf_elem p" by auto
    qed
    from wf_pi have "wf_elem (fst (pi (f,length ts)))" by auto
    with Pair `wf_elem ?ls` have "wf_elem (a + ?ls)" by auto
    thus ?thesis using Pair by (auto simp: wf_max0 max0_pos)
  qed
qed

definition apos_ass :: "('v,'a)p_ass \<Rightarrow> bool" where "apos_ass \<alpha> \<equiv> \<forall> x. arc_pos (\<alpha> x)"

lemma apos_term[simp]: assumes wf_ass: "wf_ass \<alpha>"
  and apos_ass: "apos_ass \<alpha>"
  shows "arc_pos (pi\<guillemotleft>\<alpha>,t\<guillemotright>)"
proof (induct t)
  case (Var x)
  from apos_ass show ?case unfolding apos_ass_def by auto
next
  case (Fun f ts)
  show ?case 
  proof (cases "pi (f,length ts)")
    case (Pair a as)
    have wf_args: "\<forall>a\<in>set (map (\<lambda>at. fst at * (pi\<guillemotleft>\<alpha>,snd at\<guillemotright>)) (zip as ts)). wf_elem a"
    proof (rule ballI)
      fix p
      assume "p \<in> set (map (\<lambda>at. fst at * (pi\<guillemotleft>\<alpha>,snd at\<guillemotright>)) (zip as ts))" (is "p \<in> set (map _ ?zip)")
      from this obtain a' t where p: "p = a' * (pi\<guillemotleft>\<alpha>,t\<guillemotright>)" and ab: "(a',t) \<in> set ?zip" by auto
      from ab have a': "a' \<in> set as" and t: "t \<in> set ts" using set_zip_leftD[where x = a' and y = t] set_zip_rightD[where y = t and x = a'] by auto
      from a' Pair have "a' \<in> set (snd (pi (f,length ts)))" by auto
      with Fun wf_pi have "wf_elem a'" by auto
      with Fun t p wf_ass wf_pi show "wf_elem p" by auto
    qed
    have "wf_elem (list_sum (map (\<lambda>at. times (fst at) (pi\<guillemotleft>\<alpha>,snd at\<guillemotright>)) (zip as ts)))" (is "wf_elem ?ls")
      by (rule wf_list_sum, rule wf_args)
    from wf_pi have "wf_elem (fst (pi (f,length ts)))" by auto
    with Pair have wf_a: "wf_elem a" by auto
    with wf_pi have "arc_pos (fst (pi (f,length ts))) \<or> (\<exists> a \<in> set (take (length ts) (snd (pi (f,length ts)))). arc_pos a)" (is "?ac \<or> ?aa") by auto
    thus ?thesis 
    proof
      assume "?ac"
      hence "arc_pos (a + ?ls)" and "wf_elem (a + ?ls)" using Pair `wf_elem ?ls` wf_a arc_pos_plus by auto
      thus ?thesis using Pair by auto
    next
      assume ?aa
      from this obtain c where c: "c \<in> set (take (length ts) as) \<and> arc_pos c" using Pair by auto
      from wf_pi have "(\<forall> b \<in> set (snd (pi (f,length ts))). wf_elem b)" by auto
      with Pair have wf_as: "\<forall> b \<in> set as. wf_elem b" by auto
      from Pair Fun have rec: "\<forall> t \<in> set ts. arc_pos (pi\<guillemotleft>\<alpha>,t\<guillemotright>)" by auto
      from c wf_as rec wf_args have aa: "arc_pos ?ls"
      proof (induct ts arbitrary: as)
	case (Cons t ts) note oCons = this
	show ?case
	proof (cases as)
	  case Nil
	  with Cons show ?thesis by auto
	next
	  case (Cons b bs)
	  let ?prod = "b * (pi\<guillemotleft>\<alpha>,t\<guillemotright>)"
	  have wf: "wf_elem (pi\<guillemotleft>\<alpha>,t\<guillemotright>) \<and> wf_elem b \<and> wf_elem ?prod" using wf_ass Cons oCons by auto
	  hence "wf_elem ?prod" by auto
	  show ?thesis 
	  proof (cases "b = c")
	    case True
	    show ?thesis  by (simp add: Cons, rule arc_pos_plus, rule arc_pos_mult, simp add: True c, auto simp: Cons oCons wf)
	  next
	    case False
	    with Cons oCons have "c \<in> set (take (length ts) bs)" by auto
	    with Cons oCons have ap: "arc_pos (list_sum (map (\<lambda> at. fst at * (pi\<guillemotleft>\<alpha>,snd at\<guillemotright>)) (zip bs ts)))" (is "arc_pos ?ls") by auto
	    with Cons oCons have "wf_elem ?ls" by auto
	    show ?thesis by (simp add: Cons, simp only: plus_comm[OF `wf_elem ?prod` `wf_elem ?ls`], rule arc_pos_plus[OF ap], auto simp: Cons oCons wf)
	  qed
	qed 
      qed simp
      have "a + ?ls = ?ls + a" by (rule plus_comm[OF wf_a `wf_elem ?ls`])
      hence "arc_pos (a + ?ls)" and "wf_elem (a + ?ls)" using aa `wf_elem ?ls` wf_a arc_pos_plus  by auto
      thus ?thesis using Pair by auto
    qed
  qed
qed

definition inter_s :: "('f,'v)trs"
  where "inter_s \<equiv> {(s,t). (\<forall> \<alpha> . wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> gt (pi\<guillemotleft>\<alpha>,s\<guillemotright>) (pi\<guillemotleft>\<alpha>,t\<guillemotright>))}"

definition inter_ns :: "('f,'v)trs"
  where "inter_ns \<equiv> {(s,t). (\<forall> \<alpha> . wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> geq (pi\<guillemotleft>\<alpha>,s\<guillemotright>) (pi\<guillemotleft>\<alpha>,t\<guillemotright>))}"

definition default_ass :: "('v,'a)p_ass" where "default_ass x \<equiv> default"

lemma wf_pos_apos_default_ass[simp]: "wf_ass default_ass \<and> pos_ass default_ass \<and> apos_ass default_ass"
  unfolding wf_ass_def pos_ass_def apos_ass_def default_ass_def
  by auto

lemma interpr_subst: assumes wf_ass: "wf_ass \<alpha>"
  shows "(pi\<guillemotleft>\<alpha>, t \<cdot> \<sigma>\<guillemotright>) = (pi\<guillemotleft>(\<lambda> x. pi\<guillemotleft>\<alpha>, \<sigma> x\<guillemotright>),t\<guillemotright>)" 
proof -
  have "(pi\<guillemotleft>\<alpha>, t \<cdot> \<sigma>\<guillemotright>) = (pi\<guillemotleft>(\<lambda> x. pi\<guillemotleft>\<alpha>, \<sigma> x\<guillemotright>),t\<guillemotright>)" (is "_ = (pi\<guillemotleft>?B, t\<guillemotright>)")
  proof (induct t)
    case (Var x)
    thus ?case by simp
  next
    case (Fun f ss)
    let ?ts = "map (\<lambda> s. s \<cdot> \<sigma>) ss"
    have wf_assB: "wf_ass ?B" unfolding wf_ass_def using wf_ass wf_pi by auto
    have map: "(map (eval_term pi \<alpha>) (?ts))
      = (map (eval_term pi ?B) ss)" using Fun by (induct ss, auto)
    let ?i = "pi (f, length ss)"
    show ?case 
    proof (cases "?i")
      case (Pair c cs)
      from wf_pi have "\<forall> d \<in> set (snd ?i). wf_elem d" by auto
      with Pair have wf_i: "\<forall> d \<in> set cs. wf_elem d" by auto
      from wf_i map have "map (\<lambda>at. fst at * (pi\<guillemotleft>\<alpha>,snd at\<guillemotright>)) (zip cs (map (\<lambda>t. t \<cdot> \<sigma>) ss)) = 
	map (\<lambda>at. fst at * (pi\<guillemotleft>\<lambda>x. pi\<guillemotleft>\<alpha>, \<sigma> x\<guillemotright>,snd at\<guillemotright>)) (zip cs ss)"
      proof (induct cs arbitrary: ss)
	case (Cons c cs) note oCons = this
	thus ?case 
	proof (cases ss, simp)
	  case (Cons s sss)
	  with oCons show ?thesis by auto
	qed
      qed simp
      with Pair show ?thesis by auto
    qed
  qed
  thus ?thesis by simp
qed

end

fun PleftI :: "'a \<Rightarrow> 'a \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> ('a \<Rightarrow> 'a) \<Rightarrow> ('f,'a)lpoly_inter \<Rightarrow> ('f,'v)term \<Rightarrow> ('v,'a)l_poly" 
where "PleftI ze on pl ti mz pi (Var x) = (LPoly ze [(x,on)] )" 
   |  "PleftI ze on pl ti mz pi (Fun f ts) = (let (c, as) = pi (f,length ts) in 
              (case (sum_lpolyI ze pl (LPoly c []) (list_sumI (LPoly ze []) (sum_lpolyI ze pl) (map (\<lambda> at . (mul_lpolyI ze ti (fst at)  (PleftI ze on pl ti mz pi (snd at)))) (zip as ts)))) :: ('v,'a)l_poly of
                  LPoly d [] \<Rightarrow> LPoly (mz d) [] | p \<Rightarrow> p))"

fun PrightI :: "'a \<Rightarrow> 'a \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> ('a \<Rightarrow> 'a) \<Rightarrow> ('f,'a)lpoly_inter \<Rightarrow> ('f,'v)term \<Rightarrow> ('v,'a)l_poly" 
where "PrightI ze on pl ti mz pi (Var x) = (LPoly ze [(x,on)] )" 
   |  "PrightI ze on pl ti mz pi (Fun f ts) = (let (c, as) = pi (f,length ts) in 
              (case (sum_lpolyI ze pl (LPoly c []) (list_sumI (LPoly ze []) (sum_lpolyI ze pl) (map (\<lambda> at . (mul_lpolyI ze ti (fst at)  (PrightI ze on pl ti mz pi (snd at)))) (zip as ts)))) :: ('v,'a)l_poly of
                  LPoly d vp \<Rightarrow> LPoly (mz d) vp))"

fun
  check_polo_nsI
where
  "check_polo_nsI ze on pl ti mz geq pi (s,t) = do {
     let left = PleftI ze on pl ti mz pi s;
     let right = PrightI ze on pl ti mz pi t;
     check_lpoly_nsI ze on geq left right 
      <+? (\<lambda>e. ''could not ensure '' +#+ shows s +@+ '' >= '' +#+ shows t +@+ shows_nl +@+ e)
   }"


context linear_poly_order
begin

abbreviation zero_lpoly :: "('v,'a)l_poly"
where "zero_lpoly \<equiv> LPoly 0 []"

abbreviation c_lpoly :: "'a \<Rightarrow> ('v,'a)l_poly"
where "c_lpoly c \<equiv> LPoly c []"

abbreviation list_sum_lpoly :: "('v,'a)l_poly list \<Rightarrow> ('v,'a)l_poly"
where "list_sum_lpoly \<equiv> list_sumI zero_lpoly sum_lpoly"

lemma wf_list_sum_lpoly: assumes "\<forall> p \<in> set ps. wf_lpoly p"
  shows "wf_lpoly (list_sum_lpoly ps)"
using assms
proof (induct ps)
  case Nil thus ?case by (simp add: wf_pvars_def)
next
  case (Cons p ps)
  show ?case by (simp only: list_sumI.simps, rule wf_sum_lpoly, auto simp: Cons)
qed

lemma pos_list_sum_lpoly: 
  assumes "\<forall> p \<in> set ps. wf_lpoly p"
    and  "\<forall> p \<in> set ps. pos_coeffs p"
  shows "pos_coeffs (list_sum_lpoly ps)"
using assms
proof (induct ps)
  case Nil thus ?case by (simp add: pos_coeffs_def pos_pvars_def)
next
  case (Cons p ps)
  show ?case by (simp only: list_sumI.simps, rule pos_sum_lpoly, auto simp: Cons, rule wf_list_sum_lpoly, auto simp: Cons)
qed

lemma list_sum_lpoly_sound:
  assumes wf_ps: "\<forall> p \<in> set ps. wf_lpoly p"
    and wf_ass: "wf_ass \<alpha>"
  shows "eval_lpoly \<alpha> (list_sum_lpoly ps) = list_sum (map (eval_lpoly \<alpha>) ps)"
using wf_ps
proof (induct ps)
  case (Cons p ps)
  have "eval_lpoly \<alpha> (list_sum_lpoly (p # ps)) = eval_lpoly \<alpha> (sum_lpoly p (list_sum_lpoly ps))" by auto
  also have "\<dots> = eval_lpoly \<alpha> p + eval_lpoly \<alpha> (list_sum_lpoly ps)" 
    by (rule sum_poly_sound, rule wf_ass, simp add: Cons, rule wf_list_sum_lpoly, auto simp: Cons)
  also have "\<dots> = eval_lpoly \<alpha> p + list_sum (map (eval_lpoly \<alpha>) ps)" using Cons by auto
  finally show ?case by (simp only: list_sumI.simps, auto simp: Cons wf_ass)
qed simp

abbreviation Pleft :: "('f,'v)term \<Rightarrow> ('v,'a)l_poly"
where "Pleft \<equiv> PleftI zero one plus times max0 pi"

abbreviation Pright :: "('f,'v)term \<Rightarrow> ('v,'a)l_poly"
where "Pright \<equiv> PrightI zero one plus times max0 pi"

lemma Pleft_both:  assumes wf_ass: "wf_ass \<alpha>"
  and pos_ass: "pos_ass \<alpha>"
  shows "((pi\<guillemotleft>\<alpha>,t\<guillemotright>) \<ge> eval_lpoly \<alpha> (Pleft t)) \<and> wf_lpoly (Pleft t)" 
proof (induct t)
  case (Var x)
  show ?case using pos_ass wf_ass unfolding pos_ass_def wf_ass_def by (auto simp: wf_pvars_def)
  case (Fun f ts)
  let ?I = "\<lambda> s. pi\<guillemotleft>\<alpha>,s\<guillemotright>"
  let ?E = "\<lambda> p. eval_lpoly \<alpha> p"
  show ?case 
  proof (cases "pi (f,length ts)")
    case (Pair c cs)
    from wf_pi have "wf_elem (fst (pi (f,length ts))) \<and> (\<forall> c \<in> set (snd (pi (f,length ts))). wf_elem c \<and> geq c zero)" by auto
    with Pair have wf_c: "wf_elem c" and wf_cs: "\<forall> c \<in> set cs. wf_elem c" and pos_cs: "\<forall> c \<in> set cs. c \<ge> 0" by auto
    let ?plist = "map (\<lambda> at. mul_lpoly (fst at) (Pleft (snd at))) (zip cs ts)"
    have wf_ps: "\<forall> p \<in> set ?plist. wf_lpoly p"
    proof
      fix p
      assume "p \<in> set ?plist"
      from this obtain a t where p: "p = mul_lpoly a (Pleft t)" and at: "(a,t) \<in> set (zip cs ts)" by auto 
      from at have a: "a \<in> set cs" and t: "t \<in> set ts" using set_zip_leftD[where x = a and y = t] set_zip_rightD[where x = a and y = t] by auto
      show "wf_lpoly p" by (simp only: p, rule wf_mul_lpoly, auto simp: wf_cs a t Fun)
    qed
    have wf_sum: "wf_lpoly (list_sum_lpoly ?plist)"
      by (rule wf_list_sum_lpoly, rule wf_ps)
    have "(list_sum (map (\<lambda> at. fst at * (?I (snd at))) (zip cs ts))) \<ge> list_sum (map ?E ?plist)" (is "geq (list_sum ?vlist) _")
    proof (simp only: o_def map_map, rule list_sum_mono, rule ballI)
      fix ct
      assume mem: "ct \<in> set (zip cs ts)"
      from this obtain c t where ct: "ct = (c,t)" by force
      from ct mem have c: "c \<in> set cs" and t: "t \<in> set ts" using set_zip_leftD[where x = c and y = t] set_zip_rightD[where x = c and y = t] by auto
      have "((c * (?I t)) \<ge> ?E (mul_lpoly c (Pleft t))) \<and> wf_elem (c * ?I t) \<and> wf_elem (?E (mul_lpoly c (Pleft t)))" 
      proof (rule conjI)
	have id: "?E (mul_lpoly c (Pleft t)) = c * (?E (Pleft t))" by (rule mul_poly_sound, auto simp: c wf_cs Fun t wf_ass)
	show "c * (?I t) \<ge> ?E (mul_lpoly c (Pleft t))" 
	  by (simp only: id, rule times_right_mono, auto simp: c wf_cs Fun  pos_cs t Fun wf_ass, rule wf_eval_lpoly, auto simp: wf_ass t Fun)
      next
	show "wf_elem (c * (?I t)) \<and> wf_elem (?E (mul_lpoly c (Pleft t)))"
	by (rule conjI, rule wf_times, auto simp: c wf_cs wf_ass, rule wf_eval_lpoly, auto simp: wf_ass t Fun, rule wf_mul_lpoly, auto simp: c wf_cs t Fun)
      qed
      with ct show " (fst ct * (pi\<guillemotleft>\<alpha>,snd ct\<guillemotright>) \<ge> eval_lpoly \<alpha> (mul_lpoly (fst ct) (Pleft (snd ct)))) \<and> wf_elem (fst ct * (pi\<guillemotleft>\<alpha>,snd ct\<guillemotright>)) \<and> wf_elem (eval_lpoly \<alpha> (mul_lpoly (fst ct) (Pleft (snd ct))))" by simp
    qed
    also have  "\<dots> = ?E (list_sum_lpoly ?plist)" by (rule list_sum_lpoly_sound[symmetric], rule wf_ps, rule wf_ass)
    finally have part1: "list_sum ?vlist \<ge> ?E (list_sum_lpoly ?plist)" . 
    have wf_csump: "wf_lpoly (sum_lpoly (c_lpoly c) (list_sum_lpoly ?plist))" (is "wf_lpoly ?csump")
      by (rule wf_sum_lpoly, auto simp: wf_c wf_pvars_def, rule wf_sum)
    from Pair Fun have id: "Pleft (Fun f ts) = (case ?csump of LPoly d [] \<Rightarrow> LPoly (max0 d) [] | p \<Rightarrow> p)" by auto    
    from wf_csump have wf_final: "wf_lpoly (Pleft (Fun f ts))" by (simp only: id, cases ?csump, cases "get_nc_lpoly ?csump", auto simp: wf_pvars_def wf_max0)
    have "?E ?csump = ?E (c_lpoly c) + ?E (list_sum_lpoly ?plist)" by (rule sum_poly_sound, auto simp: wf_ass wf_c wf_pvars_def wf_sum)
    also have "\<dots> = c + ?E (list_sum_lpoly ?plist)" by (simp add: wf_c)
    finally have idr: "?E ?csump = c + ?E (list_sum_lpoly ?plist)" .    
    have wf_suml: "wf_elem (list_sum ?vlist)"
    proof (rule wf_list_sum, rule ballI)
      fix v
      assume "v \<in> set ?vlist"
      from this obtain c t where v: "v = c * (?I t)" and ct: "(c,t) \<in> set (zip cs ts)" by auto 
      from ct have c: "c \<in> set cs" and t: "t \<in> set ts" using set_zip_leftD[where x = c and y = t] set_zip_rightD[where x = c and y = t] by auto
      show "wf_elem v" by (simp only: v, rule wf_times, auto simp: wf_cs c wf_ass)
    qed
    have ge_csum: "c + list_sum ?vlist \<ge> ?E ?csump" 
      by (simp only: idr, rule plus_right_mono, rule part1, rule wf_c, rule wf_suml,
      rule wf_eval_lpoly, rule wf_ass, rule wf_sum)
    have wf_csuml: "wf_elem (c + list_sum ?vlist)" using wf_suml wf_c by auto
    from Pair have left: "?I (Fun f ts) = max0 ( c + list_sum ?vlist)" by auto 
    have "\<dots> \<ge> max0 ( ?E ?csump)"
      by (rule max0_mono, rule ge_csum, rule wf_csuml, rule wf_eval_lpoly, rule wf_ass, rule wf_csump)
    hence part2: "?I (Fun f ts) \<ge> max0 ( ?E ?csump)" by (simp only: left)    
    have part3: "max0 ( ?E ?csump) \<ge> ?E (Pleft (Fun f ts))"
      using wf_csump by (simp only: id, cases ?csump, cases "get_nc_lpoly ?csump", auto simp: max0_x wf_ass wf_max0)
    have ge_final: "?I (Fun f ts) \<ge> ?E (Pleft (Fun f ts))"
      by (rule geq_trans, rule part2, rule part3, rule wf_terms, rule wf_ass, rule wf_max0, 
	rule wf_eval_lpoly, rule wf_ass, rule wf_csump, rule wf_eval_lpoly, rule wf_ass, rule wf_final)    
    from wf_final ge_final show ?thesis by auto
  qed
qed
  

lemma Pleft_sound:   assumes wf_ass: "wf_ass \<alpha>"
  and pos_ass: "pos_ass \<alpha>"
  shows "(pi\<guillemotleft>\<alpha>,t\<guillemotright>) \<ge> eval_lpoly \<alpha> (Pleft t)"
  using Pleft_both assms by auto

lemma wf_Pleft[simp]: "wf_lpoly (Pleft t)"
  using Pleft_both[where \<alpha> = zero_ass] by auto

lemma pos_eval_pvars: assumes wf_ass: "wf_ass \<alpha>"
  and pos_ass: "pos_ass \<alpha>"
  and wf_vas: "wf_pvars vas"
  and pos_vas: "pos_pvars vas"
  shows "eval_pvars \<alpha> vas \<ge> 0"
using wf_vas pos_vas 
proof (induct vas)
  case (Cons xa vas)
  show ?case 
  proof (cases xa)
    case (Pair x a)
    from pos_ass wf_ass have pos_x: "\<alpha> x \<ge> 0" and wf_x: "wf_elem (\<alpha> x)" unfolding pos_ass_def wf_ass_def by auto
    from Pair Cons have pos_a: "a \<ge> 0" and wf_a: "wf_elem a" unfolding wf_pvars_def pos_pvars_def by auto
    have "a * \<alpha> x \<ge> 0 * 0" by (rule geq_trans[where y = "a * 0"], rule times_right_mono, auto simp: wf_a pos_a wf_x pos_x)
    hence pos_ax: "a * \<alpha> x \<ge> 0" by auto
    have wf_ax: "wf_elem (a * \<alpha> x)" by (auto simp: wf_x wf_a)
    from Cons have pos_rec: "eval_pvars \<alpha> vas \<ge> 0" unfolding wf_pvars_def pos_pvars_def by auto
    from Cons have "wf_pvars vas" unfolding wf_pvars_def by auto
    with Cons have wf_rec: "wf_elem (eval_pvars \<alpha> vas)" using wf_ass by auto
    have "a * \<alpha> x + eval_pvars \<alpha> vas \<ge> 0 + 0" by (rule plus_left_right_mono, auto simp: pos_ax wf_ax pos_rec wf_rec)
    thus ?thesis by (simp add: Pair)
  qed
qed simp
    

lemma Pright_both:  assumes wf_ass: "wf_ass \<alpha>"
  and pos_ass: "pos_ass \<alpha>"
  shows "(eval_lpoly \<alpha> (Pright t) \<ge> (pi\<guillemotleft>\<alpha>,t\<guillemotright>)) \<and> wf_lpoly (Pright t) \<and> pos_coeffs (Pright t)" 
proof (induct t)
  case (Var x)
  show ?case using pos_ass wf_ass unfolding pos_ass_def wf_ass_def pos_coeffs_def pos_pvars_def by (auto simp: wf_pvars_def)
  case (Fun f ts)
  let ?I = "\<lambda> s. pi\<guillemotleft>\<alpha>,s\<guillemotright>"
  let ?E = "\<lambda> p. eval_lpoly \<alpha> p"
  show ?case 
  proof (cases "pi (f,length ts)")
    case (Pair c cs)
    from wf_pi have "wf_elem (fst (pi (f,length ts))) \<and> (\<forall> c \<in> set (snd (pi (f,length ts))). wf_elem c \<and> geq c zero)" by auto
    with Pair have wf_c: "wf_elem c" and wf_cs: "\<forall> c \<in> set cs. wf_elem c" and pos_cs: "\<forall> c \<in> set cs. c \<ge> 0" by auto
    let ?plist = "map (\<lambda> at. mul_lpoly (fst at) (Pright (snd at))) (zip cs ts)"
    have wf_ps: "\<forall> p \<in> set ?plist. wf_lpoly p"
    proof
      fix p
      assume "p \<in> set ?plist"
      from this obtain a t where p: "p = mul_lpoly a (Pright t)" and at: "(a,t) \<in> set (zip cs ts)" by auto 
      from at have a: "a \<in> set cs" and t: "t \<in> set ts" using set_zip_leftD[where x = a and y = t] set_zip_rightD[where x = a and y = t] by auto
      show "wf_lpoly p" by (simp only: p, rule wf_mul_lpoly, auto simp: wf_cs a t Fun)
    qed
    have wf_sum: "wf_lpoly (list_sum_lpoly ?plist)"
      by (rule wf_list_sum_lpoly, rule wf_ps)
    have pos_ps: "\<forall> p \<in> set ?plist. pos_coeffs p"
    proof
      fix p
      assume "p \<in> set ?plist"
      from this obtain a t where p: "p = mul_lpoly a (Pright t)" and at: "(a,t) \<in> set (zip cs ts)" by auto 
      from at have a: "a \<in> set cs" and t: "t \<in> set ts" using set_zip_leftD[where x = a and y = t] set_zip_rightD[where x = a and y = t] by auto
      show "pos_coeffs p" by (simp only: p, rule pos_mul_lpoly, auto simp: a pos_cs wf_cs t Fun)
    qed
    have pos_sum: "pos_coeffs (list_sum_lpoly ?plist)"
      by (rule pos_list_sum_lpoly, rule wf_ps, rule pos_ps)
    have id1: "?E (list_sum_lpoly ?plist) = list_sum (map ?E ?plist)" by (rule list_sum_lpoly_sound, rule wf_ps, rule wf_ass)
    have "\<dots> \<ge> (list_sum (map (\<lambda> at. fst at * (?I (snd at))) (zip cs ts)))" (is "geq _ (list_sum ?vlist)")
    proof (simp only: o_def map_map, rule list_sum_mono, rule ballI)
      fix ct
      assume mem: "ct \<in> set (zip cs ts)"
      from this obtain c t where ct: "ct = (c,t)" by force
      from ct mem have c: "c \<in> set cs" and t: "t \<in> set ts" using set_zip_leftD[where x = c and y = t] set_zip_rightD[where x = c and y = t] by auto
      have "(?E (mul_lpoly c (Pright t)) \<ge> (c * (?I t))) \<and> wf_elem (c * ?I t) \<and> wf_elem (?E (mul_lpoly c (Pright t)))" 
      proof (rule conjI)
	have id: "?E (mul_lpoly c (Pright t)) = c * (?E (Pright t))" by (rule mul_poly_sound, auto simp: c wf_cs Fun t wf_ass)
	show "?E (mul_lpoly c (Pright t)) \<ge> c * (?I t)" 
	  by (simp only: id, rule times_right_mono[OF _ _ _ wf_eval_lpoly], auto simp: c wf_cs Fun[OF t]  pos_cs  wf_ass)
      next
	show "wf_elem (c * (?I t)) \<and> wf_elem (?E (mul_lpoly c (Pright t)))"
	  by (rule conjI[OF wf_times wf_eval_lpoly[OF _ wf_mul_lpoly]], auto simp: c wf_cs wf_ass Fun[OF t])
      qed	
      with ct show " (eval_lpoly \<alpha> (mul_lpoly (fst ct) (Pright (snd ct))) \<ge> fst ct * (pi\<guillemotleft>\<alpha>,snd ct\<guillemotright>)) \<and> wf_elem (eval_lpoly \<alpha> (mul_lpoly (fst ct) (Pright (snd ct)))) \<and> wf_elem (fst ct * (pi\<guillemotleft>\<alpha>,snd ct\<guillemotright>))" by simp
    qed
    hence part1: "?E (list_sum_lpoly ?plist) \<ge> list_sum ?vlist" by (simp only: id1)  
    have wf_csump: "wf_lpoly (sum_lpoly (c_lpoly c) (list_sum_lpoly ?plist))" (is "wf_lpoly ?csump")
      by (rule wf_sum_lpoly[OF _ wf_sum], auto simp: wf_c wf_pvars_def)
    have pos_csump: "pos_coeffs ?csump"
      by (rule pos_sum_lpoly, auto simp: wf_sum wf_c pos_sum, auto simp: wf_c wf_pvars_def pos_coeffs_def pos_pvars_def)
    from Pair Fun have id: "Pright (Fun f ts) = (case ?csump of LPoly d nc \<Rightarrow> LPoly (max0 d) nc)" by auto    
    from wf_csump have wf_final: "wf_lpoly (Pright (Fun f ts))" by (simp only: id, cases ?csump, auto simp: wf_pvars_def wf_max0)
    from pos_csump have pos_final: "pos_coeffs (Pright (Fun f ts))" unfolding pos_coeffs_def by (simp only: id, cases ?csump, auto)
    have "?E ?csump = ?E (c_lpoly c) + ?E (list_sum_lpoly ?plist)" by (rule sum_poly_sound, auto simp: wf_ass wf_c wf_pvars_def wf_sum)
    also have "\<dots> = c + ?E (list_sum_lpoly ?plist)" by (simp add: wf_c)
    finally have idr: "?E ?csump = c + ?E (list_sum_lpoly ?plist)" .    
    have wf_sumr: "wf_elem (list_sum ?vlist)"
    proof (rule wf_list_sum, rule ballI)
      fix v
      assume "v \<in> set ?vlist"
      from this obtain c t where v: "v = c * (?I t)" and ct: "(c,t) \<in> set (zip cs ts)" by auto 
      from ct have c: "c \<in> set cs" and t: "t \<in> set ts" using set_zip_leftD[where x = c and y = t] set_zip_rightD[where x = c and y = t] by auto
      show "wf_elem v" by (simp only: v, rule wf_times, auto simp: wf_cs c wf_ass)
    qed
    have ge_csum: "?E ?csump \<ge> c + list_sum ?vlist" 
      by (simp only: idr, rule plus_right_mono, rule part1, rule wf_c,
      rule wf_eval_lpoly, rule wf_ass, rule wf_sum, rule wf_sumr)
    have wf_csuml: "wf_elem (c + list_sum ?vlist)" using wf_sumr wf_c by auto
    from Pair have right: "?I (Fun f ts) = max0 ( c + list_sum ?vlist)" by auto 
    have "max0 ( ?E ?csump) \<ge> \<dots>"
      by (rule max0_mono, rule ge_csum, rule wf_eval_lpoly, rule wf_ass, rule wf_csump, rule wf_csuml)
    hence part2: "max0 ( ?E ?csump) \<ge> ?I (Fun f ts)" by (simp only: right)    
    have part3: "?E (Pright (Fun f ts)) \<ge> max0 (?E ?csump)"
    proof (cases ?csump)
      case (LPoly d nc)
      have wf_nc: "wf_pvars nc" using wf_csump LPoly by (simp only: id, auto)
      have wf_enc: "wf_elem (eval_pvars \<alpha> nc)" using wf_csump LPoly by (simp only: id, auto simp: wf_ass)
      have wf_d: "wf_elem d" using wf_csump LPoly by (simp only: id, auto)
      have pos_nc: "pos_pvars nc" using pos_csump LPoly unfolding pos_coeffs_def by (simp only: id, auto)
      have nc_pos: "eval_pvars \<alpha> nc \<ge> 0" by (rule pos_eval_pvars, auto simp: wf_ass pos_ass pos_nc wf_nc)
      have md_pos: "max0 d \<ge> 0" by (rule max0_pos, rule wf_d)
      have "max0 d + eval_pvars \<alpha> nc \<ge> 0 + 0" (is "geq ?pr _") by (rule plus_left_right_mono, rule md_pos, rule nc_pos, auto simp: wf_enc wf_d wf_max0)
      hence "?pr \<ge> 0" by simp
      hence max_id: "max0 (?pr) = ?pr" by (rule max0_id_pos, auto simp: wf_enc wf_d wf_max0)
      have "?pr \<ge> max0 (d + eval_pvars \<alpha> nc)"
      proof (rule geq_trans[where ?y = "max0 (?pr)"], simp only: max_id, rule geq_refl) 
        show "max0 ?pr \<ge> max0 (d + eval_pvars \<alpha> nc)"
          by (rule max0_mono[OF plus_left_mono[OF max0_x]], auto simp: wf_d wf_max0 wf_enc)
      qed (auto simp: wf_d wf_max0 wf_enc)
      thus ?thesis by (simp only: id, simp only: LPoly, simp)
    qed
    have ge_final: "?E (Pright (Fun f ts)) \<ge> ?I (Fun f ts)"
      by (rule geq_trans, rule part3, rule part2, rule wf_eval_lpoly, rule wf_ass, rule wf_final, rule wf_max0,
	rule wf_eval_lpoly, rule wf_ass, rule wf_csump, rule wf_terms, rule wf_ass)
    from wf_final ge_final pos_final show ?thesis by auto
  qed
qed


lemma Pright_sound:   assumes wf_ass: "wf_ass \<alpha>"
  and pos_ass: "pos_ass \<alpha>"
  shows "eval_lpoly \<alpha> (Pright t) \<ge> (pi\<guillemotleft>\<alpha>,t\<guillemotright>)"
  using Pright_both assms by auto

lemma wf_Pright[simp]: "wf_lpoly (Pright t)"
  using Pright_both[where \<alpha> = zero_ass] by auto

abbreviation check_polo_ns :: "('f,'v::show)rule \<Rightarrow> shows check"
where "check_polo_ns \<equiv> check_polo_nsI zero one plus times max0 geq pi"
end

type_synonym ('f,'a)lpoly_interL = "(('f \<times> nat) \<times> ('a \<times> 'a list)) list"

(* default interpretation f/n \<rightarrow> d + x_1 + ... + x_n *)
fun to_lpoly_inter :: "'a \<Rightarrow> 'a \<Rightarrow> ('f :: key, 'a) lpoly_interL \<Rightarrow> ('f, 'a) lpoly_inter" where
  "to_lpoly_inter d on I = fun_of_map_fun (ceta_map_of I) (\<lambda> fn. (d,replicate (snd fn) on))"

(* create an arity that is not used *)
fun create_arity :: "('f, 'a) lpoly_interL \<Rightarrow> nat" where
  "create_arity fcns = Suc (foldr (\<lambda>(fn, ccs) m. max m (snd fn)) fcns 0)"

lemma create_arity_sound:
  assumes m: "m \<ge> create_arity I"
  shows "to_lpoly_inter d on I (f, m) = (d, replicate m on)" (is "?I (f, m) = _")
proof (cases "map_of I (f, m)")
  case None thus ?thesis by auto
next
  case (Some val)
  hence one: "((f,m),val) \<in> set I" by (rule map_of_SomeD)
  hence "m \<le> foldr (\<lambda> (fn, ccs) n. max n (snd fn)) I 0" 
  proof (induct I)
    case (Cons fnccs I)
    thus ?case
        by (cases fnccs, cases "snd fnccs", cases "fst fnccs", auto)
  qed simp
  with m show ?thesis by auto
qed

declare create_arity.simps [simp del]

lemma replicate_prop:
  assumes "P x"
  shows "\<forall>y\<in>set (replicate n x). P y"
  using assms by (induct n) simp_all

context linear_poly_order
begin

lemma inter_ns_ce_compat:
  assumes pi: "pi = to_lpoly_inter default one I" (is "_ = ?pi")
  shows "\<exists> n. \<forall> m. m \<ge> n \<longrightarrow> (\<forall> c c'. ce_trs (c, c', m) \<subseteq> inter_ns)"
unfolding inter_ns_def
proof (rule exI[of _ "create_arity I"], intro allI impI, rule subsetI, simp only: ce_trs.simps)
  fix m c c' x
  let ?n = "create_arity I"
  assume m: "?n \<le> m"
  let ?m = "Suc (Suc m)"
  from m have "?m \<ge> ?n" by auto
  hence inter: "pi (c, ?m) = (default, replicate ?m 1)" by (simp only: pi, rule create_arity_sound)
  let ?sum = "list_sum (replicate m (1 * (\<lambda>(a, as). max0 (a + 0)) (pi (c', 0))))"
  obtain a as where pi: "pi (c', 0) = (a,as)" by force
  have "wf_elem (fst (pi (c', 0)))" using wf_pi by auto
  with pi have wf_a: "wf_elem a" by auto
  obtain b where b: "b = max0 (a + 0)" by auto
  have idb: "1 * (\<lambda>(a, as). max0 (a + 0)) (pi (c', 0)) = b"
    by (simp add: pi b wf_a wf_max0)
  have wf_pos_b: "wf_elem b \<and> geq b 0" by (simp add: b, auto simp: wf_a wf_max0 max0_pos)
  have wf_sum: "wf_elem ?sum"
    by (simp only: idb, rule wf_list_sum, rule replicate_prop, auto simp: wf_pos_b)
  have pos_sum: "?sum \<ge> 0"
    by (simp only: idb, rule pos_list_sum, rule replicate_prop, auto simp: wf_pos_b)
  from wf_sum pos_sum obtain d where d: "?sum = d" and wf_pos_d: "wf_elem d \<and> geq d 0" by force
  assume "x \<in> {(Fun c (t # s # replicate m (Fun c' [])), t)| t s. True} \<union> 
              {(Fun c (t # s # replicate m (Fun c' [])), s)| t s. True}" (is "_ \<in> ?tc \<union> ?sc")
  thus "x \<in> {(s,t). \<forall> \<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,s\<guillemotright> \<ge> pi\<guillemotleft>\<alpha>,t\<guillemotright>)}"
  proof
    assume "x \<in> ?tc"
    from this obtain t s where id: "x = (Fun c (t # s # replicate m (Fun c' [])), t)" (is "x = (?l, _)") by auto    
    have "\<forall> \<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,?l\<guillemotright> \<ge> pi\<guillemotleft>\<alpha>,t\<guillemotright>)" 
    proof (intro allI)
      fix \<alpha>
      show "wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,?l\<guillemotright> \<ge> pi\<guillemotleft>\<alpha>,t\<guillemotright>)" (is "?wf_pos \<longrightarrow> ?goal")
      proof
	assume wf_pos_ass: ?wf_pos
	let ?term = "((default + (pi\<guillemotleft>\<alpha>,t\<guillemotright>)) + (pi\<guillemotleft>\<alpha>,s\<guillemotright>)) + d"
	have max: "max0 ?term = ?term" 
          by (rule max0_id_pos, auto simp: wf_pos_ass wf_pos_d, (rule sum_pos)+, auto simp: wf_pos_ass wf_pos_d)
	have "max0 ?term \<ge> ((0 + (pi\<guillemotleft>\<alpha>,t\<guillemotright>)) + 0) + 0"
	  by (simp only: max, (rule plus_left_right_mono)+, auto simp: wf_pos_ass wf_pos_d)
	hence "max0 ?term \<ge> (pi\<guillemotleft>\<alpha>,t\<guillemotright>)" by (auto simp: wf_pos_ass)
	with wf_pos_ass show ?goal using inter 
	  by (auto simp: d wf_pos_d wf_max0)
      qed      
    qed
    thus ?thesis using id by simp
  next
    assume "x \<in> ?sc"
    from this obtain t s where id: "x = (Fun c (t # s # replicate m (Fun c' [])), s)" (is "x = (?l, _)") by auto    
    have "\<forall> \<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,?l\<guillemotright> \<ge> pi\<guillemotleft>\<alpha>,s\<guillemotright>)" 
    proof (intro allI)
      fix \<alpha>
      show "wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,?l\<guillemotright> \<ge> pi\<guillemotleft>\<alpha>,s\<guillemotright>)" (is "?wf_pos \<longrightarrow> ?goal")
      proof
	assume wf_pos_ass: ?wf_pos
	let ?term = "((default + (pi\<guillemotleft>\<alpha>,t\<guillemotright>)) + (pi\<guillemotleft>\<alpha>,s\<guillemotright>)) + d"
	have max: "max0 ?term = ?term" 
          by (rule max0_id_pos, auto simp: wf_pos_ass wf_pos_d, (rule sum_pos)+, auto simp: wf_pos_ass wf_pos_d)
	have "max0 ?term \<ge> ((0 + 0) + (pi\<guillemotleft>\<alpha>,s\<guillemotright>)) + 0" 
	  by (simp only: max, (rule plus_left_right_mono)+, auto simp: wf_pos_ass wf_pos_d) 
	hence "max0 ?term \<ge> (pi\<guillemotleft>\<alpha>,s\<guillemotright>)" by (auto simp: wf_pos_ass)
	with wf_pos_ass show ?goal using inter 
	  by (auto simp: d wf_pos_d wf_max0)
      qed
    qed
    thus ?thesis using id by simp
  qed
qed
end

definition create_afI :: "'a \<Rightarrow> ('f :: key, 'a) lpoly_interL \<Rightarrow> 'f af" where
  "create_afI ze I \<equiv> fun_of_map_fun' (ceta_map_of I) (\<lambda> _ _. True) (\<lambda> (c,coeffs). (let len = length coeffs in (\<lambda> i. i < len \<and> coeffs ! i \<noteq> ze)))" 


lemma create_af_sound:
  assumes not: "\<not> create_afI ze I fn i"
  shows "length (snd (to_lpoly_inter def on I fn)) \<le> i \<or> (snd (to_lpoly_inter def on I fn)) ! i = ze"
proof -
  note not = not[unfolded create_afI_def Let_def]
  from not obtain c coeffs where look: "map_of I fn = Some (c, coeffs)"
    by (cases "map_of I fn", force+)
  from not show ?thesis
    unfolding to_lpoly_inter.simps Let_def ceta_map_of fun_of_map_fun.simps fun_of_map_fun'.simps look option.simps split by auto
qed
      
context linear_poly_order
begin
abbreviation create_af :: "('f,'a)lpoly_interL \<Rightarrow> 'f af"
  where "create_af \<equiv> create_afI zero"

lemma inter_ns_af_compat: assumes pi: "pi = to_lpoly_inter def on I" (is "_ = ?pi")
  shows "af_compatible (create_af I) inter_ns"
proof (unfold af_compatible_def, intro allI)
  fix f :: 'f and bef and s t :: "('f,'v)term" and aft
  obtain v :: 'v where True by simp
  let ?st = "Fun f (bef @ s # aft)"
  let ?tt = "Fun f (bef @ t # aft)"
  let ?pair = "(?st, ?tt)"
  let ?n = "Suc (length bef + length aft)"
  let ?i = "length bef"
  show "create_af I (f, ?n) ?i \<or> ?pair \<in> inter_ns"
  proof (cases "create_af I (f, ?n) ?i")
    case True thus ?thesis ..
  next
    case False
    hence "length (snd (?pi (f,?n))) \<le> ?i \<or> snd (?pi (f, ?n)) ! ?i = 0" by (rule create_af_sound)
    with pi have pi3: "length (snd (pi (f,?n))) \<le> ?i \<or> snd (pi (f,?n)) ! ?i = 0" by simp
    obtain c cs where pi2: "pi (f, ?n) = (c,cs)" by force
    with pi3 have short_zero: "length cs \<le> ?i \<or> cs ! ?i = 0" by auto
    have "\<forall> \<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<longrightarrow> ((pi\<guillemotleft>\<alpha>,?st\<guillemotright>) \<ge> (pi\<guillemotleft>\<alpha>,?tt\<guillemotright>))" (is "\<forall> \<alpha>. ?wfpos \<alpha> \<longrightarrow> ?rel \<alpha>")
    proof
      fix \<alpha>
      show "?wfpos \<alpha> \<longrightarrow> ?rel \<alpha>"
      proof
	assume wfpos: "?wfpos \<alpha>"
	hence wf_ass: "wf_ass \<alpha>" ..
	let ?in = "\<lambda> t :: ('f,'v) term. (pi\<guillemotleft>\<alpha>,t\<guillemotright>)"
	let ?inter = "\<lambda> s bef cs. list_sum (map (\<lambda> at. fst at * (?in (snd at))) (zip cs (bef @ s # aft)))"
	let ?intre = "\<lambda> aft cs. list_sum (map (\<lambda> at. fst at * (?in (snd at))) (zip cs aft))"
	have int_s: "?in ?st = max0 (c + ?inter s bef cs)" by (auto simp: pi2)
	have int_t: "?in ?tt = max0 (c + ?inter t bef cs)" by (auto simp: pi2)	
	from short_zero have "?inter s bef cs = ?inter t bef cs" 
	proof (induct cs arbitrary: bef)
	  case Nil thus ?case by auto
	next
	  case (Cons c cs) note oCons = this
	  show ?case 
	  proof (cases bef)
	    case Nil
	    with Cons show ?thesis by (simp add: wf_ass)
	  next
	    case (Cons b befs)
	    with oCons have "length cs \<le> length befs \<or> cs ! length befs = 0" by auto
	    with Cons oCons show ?thesis by auto
	  qed
	qed
	with int_s int_t have eq: "?in ?st = ?in ?tt" by auto
	show "?rel \<alpha>" by (simp only: eq, rule geq_refl, rule wf_terms, auto simp: wf_ass)
      qed
    qed
    hence "?pair \<in> inter_ns" unfolding inter_ns_def by auto
    thus ?thesis .. 
  qed
qed
end

definition
  check_poly_mono_npsmI :: "'a::show \<Rightarrow> ('f \<times> nat) list \<Rightarrow> ('f::show, 'a) lpoly_interL \<Rightarrow> shows check"
where
  "check_poly_mono_npsmI ze F pi \<equiv> check_allm (\<lambda>((f, n), (c, cs)). do {
     check (n = Suc 0 \<longrightarrow> c = ze) (''constant part '' +#+ shows c +@+ '' must be zero '' +#+ shows_nl);
     check (n = length cs) (''the arity is not the same as the number of arguments'' +#+ shows_nl);
     check (n \<le> Suc 0) (''symbol has arity larger than 1'' +#+ shows_nl)
   } <+? (\<lambda>s. ''problem with monotonicity due to interpretation of '' +#+ shows f +@+ ''/''  +#+ shows n +@+ shows_nl +@+ s)   
   ) pi \<guillemotright>
   check_subseteq F (map fst pi) <+? (\<lambda> fn. ''unknown interpretation for '' +#+ shows fn +@+ shows_nl)"

context linear_poly_order_carrier
begin
abbreviation check_poly_mono_npsm :: "('f \<times> nat)list \<Rightarrow> ('f::show,'a::show)lpoly_interL \<Rightarrow> shows check"
where "check_poly_mono_npsm \<equiv> check_poly_mono_npsmI 0"

abbreviation pi_of where "pi_of \<equiv> to_lpoly_inter default 1"

lemma check_poly_mono_npsm_sound: 
  assumes ok: "isOK(check_poly_mono_npsm F I)"
  shows "\<forall> fn \<in> set F \<union> set (map fst I). snd fn \<le> Suc 0 \<and> (snd fn = Suc 0 \<longrightarrow> fst (pi_of I fn) = 0 \<and> length (snd (pi_of I fn)) = Suc 0)" (is "\<forall> fn \<in> set F \<union> set (map fst I). ?P fn")
proof 
  fix fn
  assume fn: "fn \<in> set F \<union> set (map fst I)"
  note ok = ok[unfolded check_poly_mono_npsmI_def, simplified]
  from ok[THEN conjunct2] fn obtain ccs where mem: "(fn,ccs) \<in> set I" by auto
  obtain f n where fn: "fn = (f,n)" by force  
  from ok[THEN conjunct1, rule_format, OF mem]
  have le_1: "snd fn \<le> Suc 0" unfolding fn by (cases ccs, auto)
  show "?P fn"
  proof (rule conjI[OF le_1], intro impI)
    assume eq_1: "snd fn = Suc 0"
    have "\<exists> ccs. map_of I fn = Some ccs"
    proof (cases "map_of I fn")
      case (Some ccs) thus ?thesis by blast
    next
      case None
      from None[unfolded map_of_eq_None_iff[of I fn]] mem show ?thesis by force
    qed
    then obtain c cs where look: "map_of I fn = Some (c,cs)" by force
    from map_of_SomeD[OF look] have mem: "(fn, (c,cs)) \<in> set I" .
    have pi: "pi_of I fn = (c,cs)" unfolding to_lpoly_inter.simps look ceta_map_of fun_of_map_fun.simps Let_def by simp
    note ok = ok[THEN conjunct1, rule_format, OF mem, unfolded fn split, simplified]
    from ok eq_1 
    show "fst (pi_of I fn) = 0 \<and> length (snd (pi_of I fn)) = Suc 0"
      unfolding pi unfolding fn by simp
  qed
qed
end

fun
  check_poly_monoI :: "'a::show \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> ('a \<Rightarrow> bool) \<Rightarrow> ('f::show, 'a) lpoly_interL \<Rightarrow> shows check"
where
  "check_poly_monoI ze geq cm = check_allm (\<lambda>((f, n), (c, cs)). do {
     check (geq c ze) (''constant part '' +#+ shows c +@+ '' must be at least '' +#+ shows ze +@+ shows_nl);
     check (n \<le> length cs) (''the last argument is ignored'' +#+ shows_nl);
     check_allm (\<lambda>d. check (cm d) (''coefficient '' +#+ shows d +@+ '' is not allowed'' +#+ shows_nl)) cs
   } <+? (\<lambda>s. ''problem with monotonicity due to interpretation of '' +#+ shows f +@+ ''/''  +#+ shows n +@+ shows_nl +@+ s)
   )"

context mono_linear_poly_order_carrier 
begin
abbreviation check_poly_mono :: "('f::show,'a::show)lpoly_interL \<Rightarrow> shows check"
where "check_poly_mono \<equiv> check_poly_monoI 0 geq check_mono"

lemma plus_gt_right_mono: "\<lbrakk>y > z; wf_elem x; wf_elem y; wf_elem z\<rbrakk> \<Longrightarrow> x + y > x + z"
  by (simp add: plus_comm[where x = x], rule plus_gt_left_mono, auto simp: plus_single_mono)

lemma check_poly_mono_sound: assumes "isOK(check_poly_mono I)"
  shows "\<forall> f n. (fst (pi_of I (f,n)) \<ge> 0) \<and> (length (snd (pi_of I (f, n))) \<ge> n \<and> (\<forall> c \<in> set (snd (pi_of I (f, n))). check_mono c \<or> c = 1))" (is "\<forall> f n. ?prop f n")
proof (intro allI)
  fix f n
  show "?prop f n" 
  proof (cases "map_of I (f, n)")
    case None
    hence id: "pi_of I (f,n) = (default, replicate n one)" by auto
    show ?thesis
      by (simp only: id, rule conjI, simp, rule conjI, simp, simp)
  next
    case (Some ccs)
    from this obtain c cs where map_of: "map_of I (f, n) = Some (c,cs)" by force
    hence "((f,n),(c,cs)) \<in> set I" by (rule map_of_SomeD)
    with assms map_of show "?prop f n" by auto
  qed
qed

declare check_poly_monoI.simps[simp del]
end

locale mono_linear_poly_order = linear_poly_order + mono_linear_poly_order_carrier + 
  assumes pi_mono: "\<forall> f n. (fst (pi (f, n)) \<ge> 0) \<and> (length (snd (pi (f, n))) \<ge> n \<and> (\<forall> c \<in> set (snd (pi (f, n))). check_mono c \<or> c = 1))"
begin

lemma inter_s_ce_compat: assumes pi: "pi = to_lpoly_inter default one I" (is "_ = ?pi")
  shows "\<exists> n. \<forall> m. m \<ge> n \<longrightarrow> (\<forall> c c'. ce_trs (c, c', m) \<subseteq> inter_s)"
unfolding inter_s_def
proof (rule exI[of _ "create_arity I"], intro allI impI, rule subsetI, simp only: ce_trs.simps)
  fix m c c' x
  let ?n = "create_arity I"
  assume m: "?n \<le> m"
  let ?m = "Suc (Suc m)"
  from m have "?m \<ge> ?n" by auto
  hence inter: "pi (c, ?m) = (default, replicate ?m 1)" by (simp only: pi, rule create_arity_sound)
  let ?sum = "list_sum (replicate m (1 * (\<lambda>(a, as). max0 (a + 0)) (pi (c',0))))"
  obtain a as where pi: "pi (c',0) = (a,as)" by force
  have "wf_elem (fst (pi (c',0)))" using wf_pi by auto
  with pi have wf_a: "wf_elem a" by auto
  obtain b where b: "b = max0 (a + 0)" by auto
  have idb: "1 * (\<lambda>(a, as). max0 (a + 0)) (pi (c',0)) = b"
    by (simp add: pi b wf_a wf_max0)
  have wf_pos_b: "wf_elem b \<and> geq b 0" by (simp add: b, auto simp: wf_a wf_max0 max0_pos)
  have wf_sum: "wf_elem ?sum"
    by (simp only: idb, rule wf_list_sum, rule replicate_prop, auto simp: wf_pos_b)
  have pos_sum: "?sum \<ge> 0"
    by (simp only: idb, rule pos_list_sum, rule replicate_prop, auto simp: wf_pos_b)
  from wf_sum pos_sum obtain d where d: "?sum = d" and wf_pos_d: "wf_elem d \<and> geq d 0" by force
  fix x
  assume "x \<in> {(Fun c (t # s # replicate m (Fun c' [])), t) | t s. True} \<union>
              {(Fun c (t # s # replicate m (Fun c' [])), s) | t s. True}" (is "_ \<in> ?tc \<union> ?sc")
  thus "x \<in> {(s,t). \<forall>\<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,s\<guillemotright> > pi\<guillemotleft>\<alpha>,t\<guillemotright>)}"
  proof
    assume "x \<in> ?tc"
    then obtain t s where id: "x = (Fun c (t # s # replicate m (Fun c' [])), t)" (is "x = (?l, _)") by auto    
    have "\<forall> \<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,?l\<guillemotright> > pi\<guillemotleft>\<alpha>,t\<guillemotright>)" 
    proof (intro allI)
      fix \<alpha>
      show "wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,?l\<guillemotright> > pi\<guillemotleft>\<alpha>,t\<guillemotright>)" (is "?wf_pos \<longrightarrow> ?goal")
      proof
	assume wf_pos_ass: ?wf_pos
	let ?term = "((default + (pi\<guillemotleft>\<alpha>,t\<guillemotright>)) + (pi\<guillemotleft>\<alpha>,s\<guillemotright>)) + d"
	have max: "max0 ?term = ?term" 
          by (rule max0_id_pos, auto simp: wf_pos_ass wf_pos_d, (rule sum_pos)+, auto simp: wf_pos_ass wf_pos_d)
	have "max0 ?term > ((0 + (pi\<guillemotleft>\<alpha>,t\<guillemotright>)) + 0) + 0"
	  by (simp only: max, (rule plus_gt_left_mono2)+, auto simp: wf_pos_ass wf_pos_d plus_single_mono default_gt_zero)
	hence "max0 ?term > (pi\<guillemotleft>\<alpha>,t\<guillemotright>)" by (auto simp: wf_pos_ass)
	with wf_pos_ass show ?goal using inter 
	  by (auto simp: d wf_pos_d wf_max0)
      qed
    qed
    thus ?thesis using id by simp
  next
    assume "x \<in> ?sc"
    then obtain t s where id: "x = (Fun c (t # s # replicate m (Fun c' [])), s)" (is "x = (?l, _)") by auto    
    have "\<forall> \<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,?l\<guillemotright> > pi\<guillemotleft>\<alpha>,s\<guillemotright>)" 
    proof (intro allI)
      fix \<alpha>
      show "wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,?l\<guillemotright> > pi\<guillemotleft>\<alpha>,s\<guillemotright>)" (is "?wf_pos \<longrightarrow> ?goal")
      proof
	assume wf_pos_ass: ?wf_pos
	let ?term = "((default + (pi\<guillemotleft>\<alpha>,t\<guillemotright>)) + (pi\<guillemotleft>\<alpha>,s\<guillemotright>)) + d"
	have max: "max0 ?term = ?term" 
          by (rule max0_id_pos, auto simp: wf_pos_ass wf_pos_d, (rule sum_pos)+, auto simp: wf_pos_ass wf_pos_d)
	have "max0 ?term > ((0 + 0) + (pi\<guillemotleft>\<alpha>,s\<guillemotright>)) + 0" 
	  by (simp only: max, (rule plus_gt_left_mono2)+, auto simp: wf_pos_ass wf_pos_d plus_single_mono default_gt_zero) 
	hence "max0 ?term > (pi\<guillemotleft>\<alpha>,s\<guillemotright>)" by (auto simp: wf_pos_ass)
	with wf_pos_ass show ?goal using inter 
	  by (auto simp: d wf_pos_d wf_max0)
      qed
    qed
    thus ?thesis using id by simp
  qed
qed

lemma inter_s_mono: 
  shows "ctxt_closed inter_s" 
proof (rule one_imp_ctxt_closed)
  fix f :: 'b and bef and s t :: "('b,'v)term" and aft
  assume st: "(s,t) \<in> inter_s"
  let ?st = "Fun f (bef @ s # aft)"
  let ?tt = "Fun f (bef @ t # aft)"
  let ?pair = "(?st, ?tt)"
  let ?n = "Suc (length bef + length aft)"
  let ?i = "length bef"
  obtain c cs where ccs: "pi (f,?n) = (c,cs)" by force
  from pi_mono have "((fst (pi (f,?n))) \<ge> 0) \<and> length (snd (pi (f,?n))) \<ge> ?n \<and> (\<forall> c \<in> set (snd (pi (f,?n))). check_mono c \<or> c = 1)" by auto
  with ccs have mono: "(c \<ge> 0) \<and> length cs \<ge> ?n \<and> (\<forall> d \<in> set cs. check_mono d \<or> d = 1)" by auto
  from wf_pi have "wf_elem (fst (pi (f,?n)))" and "\<forall> d \<in> set (snd (pi (f,?n))). wf_elem d \<and> geq d 0" by auto
  with ccs have wf_c: "wf_elem c" and wf_cs: "\<forall> d \<in> set cs. wf_elem d \<and> geq d 0" by auto
  show "(Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> inter_s"
    unfolding inter_s_def
  proof (clarify)
    fix \<alpha> :: "('v,'a)p_ass"
    assume wf_ass: "wf_ass \<alpha>" and pos_ass: "pos_ass \<alpha>" and apos_ass: "apos_ass \<alpha>"
    let ?in = "\<lambda> t :: ('b,'v) term. (pi\<guillemotleft>\<alpha>,t\<guillemotright>)"
    let ?inter = "\<lambda> s bef cs. list_sum (map (\<lambda> at. fst at * (?in (snd at))) (zip cs (bef @ s # aft)))"
    have int_s: "?in ?st = max0 (c + ?inter s bef cs)" by (auto simp: ccs)
    have int_t: "?in ?tt = max0 (c + ?inter t bef cs)" by (auto simp: ccs)	
    from mono wf_cs have main: "?inter s bef cs > ?inter t bef cs" 
    proof (induct cs arbitrary: bef)
      case Nil thus ?case by auto
    next
      case (Cons c cs) note oCons = this
      hence wf_c: "wf_elem c" and c_zero: "c \<ge> 0" by auto
      {
	fix aft and a :: 'a and b :: "('b,'v)term"
	assume "(a,b) \<in> set (zip cs aft)"
	hence "a \<in> set cs" by (rule set_zip_leftD)
	with Cons have "wf_elem a" by auto
	hence "wf_elem (a * ?in b)" using wf_ass by auto
      } hence wf_rec: "\<And> aft. wf_elem (list_sum (map (\<lambda>at. fst at * (pi\<guillemotleft>\<alpha>,snd at\<guillemotright>)) (zip cs aft)))"
	by (intro wf_list_sum, auto)
      show ?case
      proof (cases bef)
	case Nil
	from st have st: "?in s > ?in t" unfolding inter_s_def using wf_ass pos_ass apos_ass by auto
        from Cons have "check_mono c \<or> c = 1" by auto
        hence cst: "c * ?in s > c * ?in t"
        proof
          assume "c = 1"
          with st show ?thesis by (simp add: wf_ass)
        next
          assume "check_mono c"
          thus ?thesis
            by (rule check_mono[OF _ st], auto simp: wf_ass wf_c c_zero)
        qed
	show ?thesis 
          by (simp add: Nil Cons, rule plus_gt_left_mono2, rule cst, auto simp: Cons wf_ass plus_single_mono, rule geq_refl,(rule wf_rec)+) 
      next
	case (Cons b befs)
	with oCons have "Suc (length befs + length aft) \<le> length cs" by auto
	with Cons oCons have rec: "?inter s befs cs > ?inter t befs cs" by auto
	show ?thesis by (simp add: Cons oCons, rule plus_gt_right_mono[OF rec, simplified, OF _ wf_rec wf_rec],  auto simp: wf_c wf_ass)
      qed
    qed
    have wf_inter: "\<And> t. wf_elem (?inter t bef cs)"
    proof (rule wf_list_sum, auto)
      fix t a b
      assume "(a,b) \<in> set (zip cs (bef @ t # aft))"
      hence "a \<in> set cs" by (rule set_zip_leftD)
      with wf_cs wf_ass show "wf_elem (a * (?in b))" by auto
    qed	
    have wf_cint: "\<And> t. wf_elem (c + ?inter t bef cs)"
      by (rule wf_plus, rule wf_c, rule wf_inter)
    have pos_inter: "\<And> t. (?inter t bef cs) \<ge> 0"
    proof (rule pos_list_sum, (clarify, simp)+)
      fix t a b 
      assume "(a,b) \<in> set (zip cs (bef @ t # aft))"
      hence "a \<in> set cs" by (rule set_zip_leftD)
      hence pos_a: "a \<ge> 0" and wf_a: "wf_elem a" using wf_cs  by auto
      show "wf_elem (a * (?in b)) \<and> (a * (?in b) \<ge> 0)"
      proof (rule conjI, auto simp: wf_ass wf_a)
	have int: "(a * (?in b) \<ge> a * 0)"
	  by (rule times_right_mono, rule pos_a, rule pos_term, auto simp: wf_a wf_ass pos_ass)
	show "(a * (?in b)) \<ge> 0" by (rule geq_trans[where y = "a * 0"], rule int, auto simp: wf_a wf_ass)
      qed
    qed
    have "\<And> t. c + (?inter t bef cs) \<ge> 0 + 0"
      by (rule plus_left_right_mono, auto simp: mono, rule pos_inter, rule wf_c, rule wf_inter)
    hence pos_cint: "\<And> t. c + (?inter t bef cs) \<ge> 0" by auto
    have id_cint: "\<And> t. max0(c + (?inter t bef cs)) = c + (?inter t bef cs)"
      by (rule max0_id_pos, rule pos_cint, rule wf_cint)
    {
      fix t aft and a :: 'a and b :: "('b,'v)term"
      assume "(a,b) \<in> set (zip cs (bef @ t # aft))"
      hence "a \<in> set cs" by (rule set_zip_leftD)
      with wf_cs have "wf_elem a" by auto
      hence "wf_elem (a * ?in b)" using wf_ass by auto
    }      
    hence wf_rec: "\<And> t aft. wf_elem (list_sum (map (\<lambda>at. fst at * (pi\<guillemotleft>\<alpha>,snd at\<guillemotright>)) (zip cs (bef @ t # aft))))"
      by (intro wf_list_sum, auto)
    show "?in ?st > ?in ?tt" by (simp only: int_s int_t id_cint, rule plus_gt_right_mono, rule main, rule wf_c, (rule wf_rec)+)
  qed
qed
end

fun
  check_lpoly_coeffs :: "('a::show \<Rightarrow> bool) \<Rightarrow> ('a \<Rightarrow> bool) \<Rightarrow> ('a \<Rightarrow> bool) \<Rightarrow> ('f::show, 'a) lpoly_interL \<Rightarrow> shows check"
where
  "check_lpoly_coeffs wfe apos ch I =
     check_allm (\<lambda>((f, n), (c, cs)). do {
       check (wfe c) (''constant part '' +#+ shows c +@+ '' is not well-formed'' +#+ shows_nl);
       check (apos c \<or> Bex (set (take n cs)) apos) (''could not find positive entry which is required for arctic interpretations'' +#+ shows_nl);
       check_allm (\<lambda>a. check (ch a \<and> wfe a) (''coefficient '' +#+ shows a +@+ '' is not allowed'' +#+ shows_nl)) cs
     } <+?  (\<lambda>s. ''problem with interpretation of '' +#+ shows f +@+ ''/'' +#+ shows n +@+ shows_nl +@+ s)
     ) I"

context linear_poly_order 
begin
lemmas compat_orig = compat
end

sublocale linear_poly_order \<subseteq> redtriple_order inter_s inter_ns inter_ns
proof
  show "SN inter_s"
  proof
    fix f assume steps: "(\<forall> i. (f i, f (Suc i)) \<in> inter_s)"
    let ?b = default_ass
    have "\<forall> i. ((pi\<guillemotleft>?b,f i\<guillemotright>) > (pi\<guillemotleft>?b, f (Suc i)\<guillemotright>)) \<and> (arc_pos (pi\<guillemotleft>?b, f (Suc i)\<guillemotright>))"
    proof 
      fix i
      from steps have "(f i, f (Suc i)) \<in> inter_s" .. 
      hence dec: "(pi\<guillemotleft>?b, f i\<guillemotright>) >  (pi\<guillemotleft>?b, f (Suc i)\<guillemotright>)" (is ?dec) unfolding inter_s_def 
        by auto
      show "?dec \<and> arc_pos (pi\<guillemotleft>?b, f (Suc i)\<guillemotright>)" by (rule conjI, rule dec, rule apos_term, auto)
    qed
    from this obtain g
      where nSN: "\<forall> i. gt ((g i) :: 'a) (g (Suc i)) \<and> wf_elem (g i) \<and> wf_elem (g (Suc i))
      \<and> geq (g (Suc i)) zero \<and> arc_pos (g (Suc i))" by force
    with SN show False by (auto simp: SN_defs)
  qed
next
  show "inter_ns O inter_s \<subseteq> inter_s"
  proof (clarify)
    fix s t u :: "('f,'v)term"
    assume st: "(s,t) \<in> inter_ns" and tu: "(t,u) \<in> inter_s"
    show "(s,u) \<in> inter_s" unfolding inter_s_def
    proof(rule, unfold split, clarify)
      fix \<alpha> :: "'v \<Rightarrow> 'a"
      assume wf: "wf_ass \<alpha>" "pos_ass \<alpha>" "apos_ass \<alpha>"
      show "pi\<guillemotleft>\<alpha>,s\<guillemotright> > pi\<guillemotleft>\<alpha>,u\<guillemotright>"
        by (rule compat, insert st tu wf, unfold inter_s_def inter_ns_def, auto intro: wf_terms)
    qed
  qed
next
  show "inter_s O inter_ns \<subseteq> inter_s"
  proof (clarify)
    fix s t u :: "('f,'v)term"
    assume st: "(s,t) \<in> inter_s" and tu: "(t,u) \<in> inter_ns"
    show "(s,u) \<in> inter_s" unfolding inter_s_def
    proof(rule, unfold split, clarify)
      fix \<alpha> :: "'v \<Rightarrow> 'a"
      assume wf: "wf_ass \<alpha>" "pos_ass \<alpha>" "apos_ass \<alpha>"
      show "pi\<guillemotleft>\<alpha>,s\<guillemotright> > pi\<guillemotleft>\<alpha>,u\<guillemotright>"
        by (rule compat2, insert st tu wf, unfold inter_s_def inter_ns_def, auto intro: wf_terms)
    qed
  qed
next
  show "ctxt_closed inter_ns" unfolding inter_ns_def
  proof (rule one_imp_ctxt_closed)
    fix f :: 'f
      and bef :: "('f,'v) term list" 
      and s :: "('f,'v) term" 
      and t :: "('f,'v) term" 
      and aft :: "('f,'v) term list" 
    let ?P = "{(s,t). \<forall> \<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,s\<guillemotright> \<ge> pi\<guillemotleft>\<alpha>,t\<guillemotright>)}"
    assume "(s, t) \<in> ?P"
    hence st: "\<forall> \<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>, s\<guillemotright> \<ge> pi\<guillemotleft>\<alpha>, t\<guillemotright>)" by (simp)
    let ?inter = "pi (f, (Suc (length bef + length aft)))"
    show "(Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> ?P"
    proof (cases "?inter")
      case (Pair c list)
      have "\<forall> \<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>, Fun f (bef @ s # aft)\<guillemotright> \<ge> pi\<guillemotleft>\<alpha>,Fun f (bef @ t # aft)\<guillemotright>)" (is "\<forall> \<alpha>. ?ass \<alpha> \<longrightarrow> ?ge \<alpha>")
      proof 
        fix \<alpha>
        show "?ass \<alpha> \<longrightarrow> ?ge \<alpha>"
        proof 
	  assume "?ass \<alpha>"
	  hence wf_ass: "wf_ass \<alpha>" and pos_ass: "pos_ass \<alpha>" and apos_ass: "apos_ass \<alpha>" by auto
	  with st have sta: "pi\<guillemotleft>\<alpha>,s\<guillemotright> \<ge> pi\<guillemotleft>\<alpha>,t\<guillemotright>" by auto
	  from wf_pi have "wf_elem (fst ?inter) \<and> (\<forall> a. a \<in> set (snd ?inter) \<longrightarrow> wf_elem a \<and> geq a zero)"  by auto
	  with Pair have wf_c: "wf_elem c" and wf_list: "\<forall> a \<in> set list. wf_elem a" and pos_list: "\<forall> a \<in> set list. geq a zero" by auto
	  let ?expr = "\<lambda> list rest.  list_sum (map (\<lambda>at. fst at * (pi\<guillemotleft>\<alpha>,snd at\<guillemotright>)) (zip list rest))"
	  from wf_list pos_list have main: "(?expr list (bef @ s # aft) \<ge> ?expr list (bef @ t # aft)) \<and> wf_elem (?expr list (bef @ s # aft)) \<and> wf_elem (?expr list (bef @ t # aft))"
	  proof (induct list arbitrary:bef)
	    case (Cons a as) note oCons = this
	    hence wf_a: "wf_elem a" and pos_a: "a \<ge> 0" and wf_as: "\<forall> a \<in> set as. wf_elem a" and pos_as: "\<forall> a \<in> set as. a \<ge> 0"  by auto
	    thus ?case 
	    proof (cases bef)
	      case Nil
              {
	        fix aas aaft
	        assume "(aas,aaft) \<in> set (zip as aft)"
	        hence "aas \<in> set as" using set_zip_leftD[where x = aas and y = aaft] by auto
	        with wf_as have wfaas: "wf_elem aas" by auto
	        hence "wf_elem (aas * (pi\<guillemotleft>\<alpha>,aaft\<guillemotright>))" by (auto simp: wf_ass wfaas)
              }
	      hence wf_as_aft: "wf_elem (?expr as aft)" 
	        by (intro wf_list_sum, auto)
	      show ?thesis
	        by (rule conjI, simp add: Nil, rule plus_left_mono, rule times_right_mono, 
		  auto simp: wf_a pos_a wf_ass Nil wf_as_aft wf_a wf_ass sta)
	    next
	      case (Cons u us)
	      with oCons wf_as pos_as have wf_rec: "wf_elem (?expr as (us @ s # aft)) \<and> wf_elem (?expr as (us @ t # aft))" by best
	      show ?thesis 
	        by (simp add: Cons, rule conjI, rule plus_right_mono, auto simp: Cons oCons wf_ass wf_as pos_as)
	    qed
	  qed simp
	  show "?ge \<alpha>" by (simp add: Pair, rule max0_mono, rule plus_right_mono, auto simp: main wf_c)
        qed
      qed
      thus ?thesis by auto
    qed
  qed
next
  show "subst_closed inter_s" 
  proof (unfold subst_closed_def inter_s_def, rule subsetI)
    fix st_sig :: "('f,'v)term \<times> ('f,'v)term"
    assume st_sig: "st_sig \<in> CS {(s,t). \<forall>\<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,s\<guillemotright> > pi\<guillemotleft>\<alpha>,t\<guillemotright>) }" (is "_ \<in> CS ?P")
    thus "st_sig \<in> ?P"
    proof (cases st_sig)
      case (Pair s_sig t_sig)
      with st_sig have "(s_sig, t_sig) \<in> CS ?P" by simp
      hence "(s_sig, t_sig) \<in> ?P" 
      proof induct
        case (subst_closureI s t \<sigma>)
        show ?case 
        proof (simp del:interpr_subst, intro allI)
	  fix \<alpha> 
	  show "wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>, s \<cdot> \<sigma>\<guillemotright> > pi\<guillemotleft>\<alpha>, t \<cdot> \<sigma>\<guillemotright>)" (is "?wf \<and> ?ps \<longrightarrow> ?go") 
	  proof -
	    {
	      assume ?wf and ?ps
	      have wf_assB: "wf_ass (\<lambda>x. pi\<guillemotleft>\<alpha>,\<sigma> x\<guillemotright>)" (is "wf_ass ?bet") unfolding wf_ass_def using `?wf`  by auto
	      have pos_assB: "pos_ass ?bet" unfolding pos_ass_def using `?ps` `?wf`  by auto
	      have apos_assB: "apos_ass ?bet" unfolding apos_ass_def using `?ps` `?wf` by auto
	      with Pair subst_closureI Cons wf_assB pos_assB have "pi\<guillemotleft>?bet,s\<guillemotright> > pi\<guillemotleft>?bet,t\<guillemotright>" by force
	      with `?wf` have "?go" by (simp add: interpr_subst)
	    }
	    thus ?thesis by blast
	  qed
        qed
      qed
      with Pair show ?thesis by simp
    qed
  qed
next
  show "subst_closed inter_ns" 
  proof (unfold subst_closed_def inter_ns_def, rule subsetI)
    fix st_sig :: "('f,'v)term \<times> ('f,'v)term"
    assume st_sig: "st_sig \<in> CS {(s,t). \<forall>\<alpha>. wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,s\<guillemotright> \<ge> pi\<guillemotleft>\<alpha>,t\<guillemotright>) }" (is "_ \<in> CS ?P")
    thus "st_sig \<in> ?P"
    proof (cases st_sig)
      case (Pair s_sig t_sig)
      with st_sig have "(s_sig, t_sig) \<in> CS ?P" by simp
      hence "(s_sig, t_sig) \<in> ?P" 
      proof induct
        case (subst_closureI s t \<sigma>)
        show ?case 
        proof (simp del:interpr_subst, intro allI)
	  fix \<alpha> 
	  show "wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>, s \<cdot> \<sigma>\<guillemotright> \<ge> pi\<guillemotleft>\<alpha>, t \<cdot> \<sigma>\<guillemotright>)" (is "?wf \<and> ?ps \<longrightarrow> ?go") 
	  proof -
	    {
	      assume ?wf and ?ps
	      have wf_assB: "wf_ass (\<lambda>x. pi\<guillemotleft>\<alpha>, \<sigma> x\<guillemotright>)" (is "wf_ass ?bet") unfolding wf_ass_def using `?wf`  by auto
	      have pos_assB: "pos_ass ?bet" unfolding pos_ass_def using `?ps` `?wf`  by auto
	      have apos_assB: "apos_ass ?bet" unfolding apos_ass_def using `?ps` `?wf` by auto
	      with Cons subst_closureI Pair wf_assB pos_assB have "pi\<guillemotleft>?bet,s\<guillemotright> \<ge> pi\<guillemotleft>?bet,t\<guillemotright>" by force
	      with `?wf` have "?go" by (simp add: interpr_subst)
	    }
	    thus ?thesis by blast
	  qed
        qed
      qed
      with Pair show ?thesis by simp
    qed
  qed
next
  show "trans inter_ns" unfolding trans_def
  proof (clarify)
    fix s t u :: "('f,'v)term"
    assume st: "(s,t) \<in> inter_ns" and tu: "(t,u) \<in> inter_ns"
    show "(s,u) \<in> inter_ns" unfolding inter_ns_def
    proof(rule, unfold split, clarify)
      fix \<alpha> :: "'v \<Rightarrow> 'a"
      assume wf: "wf_ass \<alpha>" "pos_ass \<alpha>" "apos_ass \<alpha>"
      show "pi\<guillemotleft>\<alpha>,s\<guillemotright> \<ge> pi\<guillemotleft>\<alpha>,u\<guillemotright>"
        by (rule geq_trans, insert st tu wf, unfold inter_ns_def, auto intro: wf_terms)
    qed
  qed
next
  show "trans inter_s" unfolding trans_def
  proof (clarify)
    fix s t u :: "('f,'v)term"
    assume st: "(s,t) \<in> inter_s" and tu: "(t,u) \<in> inter_s"
    show "(s,u) \<in> inter_s" unfolding inter_s_def
    proof(rule, unfold split, clarify)
      fix \<alpha> :: "'v \<Rightarrow> 'a"
      assume wf: "wf_ass \<alpha>" "pos_ass \<alpha>" "apos_ass \<alpha>"
      show "pi\<guillemotleft>\<alpha>,s\<guillemotright> > pi\<guillemotleft>\<alpha>,u\<guillemotright>"
        by (rule gt_trans, insert st tu wf, unfold inter_s_def, auto intro: wf_terms)
    qed
  qed
next
  show "refl inter_ns"
    unfolding inter_ns_def refl_on_def using geq_refl by (auto intro: wf_terms)
qed

context linear_poly_order
begin

lemma eval_term_subst_eq: "\<lbrakk>\<And> x. x \<in> vars_term t \<Longrightarrow> \<alpha> x = \<beta> x\<rbrakk> \<Longrightarrow> (pi\<guillemotleft>\<alpha>,t\<guillemotright>) = (pi\<guillemotleft>\<beta>,t\<guillemotright>)"
proof (induction t)
  case (Var x)
  thus ?case by simp
next
  case (Fun f ts)
  {
    fix t x
    assume t: "t \<in> set ts" and x: "x \<in> vars_term t"
    with Fun.prems[of x] have "\<alpha> x = \<beta> x" by auto
  } note id = this
  {
    fix t
    assume t: "t \<in> set ts"
    from Fun.IH[OF t id[OF t]]
    have "(pi\<guillemotleft>\<alpha>,t\<guillemotright>) = (pi\<guillemotleft>\<beta>,t\<guillemotright>)" .
  } note IH = this
  obtain d ds where pi: "pi (f,length ts) = (d,ds)" by force
  show ?case unfolding eval_termI.simps Let_def pi split
  proof (rule arg_cong[where f = "\<lambda> a. max0 (d + list_sum a)"],
    rule map_cong[OF refl])
    fix x
    assume "x \<in> set (zip ds ts)"
    from zip_snd[OF this] have "snd x \<in> set ts" .
    from IH[OF this]
    show "fst x * (pi\<guillemotleft>\<alpha>,snd x\<guillemotright>) = fst x * (pi\<guillemotleft>\<beta>,snd x\<guillemotright>)" by simp
  qed
qed
end
  
locale npsm_mono_linear_poly_order = linear_poly_order +
  fixes F :: "'b sig"
  assumes F_unary: "\<And> fn. fn \<in> F \<Longrightarrow> snd fn \<le> Suc 0"
      and pi_mono: "\<And> fn. fn \<in> F \<Longrightarrow> snd fn = Suc 0 \<Longrightarrow> fst (pi fn) = 0 \<and> length (snd (pi fn)) = Suc 0"
      and mode: "\<not> plus_single_mono"
begin

lemma inter_s_F_mono: 
  fixes s :: "('b,'v)term"
  assumes st: "(s,t) \<in> inter_s"
  and F: "(f,Suc (length bef + length aft)) \<in> F"
  shows "(Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> inter_s"
proof -
  let ?f = "(f,Suc 0)"
  from F_unary[OF F] have empty: "bef = []" "aft = []" by auto
  from F empty have "?f \<in> F" by auto
  from pi_mono[OF this] have pi: "fst (pi ?f) = 0" "length (snd (pi ?f)) = Suc 0" by auto
  let ?st' = "Fun f (bef @ s # aft)"
  let ?tt' = "Fun f (bef @ t # aft)"
  let ?pair' = "(?st', ?tt')"
  let ?s = "Fun f [s]"
  let ?t = "Fun f [t]"
  let ?pair = "(?s, ?t)"
  have pair: "?pair' = ?pair" unfolding empty by simp
  obtain c cs where ccs: "pi ?f = (c,cs)" by force
  with pi obtain c where pi: "pi ?f = (0,[c])" by (cases "snd (pi ?f)", auto)
  from wf_pi[THEN conjunct1] have "\<forall> d \<in> set (snd (pi ?f)). wf_elem d \<and> geq d 0" by auto
  with pi have wf_c: "wf_elem c" and c_0: "c \<ge> 0" by auto
  from wf_pi[THEN conjunct2] have "arc_pos (fst (pi ?f)) \<or> (\<exists> a \<in> set (take (snd ?f) (snd (pi ?f))). arc_pos a)" by auto
  show "(Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> inter_s"
    unfolding inter_s_def pair
  proof (clarify)
    fix \<alpha> :: "('v,'a)p_ass"
    assume wf_ass: "wf_ass \<alpha>" and pos_ass: "pos_ass \<alpha>" and apos_ass: "apos_ass \<alpha>"
    let ?in = "\<lambda> t :: ('b,'v) term. (pi\<guillemotleft>\<alpha>,t\<guillemotright>)"
    let ?inter = "\<lambda> s. c * ?in s"
    from st[unfolded inter_s_def] wf_ass pos_ass apos_ass have gt: "?in s > ?in t"
      by simp
    from times_gt_right_mono[OF gt _ _ wf_c mode]
    have gt: "?inter s > ?inter t" using wf_ass by simp
    from wf_c wf_ass have wf_s: "wf_elem (?inter s)" by simp
    from wf_c wf_ass have wf_t: "wf_elem (?inter t)" by simp
    note max0 = max0_npsm_id[OF mode]
    have int_s: "?in ?s = ?inter s" using max0 wf_s by (simp add: pi)
    have int_t: "?in ?t = ?inter t" using max0 wf_t by (simp add: pi)
    show "?in ?s > ?in ?t" unfolding int_s int_t by (rule gt)
  qed
qed

lemma rel_ce_terminating: fixes R Rw :: "('b,'v)trs"
  assumes F: "funas_trs R \<subseteq> F" "funas_trs Rw \<subseteq> F"
  and orient: "R \<subseteq> inter_s" "Rw \<subseteq> inter_ns"
  shows "SN_rel (rstep (R \<union> \<Union> ce_trs ` UNIV)) (rstep (Rw \<union> \<Union> ce_trs ` UNIV))"
proof (rule sig_mono_imp_SN_rel_ce[OF inter_s_F_mono F_unary orient _ F])
  fix l r
  assume "(l,r) \<in> inter_ns \<inter> Rw"
  with F have ns: "(l,r) \<in> inter_ns" and Fl: "funas_term l \<subseteq> F" and Fr: "funas_term r \<subseteq> F" unfolding funas_trs_def funas_rule_def by force+  
  show "vars_term r \<subseteq> vars_term l" 
  proof (rule ccontr)
    assume "\<not> ?thesis"
    then obtain x where xr: "x \<in> vars_term r" and xl: "x \<notin> vars_term l" by auto
    from Fr xr have "\<exists> c. wf_elem c \<and> arc_pos c \<and> (\<forall> \<alpha>. wf_ass \<alpha> \<longrightarrow> pi\<guillemotleft>\<alpha>,r\<guillemotright> = c * \<alpha> x)"
    proof (induction r)
      case (Var y)
      from Var.prems have xy: "x = y" by simp
      show ?case
        by (rule exI[of _ one], unfold xy, auto simp: wf_ass_def arc_pos_one)
    next
      case (Fun f ss)
      from Fun.prems(1) have "(f,length ss) \<in> F" by simp
      with F_unary[OF this] Fun.prems(2) obtain s where ss: "ss = [s]"
        and F: "(f,Suc 0) \<in> F"
        by (cases ss, auto)
      with Fun.prems have "funas_term s \<subseteq> F" "x \<in> vars_term s" by auto
      from Fun.IH[OF _ this, unfolded ss] obtain c where c: "wf_elem c"
        and ac: "arc_pos c"
        and eval: "\<And> \<alpha>. wf_ass \<alpha> \<Longrightarrow> (pi\<guillemotleft>\<alpha>,s\<guillemotright>) = c * \<alpha> x" by auto
      obtain d ds where "pi (f,Suc 0) = (d,ds)" by force
      with pi_mono[OF F] obtain d where pi: "pi (f, length [s]) = (0,[d])" by (cases ds, auto)      
      from wf_pi[THEN conjunct1, rule_format, of "(f,Suc 0)"] pi 
      have d: "wf_elem d" by auto
      from wf_pi[THEN conjunct2, rule_format, of "(f,Suc 0)"] pi
        arc_pos_zero[OF mode] have ad: "arc_pos d" by auto
      show ?case unfolding ss
      proof (rule exI[of _ "d * c"], intro conjI allI impI)
        show "wf_elem (d * c)" using d c by simp
      next
        show "arc_pos (d * c)" using ad ac d c by simp
      next
        fix \<alpha> :: "('v,'a)p_ass"
        assume wf: "wf_ass \<alpha>"
        hence x: "wf_elem (\<alpha> x)" unfolding wf_ass_def by auto
        have "(pi\<guillemotleft>\<alpha>,Fun f [s]\<guillemotright>) = max0 (0 + (d * (c * \<alpha> x) + 0))"
          unfolding eval_termI.simps Let_def pi split 
          using eval[OF wf] by simp
        also have "... = d * (c * \<alpha> x)" 
          unfolding max0_npsm_id[OF mode] using d c x by simp
        finally show "(pi\<guillemotleft>\<alpha>,Fun f [s]\<guillemotright>) = (d * c) * \<alpha> x"
          using d c x by simp
      qed
    qed
    then obtain c where c: "wf_elem c" and ac: "arc_pos c" and
      evalr: "\<And> \<alpha>. wf_ass \<alpha> \<Longrightarrow> (pi\<guillemotleft>\<alpha>,r\<guillemotright>) = c * \<alpha> x"
      by auto
    let ?d = "default_ass :: ('v,'a)p_ass"
    from wf_pos_apos_default_ass
    have wd: "wf_ass ?d" and pd: "pos_ass ?d" and ad: "apos_ass ?d" by auto
    {
      fix e :: 'a
      assume we: "wf_elem e" and ae: "arc_pos e" and pe: "e \<ge> 0"
      let ?e = "\<lambda> y. if y = x then e else ?d y"
      from wd we have wae: "wf_ass ?e" unfolding wf_ass_def by auto
      from pd pe have pae: "pos_ass ?e" unfolding pos_ass_def by auto
      from ad ae have aae: "apos_ass ?e" unfolding apos_ass_def by auto
      have id: "(pi\<guillemotleft>?d,l\<guillemotright>) = (pi\<guillemotleft>?e,l\<guillemotright>)"
        by (rule eval_term_subst_eq, insert xl, auto) 
      have "... \<ge> pi\<guillemotleft>?e,r\<guillemotright>" using  ns[unfolded inter_ns_def] wae pae aae by auto
      also have "... = c * e" unfolding evalr[OF wae] by simp
      finally have "(pi\<guillemotleft>?d,l\<guillemotright>) \<ge> c * e" unfolding id .
    }
    with not_all_ge[OF mode wf_terms[OF wd] c ac] show False
      by blast
  qed
qed    
end

fun
  check_polo_sI
where
  "check_polo_sI mode ze on pl ti mz geg gt pi (s,t) = do {
     let left = PleftI ze on pl ti mz pi s;
     let right = PrightI ze on pl ti mz pi t;
     check_lpoly_sI mode ze on geg gt left right
       <+? (\<lambda>e. ''could not ensure '' +#+ shows s +@+ '' > '' +#+ shows t +@+ shows_nl +@+ e)
   }"

context linear_poly_order
begin
definition coeffs_of_constraint :: "('f,'v)rule \<Rightarrow> ('a \<times> 'a) list"
  where "coeffs_of_constraint st \<equiv> [(a,b). a \<leftarrow> coeffs_of_lpoly (Pleft (fst st)), b \<leftarrow> coeffs_of_lpoly (Pright (snd st))]"

abbreviation check_polo_s_constrained :: "('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> ('f, 'v::show) rule \<Rightarrow> shows check"
where "check_polo_s_constrained weak_gt st \<equiv> if (\<forall> (a,b) \<in> set (coeffs_of_constraint st). wf_elem a \<longrightarrow> wf_elem b \<longrightarrow> weak_gt a b \<longrightarrow> gt a b) then check_polo_sI plus_single_mono zero one plus times max0 geq weak_gt pi st else error (shows ''internal error in check_polo_s'')" 

(* TODO: move *)
lemma wf_lpoly_coeff: assumes wf: "wf_lpoly p" and mem: "a \<in> set (coeffs_of_lpoly p)" shows "wf_elem a"
proof -
  obtain b as where p: "p = LPoly b as" by (cases p, auto)
  note wf = wf[unfolded p, simplified]
  show ?thesis
  proof (cases "a = b")
    case True 
    with wf show ?thesis by auto
  next
    case False
    with mem[unfolded p] have mem: "a \<in> set (coeffs_of_pvars as)" by auto
    from wf have "wf_pvars as" ..
    with mem show ?thesis
    proof (induct as)
      case Nil
      thus ?case unfolding coeffs_of_pvars_def by auto
    next
      case (Cons b bs)
      show ?case
      proof (cases "a = snd b")
        case True
        with Cons(3) show ?thesis unfolding wf_pvars_def  by auto
      next
        case False
        with Cons(2) have mem: "a \<in> set (coeffs_of_pvars bs)" unfolding coeffs_of_pvars_def by auto
        from False Cons(3) have wf: "wf_pvars bs" unfolding wf_pvars_def by auto
        from Cons(1)[OF mem wf] show ?thesis .
      qed
    qed
  qed
qed
        

lemma wf_elem_coeff_left: assumes "a \<in> set (coeffs_of_lpoly (Pleft t))" shows "wf_elem a"
proof -
  have "wf_ass default_ass" and "pos_ass default_ass" by auto
  from Pleft_both[OF this] have "wf_lpoly (Pleft t)" ..
  from wf_lpoly_coeff[OF this] show ?thesis using assms by auto
qed

lemma wf_elem_coeff_right: assumes "a \<in> set (coeffs_of_lpoly (Pright t))" shows "wf_elem a"
proof -
  have "wf_ass default_ass" and "pos_ass default_ass" by auto
  from Pright_both[OF this] have "wf_lpoly (Pright t)" by auto
  from wf_lpoly_coeff[OF this] show ?thesis using assms by auto
qed
end

sublocale linear_poly_order \<subseteq> redtriple_impl "inter_s :: ('f :: {show,key},'v :: show)trs" inter_ns inter_ns "check_polo_s_constrained weak_gt" check_polo_ns check_polo_ns
proof
  fix st :: "('f,'v)rule"
  assume check: "isOK(check_polo_s_constrained weak_gt st)"
  obtain s t where st: "st = (s,t)" by force
  note check = check[unfolded st]
  from check have weak_gt: "\<And> a b. (a,b) \<in> set (coeffs_of_constraint (s,t)) \<Longrightarrow> wf_elem a \<Longrightarrow> wf_elem b \<Longrightarrow> weak_gt a b \<Longrightarrow> a > b"  by auto
  from check have "isOK(check_polo_sI plus_single_mono 0 1 op + op * max0 geq weak_gt pi (s,t))" by simp
  show "st \<in> inter_s" unfolding st inter_s_def 
  proof (intro CollectI, simp, intro allI) 
    fix \<alpha>
    show "wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,s\<guillemotright> > pi\<guillemotleft>\<alpha>,t\<guillemotright>)" (is "?ass \<longrightarrow> ?goal")
    proof
      assume ?ass
      from check have check1: "isOK(check_lpoly_s weak_gt one (Pleft s) (Pright t))" by (auto simp: Let_def)
      have strict: "(Pleft s, Pright t) \<in> poly_s" 
        by (rule check_lpoly_s_sound[OF check1 _ _ weak_gt], auto simp: coeffs_of_constraint_def wf_elem_coeff_left wf_elem_coeff_right)  
      have ge1: "pi\<guillemotleft>\<alpha>,s\<guillemotright> \<ge> eval_lpoly \<alpha> (Pleft s)" by (rule Pleft_sound, auto simp: `?ass`)
      have gt: "\<dots> > eval_lpoly \<alpha> (Pright t)" using strict unfolding poly_s_def using `?ass` by auto
      have ge2: "\<dots> \<ge> pi\<guillemotleft>\<alpha>,t\<guillemotright>" by (rule Pright_sound, auto simp: `?ass`)
      show ?goal 
        by (rule compat_orig[OF ge1 compat2[OF gt ge2]], auto simp: `?ass` wf_eval_lpoly) 
    qed
  qed
next
  fix st :: "('f,'v)rule"
  assume check: "isOK(check_polo_ns st)"
  show "st \<in> inter_ns"
  proof (cases st)
    case (Pair s t)
    thus ?thesis unfolding inter_ns_def
    proof (simp, intro allI)
      fix \<alpha>
      show "wf_ass \<alpha> \<and> pos_ass \<alpha> \<and> apos_ass \<alpha> \<longrightarrow> (pi\<guillemotleft>\<alpha>,s\<guillemotright> \<ge> pi\<guillemotleft>\<alpha>,t\<guillemotright>)" (is "?ass \<longrightarrow> ?goal")
      proof
        assume ?ass
        from check Pair have check1: "isOK(check_lpoly_ns one (Pleft s) (Pright t))" by (auto simp: Let_def)
        have strict: "(Pleft s, Pright t) \<in> poly_ns" 
          by (rule check_lpoly_ns_sound, rule check1, auto)
        have ge1: "pi\<guillemotleft>\<alpha>,s\<guillemotright> \<ge> eval_lpoly \<alpha> (Pleft s)" by (rule Pleft_sound, auto simp: `?ass`)
        have gt: "\<dots> \<ge> eval_lpoly \<alpha> (Pright t)" using strict unfolding poly_ns_def using `?ass` by auto
        have ge2: "\<dots> \<ge> pi\<guillemotleft>\<alpha>,t\<guillemotright>" by (rule Pright_sound, auto simp: `?ass`)
        show ?goal by (rule geq_trans, rule ge1, rule geq_trans, rule gt, rule ge2, auto simp: `?ass` wf_eval_lpoly)
      qed
    qed
  qed
qed


record 'a poly_c =
  wf :: "'a \<Rightarrow> bool"
  zero :: 'a
  one :: 'a
  default :: 'a
  plus :: "'a \<Rightarrow> 'a \<Rightarrow> 'a"
  times :: "'a \<Rightarrow> 'a \<Rightarrow> 'a"
  max0 :: "'a \<Rightarrow> 'a"
  ge :: "'a \<Rightarrow> 'a \<Rightarrow> bool"
  gt :: "'a \<Rightarrow> 'a \<Rightarrow> bool"
  mono :: bool
  description :: string
  arc_pos :: "'a \<Rightarrow> bool"
  check_mono :: "'a \<Rightarrow> bool"
  max :: "'a \<Rightarrow> 'a \<Rightarrow> 'a"
  derive_complexity :: "'a \<Rightarrow> shows + nat"


hide_const(open) wf zero one default plus times max0 ge gt mono description arc_pos check_mono max derive_complexity

abbreviation
  check_polo_s_comp :: "('a::show) poly_c \<Rightarrow> ('f, 'a) lpoly_inter \<Rightarrow> ('f :: show, 'v::show) rule \<Rightarrow> shows check"
where
  "check_polo_s_comp c \<equiv>
     check_polo_sI (poly_c.mono c) (poly_c.zero c) (poly_c.one c) (poly_c.plus c) (poly_c.times c) (poly_c.max0 c) (poly_c.ge c) (poly_c.gt c)"

abbreviation check_polo_ns_comp :: "('a::show)poly_c \<Rightarrow> ('f,'a)lpoly_inter \<Rightarrow> ('f::show,'v::show)rule \<Rightarrow> shows check"
where "check_polo_ns_comp c \<equiv> check_polo_nsI (poly_c.zero c) (poly_c.one c) (poly_c.plus c) (poly_c.times c) (poly_c.max0 c) (poly_c.ge c)"

abbreviation check_polo_coeffs_comp :: "('a :: show)poly_c \<Rightarrow> ('f::show,'a)lpoly_interL \<Rightarrow> shows check"
where "check_polo_coeffs_comp c \<equiv> check_lpoly_coeffs (poly_c.wf c) (poly_c.arc_pos c) (\<lambda> x. (poly_c.ge c) x (poly_c.zero c))"

abbreviation check_polo_mono_comp :: "('a :: show)poly_c \<Rightarrow> ('f :: show, 'a)lpoly_interL \<Rightarrow> shows check" 
  where "check_polo_mono_comp c \<equiv> check_poly_monoI (poly_c.zero c) (poly_c.ge c) (poly_c.check_mono c)"

context linear_poly_order_carrier
begin
definition bound_entry :: "'a \<Rightarrow> 'a \<Rightarrow> 'a \<times> 'a list \<Rightarrow> bool"
  where "bound_entry bcoeff bconst \<equiv> \<lambda> (a,as). geq bconst a \<and> (\<forall> a \<in> set as. geq bcoeff a \<or> geq one a)"

definition bound_entry_strict :: "'a \<Rightarrow> 'a \<Rightarrow> 'a \<times> 'a list \<Rightarrow> bool"
  where "bound_entry_strict bcoeff bconst \<equiv> \<lambda> (a,as). geq bconst a \<and> (\<forall> a \<in> set as. geq bcoeff a)"

fun pow :: "'a \<Rightarrow> nat \<Rightarrow> 'a"
where "pow a 0 = one"
    | "pow a (Suc n) = times (pow a n) a"

lemma wf_pow[simp]: "wf_elem a \<Longrightarrow> wf_elem (pow a n)"
  by (induct n, auto)

lemma pow_ge_zero[intro]: assumes a: "geq a 0"
  and wf: "wf_elem a"
  shows "pow a n \<ge> 0"
proof (induct n)
  case (Suc n)
  have "pow a (Suc n) \<ge> 0 * a" unfolding pow.simps
    by (rule times_left_mono[OF a Suc], insert wf, auto)
  thus ?case using wf by auto
qed auto

lemma pow_commute_left: assumes wf: "wf_elem a"
  shows "pow a n * a = a * pow a n"
proof (induct n)
  case 0
  show ?case unfolding pow.simps using wf by simp
next
  case (Suc n)
  have "pow a (Suc n) * a = (a * pow a n) * a"
    unfolding pow.simps unfolding Suc ..
  also have "... = a * (pow a n * a)"
    by (rule times_ass[symmetric, OF _ wf_pow], insert wf, blast+)
  finally show ?case by simp
qed

lemma pow_ge_1: assumes 1: "geq 1 a"
  and 0: "geq a 0"
  and wf: "wf_elem a"
  shows "geq 1 (pow a n)"
proof (induct n)
  case 0
  show ?case by simp
next
  case (Suc n)
  have "geq 1 (pow a (Suc n)) = geq (1 * 1) (pow a n * a)" by simp
  also have "..."
    by (rule geq_trans[OF times_right_mono[OF one_geq_zero 1] times_left_mono[OF 0 Suc]], insert wf, auto)
  thus ?case by simp
qed

lemma list_sum_append[simp]: assumes as: "\<And> a. a \<in> set as \<Longrightarrow> wf_elem a"
  and bs: "\<And> b. b \<in> set bs \<Longrightarrow> wf_elem b"
  shows "list_sum (as @ bs) = list_sum as + list_sum bs"
  using as 
  by (induct as, insert bs, auto)
end

context linear_poly_order
begin

lemma bound_eval_term_main: assumes bF: "\<And> fn. fn \<in> F \<Longrightarrow> bound_entry bc bcv (pi fn)"
  and bc: "geq bc 0"
  and bcv: "geq bcv 0"
  and wfbcv: "wf_elem bcv"
  and wfbc: "wf_elem bc"
  and t: "funas_term (t :: ('f,'v)term) \<subseteq> F"  
  shows "\<exists> bs. length bs \<le> size t \<and> (\<forall> b \<in> set bs. wf_elem b \<and> (\<exists> n < size t. b = pow bc n * bcv)) \<and> (list_sum bs \<ge> pi\<guillemotleft>zero_ass,t\<guillemotright>)"
  using t
proof (induct t)
  case (Var x)
  show ?case
    by (rule exI[of _ Nil], auto simp: zero_ass_def)
next
  case (Fun f ts)
  let ?e = "\<lambda> t. pi\<guillemotleft>zero_ass,t\<guillemotright>"
  let ?n = "length ts"
  obtain a as where pi: "pi (f, ?n) = (a,as)" by force
  let ?map = "map (\<lambda> at. fst at * ?e (snd at)) (zip as ts)"
  let ?nn = "length (zip as ts)"
  note wf = wf_pi[THEN conjunct1, rule_format, of "(f, ?n)", unfolded pi fst_conv snd_conv]
  from wf have wfa: "wf_elem a" and wfas: "\<And> a. a \<in> set as \<Longrightarrow> wf_elem a"
    and as0: "\<And> a. a \<in> set as \<Longrightarrow> geq a 0" by auto
  from Fun(2) have "(f,?n) \<in> F" by auto
  from bF[OF this, unfolded pi bound_entry_def split] 
  have b_a: "bcv \<ge> a" and b_as: "\<And> a. a \<in> set as \<Longrightarrow> (bc \<ge> a) \<or> (1 \<ge> a)" by auto
  obtain Pow where Pow: "Pow = (\<lambda> b (t :: ('f,'v)term). (\<exists> n < size t. b = pow bc n * bcv))" by auto
  let ?P = "\<lambda> bs i. length bs \<le> size (ts ! i) \<and> (\<forall> b \<in> set bs. Pow b (Fun f ts)) \<and> (\<forall> b \<in> set bs. wf_elem b) \<and> (list_sum bs \<ge> ?map ! i)"
  {
    fix i
    assume i: "i < ?nn"
    hence ias: "i < length as" and its: "i < ?n" by auto
    from ias have asi: "as ! i \<in> set as" by auto
    from its have tsi: "ts ! i \<in> set ts" by auto
    from Fun(2) tsi have "funas_term (ts ! i) \<subseteq> F" by auto
    note IH = Fun(1)[OF tsi this]
    from as0[OF asi] have asi0: "as ! i \<ge> 0" .
    from b_as[OF asi] have b_asi: "(bc \<ge> as ! i) \<or> (1 \<ge> as ! i)" .
    from wfas[OF asi] have wfasi: "wf_elem (as ! i)" .
    from i have id: "?map ! i = (as ! i * ?e (ts ! i))" by auto
    have wf_map: "wf_elem (?map ! i)" unfolding id using wfasi by auto
    from IH obtain bs where len: "length bs \<le> size (ts ! i)"
      and bs: "\<And> b. b \<in> set bs \<Longrightarrow> Pow b (ts ! i)"
      and wf_bs: "\<And> b. b \<in> set bs \<Longrightarrow> wf_elem b"
      and ge2: "list_sum bs \<ge> ?e (ts ! i)" unfolding Pow by auto
    let ?bs = "map (times bc) bs"
    have lsum: "list_sum ?bs = bc * list_sum bs" using wf_bs
    proof (induct bs)
      case Nil show ?case by (simp add: wfbc)
    next
      case (Cons b bs)
      hence wfb: "wf_elem b" and wfbs: "\<And> b. b \<in> set bs \<Longrightarrow> wf_elem b" and 
        IH: "list_sum (map (times bc) bs) = times bc (list_sum bs)" by auto
      show ?case unfolding List.map.simps list_sumI.simps
        unfolding IH
        by (rule distrib_left[symmetric], insert wfb wfbc wfbs, auto)
    qed
    {
      fix bbc
      assume "bbc \<in> set ?bs"
      then obtain b where bbc: "bbc = bc * b" and b: "b \<in> set bs" by auto
      hence "wf_elem bbc" using wfbc wf_bs by auto
    } note wf_bs2 = this
    have size_tsi: "size (ts ! i) < size (Fun f ts)" 
      by (rule supt_size, insert tsi, auto)
    {
      fix bbc
      assume "bbc \<in> set ?bs"
      then obtain b where bbc: "bbc = bc * b" and b: "b \<in> set bs" by auto
      from bs[OF b] obtain n where b: "b = pow bc n * bcv" and n: "n < size (ts ! i)" unfolding Pow by auto
      have bbc: "bbc = pow bc (Suc n) * bcv" unfolding bbc b pow.simps pow_commute_left[OF wfbc]
        by (rule times_ass, insert wfbc wfbcv, auto)
      from size_tsi n have "Suc n < size (Fun f ts)" by auto
      with bbc have "\<exists> n < size (Fun f ts). bbc = pow bc n * bcv" by blast
      hence "Pow bbc (Fun f ts)" unfolding Pow .
    } note bs2 = this
    from b_asi have "\<exists> bs2. ?P bs2 i"
    proof
      assume b_asi: "bc \<ge> as ! i"
      have ge: "(bc * ?e (ts ! i)) \<ge> ?map ! i" unfolding id
        by (rule times_left_mono[OF _ b_asi wfbc wfasi], auto)
      have ge2: "list_sum ?bs \<ge> (times bc (?e (ts ! i)))" unfolding lsum
        by (rule times_right_mono[OF bc ge2 wfbc wf_list_sum], insert wf_bs wfbcv, auto)
      have ge: "list_sum ?bs \<ge> ?map ! i"
        by (rule geq_trans[OF ge2 ge wf_list_sum _ wf_map], insert wf_bs2 wfbc, auto)
      show "\<exists> bs2. ?P bs2 i"
        by (rule exI[of _ ?bs], unfold length_map, insert len ge bs2 wf_bs2, blast)
    next
      assume asi: "1 \<ge> as ! i"
      show "\<exists> bs2. ?P bs2 i" 
      proof (rule exI[of _ bs], rule conjI[OF len conjI[OF ballI conjI[OF ballI[OF wf_bs]]]])
        fix b
        assume "b \<in> set bs"
        from bs[OF this, unfolded Pow] obtain n where n: "n < size (ts ! i)" and b: "b = pow bc n * bcv"
          by auto 
        with size_tsi have "n < size (Fun f ts)" by auto
        with b show "Pow b (Fun f ts)" unfolding Pow by blast
      next
        have "(1 * ?e (ts ! i)) \<ge> ?map ! i" unfolding id
          by (rule times_left_mono[OF _ asi _ wfasi], auto)
        hence ge: "?e (ts ! i) \<ge> ?map ! i" by simp
        show "list_sum bs \<ge> ?map ! i"
          by (rule geq_trans[OF ge2 ge wf_list_sum _ wf_map], insert wf_bs, auto)
      qed
    qed
    note this `wf_elem (?map ! i)`
  }
  hence bsi: "\<forall> i. \<exists> bs. i < ?nn \<longrightarrow> ?P bs i" and wfmi: "\<And> i. i < ?nn \<Longrightarrow> wf_elem (?map ! i)"
    by blast+
  from choice[OF bsi] obtain bs where bs: "\<And> i. i < ?nn \<Longrightarrow> ?P (bs i) i" by auto
  let ?bs' = "concat (map bs [0 ..< ?nn])"
  let ?bs = "bcv # ?bs'"
  have "length ?bs' \<le> listsum (map size ts)" unfolding length_concat
    unfolding map_map o_def 
    using bs[THEN conjunct1]
  proof (induct ts arbitrary: as bs)
    case Nil 
    show ?case by auto
  next
    case (Cons t ts aas bs)
    note oCons = this
    show ?case
    proof (cases aas)
      case Nil thus ?thesis by auto
    next
      case (Cons a as)
      note oCons = oCons[unfolded Cons]
      note IH = oCons(1)[of as "\<lambda> i. bs (Suc i)"] 
      note pre = oCons(2)
      from pre[of "0 :: nat"] have le: "length (bs 0) \<le> size t" by auto
      {
        fix i
        assume i: "i < length (zip as ts)"
        hence i: "Suc i < length (zip (a # as) (t # ts))" by simp
        from pre[OF this] have "length (bs (Suc i)) \<le> size (ts ! i)" by auto
      }
      from IH[OF this] have 
        le2: " (\<Sum>x\<leftarrow>[0..<length (zip as ts)]. length (bs (Suc x))) \<le> listsum (map size ts)" 
        by simp
      have "(\<Sum>x\<leftarrow>[0..<length (zip aas (t # ts))]. length (bs x)) = 
        (\<Sum>x\<leftarrow>[0..< Suc (length (zip as ts))]. length (bs x))" 
        unfolding Cons by simp
      also have "... = length (bs 0) + (\<Sum>x\<leftarrow>[0..< length (zip as ts)]. length (bs (Suc x)))" 
        unfolding map_upt_Suc by simp
      also have "... \<le> size t + listsum (map size ts)" using le le2 by auto
      finally
      show ?thesis by simp
    qed
  qed
  also have "... \<le> list_size size ts" 
    unfolding list_size_conv_listsum by simp       
  finally have len: "length ?bs \<le> size (Fun f ts)" by auto
  {
    fix b
    assume "b \<in> set ?bs"
    hence "b = bcv \<or> b \<in> set ?bs'" by auto
    hence "wf_elem b \<and> Pow b (Fun f ts) \<and> (b \<ge> 0)" 
    proof
      assume b: "b = bcv"
      show ?thesis
        unfolding b Pow
        by (rule conjI[OF wfbcv], rule conjI, rule exI[of _ "0 :: nat"], insert wfbcv bcv, auto)
    next
      assume "b \<in> set ?bs'"
      then obtain i where i: "i < ?nn" and b: "b \<in> set (bs i)" by auto
      from i have ias: "i < length as" by auto
      hence asi: "as ! i \<in> set as" by auto
      from bs[OF i] b have "Pow b (Fun f ts)" by auto
      from this[unfolded Pow] obtain n where bpow: "b = pow bc n * bcv" by auto
      have "b \<ge> 0 * bcv" unfolding bpow
        by (rule times_left_mono[OF bcv pow_ge_zero], insert wfbc wfbcv bc, auto)
      hence "b \<ge> 0" using wfbcv by auto
      with bs[OF i] b show ?thesis by auto
    qed
  } note bs_wf_pow = this
  show ?case
  proof (rule exI[of _ ?bs], intro conjI, rule len, intro ballI)
    fix b
    assume "b \<in> set ?bs"
    from bs_wf_pow[OF this]
    show "wf_elem b \<and> (\<exists> n < size (Fun f ts). b = pow bc n * bcv)"
      unfolding Pow by simp
  next
    have wf_bs: "wf_elem (list_sum ?bs)"
      using bs_wf_pow by auto
    let ?sum = "a + list_sum ?map"
    have e: "?e (Fun f ts) = max0 ?sum" unfolding eval_termI.simps Let_def pi split ..    
    {
      fix p
      assume "p \<in> set ?map"
      from this[unfolded set_map set_zip]
      obtain i where i: "i < length as" and p: "p = (as ! i * ?e (ts ! i))" by auto
      from i have asi: "as ! i \<in> set as" by auto
      from wfas[OF asi] have "wf_elem p" unfolding p by auto
    } note wf_prods = this
    hence wf_sum: "wf_elem (list_sum ?map)" by auto
    with wfa have wf_asum: "wf_elem ?sum" by auto
    have "list_sum ?bs = max0 (list_sum ?bs)"
      by (rule max0_id_pos[symmetric, OF pos_list_sum wf_bs], insert bs_wf_pow, auto)
    also have "... \<ge> ?e (Fun f ts)" unfolding e
    proof (rule max0_mono[OF _ wf_bs wf_asum])
      have wf_bs': "wf_elem (list_sum ?bs')" using bs_wf_pow by auto
      have id: "list_sum ?bs = bcv + list_sum ?bs'" by simp
      have ge1: "... \<ge> a + list_sum ?bs'" 
        by (rule plus_left_mono[OF b_a wfbcv wfa wf_bs'])
      have ge2: "... \<ge> ?sum"
      proof (rule plus_right_mono[OF _ wfa wf_bs' wf_sum])
        obtain z where z: "z = zip as ts" by auto
        obtain f where f: "f = (\<lambda> (at :: 'a \<times> ('f,'v)term). fst at * ?e (snd at))" by auto
        from wf_prods have "\<And> p. p \<in> set (map f z) \<Longrightarrow> wf_elem p"
          unfolding f z .
        thus "list_sum ?bs' \<ge> list_sum ?map" using bs[THEN conjunct2, THEN conjunct2] unfolding z[symmetric] f[symmetric]
        proof (induct z arbitrary: bs)
          case Nil
          thus ?case by simp
        next
          case (Cons a z bs)
          note IH = Cons(1)[of "\<lambda> i. bs (Suc i)"]
          note wfaz = Cons(2)
          hence wfa: "wf_elem (f a)" and wfz: "\<And> p. p \<in> set (map f z) \<Longrightarrow> wf_elem p" by auto
          note bs = Cons(3)
          {
            fix i
            assume "i < length z"
            hence "Suc i < length (a # z)" by auto
            from bs[OF this]
            have "(\<forall> a \<in> set (bs (Suc i)). wf_elem a) \<and> (list_sum (bs (Suc i)) \<ge> map f z ! i)" by auto
          } note bsz = this
          let ?cc = "concat (map (\<lambda> i. bs (Suc i)) [0 ..< length z])"
          from IH[OF wfz bsz] have ge2: "list_sum ?cc \<ge> list_sum (map f z)" .
          from bs[of "0 :: nat"] have wfbs0: "\<And> a. a \<in> set (bs 0) \<Longrightarrow> wf_elem a"
            and ge1: "list_sum (bs 0) \<ge> f a" by auto
          let ?start = "list_sum (concat (map bs [0..<length (a # z)]))"
          have "list_sum (concat (map bs [0..<length (a # z)]))
            = list_sum (concat (map bs [0..< Suc (length z)]))" by simp
          also have "... = list_sum (bs 0 @ ?cc)" 
            unfolding map_upt_Suc by auto
          also have "... = list_sum (bs 0) + list_sum ?cc"
            by (rule list_sum_append[OF wfbs0], insert bsz, auto)
          finally have id: "?start = list_sum (bs 0) + list_sum ?cc" "list_sum (map f (a # z)) = f a + list_sum (map f z)" by auto
          show ?case unfolding id
            by (rule plus_left_right_mono[OF ge1 ge2], insert wfbs0 wfa wfz bsz, auto)
        qed
      qed
      show "list_sum ?bs \<ge> ?sum" unfolding id
        by (rule geq_trans[OF ge1 ge2 _ _ wf_asum], insert wfa wfbcv bs_wf_pow, auto)
    qed
    finally
    show "list_sum ?bs \<ge> ?e (Fun f ts)" .
  qed
qed
end    

locale complexity_linear_poly_order = linear_poly_order + complexity_linear_poly_order_carrier
begin

lemma bound_eval_term_derivational: assumes bF: "\<And> fn. fn \<in> set F \<Longrightarrow> bound_entry bc bcv (pi fn)"
  and bc: "bc \<ge> zero"
  and bcv: "bcv \<ge> zero"
  and wfbcv: "wf_elem bcv"
  and wfbc: "wf_elem bc"
  and mono: "\<And> n m. n \<le> m \<Longrightarrow> g n \<le> g m"
  and bG: "\<And> n. bound (pow bc n * bcv) \<le> g n"
  and deriv: "t \<in> terms_of_nat (Derivational_Complexity F) N"
  shows "g N * N \<ge> bound (pi\<guillemotleft>zero_ass,t\<guillemotright>)"
proof -
  from deriv have t: "funas_term t \<subseteq> set F" and N: "size t \<le> N" by auto
  from bound_eval_term_main[OF bF bc bcv wfbcv wfbc t]
  obtain bs where len: "length bs \<le> size t"
    and bs: "\<And> b. b \<in> set bs \<Longrightarrow> wf_elem b \<and> (\<exists> n < size t. b = pow bc n * bcv)"
    and ge: "list_sum bs \<ge> pi\<guillemotleft>zero_ass,t\<guillemotright>"
    by auto
  {
    fix b
    assume "b \<in> set bs"
    from bs[OF this] obtain n where wf: "wf_elem b" and n: "n < size t" and b: "b = pow bc n * bcv"
      by auto
    from bG[of n] have ge: "g n \<ge> bound b" unfolding b .
    from n N have "n \<le> N" by auto
    from mono[OF this] ge wf have "g N \<ge> bound b" "wf_elem b" by auto
  } note bs = this  
  hence "bound (list_sum bs) \<le> g N * length bs"
  proof (induct bs)
    case (Cons b bs)
    from Cons have ge1: "g N \<ge> bound b" by auto
    from Cons have ge2: "g N * length bs \<ge> bound (list_sum bs)" by auto
    from Cons have id: "bound (list_sum (b # bs)) \<le> bound b + bound (list_sum bs)" using bound_plus by simp
    also have "... \<le> g N + g N * length bs" using ge1 ge2 by arith
    finally show ?case by simp
  qed simp
  also have "... \<le> g N * size N"
    using len N by auto
  finally show ?thesis using bound_mono[OF ge] bs by auto
qed    

lemma bound_eval_term_runtime: assumes bC: "\<And> fn. fn \<in> set C \<Longrightarrow> bound_entry bc bcv (pi fn)"
  and bc: "bc \<ge> zero"
  and bcv: "bcv \<ge> zero"
  and wfbcv: "wf_elem bcv"
  and wfbc: "wf_elem bc"
  and bD: "\<And> fn. fn \<in> set D \<Longrightarrow> bound_entry_strict bd bdv (pi fn)"
  and bn: "\<And> fn. fn \<in> set D \<Longrightarrow> snd fn \<le> bn"
  and bd: "bd \<ge> zero"
  and bdv: "bdv \<ge> zero"
  and wfbdv: "wf_elem bdv"
  and wfbd: "wf_elem bd"
  and mono: "\<And> n m. n \<le> m \<Longrightarrow> g n \<le> g m"
  and bG: "\<And> n. bound (times bd  (pow bc n * bcv)) \<le> g n"
  and deriv: "t \<in> terms_of_nat (Runtime_Complexity C D) N"
  shows "bound bdv + g N * N * bn \<ge> bound (pi\<guillemotleft>zero_ass,t\<guillemotright>)"
proof -
  from deriv[simplified] obtain f ts where t: "t = Fun f ts" and f: "(f,length ts) \<in> set D" and 
    ts: "\<And> ti. ti \<in> set ts \<Longrightarrow> funas_term ti \<subseteq> set C" and size: "size t \<le> N" by (cases t, auto)
  obtain a as where pi: "pi (f,length ts) = (a,as)" by force
  from bn[OF f] have bn: "length ts \<le> bn" by simp
  from bD[OF f] have ge_a: "bdv \<ge> a" and ge_as: "\<And> a . a \<in> set as \<Longrightarrow> bd \<ge> a"  
    unfolding pi bound_entry_strict_def by auto
  from wf_pi[THEN conjunct1, rule_format, of "(f,length ts)", unfolded pi] 
  have wfa: "wf_elem a" and wfas: "\<And> a. a \<in> set as \<Longrightarrow> wf_elem a" and as0: "\<And> a. a \<in> set as \<Longrightarrow> a \<ge> 0" by auto
  let ?eval = "\<lambda> t. pi\<guillemotleft>zero_ass,t\<guillemotright>"
  let ?p = "\<lambda> at. fst at * ?eval (snd at)"
  let ?map = "map ?p (zip as ts)"
  obtain prods where prods: "prods = ?map" by auto
  {
    fix p
    assume "p \<in> set prods"
    from this[unfolded prods set_conv_nth] obtain i where ias: "i < length as" and its: "i < length ts"
      and p: "p = ?map ! i" by auto
    let ?asi = "as ! i"
    let ?tsi = "ts ! i"
    from p ias its have p: "p = (?asi * ?eval ?tsi)" by auto
    from ias have asi: "?asi \<in> set as" by auto
    from wfas[OF this] as0[OF this] ge_as[OF this] have wfasi: "wf_elem ?asi" and asi0: "?asi \<ge> 0" 
      and bd_ge: "bd \<ge> as ! i" .
    from its have tsi: "?tsi \<in> set ts" by auto
    from ts[OF tsi] have tsiC: "funas_term ?tsi \<subseteq> set C" .
    from tsi have "t \<rhd> ?tsi" unfolding t by auto
    from supt_size[OF this] size have size: "size ?tsi \<le> N" by simp
    from bound_eval_term_main[OF bC bc bcv wfbcv wfbc tsiC]
    obtain bs where len: "length bs \<le> size ?tsi"
      and bs: "\<And> b. b \<in> set bs \<Longrightarrow> wf_elem b \<and> (\<exists> n < size ?tsi. b = pow bc n * bcv)"
      and ge: "list_sum bs \<ge> ?eval ?tsi"
      by auto
    have bd_p: "bd * list_sum bs \<ge> p" unfolding p
      by (rule geq_trans[OF times_left_mono[OF geq_trans[OF ge pos_term] bd_ge wfbd wfasi] times_right_mono[OF asi0 ge]], insert bs wfasi wfbd, auto)
    from bs have gn_bd: "g N * length bs \<ge> bound (bd * list_sum bs)"
    proof (induct bs)
      case Nil
      show ?case by (simp add: wfbd)
    next
      case (Cons b bs)
      from Cons(2)[of b] obtain n where n: "n < size ?tsi" and b: "b = pow bc n * bcv" by auto
      from n size have n: "n \<le> N" by auto
      have "bound (bd * list_sum (b # bs)) = bound (bd * b + bd * list_sum bs)"
        using Cons(2) wfbd by simp
      also have "... \<le> bound (bd * b) + bound (bd * list_sum bs)"
        by (rule bound_plus, insert Cons(2) wfbd, auto)
      also have "... \<le> bound (bd * b) + g N * length bs"
        using Cons by auto
      also have "bound (bd * b) \<le> g n" unfolding b by (rule bG)
      also have "... \<le> g N"
        by (rule mono[OF n])
      finally show ?case by simp
    qed
    from le_trans[OF bound_mono[OF bd_p] gn_bd]
    have "bound p \<le> g N * length bs" using bs wfbd wfasi unfolding p by auto
    also have "... \<le> g N * N"
      using le_trans[OF len size]  by auto
    finally have main: "bound p \<le> g N * N" "wf_elem p" unfolding p using wfasi by auto
    have "p \<ge> 0 * 0" unfolding p
      by (rule geq_trans[OF times_left_mono[OF pos_term asi0]], insert wfasi, auto)
    hence "p \<ge> 0" by simp
    note main this
  } note bound_prods = this
  have "bound (?eval t) = bound (max0 (a + list_sum prods))"
    unfolding t using pi by (simp add: Let_def prods)
  also have "... \<le> bound (max0 ( bdv + list_sum prods))"
    by (rule bound_mono[OF max0_mono[OF plus_left_mono[OF ge_a]]], insert wfa wfbdv bound_prods, auto simp: wf_max0)
  also have "max0 (bdv + list_sum prods) = bdv + list_sum prods"
  proof (rule max0_id_pos)
    have "bdv + list_sum prods \<ge> 0 + 0"
      by (rule geq_trans[OF plus_right_mono[OF pos_list_sum[OF ballI]] plus_left_mono[OF bdv]], insert wfa wfbdv bound_prods, auto)
    thus "bdv + list_sum prods \<ge> 0" by simp
  qed (insert bound_prods wfbdv, auto)
  also have "bound (bdv + list_sum prods) \<le> bound bdv + bound (list_sum prods)" 
    by (rule bound_plus, insert bound_prods wfbdv, auto)
  also have "bound (list_sum prods) \<le> g N * N * length (prods)"
    using bound_prods(1) bound_prods(2)
  proof (induct prods)
    case Nil
    show ?case by simp
  next
    case (Cons p prods)
    have "bound (list_sum (p # prods)) \<le> bound p + bound (list_sum prods)"
      unfolding list_sumI.simps 
      by (rule bound_plus, insert Cons(3), auto)
    also have "bound (list_sum prods) \<le> g N * N * length prods" 
      using Cons by auto
    also have "bound p \<le> g N * N" using Cons(2)[of p] by auto
    finally show ?case by simp
  qed
  also have "... \<le> g N * N * bn" 
  proof -
    from prods bn have "length prods \<le> bn" by simp
    thus ?thesis by simp
  qed
  finally show "bound (?eval t) \<le> bound bdv + g N * N * bn" by auto
qed
end

context mono_linear_poly_order_carrier
begin
lemma poly_c_max_list: assumes init: "wf_elem init"
  shows "\<lbrakk> \<And> a. a \<in> set as \<Longrightarrow> wf_elem a\<rbrakk> \<Longrightarrow> 
  wf_elem (foldr max as init) \<and> geq (foldr max as init) init \<and> (\<forall> a \<in> set as. geq (foldr max as init) a)"
proof (induct as)
  case Nil
  thus ?case using init by simp
next
  case (Cons a as)
  let ?w = "wf_elem"
  from Cons(2) have a: "?w a" and as: "\<And> a. a \<in> set as \<Longrightarrow> ?w a" by auto
  let ?mf = "foldr max as init"
  let ?mm = "max a ?mf"
  note IH = Cons(1)[OF as]
  from IH have mf: "?w ?mf" and ge_init: "geq ?mf init" and ge_as: "\<And> a. a \<in> set as \<Longrightarrow> geq ?mf a" by auto  
  from wf_max[OF a mf] have mm: "?w ?mm" .
  from max_left[OF a mf] have mm_a: "geq ?mm a" .
  from max_right[OF a mf] have mm_mf: "geq ?mm ?mf" .
  note mm_mf = geq_trans[OF mm_mf _ mm mf]
  from mm_mf[OF ge_init init] have mm_init: "geq ?mm init" .
  from mm_mf[OF ge_as as] have mm_as: "\<And> a. a \<in> set as \<Longrightarrow> geq ?mm a" .
  from mm mm_init mm_as mm_a
  show ?case by simp
qed
end

definition poly_c_max_inter_bcoeff :: "'a \<Rightarrow> 'a \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> ('f \<times> nat)list \<Rightarrow> ('f,'a)lpoly_inter \<Rightarrow> 'a"
  where "poly_c_max_inter_bcoeff ze on geq ma F pi \<equiv> foldr ma (concat (map (\<lambda> fn. filter (\<lambda> b. \<not> (geq on b)) (snd (pi fn))) F)) ze"

definition poly_c_max_inter_bcoeff_strict :: "'a \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> ('f \<times> nat)list \<Rightarrow> ('f,'a)lpoly_inter \<Rightarrow> 'a"
  where "poly_c_max_inter_bcoeff_strict ze ma F pi \<equiv> foldr ma (concat (map (\<lambda> fn. snd (pi fn)) F)) ze"

definition poly_c_max_inter_bconst :: "'a \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> ('f \<times> nat)list \<Rightarrow> ('f,'a)lpoly_inter \<Rightarrow> 'a"
  where "poly_c_max_inter_bconst ze ma F pi \<equiv> foldr ma (map (\<lambda> fn. fst (pi fn)) F) ze"

context complexity_linear_poly_order
begin

lemma poly_c_max_inter_bcoeff: 
  "wf_elem (poly_c_max_inter_bcoeff 0 1 geq max F pi) \<and> geq (poly_c_max_inter_bcoeff 0 1 geq max F pi) 0 \<and> 
  (\<forall> a \<in> set (concat (map (\<lambda> fn. filter (\<lambda> b. \<not> (geq 1 b)) (snd (pi fn))) F)). geq (poly_c_max_inter_bcoeff 0 1 geq max F pi) a)"
  unfolding poly_c_max_inter_bcoeff_def
  by (rule poly_c_max_list[OF wf_zero], insert wf_pi, auto)

lemma poly_c_max_inter_bcoeff_strict: 
  "wf_elem (poly_c_max_inter_bcoeff_strict 0 max F pi) \<and> geq (poly_c_max_inter_bcoeff_strict 0 max F pi) 0 \<and> 
  (\<forall> a \<in> set (concat (map (\<lambda> fn. snd (pi fn)) F)). geq (poly_c_max_inter_bcoeff_strict 0 max F pi) a)"
  unfolding poly_c_max_inter_bcoeff_strict_def
  by (rule poly_c_max_list[OF wf_zero], insert wf_pi, auto)

lemma poly_c_max_inter_bconst: 
  "wf_elem (poly_c_max_inter_bconst 0 max F pi) \<and> geq (poly_c_max_inter_bconst 0 max F pi) 0 \<and> 
  (\<forall> a \<in> set (map (\<lambda> fn. fst (pi fn)) F). geq (poly_c_max_inter_bconst 0 max F pi) a)"
  unfolding poly_c_max_inter_bconst_def
  by (rule poly_c_max_list[OF wf_zero], insert wf_pi, auto)

lemma bound_eval_term_max_derivational: 
  assumes mono: "\<And> n m. n \<le> m \<Longrightarrow> g n \<le> g m"
  and bG: "\<And> n. bound (pow (poly_c_max_inter_bcoeff 0 1 geq max F pi) n * (poly_c_max_inter_bconst 0 max F pi)) \<le> g n"
  and t: "t \<in> terms_of_nat (Derivational_Complexity F) N"
  shows "g N * N \<ge> bound (pi\<guillemotleft>zero_ass,t\<guillemotright>)"
proof - 
  let ?bc = "poly_c_max_inter_bcoeff 0 1 geq max F pi"
  let ?bcv = "poly_c_max_inter_bconst 0 max F pi"
  let ?c = "concat (map (\<lambda> fn. filter (\<lambda> b. \<not> (geq 1 b)) (snd (pi fn))) F)"
  let ?cv = "map (\<lambda> fn. fst (pi fn)) F"
  have bc: "wf_elem ?bc \<and> geq ?bc 0 \<and> (\<forall> a \<in> set ?c. geq ?bc a)"
    by (rule poly_c_max_inter_bcoeff) 
  have bcv: "wf_elem ?bcv \<and> geq ?bcv 0 \<and> (\<forall> a \<in> set ?cv. geq ?bcv a)"
    by (rule poly_c_max_inter_bconst) 
  {
    fix fn
    assume fn: "fn \<in> set F"
    obtain a as where pi: "pi fn = (a,as)" by force
    have "bound_entry ?bc ?bcv (pi fn)"
      unfolding bound_entry_def pi split
      using fn bcv bc pi by force+
  } note bound_entry = this
  show ?thesis 
    by (rule bound_eval_term_derivational[OF bound_entry _ _ _ _ mono bG t], insert bc bcv, auto)
qed

lemma bound_eval_term_max_runtime:
  assumes mono: "\<And> n m. n \<le> m \<Longrightarrow> g n \<le> g m"
  and bG: "\<And> n. bound (poly_c_max_inter_bcoeff_strict 0 max D pi * (pow (poly_c_max_inter_bcoeff 0 1 geq max C pi) n * (poly_c_max_inter_bconst 0 max C pi))) \<le> g n"
  and t: "t \<in> terms_of_nat (Runtime_Complexity C D) N"
  shows "bound (poly_c_max_inter_bconst 0 max D pi) + g N * N * max_list (map snd D) \<ge> bound (pi\<guillemotleft>zero_ass,t\<guillemotright>)"
proof - 
  let ?bc = "poly_c_max_inter_bcoeff 0 1 geq max C pi"
  let ?bcv = "poly_c_max_inter_bconst 0 max C pi"
  let ?bd = "poly_c_max_inter_bcoeff_strict 0 max D pi"
  let ?bdv = "poly_c_max_inter_bconst 0 max D pi"
  let ?c = "concat (map (\<lambda> fn. filter (\<lambda> b. \<not> (geq 1 b)) (snd (pi fn))) C)"
  let ?d = "concat (map (\<lambda> fn. snd (pi fn)) D)"
  let ?cv = "map (\<lambda> fn. fst (pi fn)) C"
  let ?dv = "map (\<lambda> fn. fst (pi fn)) D"
  have bc: "wf_elem ?bc \<and> geq ?bc 0 \<and> (\<forall> a \<in> set ?c. geq ?bc a)"
    by (rule poly_c_max_inter_bcoeff) 
  have bcv: "wf_elem ?bcv \<and> geq ?bcv 0 \<and> (\<forall> a \<in> set ?cv. geq ?bcv a)"
    by (rule poly_c_max_inter_bconst) 
  have bd: "wf_elem ?bd \<and> geq ?bd 0 \<and> (\<forall> a \<in> set ?d. geq ?bd a)"
    by (rule poly_c_max_inter_bcoeff_strict) 
  have bdv: "wf_elem ?bdv \<and> geq ?bdv 0 \<and> (\<forall> a \<in> set ?dv. geq ?bdv a)"
    by (rule poly_c_max_inter_bconst) 
  {
    fix fn
    assume fn: "fn \<in> set C"
    obtain a as where pi: "pi fn = (a,as)" by force
    have "bound_entry ?bc ?bcv (pi fn)"
      unfolding bound_entry_def pi split
      using fn bcv bc pi by force+
  } note bound_entry = this
  {
    fix fn
    assume fn: "fn \<in> set D"
    obtain a as where pi: "pi fn = (a,as)" by force
    have "bound_entry_strict ?bd ?bdv (pi fn)"
      unfolding bound_entry_strict_def pi split
      using fn bdv bd pi by force+
  } note bound_entry_strict = this
  let ?bn = "max_list (map snd D)"
  {
    fix fn
    assume fn: "fn \<in> set D"
    hence "snd fn \<in> set (map snd D)" by simp
    from max_list[OF this] have "snd fn \<le> ?bn" .
  } note bn = this
  show ?thesis 
    by (rule bound_eval_term_runtime[OF bound_entry _ _ _ _ bound_entry_strict _ _ _ _ _ mono bG t],
      insert bc bcv bd bdv bn, auto)
qed
end

definition complexity_cond :: "('a :: show)poly_c \<Rightarrow> ('a \<Rightarrow> nat) \<Rightarrow> bool"
  where "complexity_cond c bound \<equiv> \<forall> bc bcv bd deg. 
  poly_c.wf c bc \<longrightarrow> poly_c.wf c bcv \<longrightarrow> poly_c.wf c bd \<longrightarrow> 
  poly_c.ge c bc (poly_c.zero c) \<longrightarrow> poly_c.ge c bcv (poly_c.zero c) \<longrightarrow> poly_c.ge c bd (poly_c.zero c) \<longrightarrow>
  poly_c.derive_complexity c bc = return deg \<longrightarrow> (\<exists> g. 
  g \<in> O_of (Comp_Poly deg) \<and>
  (\<forall> n. bound (poly_c.times c bd (poly_c.times c (linear_poly_order_carrier.pow (poly_c.times c) (poly_c.one c) bc n) bcv)) \<le> g n)
  )
  "

fun smaller_one_complexity where "smaller_one_complexity c bc = (if poly_c.ge c (poly_c.one c) bc then return 0 else poly_c.derive_complexity c bc)"
fun 
convert_lpoly_complexity where 
   "convert_lpoly_complexity c pi cc = (let 
        F = (case cc of Derivational_Complexity F \<Rightarrow> F | Runtime_Complexity C D \<Rightarrow> C);        
        bc = poly_c_max_inter_bcoeff (poly_c.zero c) (poly_c.one c) (poly_c.ge c) (poly_c.max c) F pi
      in sum_map (\<lambda> x. x) (\<lambda> deg. Comp_Poly (Suc deg)) (smaller_one_complexity c bc))"

abbreviation check_polo_mono_npsm_comp :: "('a :: show)poly_c \<Rightarrow> ('f \<times> nat)list \<Rightarrow> ('f :: show, 'a)lpoly_interL \<Rightarrow> shows check" 
  where "check_polo_mono_npsm_comp c \<equiv> check_poly_mono_npsmI (poly_c.zero c)"

locale linear_poly_order_impl = 
  fixes c :: "('a :: show) poly_c"
    and I :: "('f :: {show,key}, 'a)lpoly_interL"
    and d :: string
  assumes check_components: "isOK(check_polo_coeffs_comp c I)"
     and carrier: "\<And> as. \<exists> gt. linear_poly_order_carrier (poly_c.wf c) (poly_c.zero c) (poly_c.plus c) (poly_c.times c) (poly_c.ge c) gt (poly_c.mono c) (poly_c.one c) (poly_c.default c) (poly_c.max0 c) (poly_c.arc_pos c) \<and> (\<forall> (a,b) \<in> set as. poly_c.wf c a \<longrightarrow> poly_c.wf c b \<longrightarrow> poly_c.gt c a b \<longrightarrow> gt a b) \<and> (poly_c.mono c \<longrightarrow>
    (\<exists> bound. complexity_linear_poly_order_carrier (poly_c.wf c) (poly_c.zero c) (poly_c.plus c) (poly_c.times c) (poly_c.ge c) gt (poly_c.mono c) (poly_c.one c) (poly_c.default c) (poly_c.max0 c) (poly_c.arc_pos c) (poly_c.check_mono c) (poly_c.max c) bound \<and>
     complexity_cond c bound))"
begin

abbreviation zero where "zero \<equiv> poly_c.zero c"
abbreviation one where "one \<equiv> poly_c.one c"
abbreviation default where "default \<equiv> poly_c.default c"
abbreviation times where "times \<equiv> poly_c.times c"
abbreviation plus where "plus \<equiv> poly_c.plus c"
abbreviation max0 where "max0 \<equiv> poly_c.max0 c"
abbreviation mono where "mono \<equiv> poly_c.mono c"
abbreviation arc_pos where "arc_pos \<equiv> poly_c.arc_pos c"
abbreviation wf where "wf \<equiv> poly_c.wf c"
abbreviation ge where "ge \<equiv> poly_c.ge c"
abbreviation check_mono where "check_mono \<equiv> poly_c.check_mono c"
abbreviation max where "max \<equiv> poly_c.max c"
abbreviation derive_complexity where "derive_complexity \<equiv> poly_c.derive_complexity c"

abbreviation(input) II :: "('f,'a)lpoly_inter" where "II \<equiv> to_lpoly_inter default one I"

lemma linear_poly_order: 
  shows "\<exists> (S :: ('f,'v :: show) trs) NS. ce_af_redtriple_order S NS NS (create_afI zero I) \<and> 
  (\<forall> st \<in> set sts. isOK(check_polo_s_comp c II st) \<longrightarrow> st \<in> S) \<and>
  (\<forall> st \<in> set sts. isOK(check_polo_ns_comp c II st) \<longrightarrow> st \<in> NS) \<and>
  (mono \<longrightarrow> (cpx_ce_af_redtriple_order S NS NS (create_afI zero I) (convert_lpoly_complexity c II)) \<and> 
           (isOK(check_polo_mono_comp c I) \<longrightarrow> mono_ce_af_redtriple_order S NS NS (create_afI zero I) \<and> ctxt_closed S)) \<and> (\<not> mono \<and> isOK(check_polo_mono_npsm_comp c (funas_trs_list sts) I) \<longrightarrow> mono_ce_af_redtriple_order S NS NS (create_afI zero I) \<and> ctxt_closed S)"
proof -
  let ?pi = "create_afI zero I"
  let ?as = "[ab. st \<leftarrow> sts, ab \<leftarrow> linear_poly_order.coeffs_of_constraint zero plus times one max0 II st]"
  from carrier[of ?as] obtain gt where carrier: 
    "linear_poly_order_carrier wf zero plus times ge gt mono one default max0 arc_pos"
    and gt: "\<forall> (a,b) \<in> set ?as. wf a \<longrightarrow> wf b \<longrightarrow> poly_c.gt c  a b \<longrightarrow> gt a b" 
    and mono: "mono \<longrightarrow> (\<exists> bound. complexity_linear_poly_order_carrier wf zero plus times ge gt mono one default max0 arc_pos check_mono max bound \<and> complexity_cond c bound)" by auto
  interpret linear_poly_order_carrier wf zero plus times ge gt mono one default max0 arc_pos by fact
  interpret linear_poly_order wf zero plus times ge gt mono one default max0 arc_pos II
  proof
    show "wf_lpoly_inter II"
    proof (rule conjI, intro allI)
      fix fn 
      let ?goal = "\<lambda>fn. wf (fst (II fn)) \<and> (\<forall>a. a \<in> set (snd (II fn)) \<longrightarrow> wf a \<and> (ge a zero))"
      show "?goal fn"
      proof (cases "map_of I fn")
        case None
        hence repl: "II fn = (default, replicate (snd fn) one)" by auto
        have "\<forall> x \<in> set (replicate (snd fn) one). x = one" by (induct "snd fn", auto)
        with repl show "?goal fn" by auto
      next
        case (Some aas)
        hence aas: "II fn = aas" by auto
        from Some have "(fn ,aas) \<in> set I" by (rule map_of_SomeD)
        with check_components have "wf (fst aas) \<and> (\<forall> a \<in> set (snd aas). wf a \<and> ge a zero)"
          by (cases fn, cases aas, auto)
        with aas show "?goal fn" by auto
      qed
    next
      let ?goal2 = "\<lambda> fn. arc_pos (fst (II fn)) \<or>
        (\<exists>a\<in>set (take (snd fn) (snd (II fn))). arc_pos a)"
      show "\<forall> fn. ?goal2 fn"
      proof (intro impI allI)
        fix fn
        show "?goal2 fn"
        proof (cases "map_of I fn")
          case None
          hence "II fn = (default, replicate (snd fn) one)" by auto
          thus "?goal2 fn" by auto
        next
          case (Some aas)
          hence aas: "II fn = aas" by auto
          from Some have "(fn, aas) \<in> set I" by (rule map_of_SomeD)
          with check_components have "arc_pos (fst aas) \<or> Bex  (set (take (snd fn) (snd aas))) arc_pos" by (cases fn, cases aas, auto)
          with aas show "?goal2 fn" by auto
        qed
      qed
    qed
  qed
  let ?inter_s = "inter_s :: ('f,'v)trs"
  interpret ce_af_redtriple_order ?inter_s inter_ns inter_ns ?pi
  proof
    show "\<exists> n. \<forall> m\<ge>n . \<forall> c d. ce_trs (c,d,m) \<subseteq> inter_ns"
      by (rule inter_ns_ce_compat[OF refl])
  next
    show "af_compatible ?pi inter_ns"
      by (rule inter_ns_af_compat[OF refl])
  qed
  have order: "ce_af_redtriple_order ?inter_s inter_ns inter_ns ?pi" ..
  {
    fix st
    assume st: "st \<in> set sts" and ok: "isOK(check_polo_s_comp c II st)"
    have "st \<in> inter_s"
      by (rule checkSSound[where weak_gt = "poly_c.gt c"], insert ok gt st, auto)
  } note S = this
  {
    fix st
    assume st: "st \<in> set sts" and ok: "isOK(check_polo_ns_comp c II st)"
    have "st \<in> inter_ns"
      by (rule checkNSSound, insert ok gt st, auto)
  } note NS = this
  let ?amono = "\<not> mono \<and> isOK(check_polo_mono_npsm_comp c (funas_trs_list sts) I)"
  show ?thesis
  proof (cases ?amono)
    case False
    hence id: "?amono = False" by simp
    show ?thesis unfolding id
    proof (intro exI conjI allI ballI, 
        rule order, 
        (intro impI, rule S, auto)[1],
        (intro impI, rule NS, auto)[1],
          intro impI, rule conjI)
      assume mono
      from mono `mono` obtain bound where "complexity_linear_poly_order_carrier wf zero plus times ge gt mono one default max0 arc_pos check_mono max bound" and cpx: "complexity_cond c bound" by auto
      then interpret complexity_linear_poly_order_carrier wf zero plus times ge gt mono one default max0 arc_pos check_mono max bound by auto
      interpret complexity_linear_poly_order wf zero plus times ge gt mono one default max0 arc_pos II check_mono max bound ..
      show "isOK(check_polo_mono_comp c I) \<longrightarrow> 
        mono_ce_af_redtriple_order ?inter_s inter_ns inter_ns ?pi \<and> ctxt_closed inter_s"
      proof
        assume ok: "isOK(check_polo_mono_comp c I)"
        interpret mono_linear_poly_order wf zero plus times ge gt mono one default max0 arc_pos II check_mono max
          by (unfold_locales, rule check_poly_mono_sound[OF ok])
        interpret mono_ce_af_redtriple_order ?inter_s inter_ns inter_ns ?pi 
          by (unfold_locales, rule inter_s_ce_compat[OF refl])
        show "mono_ce_af_redtriple_order ?inter_s inter_ns inter_ns ?pi \<and> ctxt_closed inter_s"
          by (rule conjI[OF _ inter_s_mono], unfold_locales)
      qed
      let ?conv = "convert_lpoly_complexity c II :: ('f,'v)complexity_measure \<Rightarrow> shows + complexity_class" 
      show "cpx_ce_af_redtriple_order ?inter_s inter_ns inter_ns ?pi ?conv" 
      proof
        fix cm cc
        assume conv: "?conv cm = return cc"
        let ?bcF = "\<lambda> F. poly_c_max_inter_bcoeff zero one ge max F II"
        let ?smone = "\<lambda> F. if (ge one (?bcF F)) then return 0 else derive_complexity (?bcF F)"  
        obtain smone where smone: "smone = ?smone" by auto
        {
          fix f g a b
          have "(sum_map f g a = Inr b) = (\<exists> c. a = Inr c \<and> g c = b)" by (cases a, auto)
        } note sum_map_Inr = this
        from conv have "\<exists> F C D nn. (cm = Derivational_Complexity F \<or> cm = Runtime_Complexity F D) \<and> 
          smone F = Inr nn \<and> cc = Comp_Poly (Suc nn)" unfolding smone
          by (cases cm, auto simp: sum_map_Inr)
        then obtain F D nn where choice: "cm = Derivational_Complexity F \<or> cm = Runtime_Complexity F D"
          and smoneF: "smone F = Inr nn" and cc: "cc = Comp_Poly (Suc nn)" by auto
        let ?bc = "?bcF F"
        let ?bcv = "poly_c_max_inter_bconst zero max F II"
        let ?bd = "poly_c_max_inter_bcoeff_strict zero max D II"
        from poly_c_max_inter_bcoeff[of F] have wfbc: "wf ?bc" and gebc: "ge ?bc zero" by auto
        from poly_c_max_inter_bcoeff_strict[of D] have wfbd: "wf ?bd" and gebd: "ge ?bd zero" by auto
        from poly_c_max_inter_bconst[of F] have wfbcv: "wf ?bcv" and gebcv: "ge ?bcv zero" by auto
        from arc_pos_True have apos: "apos_ass zero_ass" unfolding apos_ass_def by simp
        let ?eval = "\<lambda> t :: ('f,'v)term. eval_term II zero_ass t"
        let ?S = "{(a,b). wf a \<and> wf b \<and> ge b zero \<and> arc_pos b \<and> gt a b}"      
        {
          fix s t 
          assume "(s,t) \<in> ?inter_s"
          from this[unfolded inter_s_def] have "\<And> \<alpha>. wf_ass \<alpha> \<Longrightarrow> pos_ass \<alpha> \<Longrightarrow> apos_ass \<alpha> \<Longrightarrow> gt (eval_term II \<alpha> s) (eval_term II \<alpha> t)"
            by auto      
          from this[OF wf_zero_ass pos_zero_ass apos]       
          have gt: "gt (?eval s) (?eval t)" by auto
          from pos_term[OF wf_zero_ass pos_zero_ass] have ge: "ge (?eval t) zero" .
          from apos_term[OF wf_zero_ass apos] have apos: "arc_pos (?eval t)" .
          from ge gt apos have "(?eval s,?eval t) \<in> ?S" by (auto simp: Let_def simp del: to_lpoly_inter.simps)
        } note image = this
        {
          fix bd
          assume wfbd: "wf bd" and gebd: "ge bd zero"
          have "\<exists> g. g \<in> O_of (Comp_Poly nn) \<and> (\<forall> n. bound (times bd (times (pow ?bc n) ?bcv)) \<le> g n)"
          proof (cases "ge one ?bc")
            case False
            with smoneF[unfolded smone] have der: "derive_complexity ?bc = return nn" by auto
            from cpx[unfolded complexity_cond_def, rule_format, OF wfbc wfbcv wfbd gebc gebcv gebd]
            show ?thesis using der by auto
          next
            case True
            with smoneF[unfolded smone] have nn: "nn = 0" by simp
            let ?b = "times bd (times one ?bcv)"
            let ?g = "\<lambda> n. bound ?b"
            show ?thesis unfolding nn
            proof (rule exI[of _ ?g], rule conjI[OF _ allI])
              show "?g \<in> O_of (Comp_Poly 0)"
                by (rule O_of_polyI[of _ 0 _ "bound ?b"], auto)
            next
              fix n
              show "bound (times bd (times (pow ?bc n) ?bcv)) \<le> ?g n"
                by (rule bound_mono[OF times_right_mono[OF gebd times_left_mono[OF gebcv pow_ge_1[OF True gebc wfbc]]]], insert wfbc wfbcv wfbd, auto)
            qed
          qed
        } note get_g = this
        from choice
        have "\<exists> c d g. g \<in> complexity_of (Comp_Poly nn) \<and> (\<forall> n t. t \<in> terms_of_nat cm n \<longrightarrow> bound (?eval t) \<le> c + g n * n * d)"
        proof
          assume cm: "cm = Derivational_Complexity F"
          from get_g[OF wf_one one_geq_zero] obtain g where g: "g \<in> O_of (Comp_Poly nn)"
            and bnd2: "\<And> n. bound (times (pow ?bc n) ?bcv) \<le> g n" using wfbc wfbcv by auto
          from g[unfolded O_of_def] obtain gg where gg: "gg \<in> complexity_of (Comp_Poly nn)" and ggg: "\<And> n. g n \<le> gg n" by auto
          from le_trans[OF bnd2 ggg] have bnd2: "\<And> n. bound (times (pow ?bc n) ?bcv) \<le> gg n" by auto     
          from complexity_poly_mono[OF gg] have gg_mono: "\<And> n m. n \<le> m \<Longrightarrow> gg n \<le> gg m" .
          from bound_eval_term_max_derivational[of gg, OF gg_mono bnd2]
          have bound2: "\<And> t N. t \<in> terms_of_nat cm N \<Longrightarrow> bound (?eval t) \<le> gg N * N" 
            unfolding cm by auto
          show ?thesis
            by (rule exI[of _ 0], rule exI[of _ 1], rule exI, rule conjI[OF gg], insert bound2, auto)
        next
          assume cm: "cm = Runtime_Complexity F D"
          from get_g[OF wfbd gebd] obtain g where g: "g \<in> O_of (Comp_Poly nn)"
            and bnd2: "\<And> n. bound (times ?bd (times (pow ?bc n) ?bcv)) \<le> g n" using wfbc wfbcv by auto
          from g[unfolded O_of_def] obtain gg where gg: "gg \<in> complexity_of (Comp_Poly nn)" and ggg: "\<And> n. g n \<le> gg n" by auto
          from le_trans[OF bnd2 ggg] have bnd2: "\<And> n. bound (times ?bd (times (pow ?bc n) ?bcv)) \<le> gg n" by auto     
          from complexity_poly_mono[OF gg] have gg_mono: "\<And> n m. n \<le> m \<Longrightarrow> gg n \<le> gg m" .
          let ?bn = "max_list (map snd D)"
          let ?bdv = "poly_c_max_inter_bconst zero max D II"
          from bound_eval_term_max_runtime[of gg, OF gg_mono bnd2]
          have bound2: "\<And> t N. t \<in> terms_of_nat cm N \<Longrightarrow> bound (?eval t) \<le> bound ?bdv + gg N * N * ?bn"
            unfolding cm by auto
          show ?thesis
            by (rule exI[of _ "bound ?bdv"], rule exI[of _ ?bn], rule exI, rule conjI[OF gg], insert bound2, auto)
        qed
        then obtain c d g where g: "g \<in> complexity_of (Comp_Poly nn)"
          and bound2: "\<And> n t. t \<in> terms_of_nat cm n \<Longrightarrow> bound (?eval t) \<le> c + g n * n * d" by blast
        from g[simplified] obtain c' e where g: "g = (\<lambda> n. c' * n ^ nn + e)" by auto
        obtain c'' where c'': "c'' = c' * d + e * d" by auto
        let ?h = "(\<lambda> n. c'' * n ^ Suc nn + c)"
        obtain h where h: "h = ?h" by auto
        {
          fix n :: nat
          {
            assume n: "1 \<le> n"
            have le: "(e * d) * (n * 1) \<le> (e * d) * (n * n ^ nn)"
              by (rule mult_left_mono[OF mult_left_mono[OF one_le_power[OF n]]], auto)
            have "c + g n * n * d = c + (c' * d) * (n ^ Suc nn) + (e * d * n)" unfolding g 
              by (simp add: field_simps)
            also have "... \<le> h n" unfolding h c'' using le by (simp add: field_simps)
            finally have "c + g n * n * d \<le> h n" .
          } note le = this
          have "c + g n * n * d \<le> h n" using le unfolding h g c'' by (cases n, auto simp: field_simps)
        }
        from le_trans[OF bound2 this] have bound2: "\<And> t n . t \<in> terms_of_nat cm n \<Longrightarrow> bound (?eval t) \<le> h n" .
        have one: "\<And> t n. \<lbrakk>t \<in> terms_of_nat cm n\<rbrakk>
          \<Longrightarrow> deriv_bound
          ?inter_s
          t (h n)" 
          by (rule deriv_bound_image[of ?S ?eval _ _ ?inter_s, OF deriv_bound_mono[OF bound bound2]], insert image, auto)
        show deriv_bound: "deriv_bound_measure_class ?inter_s cm cc" unfolding cc
          unfolding deriv_bound_measure_class_def
          by (rule deriv_bound_rel_class_polyI[of _ _ c'' _ c], insert one[unfolded h], auto)  
      qed    
    qed simp
  next
    case True
    hence npsm: "isOK(check_poly_mono_npsm (funas_trs_list sts) I)" and nmono: "\<not> mono" by auto
    let ?F = "funas_trs (set sts)"
    interpret npsm_mono_linear_poly_order wf zero plus times ge gt mono one default max0 arc_pos II ?F
      by (unfold_locales, insert check_poly_mono_npsm_sound[OF npsm] nmono, auto)
    obtain R where R: "R = set (filter (\<lambda> st. isOK(check_polo_s_comp c II st)) sts)" by auto
    obtain Rw where Rw: "Rw = set (filter (\<lambda> st. isOK(check_polo_ns_comp c II st)) sts)" by auto
    from S R have RS: "R \<subseteq> inter_s" by auto
    from NS Rw have RwNS: "Rw \<subseteq> inter_ns" by auto
    from R Rw have F: "funas_trs R \<subseteq> ?F" "funas_trs Rw \<subseteq> ?F" unfolding funas_trs_def by auto
    let ?Ce = "\<Union> ce_trs ` UNIV :: ('f,'v)trs"
    let ?R = "rstep (R \<union> ?Ce)"
    let ?Rw = "rstep (Rw \<union> ?Ce)"
    from rel_ce_terminating[OF F RS RwNS] have 
      SN_rel: "SN_rel (?R) (?Rw)" .
    let ?S = "(relto (?R) (?Rw))^+"
    let ?NS = "?Rw^*"
    interpret mono_ce_af_redtriple_order ?S ?NS ?NS ?pi 
    proof (rule ce_SN_rel_imp_redtriple[OF SN_rel])
      fix f and n i :: nat
      assume i: "i < n"
      show "?pi (f, n) i"
      proof (cases "map_of I (f, n)")
        case None
        thus ?thesis unfolding create_afI_def by auto
      next
        case (Some ccs)
        then obtain c cs where look: "map_of I (f, n) = Some (c,cs)" by force
        from map_of_SomeD[OF look] have mem: "((f,n),(c,cs)) \<in> set I" by auto
        with check_components have apos: "arc_pos c \<or> Bex (set (take n cs)) arc_pos" by auto
        from mem have "(f,n) \<in> set (map fst I)" by force
        with check_poly_mono_npsm_sound[OF npsm] 
        have mono: "n \<le> Suc 0" "n = Suc 0 \<Longrightarrow> fst (II (f,n)) = zero \<and> length (snd (II (f,n))) = Suc 0" by force+
        from look have II: "II (f,n) = (c,cs)" unfolding to_lpoly_inter.simps by auto        
        note mono = mono[unfolded II fst_conv snd_conv]
        from mono i have n: "n = Suc 0" by auto
        from mono(2)[OF this] obtain d where c: "c = zero" and cs: "cs = [d]" by (cases cs, auto)
        from apos[unfolded c n cs] arc_pos_zero[OF mode]
        have ad: "arc_pos d" by auto
        with arc_pos_zero[OF mode] have d: "d \<noteq> zero" by auto
        from n i have i: "i = 0" by simp
        show ?thesis unfolding create_afI_def look cs i Let_def ceta_map_of fun_of_map_fun'.simps using d by simp
      qed
    qed auto
    show ?thesis 
    proof (rule exI[of _ ?S], rule exI[of _ ?NS], intro conjI impI ballI)
      show "ce_af_redtriple_order ?S ?NS ?NS ?pi" ..      
    next
      fix st
      assume "st \<in> set sts" and "isOK(check_polo_s_comp c II st)"
      hence "st \<in> ?R" unfolding R by (cases st, auto)
      also have "?R \<subseteq> ?S" unfolding relto_trancl_conv by regexp
      finally show "st \<in> ?S" .
    next
      fix st
      assume "st \<in> set sts" and "isOK(check_polo_ns_comp c II st)"
      hence "st \<in> ?Rw" unfolding Rw by (cases st, auto)
      also have "?Rw \<subseteq> ?NS" by regexp
      finally show "st \<in> ?NS" .
    next
      show "mono_ce_af_redtriple_order ?S ?NS ?NS ?pi" ..
      show "mono_ce_af_redtriple_order ?S ?NS ?NS ?pi" ..
      show "ctxt_closed ?S" unfolding relto_trancl_conv by blast
    qed (insert nmono, auto)
  qed
qed
end  

definition create_lpoly_repr :: "('a) poly_c \<Rightarrow> ('f::{key,show},'a::show)lpoly_interL \<Rightarrow> shows"
where "create_lpoly_repr c I = (let pi = to_lpoly_inter (poly_c.default c) (poly_c.one c) I in  (
  ''polynomial interpretration over '' +#+ poly_c.description c +#+ shows_nl +@+ (shows_sep (\<lambda>(f,n).
    let t = Fun f (map Var (fresh_strings_list ''x_'' 1 [] n)) in (
      ''Pol('' +#+ shows t +@+ '') = '' +#+
      shows_lpoly (poly_c.zero c) (poly_c.one c) (PleftI (poly_c.zero c) (poly_c.one c) (poly_c.plus c) (poly_c.times c) (poly_c.max0 c) pi t)
    )) shows_nl (remdups (map fst I)))))"

abbreviation create_polo_af_comp :: "('a :: show)poly_c \<Rightarrow> ('f :: key,'a)lpoly_interL \<Rightarrow> 'f af"
  where "create_polo_af_comp c I \<equiv> create_afI (poly_c.zero c) I"

definition create_poly_redtriple :: "(('f,'a)lpoly_interL \<Rightarrow> shows check) \<Rightarrow> ('a) poly_c \<Rightarrow> ('f :: {show,key}, 'a :: show)lpoly_interL \<Rightarrow> ('f,'v :: show)redtriple"
where "create_poly_redtriple cI c I = (let pi = to_lpoly_inter (poly_c.default c) (poly_c.one c) I; ns = check_polo_ns_comp c pi in 
    \<lparr>redtriple.valid = do {cI I; check_polo_coeffs_comp c I}, 
     s = check_polo_s_comp c pi, 
     ns = ns, nst = ns, 
     af = create_polo_af_comp c I,
     mono = (\<lambda> s_ns_nst. if (poly_c.mono c) then check_polo_mono_comp c I else check_polo_mono_npsm_comp c (funas_trs_list s_ns_nst) I), 
     desc = create_lpoly_repr c I,
     cpx = (if poly_c.mono c then convert_lpoly_complexity c pi else no_complexity) \<rparr>)"

lemma create_poly_redtriple:
  fixes cI :: "('f :: {key,show}, 'a :: show)lpoly_interL \<Rightarrow> shows check"
  assumes carrier: "\<And> I as. isOK(cI I) \<Longrightarrow> \<exists> gt. linear_poly_order_carrier 
  (poly_c.wf c)
  (poly_c.zero c)
  (poly_c.plus c)
  (poly_c.times c)
  (poly_c.ge c)
  gt
  (poly_c.mono c)
  (poly_c.one c)
  (poly_c.default c)
  (poly_c.max0 c)
  (poly_c.arc_pos c) \<and> 
  (\<forall>(a, b)\<in>set as. poly_c.wf c a \<longrightarrow> poly_c.wf c b \<longrightarrow> poly_c.gt c a b \<longrightarrow> gt a b) \<and>
  (poly_c.mono c \<longrightarrow>
    (\<exists> bound. complexity_linear_poly_order_carrier (poly_c.wf c) (poly_c.zero c) (poly_c.plus c) (poly_c.times c) (poly_c.ge c) gt (poly_c.mono c) (poly_c.one c) (poly_c.default c) (poly_c.max0 c) (poly_c.arc_pos c) (poly_c.check_mono c) (poly_c.max c) bound \<and> complexity_cond c bound))"
  shows "generic_redtriple_impl (create_poly_redtriple cI c)"
proof
  fix I :: "('f,'a)lpoly_interL" and s_list ns_list nst_list :: "('f,'v :: show)rule list" and  \<pi>
  let ?c = "create_poly_redtriple cI c I :: ('f,'v)redtriple"
  let ?pi = "redtriple.af ?c :: 'f af"
  assume valid: "isOK (redtriple.valid ?c)" and 
    check_s: "isOK(check_allm (redtriple.s ?c) s_list)" and 
    check_ns: "isOK(check_allm (redtriple.ns ?c) ns_list)" and
    check_nst: "isOK(check_allm (redtriple.nst ?c) nst_list)"
  from valid have carrier_valid: "isOK (cI I)" and polo_valid: "isOK(check_polo_coeffs_comp c I)" unfolding create_poly_redtriple_def by (auto simp: Let_def)
  let ?I = "to_lpoly_inter (poly_c.default c) (poly_c.one c) I :: ('f,'a)lpoly_inter"
  let ?conv = "convert_lpoly_complexity c ?I :: ('f,'v)complexity_measure \<Rightarrow> shows + complexity_class"
  let ?cpx = "redtriple.cpx ?c" 
  let ?all = "s_list @ ns_list @ nst_list"
  interpret linear_poly_order_impl c I 
    by (unfold_locales, rule polo_valid, rule carrier[OF carrier_valid]) 
  from linear_poly_order[of ?all] obtain S NS cpx where order: "ce_af_redtriple_order S NS NS ?pi" 
    and S: "(\<forall> st \<in> set ?all. isOK(check_polo_s_comp c ?I st) \<longrightarrow> st \<in> S)" 
    and NS: "(\<forall> st \<in> set ?all. isOK(check_polo_ns_comp c ?I st) \<longrightarrow> st \<in> NS)" 
    and cpx: "mono \<Longrightarrow> cpx_ce_af_redtriple_order S NS NS ?pi (convert_lpoly_complexity c ?I)"
    and mono: "mono \<Longrightarrow> isOK(check_polo_mono_comp c I) \<Longrightarrow> mono_ce_af_redtriple_order S NS NS ?pi \<and> ctxt_closed S" 
    and nmono: "\<not> mono \<Longrightarrow> isOK(check_polo_mono_npsm_comp c (funas_trs_list ?all) I) \<Longrightarrow> mono_ce_af_redtriple_order S NS NS ?pi \<and> ctxt_closed S" unfolding create_poly_redtriple_def Let_def redtriple.simps by blast
  interpret ce_af_redtriple_order S NS NS ?pi by fact
  have order: "cpx_ce_af_redtriple_order S NS NS ?pi ?cpx"
  proof (cases mono)
    case True
    hence id: "?cpx = ?conv" unfolding create_poly_redtriple_def Let_def by simp
    show ?thesis unfolding id using cpx[OF True] .
  next
    case False
    hence id: "?cpx = no_complexity" unfolding create_poly_redtriple_def Let_def by simp
    show ?thesis unfolding id ..
  qed
  from check_s[unfolded create_poly_redtriple_def Let_def] S have S: "set s_list \<subseteq> S" by auto
  from check_nst[unfolded create_poly_redtriple_def Let_def] NS have NST: "set nst_list \<subseteq> NS" by auto
  from check_ns[unfolded create_poly_redtriple_def Let_def] NS have NS: "set ns_list \<subseteq> NS" by auto
  show "\<exists> S NS NST. cpx_ce_af_redtriple_order S NS NST ?pi ?cpx \<and> set s_list \<subseteq> S \<and> set ns_list \<subseteq> NS \<and> set nst_list \<subseteq> NST \<and> (isOK (redtriple.mono ?c ?all) \<longrightarrow> mono_ce_af_redtriple_order S NS NST ?pi \<and> ctxt_closed S)"
  proof (intro exI conjI, rule order, rule S, rule NS, rule NST, intro impI)
    assume ok: "isOK(redtriple.mono ?c ?all)"
    note ok = ok[unfolded create_poly_redtriple_def Let_def, simplified]
    show "mono_ce_af_redtriple_order S NS NS ?pi \<and> ctxt_closed S" 
    proof (cases mono)
      case True
      with ok have "isOK (check_polo_mono_comp c I)" by simp
      from mono[OF True this] show ?thesis .
    next
      case False
      with ok have "isOK(check_polo_mono_npsm_comp c (funas_trs_list ?all) I)" 
        by simp
      from nmono[OF False this] show ?thesis .
    qed
  qed
qed
    
context linear_poly_order_carrier
begin
declare poly_simps[simp del] poly_order_simps[simp del]
end

end
