(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Parser_Monad2
imports Monad_Syntax 
begin

datatype ('e,'a) parse_result = PNonterm | PError 'e | PResult 'a

abbreviation "parse_result_ord \<equiv> flat_ord PNonterm"

interpretation parse_result!:
  partial_function_definitions parse_result_ord "flat_lub PNonterm"
by (rule flat_interpretation)

declaration {* Partial_Function.init "parse_result" @{term parse_result.fixp_fun}
  @{term parse_result.mono_body} @{thm parse_result.fixp_rule_uc} 
  NONE *}


fun bind :: "('e,'a) parse_result \<Rightarrow> ('a \<Rightarrow> ('e,'b) parse_result) \<Rightarrow> ('e,'b) parse_result"
  where "bind PNonterm f = PNonterm"
      | "bind (PError e) f = PError e"
      | "bind (PResult x) f = f x"

lemma bind_cong[fundef_cong]:
  assumes "xs = ys" and "\<And>x. ys = PResult x \<Longrightarrow> f x = g x"
  shows "bind xs f = bind ys g"
  unfolding assms(1) using assms(2)
  by (cases ys, auto)

abbreviation "mono_parse_result \<equiv> monotone (fun_ord parse_result_ord) parse_result_ord"

(* TODO: perhaps use Partial_Function.bind_mono to proof this result immediately *)
lemma bind_mono[partial_function_mono]:
assumes mf: "mono_parse_result B" and mg: "\<And>y. mono_parse_result (\<lambda>f. C y f)"
shows "mono_parse_result (\<lambda>f. bind (B f) (\<lambda>y. C y f))"
proof (rule monotoneI)
  fix f g :: "'a \<Rightarrow> ('b,'c) parse_result" assume fg: "fun_ord parse_result_ord f g"
  with mf
  have "parse_result_ord (B f) (B g)" by (rule monotoneD[of _ _ _ f g])
  then have "parse_result_ord (bind (B f) (\<lambda>y. C y f)) (bind (B g) (\<lambda>y. C y f))"
    unfolding flat_ord_def by auto    
  also from mg
  have "\<And>y'. parse_result_ord (C y' f) (C y' g)"
    by (rule monotoneD) (rule fg)
  then have "parse_result_ord (bind (B g) (\<lambda>y'. C y' f)) (bind (B g) (\<lambda>y'. C y' g))"
    unfolding flat_ord_def by (cases "B g") auto
  finally (parse_result.leq_trans)
  show "parse_result_ord (bind (B f) (\<lambda>y. C y f)) (bind (B g) (\<lambda>y'. C y' g))" .
qed

setup {*
  Adhoc_Overloading.add_variant @{const_name Monad_Syntax.bind} @{const_name bind}
*}

hide_const (open) bind

fun
  catch :: "('e,'a) parse_result \<Rightarrow> ('e \<Rightarrow> ('f,'a) parse_result) \<Rightarrow> ('f,'a) parse_result"
     ("(try(/ _)/ catch(/ _))" [12, 12] 13)
  where "(try PNonterm catch f) = PNonterm "
      | "(try (PError a) catch f) = f a"
      | "(try (PResult a) catch f) = PResult a"

lemma catch_mono[partial_function_mono]:
assumes mf: "mono_parse_result B" and mg: "\<And>y. mono_parse_result (\<lambda>f. C y f)"
shows "mono_parse_result (\<lambda>f. try (B f) catch (\<lambda>y. C y f))"
proof (rule monotoneI)
  fix f g :: "'a \<Rightarrow> ('b,'c) parse_result" assume fg: "fun_ord parse_result_ord f g"
  with mf
  have "parse_result_ord (B f) (B g)" by (rule monotoneD[of _ _ _ f g])
  then have "parse_result_ord (try (B f) catch (\<lambda>y. C y f)) (try (B g) catch (\<lambda>y. C y f))"
    unfolding flat_ord_def by auto    
  also from mg
  have "\<And>y'. parse_result_ord (C y' f) (C y' g)"
    by (rule monotoneD) (rule fg)
  then have "parse_result_ord (try (B g) catch (\<lambda>y'. C y' f)) (try (B g) catch (\<lambda>y'. C y' g))"
    unfolding flat_ord_def by (cases "B g") auto
  finally (parse_result.leq_trans)
  show "parse_result_ord (try (B f) catch (\<lambda>y. C y f)) (try (B g) catch (\<lambda>y'. C y' g))" .
qed

definition error :: "'e \<Rightarrow> ('e,'a) parse_result"
  where [simp]: "error x \<equiv> PError x"

definition return :: "'a \<Rightarrow> ('e,'a) parse_result"
  where [simp]: "return x \<equiv> PResult x"

fun map_parse_result :: "('a \<Rightarrow> ('e,'b) parse_result) \<Rightarrow> 'a list \<Rightarrow> ('e,'b list) parse_result" where
  "map_parse_result f [] = return []"
| "map_parse_result f (x#xs) = do {
    y \<leftarrow> f x;
    ys \<leftarrow> map_parse_result f xs;
    return (y # ys)
  }"

lemma map_parse_result_cong[fundef_cong]:
  assumes "xs = ys" and "\<And>x. x \<in> set ys \<Longrightarrow> f x = g x"
  shows "map_parse_result f xs = map_parse_result g ys"
  unfolding assms(1) using assms(2) by (induct ys) auto

lemmas parse_result_const_mono = parse_result.const_mono[of "fun_ord parse_result_ord"]

lemma map_parse_result_mono[partial_function_mono]:
  fixes C :: "'a \<Rightarrow> ('b \<Rightarrow> ('e,'c) parse_result) \<Rightarrow> ('e,'d) parse_result"
  assumes C: "\<And>y. y \<in> set B \<Longrightarrow> mono_parse_result (C y)"
  shows "mono_parse_result (\<lambda>f. map_parse_result (\<lambda>y. C y f) B)" 
  using C
  by (induct B, auto intro!: partial_function_mono)

abbreviation update_error :: "('e,'a) parse_result \<Rightarrow> ('e \<Rightarrow> 'f) \<Rightarrow> ('f,'a) parse_result"
 (infixl "<+?" 61)
  where "update_error r f \<equiv> try r catch (\<lambda> e. error (f e))"

fun parseresult :: "'e + 'a \<Rightarrow> ('e,'a)parse_result"
  where "parseresult (Inl x) = PError x"
     |  "parseresult (Inr x) = PResult x"

code_datatype parseresult

lemma [code]: "bind (parseresult a) f = (case a of Inl b \<Rightarrow> parseresult (Inl b) | Inr a \<Rightarrow> f a)" by (cases a, auto)
lemma [code]: "(try (parseresult a) catch f) = (case a of Inl b \<Rightarrow> f b | Inr a \<Rightarrow> parseresult (Inr a))" by (cases a, auto)

lemma [code]: "PResult x = parseresult (Inr x)" by simp
lemma [code]: "PError x = parseresult (Inl x)" by simp
lemma [code]: "return x = parseresult (Inr x)" by simp
lemma [code]: "error x = parseresult (Inl x)" by simp
lemma [code]: "parse_result_case f g h (parseresult p) = sum_case g h p" 
  by (cases p, auto)

end
