(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Q_Reduction_Impl
imports
  Q_Reduction
  QDP_Framework_Impl
  Innermost_Impl
begin

definition
  q_reduction_proc ::
    "('dpp, 'f::show, 'v::show) dpp_ops \<Rightarrow>
    ('f,'v)term list \<Rightarrow> 
    'dpp proc"
where
  "q_reduction_proc I Q' dpp \<equiv> 
    let Pb  = dpp_ops.pairs I dpp;
        Rb = dpp_ops.rules I dpp;
        F = map Some (funas_trs_list (Pb @ Rb));
        Q = dpp_ops.Q I dpp;
        isnf = dpp_ops.is_QNF I dpp;
        QQ = filter (\<lambda> q. Ball (set (args q)) isnf) Q;
        rQ = filter (\<lambda> q. root q \<in> set F) QQ
 in check_return (do {
  check (dpp_ops.NFQ_subset_NF_rules I dpp) (shows ''innermost rewriting required'');  
  check_wf_trs Rb;
  check_NF_terms_subset Q' rQ <+? (\<lambda> q. shows ''the term '' +@+ shows q +@+ shows '' is missing in Q' '');
  check_NF_terms_subset QQ Q' <+? (\<lambda> q. shows ''the term '' +@+ shows q +@+ shows '' is not allowed in Q' '');
  check_allm (\<lambda> lr. check_subseteq (vars_term_list (snd lr)) (vars_term_list (fst lr)) <+? (\<lambda> x. shows ''variable condition in rule violated'')) Pb
  } <+?
  (\<lambda> e. shows ''problem when reducing Q in the DP problem '' +@+ shows_nl +@+ 
  shows_dpp I dpp +@+ shows_nl +@+ shows ''to the set '' +@+ shows_nl 
  +@+ shows_terms' shows shows ''Q':'' Q' +@+ shows_nl +@+ e)) 
     (dpp_ops.mk I (dpp_ops.nfs I dpp) (dpp_ops.minimal I dpp) (dpp_ops.P I dpp) (dpp_ops.Pw I dpp) Q' (dpp_ops.R I dpp) (dpp_ops.Rw I dpp))"

lemma q_reduction_proc: assumes I: "dpp_spec I"
 shows "dpp_spec.sound_proc_impl I (q_reduction_proc I (Q' :: ('f :: show,'v :: show)term list))"
proof -
  from assms interpret dpp_spec I .
  show ?thesis
  proof
    fix d d'    
    assume ok: "q_reduction_proc I Q' d = Inr d'" and fin: "finite_dpp (dpp d')"
    let ?P = "set (P d)"
    let ?Pw = "set (Pw d)"
    let ?Pb = "?P \<union> ?Pw"
    let ?Q = "set (Q d)"
    let ?QQ = "{q . set (args q) \<subseteq> NF_terms ?Q \<and> q \<in> ?Q}"
    let ?Q' = "set Q'"
    let ?R = "set (R d)"
    let ?Rw = "set (Rw d)"
    let ?Rb = "?R \<union> ?Rw"
    let ?nfs = "NFS d"
    let ?m = "M d"
    let ?F = "funas_trs ?Pb \<union> funas_trs ?Rb"
    let ?rQ = "{ q. q \<in> ?QQ \<and> root q \<in> Some ` ?F}"
    let ?dpp = "\<lambda> q. (?nfs,?m,?P,?Pw,q,?R,?Rw)"
    note ok = ok[unfolded q_reduction_proc_def Let_def dpp_spec_sound, simplified] 
    have NFQQ: "NF_terms ?Q = NF_terms ?QQ"
    proof
      show "NF_terms ?QQ \<supseteq> NF_terms ?Q"
        by (rule NF_terms_anti_mono, auto)
      show "NF_terms ?QQ \<subseteq> NF_terms ?Q"
      proof -
        {
          fix q
          assume "q \<in> ?Q" and "q \<in> NF_terms ?QQ"
          hence False
          proof (induct q rule: wf_induct[OF wf_measure[of size]])
            case (1 q)
            {
              assume "set (args q) \<subseteq> NF_terms ?Q"
              with 1(2) have "q \<in> ?QQ" by auto
              hence "q \<notin> NF_terms ?QQ" by blast
              with 1(3) have False by auto
            }
            then obtain p where pq: "p \<in> set (args q)" and NF: "p \<notin> NF_terms ?Q" by auto
            from NF obtain C q' \<sigma> where p: "p = C \<langle> q' \<cdot> \<sigma> \<rangle>" and q': "q' \<in> ?Q" by auto
            from pq obtain f ts where q: "q = Fun f ts" and pts: "p \<in> set ts" by (cases q, auto)
            from pts[unfolded in_set_conv_decomp] obtain bef aft where ts: "ts = bef @ p # aft" by blast
            let ?C = "More f bef C aft"
            from q ts p have qq': "q = ?C \<langle> q' \<cdot> \<sigma> \<rangle>" by simp
            hence "q \<unrhd> q' \<cdot> \<sigma>" by auto
            from NF_instance[OF NF_subterm[OF 1(3) this]] have q'NF: "q' \<in> NF_terms ?QQ" .
            have "size q' \<le> size (q' \<cdot> \<sigma>)" by (rule size_subst)
            also have "\<dots> \<le> size (C \<langle>q' \<cdot> \<sigma>\<rangle>)" by (rule supteq_size, auto)
            also have "\<dots> < size q" unfolding qq' by simp
            finally show False using 1(1) q' q'NF by auto
          qed
        }
        thus ?thesis unfolding NF_terms_subset_criterion[symmetric]
          by auto
      qed
    qed
    hence nfqq : "NF_terms ?QQ = NF_terms ?Q" by simp
    from ok have inn: "NF_terms ?QQ \<subseteq> NF_trs ?Rb" unfolding nfqq by auto
    from ok have wf: "wf_trs ?Rb" by auto
    from ok have NFQ: "NF_terms ?QQ \<subseteq> NF_terms ?Q'" by force
    from ok have "NF_terms ?Q'
      \<subseteq> NF_terms {q \<in> ?Q.
      (\<forall>t\<in>set (args q). t \<in> NF_terms ?Q) \<and> root q \<in> Some ` ?F}" (is "_ \<subseteq> NF_terms ?QQQ") by auto
    also have "?QQQ = ?rQ" by auto
    finally have NFrQ: "NF_terms ?Q' \<subseteq> NF_terms ?rQ" unfolding nfqq .
    from ok have vars: "\<And> lr. lr \<in> ?Pb \<Longrightarrow> vars_term (snd lr) \<subseteq> vars_term (fst lr)" by auto
    hence vars: "\<And> l r. (l,r) \<in> ?Pb \<Longrightarrow> vars_term r \<subseteq> vars_term l" by force
    from ok have id: "dpp d = ?dpp ?Q" "dpp d' = ?dpp ?Q'" by auto
    {
      fix s t \<sigma>
      assume Qchain: "min_ichain (?dpp ?Q) s t \<sigma>"
      from min_ichain_mono[OF this subset_refl subset_refl NFQQ subset_refl refl]
      have Qchain: "min_ichain (?dpp ?QQ) s t \<sigma>" .
      from min_ichain_reduced_Q[OF inn refl _ _ wf vars Qchain, of ?F]
      have rQchain: "min_ichain (?dpp ?rQ) s t \<sigma>" by auto
      from Qchain have Qchain: "ichain (?dpp ?QQ) s t \<sigma>" by auto
      have Q'chain: "ichain (?dpp ?Q') s t \<sigma>"
        by (rule ichain_mono[OF Qchain], insert NFQ, auto)
      {
        fix i
        from rQchain have SN: "?m \<Longrightarrow> SN_on (qrstep ?nfs ?rQ ?Rb) {t i \<cdot> \<sigma> i}" by (auto simp: minimal_cond_def)
        have SN: "?m \<Longrightarrow> SN_on (qrstep ?nfs ?Q' ?Rb) {t i \<cdot> \<sigma> i}"
          by (rule SN_on_mono[OF SN qrstep_mono], insert NFrQ, auto)
      }
      with Q'chain
      have "min_ichain (?dpp ?Q') s t \<sigma>" by (auto simp: minimal_cond_def)
    }
    with fin      
    show "finite_dpp (dpp d)" unfolding id finite_dpp_def by blast
  qed
qed

end