(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Bin_Tree
imports "../TermFun"
begin

hide_const (open) List.remdups

datatype tree = E | N tree nat tree
setup {* Term_Fun.register_datatypes ["Bin_Tree.tree"] *}

fun getmax where
  "getmax E = 0"
| "getmax (N l x E) = x"
| "getmax (N l x r) = getmax r"

fun delmax where
  "delmax E = E"
| "delmax (N l x E) = l"
| "delmax (N l x r) = N l x (delmax r)"

fun del where
  "del x E = E"
| "del x (N E y r) = 
   (if x < y then (N E y r)
    else if eq x y then del x r
    else N E y (del x r))"
| "del x (N l y r) = 
   (if x < y then N (del x l) y r
    else if x > y then N l y (del x r)
    else N (delmax (del x l)) (getmax (del x l)) (del x r))"

(* Induction proof that we are avoiding: *)
lemma size_delmax: "size (delmax t) \<le> size t"
  by (induct t rule: delmax.induct) auto

lemma size_del: "size (del x t) \<le> size t"
proof (induct t arbitrary: x)
  case E show ?case by simp
next
  case (N l x r y) note IH = this
  note size_delmax[simp] = size_delmax
  show ?case
  proof (cases l)
    case E 
    with IH(2)[of y]
    show ?thesis by auto
  next
    case N
    have "size (delmax (del x l)) \<le> size (del x l)" by simp
    also have "... \<le> size l" by (rule IH)
    finally 
    show ?thesis
      unfolding N
      apply (unfold del.simps, fold N)
      apply (auto simp: IH)
      apply (insert IH(2)[of x]) 
      by auto
  qed
qed

function remdups where
  "remdups E = E"
| "remdups (N l x r) = 
   N (remdups (del x l)) x (remdups (del x r))"
  by pat_completeness auto

define_trs getmax.simps
define_trs delmax.simps
define_trs del.simps
define_trs (no_sim) remdups.psimps

termination remdups
  apply (trs_reduce remdups)
  apply expand_trs_problem
  export_termination_problem "remdups.xml" 
  apply (ceta_import_proof "proofs/remdups.proof.xml")
  apply ceta_eval
done

end
