(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Regression
imports "../TermFun"
begin

text {* simple recursive identity *}

function f :: "nat \<Rightarrow> nat"
where
  "f 0 = 0"
| "f (Suc n) = Suc (f n)"
by pat_completeness auto

define_trs (no_sim) f.psimps

termination f
apply (trs_reduce f)
apply expand_trs_problem
export_termination_problem "f.xml"
apply (ceta_import_proof "proofs/f.proof.xml")
apply ceta_eval 
done

text {* and again, using dp-extraction *}

function ff :: "nat \<Rightarrow> nat"
where
  "ff 0 = 0"
| "ff (Suc n) = Suc (ff n)"
by pat_completeness auto

define_dps ff.psimps

termination ff
apply (dps_reduce ff)
apply expand_dp_problem
export_dp_problem "ff.xml"
apply (ceta_import_dp_proof "proofs/ff.proof.xml")
apply ceta_dp_eval 
done

define_trs ff.simps (* generate simulation rules for ff *)

text {* nested case *}

definition predpred :: "nat \<Rightarrow> nat"
where
  "predpred n = (case n of 
    0 \<Rightarrow> 0
  | Suc m \<Rightarrow> case m of 
       0 \<Rightarrow> 0
     | Suc m' \<Rightarrow> m')"

define_trs predpred_def

text {* if *}

function test1 :: "nat \<Rightarrow> nat"
where
  "test1 n = (if n = 0 then 0 else Suc (test1 (n - Suc 0)))"
by pat_completeness auto

define_trs (no_sim) test1.psimps

termination test1
apply (trs_reduce test1) 
apply expand_trs_problem
export_termination_problem "test1.xml"
apply (ceta_import_proof "proofs/test1.proof.xml") 
apply ceta_eval
done

end