(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Xml
imports
  XML_Parser
  Parser_Monad2
  Mini_Algebraic_Numbers
begin

subsection {* Pretty Printing of XML nodes and documents *}

instantiation xml :: "show" begin
  definition
    shows_attr :: "string \<times> string \<Rightarrow> shows"
  where
    "shows_attr av = (shows (fst av) +@+ ''='' +#+ shows_string (dquote # (snd av) @ [dquote]))"

  definition
    shows_attrs :: "(string\<times>string)list \<Rightarrow> shows"
  where
    "shows_attrs as = shows_map (\<lambda>a. '' '' +#+ shows_attr a) as"

  fun
    shows_XML_indent :: "string \<Rightarrow> nat \<Rightarrow> xml \<Rightarrow> shows"
  where
    "shows_XML_indent ind i (XML n a c t) = (
    ind +#+ ''<'' +#+ shows n +@+ shows_attrs a +@+ (
      if c = [] \<and> t = None then shows_string ''/>'' else (
        ''>'' +#+ (
          if c \<noteq> [] then (
            shows_map (shows_XML_indent (ind @ replicate i (CHR '' '')) i) c +@+ shows_string ind
          ) else id
        ) +@+ (if t = None then id else shows_string(the t)) +@+
          ''</'' +#+ shows n +@+ shows_string ''>''
      )
    )
  )"
  definition "shows_prec (d::nat) xml \<equiv> shows_XML_indent [newline] 2 xml"
  
  lemma shows_attr_assoc: "(s +#+ shows_attr av) r @ t = (s +#+ shows_attr av) (r@t)"
    unfolding shows_attr_def by (cases av) (auto simp: show_defs o_def dquote_def)
  lemma shows_attr_assoc_elt: "\<forall>av\<in>set avs. (s +#+ shows_attr av) r @ t = (s +#+ shows_attr av) (r@t)"
    using shows_attr_assoc by best
  lemma shows_attrs_assoc: "shows_attrs as r @ s = shows_attrs as (r @ s)"
    unfolding shows_attrs_def shows_map_assoc[OF shows_attr_assoc_elt[of as "'' ''"]] by simp
  lemma assoc_xml': "shows_XML_indent ind i xml r @ s = shows_XML_indent ind i xml (r @ s)"
  proof (induct xml arbitrary: ind i r s)
    let ?elt = "shows_XML_indent"
    case (XML n a cs t ind i)
    hence elt: "\<And>r s.\<forall>c\<in>set cs. ?elt (ind @ replicate i CHR '' '') i c r @ s = ?elt (ind @ replicate i CHR '' '') i c (r@s)" by auto
    have IH: "\<And>r s. shows_map (?elt (ind @ replicate i CHR '' '') i) cs r @ s = shows_map (?elt (ind @ replicate i CHR '' '') i) cs (r@s)"
      by (rule shows_map_assoc[OF elt])
    show ?case
    proof (cases "cs = []")
      case True show ?thesis
      proof (cases t)
        case None
        with True have empty: "cs = [] \<and> t = None" by simp
        show ?thesis by (auto simp: shows_attrs_assoc if_P[OF empty])
      next
        case (Some ts)
        hence t: "\<not>(cs = [] \<and> t = None)" and some_t: "\<not> t = None" by auto
        from True have empty: "\<not> cs \<noteq> []" by simp
        show ?thesis by (auto simp: if_not_P[OF t] if_not_P[OF empty] if_not_P[OF some_t] shows_attrs_assoc)
      qed
    next
      case False
      hence ne: "\<not>(cs = [] \<and> t = None)" by simp
      show ?thesis
      proof (cases t)
        case None show ?thesis by (auto simp: if_not_P[OF ne] if_P[OF False] if_P[OF None] shows_attrs_assoc IH)
      next
        case (Some ts)
        hence some_t: "\<not> t = None" by simp
        show ?thesis by (auto simp: if_not_P[OF ne] if_P[OF False] if_not_P[OF some_t] shows_attrs_assoc IH)
      qed
    qed
  qed

  lemma assoc_xml: "shows_prec d (xml::xml) r @ s = shows_prec d xml (r @ s)" unfolding shows_prec_xml_def by (rule assoc_xml')
  standard_shows_list assoc_xml
end

instantiation xmldoc :: "show" begin
  fun shows_xmldoc where "shows_xmldoc(XMLDOC h x) = (h +#+ shows_nl +@+ shows x)"
  definition "shows_prec (d::nat) doc = shows_xmldoc doc"

  lemma assoc_xmldoc: "shows_prec d (x::xmldoc) r @ s = shows_prec d x (r @ s)" by (cases x) (auto simp: shows_prec_xmldoc_def)
  standard_shows_list assoc_xmldoc
end

subsection {* XML Transformers *}

(* transformers on xml nodes *)
type_synonym 'a sparse_result = "(string,'a) parse_result"
type_synonym
  'a xmlt = "xml \<Rightarrow> 'a sparse_result"

definition
  map :: "(xml \<Rightarrow> ('e,'a)parse_result) \<Rightarrow> xml list \<Rightarrow> ('e,'a list) parse_result"
where
  [code_unfold]: "map = map_parse_result"

lemma map_mono[partial_function_mono]:
  fixes C :: "xml \<Rightarrow> ('b \<Rightarrow> ('e,'c) parse_result) \<Rightarrow> ('e,'d) parse_result"
  assumes C: "\<And>y. y \<in> set B \<Longrightarrow> mono_parse_result (C y)"
  shows "mono_parse_result (\<lambda>f. map (\<lambda>y. C y f) B)" 
  unfolding map_def by (auto intro: partial_function_mono C)

hide_const (open) map



fun
  get_name :: "xml \<Rightarrow> string" 
where
  "get_name (XML name _ _ _) = name"
hide_const (open) get_name

fun
  xml_children :: "xml \<Rightarrow> xml list" 
where
  "xml_children (XML _ _ cs _) = cs"

definition
  string2bool :: "string \<Rightarrow> bool sparse_result"
where
  "string2bool s \<equiv> (if s = ''true''
     then return True
     else (if s = ''false''
       then return False
       else error (''cannot convert '' @ s @ '' into Boolean'')))"

fun
  "text" :: "string \<Rightarrow> string xmlt"
where
  "text tag (XML n atts cs t) = 
     (if n = tag \<and> atts = [] \<and> cs = [] \<and> t \<noteq> None \<and> the t \<noteq> []
       then return (the t)
       else error (concat [''could not extract text for '', tag,'' from '',
         [newline], show (XML n atts cs t)]))"
hide_const (open) "text"

fun
 bool :: "string \<Rightarrow> bool xmlt"
where
  "bool tag node = (Xml.text tag node \<guillemotright>= string2bool)"
hide_const (open) bool

definition
  no_transform :: "string \<Rightarrow> 'a xmlt"
where
  "no_transform tag xml = error (concat [''could not transform the following
     xml element (expected '', tag, '')'', [newline], show xml])"

definition xml_if :: "(xml \<Rightarrow> bool) \<Rightarrow> 'a xmlt \<Rightarrow> 'a xmlt \<Rightarrow> 'a xmlt"
  where "xml_if p p1 p2 x = (if p x then p1 x else p2 x)" 

fun xml_elem_len :: "xml \<Rightarrow> nat" where "xml_elem_len (XML _ _ xs _) = length xs"

lemma xml_if_mono[partial_function_mono]:
  assumes p1: "\<And>y. mono_parse_result (p1 y)" and
  p2: "\<And>y. mono_parse_result (p2 y)"
  shows "mono_parse_result (\<lambda>g. xml_if p (\<lambda>y. p1 y g) (\<lambda>y. p2 y g) x)" 
  by (cases x, auto intro!: partial_function_mono p1 p2 simp: xml_if_def)

fun
  xml0elements :: "string \<Rightarrow> 'a \<Rightarrow> 'a xmlt"
where
  "xml0elements tag x (XML name atts cs text) = 
     (if name = tag \<and> atts = [] \<and> cs = [] \<and> text = None
       then return x 
       else no_transform tag (XML name atts cs text))"

fun list1element :: "'a list \<Rightarrow> ('a) option" where
  "list1element [x] = Some (x)"
| "list1element _ = None"

fun
  xml1element ::
    "string \<Rightarrow> 'a xmlt \<Rightarrow> ('a \<Rightarrow> 'b) \<Rightarrow> 'b xmlt"
where
  "xml1element tag p1 f xml = (case xml of XML name atts cs text \<Rightarrow>
     (if name = tag \<and> atts = [] \<and> text = None
       then case list1element cs of 
         Some (cs1) \<Rightarrow> do {
           a \<leftarrow> p1 cs1;
           return (f a)
         }
       | None \<Rightarrow> no_transform tag xml 
      else no_transform tag xml))"

lemma xml1element_mono[partial_function_mono]:
  assumes p: "\<And>y. mono_parse_result (p1 y)"
  shows "mono_parse_result (\<lambda>g. xml1element t (\<lambda>y. p1 y g) f x)" 
    by (cases x, cases "list1element (xml_children x)", auto intro!: partial_function_mono p)

fun list2elements :: "'a list \<Rightarrow> ('a \<times> 'a) option" where
  "list2elements [x,y] = Some (x,y)"
| "list2elements _ = None"

fun
  xml2elements ::
    "string \<Rightarrow> 'a xmlt \<Rightarrow> 'b xmlt \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> 'c) \<Rightarrow> 'c xmlt"
where
  "xml2elements tag p1 p2 f xml = (case xml of XML name atts cs text \<Rightarrow>
     (if name = tag \<and> atts = [] \<and> text = None
       then case list2elements cs of 
         Some (cs1,cs2) \<Rightarrow> do {
           a \<leftarrow> p1 cs1;
           b \<leftarrow> p2 cs2;
           return (f a b)
         }
       | None \<Rightarrow> no_transform tag xml 
      else no_transform tag xml))"

lemma xml2elements_mono[partial_function_mono]:
  assumes p: "\<And>y. mono_parse_result (p1 y)"
             "\<And>y. mono_parse_result (p2 y)"
  shows "mono_parse_result (\<lambda>g. xml2elements t (\<lambda>y. p1 y g) (\<lambda> y. p2 y g) f x)" 
    by (cases x, cases "list2elements (xml_children x)", auto intro!: partial_function_mono p)

fun list3elements :: "'a list \<Rightarrow> ('a \<times> 'a \<times> 'a) option" where
  "list3elements [x,y,z] = Some (x,y,z)"
| "list3elements _ = None"

fun
  xml3elements ::
    "string \<Rightarrow> 'a xmlt \<Rightarrow> 'b xmlt \<Rightarrow> 'c xmlt \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> 'c \<Rightarrow> 'd) \<Rightarrow> 'd xmlt"
where
  "xml3elements tag p1 p2 p3 f xml = (case xml of XML name atts cs text \<Rightarrow>
     (if name = tag \<and> atts = [] \<and> text = None
       then case list3elements cs of 
         Some (cs1,cs2,cs3) \<Rightarrow> do {
           a \<leftarrow> p1 cs1;
           b \<leftarrow> p2 cs2;
           c \<leftarrow> p3 cs3;
           return (f a b c)
         }
       | None \<Rightarrow> no_transform tag xml 
      else no_transform tag xml))"

lemma xml3elements_mono[partial_function_mono]:
  assumes p: "\<And>y. mono_parse_result (p1 y)"
             "\<And>y. mono_parse_result (p2 y)"
             "\<And>y. mono_parse_result (p3 y)"
  shows "mono_parse_result (\<lambda>g. xml3elements t (\<lambda>y. p1 y g) (\<lambda> y. p2 y g) (\<lambda> y. p3 y g) f x)" 
    by (cases x, cases "list3elements (xml_children x)", auto intro!: partial_function_mono p)

fun list4elements :: "'a list \<Rightarrow> ('a \<times> 'a \<times> 'a \<times> 'a) option" where
  "list4elements [x,y,z,u] = Some (x,y,z,u)"
| "list4elements _ = None"

fun
  xml4elements ::
    "string \<Rightarrow> 'a xmlt \<Rightarrow> 'b xmlt \<Rightarrow> 'c xmlt \<Rightarrow> 'd xmlt \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> 'c \<Rightarrow> 'd \<Rightarrow> 'e) \<Rightarrow> 'e xmlt"
where
  "xml4elements tag p1 p2 p3 p4 f xml = (case xml of XML name atts cs text \<Rightarrow>
     (if name = tag \<and> atts = [] \<and> text = None
       then case list4elements cs of 
         Some (cs1,cs2,cs3,cs4) \<Rightarrow> do {
           a \<leftarrow> p1 cs1;
           b \<leftarrow> p2 cs2;
           c \<leftarrow> p3 cs3;
           d \<leftarrow> p4 cs4;
           return (f a b c d)
         }
       | None \<Rightarrow> no_transform tag xml 
      else no_transform tag xml))"

lemma xml4elements_mono[partial_function_mono]:
  assumes p: "\<And>y. mono_parse_result (p1 y)"
             "\<And>y. mono_parse_result (p2 y)"
             "\<And>y. mono_parse_result (p3 y)"
             "\<And>y. mono_parse_result (p4 y)"
  shows "mono_parse_result (\<lambda>g. xml4elements t (\<lambda>y. p1 y g) (\<lambda> y. p2 y g) (\<lambda> y. p3 y g) (\<lambda> y. p4 y g) f x)" 
    by (cases x, cases "list4elements (xml_children x)", auto intro!: partial_function_mono p)

fun list5elements :: "'a list \<Rightarrow> ('a \<times> 'a \<times> 'a \<times> 'a \<times> 'a) option" where
  "list5elements [x,y,z,u,v] = Some (x,y,z,u,v)"
| "list5elements _ = None"

fun
  xml5elements ::
    "string \<Rightarrow> 'a xmlt \<Rightarrow> 'b xmlt \<Rightarrow> 'c xmlt \<Rightarrow> 'd xmlt \<Rightarrow> 'e xmlt \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> 'c \<Rightarrow> 'd \<Rightarrow> 'e \<Rightarrow> 'f) \<Rightarrow> 'f xmlt"
where
  "xml5elements tag p1 p2 p3 p4 p5 f xml = (case xml of XML name atts cs text \<Rightarrow>
     (if name = tag \<and> atts = [] \<and> text = None
       then case list5elements cs of 
         Some (cs1,cs2,cs3,cs4,cs5) \<Rightarrow> do {
           a \<leftarrow> p1 cs1;
           b \<leftarrow> p2 cs2;
           c \<leftarrow> p3 cs3;
           d \<leftarrow> p4 cs4;
           e \<leftarrow> p5 cs5;
           return (f a b c d e)
         }
       | None \<Rightarrow> no_transform tag xml 
      else no_transform tag xml))"

lemma xml5elements_mono[partial_function_mono]:
  assumes p: "\<And>y. mono_parse_result (p1 y)"
             "\<And>y. mono_parse_result (p2 y)"
             "\<And>y. mono_parse_result (p3 y)"
             "\<And>y. mono_parse_result (p4 y)"
             "\<And>y. mono_parse_result (p5 y)"
  shows "mono_parse_result (\<lambda>g. xml5elements t (\<lambda>y. p1 y g) (\<lambda> y. p2 y g) (\<lambda> y. p3 y g) (\<lambda> y. p4 y g) (\<lambda> y. p5 y g) f x)" 
  by (cases x, cases "list5elements (xml_children x)", auto intro!: partial_function_mono p)

fun list6elements :: "'a list \<Rightarrow> ('a \<times> 'a \<times> 'a \<times> 'a \<times> 'a \<times> 'a) option" where
  "list6elements [x,y,z,u,v,w] = Some (x,y,z,u,v,w)"
| "list6elements _ = None"

fun
  xml6elements ::
    "string \<Rightarrow> 'a xmlt \<Rightarrow> 'b xmlt \<Rightarrow> 'c xmlt \<Rightarrow> 'd xmlt \<Rightarrow> 'e xmlt \<Rightarrow> 'f xmlt \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> 'c \<Rightarrow> 'd \<Rightarrow> 'e \<Rightarrow> 'f \<Rightarrow> 'g) \<Rightarrow> 'g xmlt"
where
  "xml6elements tag p1 p2 p3 p4 p5 p6 f xml = (case xml of XML name atts cs text \<Rightarrow>
     (if name = tag \<and> atts = [] \<and> text = None
       then case list6elements cs of 
         Some (cs1,cs2,cs3,cs4,cs5,cs6) \<Rightarrow> do {
           a \<leftarrow> p1 cs1;
           b \<leftarrow> p2 cs2;
           c \<leftarrow> p3 cs3;
           d \<leftarrow> p4 cs4;
           e \<leftarrow> p5 cs5;
           ff \<leftarrow> p6 cs6;
           return (f a b c d e ff)
         }
       | None \<Rightarrow> no_transform tag xml 
      else no_transform tag xml))"

lemma xml6elements_mono[partial_function_mono]:
  assumes p: "\<And>y. mono_parse_result (p1 y)"
             "\<And>y. mono_parse_result (p2 y)"
             "\<And>y. mono_parse_result (p3 y)"
             "\<And>y. mono_parse_result (p4 y)"
             "\<And>y. mono_parse_result (p5 y)"
             "\<And>y. mono_parse_result (p6 y)"
  shows "mono_parse_result (\<lambda>g. xml6elements t (\<lambda>y. p1 y g) (\<lambda> y. p2 y g) (\<lambda> y. p3 y g) (\<lambda> y. p4 y g) (\<lambda> y. p5 y g) (\<lambda> y. p6 y g) f x)" 
  by (cases x, cases "list6elements (xml_children x)", auto intro!: partial_function_mono p)

fun
  xml0to1elements :: "string \<Rightarrow> 'a xmlt \<Rightarrow> ('a option \<Rightarrow> 'b) \<Rightarrow> 'b xmlt"
where
  "xml0to1elements tag p1 f (XML name atts cs text) = (
     let l = length cs in
     (if name = tag \<and> atts = [] \<and> l \<ge> 0 \<and> l \<le> 1 \<and> text = None
       then do {
         (if l = 1
           then do {
             x1 \<leftarrow> p1 (cs ! 0);
             return (f (Some x1))
           } else return (f None))
       } else no_transform tag (XML name atts cs text)))"

lemma xml0to1elements_mono[partial_function_mono]:
  assumes p1: "\<And>y. mono_parse_result (p1 y)"
  shows "mono_parse_result (\<lambda>g. xml0to1elements t (\<lambda>y. p1 y g) f x)" 
  by (cases x, auto intro!: partial_function_mono p1)


fun
  xml1to2elements :: "string \<Rightarrow> 'a xmlt \<Rightarrow> 'b xmlt \<Rightarrow> ('a \<Rightarrow> 'b option \<Rightarrow> 'c) \<Rightarrow> 'c xmlt"
where
  "xml1to2elements tag p1 p2 f (XML name atts cs text) = (
     let l = length cs in
     (if name = tag \<and> atts = [] \<and> l \<ge> 1 \<and> l \<le> 2 \<and> text = None
       then do {
         x1 \<leftarrow> p1 (cs ! 0);
         (if l = 2
           then do {
             x2 \<leftarrow> p2 (cs ! 1);
             return (f x1 (Some x2))
           } else return (f x1 None))
       } else no_transform tag (XML name atts cs text)))"

lemma xml1to2elements_mono[partial_function_mono]:
  assumes p1: "\<And>y. mono_parse_result (p1 y)"
              "\<And>y. mono_parse_result (p2 y)"
  shows "mono_parse_result (\<lambda>g. xml1to2elements t (\<lambda>y. p1 y g) (\<lambda>y. p2 y g) f x)" 
  by (cases x, auto intro!: partial_function_mono p1)

(* parses the first element, the second argument must be a boolean. If the boolean is true, then the last element is parsed additionally *)
fun
  xml2nd_choice :: "string \<Rightarrow> 'a xmlt \<Rightarrow> string \<Rightarrow> 'b xmlt \<Rightarrow> ('a \<Rightarrow> 'b option \<Rightarrow> 'c) \<Rightarrow> 'c xmlt"
where
  "xml2nd_choice tag p1 cn p2 f (XML name atts cs text) = (
     let l = length cs in
     (if name = tag \<and> atts = [] \<and> l \<ge> 2 \<and> text = None
       then do {
         x1 \<leftarrow> p1 (cs ! 0);
         b \<leftarrow> Xml.bool cn (cs ! 1);
         (if b
           then do {
             x2 \<leftarrow> p2 (cs ! (l - 1));
             return (f x1 (Some x2))
           } else return (f x1 None))
       } else no_transform tag (XML name atts cs text)))"

lemma xml2nd_choice_mono[partial_function_mono]:
  assumes p1: "\<And>y. mono_parse_result (p1 y)"
              "\<And>y. mono_parse_result (p2 y)"
  shows "mono_parse_result (\<lambda>g. xml2nd_choice t (\<lambda>y. p1 y g) h (\<lambda>y. p2 y g) f x)" 
  by (cases x, auto intro!: partial_function_mono p1)

fun
  xml2to3elements :: "string \<Rightarrow> 'a xmlt \<Rightarrow> 'b xmlt \<Rightarrow> 'c xmlt \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> 'c option \<Rightarrow> 'd) \<Rightarrow> 'd xmlt"
where
  "xml2to3elements tag p1 p2 p3 f (XML name atts cs text) = (
     let l = length cs in
     (if name = tag \<and> atts = [] \<and> l \<ge> 2 \<and> l \<le> 3 \<and> text = None
       then do {
         x1 \<leftarrow> p1 (cs ! 0);
         x2 \<leftarrow> p2 (cs ! 1);
         (if l = 3
           then do {
             x3 \<leftarrow> p3 (cs ! 2);
             return (f x1 x2 (Some x3))
           } else return (f x1 x2 None))
       } else no_transform tag (XML name atts cs text)))"

lemma xml2to3elements_mono[partial_function_mono]:
  assumes p1: "\<And>y. mono_parse_result (p1 y)"
              "\<And>y. mono_parse_result (p2 y)"
              "\<And>y. mono_parse_result (p3 y)"
  shows "mono_parse_result (\<lambda>g. xml2to3elements t (\<lambda>y. p1 y g) (\<lambda>y. p2 y g) (\<lambda>y. p3 y g) f x)" 
  by (cases x, auto intro!: partial_function_mono p1)

fun
  xml3to4elements :: "string \<Rightarrow> 'a xmlt \<Rightarrow> 'b xmlt \<Rightarrow> 'c xmlt \<Rightarrow> 'd xmlt \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> 'c option \<Rightarrow> 'd \<Rightarrow> 'e) \<Rightarrow> 'e xmlt"
where
  "xml3to4elements tag p1 p2 p3 p4 f (XML name atts cs text) = (
     let l = length cs in
     (if name = tag \<and> atts = [] \<and> l \<ge> 3 \<and> l \<le> 4 \<and> text = None
       then do {
         x1 \<leftarrow> p1 (cs ! 0);
         x2 \<leftarrow> p2 (cs ! 1);
         (if l = 4
           then do {
             x3 \<leftarrow> p3 (cs ! 2);
             x4 \<leftarrow> p4 (cs ! 3);
             return (f x1 x2 (Some x3) x4)
           } else do {
             x4 \<leftarrow> p4 (cs ! 2);
             return (f x1 x2 None x4)
           } )
       } else no_transform tag (XML name atts cs text)))"

lemma xml3to4elements_mono[partial_function_mono]:
  assumes p1: "\<And>y. mono_parse_result (p1 y)"
              "\<And>y. mono_parse_result (p2 y)"
              "\<And>y. mono_parse_result (p3 y)"
              "\<And>y. mono_parse_result (p4 y)"
  shows "mono_parse_result (\<lambda>g. xml3to4elements t (\<lambda>y. p1 y g) (\<lambda>y. p2 y g) (\<lambda>y. p3 y g) (\<lambda>y. p4 y g) f x)" 
  by (cases x, auto intro!: partial_function_mono p1)

fun xml_cs :: "xml \<Rightarrow> xml list"
  where "xml_cs (XML _ _ cs _) = cs" 

fun
  xml_many_elements :: "string \<Rightarrow> 'a xmlt \<Rightarrow> ('a list \<Rightarrow> 'b) \<Rightarrow> 'b xmlt"
where
  "xml_many_elements tag p f (XML name atts cs text) =
     (if name = tag \<and> atts = [] \<and> text = None
       then (Xml.map p cs \<guillemotright>= (return \<circ> f))
       else no_transform tag (XML name atts cs text))"

lemma xml_many_elements_mono[partial_function_mono]:
  fixes p1 :: "xml \<Rightarrow> ('b \<Rightarrow> (string,'c) parse_result) \<Rightarrow> (string,'d) parse_result"
  assumes p1: "\<And>y. y \<in> set (xml_cs x) \<Longrightarrow> mono_parse_result (p1 y)"
  shows "mono_parse_result (\<lambda>g. xml_many_elements t (\<lambda>y. p1 y g) f x)" 
  using p1 
  by (cases x, auto intro!: partial_function_mono)

fun
  xml1many_elements_gen :: "string \<Rightarrow> 'a xmlt \<Rightarrow> ('a \<Rightarrow> 'b xmlt) \<Rightarrow> ('a \<Rightarrow> 'b list \<Rightarrow> 'c) \<Rightarrow> 'c xmlt"
where
  "xml1many_elements_gen tag p1 p2 f (XML name atts cs text) =
     (if name = tag \<and> atts = [] \<and> text = None \<and> cs \<noteq> []
       then case cs of h # t \<Rightarrow> do {
         x \<leftarrow> p1 h;
         xs \<leftarrow> Xml.map (p2 x) t;
         return (f x xs)
       } else no_transform tag (XML name atts cs text))"

(* TODO 
lemma xml1many_elements_gen_mono[partial_function_mono]:
  fixes p1 :: "xml \<Rightarrow> ('b \<Rightarrow> 'c parse_result) \<Rightarrow> 'd parse_result"
  assumes p1: "\<And>y. mono_parse_result (p1 y)"
              "\<And>y. mono_parse_result (p2 y)"
  shows "mono_parse_result (\<lambda>g. xml1many_elements_gen t (\<lambda>y. p1 y g) (\<lambda>y. p2 y g) f x)" 
  by (cases x, auto intro!: partial_function_mono p1)
*)

definition
  xml1many_elements :: "string \<Rightarrow> 'a xmlt \<Rightarrow> 'b xmlt \<Rightarrow> ('a \<Rightarrow> 'b list \<Rightarrow> 'c) \<Rightarrow> 'c xmlt"
where
  "xml1many_elements tag p1 p2 = xml1many_elements_gen tag p1 (\<lambda>_. p2)"

lemma xml1many_elements_mono[partial_function_mono]:
  fixes p1 :: "xml \<Rightarrow> ('b \<Rightarrow> (string,'c) parse_result) \<Rightarrow> (string,'d) parse_result"
  assumes p1: "\<And>y. mono_parse_result (p1 y)"
              "\<And>y. y \<in> set (tl (xml_cs x)) \<Longrightarrow> mono_parse_result (p2 y)"
  shows "mono_parse_result (\<lambda>g. xml1many_elements t (\<lambda>y. p1 y g) (\<lambda>y. p2 y g) f x)" 
  unfolding xml1many_elements_def using p1
  by (cases x, cases "xml_children x", auto intro!: partial_function_mono)

fun length_ge_2 :: "'a list \<Rightarrow> bool" where 
  "length_ge_2 (_ # _ # _) = True"
| "length_ge_2 _ = False"

fun
  xml2many_elements :: "string \<Rightarrow> 'a xmlt \<Rightarrow> 'b xmlt \<Rightarrow> 'c xmlt \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> 'c list \<Rightarrow> 'd) \<Rightarrow> 'd xmlt"
where
  "xml2many_elements tag p1 p2 p3 f (XML name atts cs text) =
     (if name = tag \<and> atts = [] \<and> text = None \<and> length_ge_2 cs 
       then case cs of cs0 # cs1 # t \<Rightarrow> do {
         x \<leftarrow> p1 cs0;
         y \<leftarrow> p2 cs1;
         xs \<leftarrow> Xml.map p3 t;
         return (f x y xs)
       } else no_transform tag (XML name atts cs text))"

lemma xml2many_elements_mono[partial_function_mono]:
  fixes p1 :: "xml \<Rightarrow> ('b \<Rightarrow> (string,'c) parse_result) \<Rightarrow> (string,'d) parse_result"
  assumes p1: "\<And>y. mono_parse_result (p1 y)"
              "\<And>y. mono_parse_result (p2 y)"
              "\<And>y. mono_parse_result (p3 y)"
  shows "mono_parse_result (\<lambda>g. xml2many_elements t (\<lambda>y. p1 y g) (\<lambda>y. p2 y g) (\<lambda>y. p3 y g) f x)" 
  by (cases x, cases "xml_children x", (auto intro!: partial_function_mono)[1], cases "tl (xml_children x)", auto intro!: partial_function_mono p1)

fun
  xml1or2many_elements :: "string \<Rightarrow> 'a xmlt \<Rightarrow> 'b xmlt \<Rightarrow> 'c xmlt \<Rightarrow> ('a \<Rightarrow> 'b option \<Rightarrow> 'c list \<Rightarrow> 'd) \<Rightarrow> 'd xmlt"
where
  "xml1or2many_elements tag p1 p2 p3 f (XML name atts cs text) =
     (if name = tag \<and> atts = [] \<and> text = None \<and> cs \<noteq> []
       then case cs of cs0 # tt \<Rightarrow> do { 
         x \<leftarrow> p1 cs0;
         (case tt of cs1 # t \<Rightarrow> do {
         try do {
           y \<leftarrow> p2 cs1;
           xs \<leftarrow> Xml.map p3 t;
           return (f x (Some y) xs)
         }  catch ( \<lambda> _. 
               do {
                 xs \<leftarrow> Xml.map p3 tt;
                 return (f x None xs)
                  })
       }
     | [] \<Rightarrow> return (f x None []))} 
     else no_transform tag (XML name atts cs text))"


fun
  xml1many2elements_gen :: "string \<Rightarrow> 'a xmlt \<Rightarrow> ('a \<Rightarrow> 'b xmlt) \<Rightarrow> 'c xmlt \<Rightarrow> 'd xmlt \<Rightarrow> ('a \<Rightarrow> 'b list \<Rightarrow> 'c \<Rightarrow> 'd \<Rightarrow> 'e) \<Rightarrow> 'e xmlt"
where
  "xml1many2elements_gen tag p1 p2 p3 p4 f (XML name atts cs text) = (
     let ds = List.rev cs; l = length cs in
     (if name = tag \<and> atts = [] \<and> text = None \<and> l \<ge> 3
       then do {
         x \<leftarrow> p1 (cs ! 0);
         xs \<leftarrow> Xml.map (p2 x) (tl (take (l - 2) cs));
         y \<leftarrow> p3 (ds ! 1);
         z \<leftarrow> p4 (ds ! 0);
         return (f x xs y z)
       } else no_transform tag (XML name atts cs text)))"

lemma xml1many2elements_gen_mono[partial_function_mono]:
  fixes p1 :: "xml \<Rightarrow> ('b \<Rightarrow> (string,'c) parse_result) \<Rightarrow> (string,'d) parse_result"
  assumes p1: "\<And>y. mono_parse_result (p1 y)"
              "\<And>y. mono_parse_result (p3 y)"
              "\<And>y. mono_parse_result (p4 y)"
  shows "mono_parse_result (\<lambda>g. xml1many2elements_gen t (\<lambda>y. p1 y g) p2 (\<lambda>y. p3 y g) (\<lambda>y. p4 y g) f x)" 
  by (cases x, auto intro!: partial_function_mono p1)

fun
  xml1many2elements :: "string \<Rightarrow> 'a xmlt \<Rightarrow> 'b xmlt \<Rightarrow> 'c xmlt \<Rightarrow> 'd xmlt \<Rightarrow> ('a \<Rightarrow> 'b list \<Rightarrow> 'c \<Rightarrow> 'd \<Rightarrow> 'e) \<Rightarrow> 'e xmlt"
where
  "xml1many2elements tag p1 p2 = xml1many2elements_gen tag p1 (\<lambda>_. p2)"

fun
  xml_many2elements :: "string \<Rightarrow> 'a xmlt \<Rightarrow> 'b xmlt \<Rightarrow> 'c xmlt \<Rightarrow> ('a list \<Rightarrow> 'b \<Rightarrow> 'c \<Rightarrow> 'd) \<Rightarrow> 'd xmlt"
where
  "xml_many2elements tag p1 p2 p3 f (XML name atts cs text) = (
     let ds = List.rev cs in
     (if name = tag \<and> atts = [] \<and> text = None \<and> length_ge_2 cs
       then do {
         xs \<leftarrow> Xml.map p1 (List.rev (tl (tl ds)));
         y \<leftarrow> p2 (ds ! 1);
         z \<leftarrow> p3 (ds ! 0);
         return (f xs y z)
       } else no_transform tag (XML name atts cs text)))"

definition
  xml_options :: "(string \<times> 'a xmlt) list \<Rightarrow> 'a xmlt"
where
  "xml_options ps x \<equiv>
    (case map_of ps (Xml.get_name x) of 
      None \<Rightarrow> error (concat [''expected one of: '', concat (map (\<lambda>p. fst p @ '' '') ps), [newline], ''but found'', [newline], show x])
    | Some p \<Rightarrow> p x)"

lemma xml_options_mono_gen[partial_function_mono]:
  assumes p: "\<And> k p. (k,p) \<in> set ps \<Longrightarrow> mono_parse_result (p x)"
  shows "mono_parse_result (\<lambda> g. xml_options (map (\<lambda> (k,p). (k, (\<lambda> y. p y g))) ps) x)"
proof -
  {
    fix g
    have "(map (\<lambda>p. fst p @ '' '')
                   (map (\<lambda>(k, p). (k, \<lambda>y. p y g)) ps)) = 
      map (\<lambda>p. fst p @ '' '') ps" by (induct ps, auto)
  } note id = this
  {
    fix z
    have "mono_parse_result
      (\<lambda>g. case map_of (map (\<lambda>(k, p). (k, \<lambda>y. p y g)) ps) (Xml.get_name x) of
        None \<Rightarrow> z
      | Some p \<Rightarrow> p x)"
      using p
    proof (induct ps)
      case Nil
      show ?case by (auto intro: partial_function_mono)
    next
      case (Cons kp ps)
      obtain k p where kp: "kp = (k,p)" by force
      note Cons = Cons[unfolded kp]
      from Cons(2) have monop: "mono_parse_result (p x)" and mono: "\<And> k p. (k,p) \<in> set ps \<Longrightarrow> mono_parse_result (p x)" by auto
      show ?case 
      proof (cases "Xml.get_name x = k")
        case True
        thus ?thesis unfolding kp using monop by auto
      next
        case False
        thus ?thesis using Cons(1) mono unfolding kp by auto
      qed
    qed
  } note main = this
  show ?thesis unfolding xml_options_def 
    unfolding id
    by (rule main)
qed

(* instantiate this lemma to have the monotonicity lemmas for lists of variable lengths which
   are then applicable, e.g., for lists of length 3 it would be

mono_parse_result (p1 x) \<Longrightarrow> mono_parse_result (p2 x) \<Longrightarrow> mono_parse_result (p3 x) 
\<Longrightarrow> mono_parse_result (\<lambda>g. xml_options [(k1, \<lambda>y. p1 y g), (k2, \<lambda>y. p2 y g), (k3, \<lambda>y. p3 y g)] x)

*)
local_setup {*
  fn lthy => 
  let
    val N = 24 (* we require monotonicity lemmas for xml-options for lists up to length N *) 
    val thy = Proof_Context.theory_of lthy
    val xml_options = @{term "xml_options :: (string \<times> (xml \<Rightarrow> (string, 'd) parse_result)) list \<Rightarrow> xml \<Rightarrow> (string, 'd) parse_result"}
    val mono_parse_result = @{term "mono_parse_result :: (('a \<Rightarrow> ('b, 'c) parse_result) \<Rightarrow> (string, 'd) parse_result) \<Rightarrow> bool"}
    val ktyp = @{typ string}
    val x = @{term "x :: xml"}
    val y = @{term "y :: xml"}
    val g = @{term "g :: 'a \<Rightarrow> ('b, 'c) parse_result"}
    val ptyp = @{typ "xml \<Rightarrow> ('a \<Rightarrow> ('b, 'c) parse_result) \<Rightarrow> (string, 'd) parse_result"}
    fun k i = Free ("k" ^ string_of_int i,ktyp)
    fun p i = Free ("p" ^ string_of_int i,ptyp)
    fun prem i = HOLogic.mk_Trueprop (mono_parse_result $ (p i $ x))
    fun prems n = 1 upto n |> map prem
    fun pair i = HOLogic.mk_prod (k i, lambda y (p i $ y $ g))
    fun pair2 i = HOLogic.mk_prod (k i, p i)
    fun list n = 1 upto n |> map pair |> HOLogic.mk_list (@{typ "(string \<times> (xml \<Rightarrow> (string, 'd) parse_result))"})
    fun list2 n = 1 upto n |> map pair2 |> HOLogic.mk_list (HOLogic.mk_prodT (ktyp,ptyp))
    fun concl n = HOLogic.mk_Trueprop (mono_parse_result $ lambda g (xml_options $ (list n) $ x))
    fun xs n = x :: (1 upto n |> map (fn i => [p i, k i]) |> List.concat)
       |> map (fst o dest_Free)
    fun tac n pc = let
      val {prems = prems, context = ctxt} = pc
      val mono_thm = Drule.instantiate' 
            (map (SOME o ctyp_of thy) [@{typ 'a},@{typ 'b},@{typ 'c},@{typ 'd}]) 
            (map (SOME o cterm_of thy) [list2 n,x]) @{thm xml_options_mono_gen}
      in 
        Method.insert_tac (mono_thm :: prems) 1 THEN force_tac ctxt 1
      end
    fun thm n = Goal.prove lthy (xs n) (prems n) (concl n) (tac n)
    val thms = map thm (0 upto N)
   in Local_Theory.note ((@{binding "xml_options_mono_thms"}, []), thms) lthy |> snd end
*}

declare xml_options_mono_thms[partial_function_mono]

fun
  xml_choice :: "string \<Rightarrow> 'a xmlt list \<Rightarrow> 'a xmlt"
where
  "xml_choice e [] x = error (concat [''error in parsing choice for '',e,[newline],show x])"
| "xml_choice e (p # ps) x = (try p x catch (\<lambda> _. xml_choice e ps x))"

lemma xml_choice_mono_2[partial_function_mono]:
  assumes p: "mono_parse_result (p1 x)"
             "mono_parse_result (p2 x)"
  shows "mono_parse_result (\<lambda> g. xml_choice e [
  (\<lambda> y. p1 y g), 
  (\<lambda> y. p2 y g) 
  ] x)"
  using p by (auto intro!: partial_function_mono) 

lemma xml_choice_mono_3[partial_function_mono]:
  assumes p: "mono_parse_result (p1 x)"
             "mono_parse_result (p2 x)"
             "mono_parse_result (p3 x)"
  shows "mono_parse_result (\<lambda> g. xml_choice e [
  (\<lambda> y. p1 y g), 
  (\<lambda> y. p2 y g), 
  (\<lambda> y. p3 y g)
  ] x)"
  using p by (auto intro!: partial_function_mono) 

fun
  xml_change :: "'a xmlt \<Rightarrow> ('a \<Rightarrow> 'b) \<Rightarrow> 'b xmlt"
where
  "xml_change p f x = (p x \<guillemotright>= (\<lambda>a. return (f a)))"

fun
  digit2int :: "char \<Rightarrow> int sparse_result"
where
  "digit2int x =
     (if x = CHR ''0'' then return 0 else
     (if x = CHR ''1'' then return 1 else
     (if x = CHR ''2'' then return 2 else
     (if x = CHR ''3'' then return 3 else
     (if x = CHR ''4'' then return 4 else
     (if x = CHR ''5'' then return 5 else
     (if x = CHR ''6'' then return 6 else
     (if x = CHR ''7'' then return 7 else
     (if x = CHR ''8'' then return 8 else
     (if x = CHR ''9'' then return 9 else 
       error ([x] @ '' is not a digit'')))))))))))"

fun
  string2int_aux :: "int \<Rightarrow> string \<Rightarrow> int sparse_result" 
where
  "string2int_aux n [] = return n" |
  "string2int_aux n (d # s) = (digit2int d \<guillemotright>= (\<lambda>m. string2int_aux (10 * n + m) s))"

definition
  string2int :: "string \<Rightarrow> int sparse_result"
where
  "string2int s =
     (if s = []
       then error ''cannot convert empty string into number'' 
       else if take 1 s = ''-''
         then string2int_aux 0 (tl s) \<guillemotright>= (\<lambda> i. return (0 - i))
         else string2int_aux 0 s)"

hide_const string2int_aux

fun
  int :: "string \<Rightarrow> int xmlt"
where
  "int tag x = (Xml.text tag x \<guillemotright>= string2int)"
hide_const (open) int

fun
  nat :: "string \<Rightarrow> nat xmlt"
where
  "nat tag x = do {
    txt \<leftarrow> Xml.text tag x;
    i \<leftarrow> string2int txt;
    return (Int.nat i)
  }"
hide_const (open) nat

definition
  rat :: "rat xmlt"
where
  "rat \<equiv> xml_options [(''integer'', xml_change (Xml.int ''integer'') of_int),
        (''rational'', xml2elements ''rational'' (Xml.int ''numerator'') (Xml.int ''denominator'') (\<lambda> x y. of_int x / of_int y))]"
hide_const (open) rat

definition
  mini_alg :: "mini_alg_number xmlt"
where
  "mini_alg x \<equiv> try (xml_change Xml.rat mini_of_rat x) catch (\<lambda> _.
     do {
       (r,c) \<leftarrow> xml3elements ''algebraic'' (Xml.rat) (Xml.rat) (Xml.rat) (\<lambda> a b c. (mini_of_rat a + mini_of_rat b * sqrtp,c)) x;
       if (c = Mini_Algebraic_Numbers.p) then return r else no_transform ((
         shows ''only sqrt('' \<circ> shows Mini_Algebraic_Numbers.p \<circ> shows '') allowed in irrational part'') []) x
     })"
hide_const (open) mini_alg

end

